/*********************************************************************/
/*                     I.B.M. CONFIDENTIAL                           */
/*      (c) COPYRIGHT IBM CORP. 2005 ALL RIGHTS RESERVED.            */
/*                                                                   */
/*                LICENSED MATERIALS-PROPERTY OF IBM.                */
/*                                                                   */
/*********************************************************************/

/*
 * Version: $Id: cg_square.cpp,v 1.4 2006/12/28 17:59:11 bikshand Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

/** 
 * Description:
 * HTA++ implementation of NAS CG benchmark 
 * CG implements the inverse power-iteration method
 * to find the most dominant eigenvalue and eigenvector.
 * The power-iteration method in-turn uses the
 * Conjugate Gradient algorithm to find the inverse
 * of matrix A. Please refer to 
 * http://www.cs.cmu.edu/~jrs/jrspapers.html for CG algorithm.
 * Please refer to Scientific Computing An Introductory Survey,
 * (Micheal T. Heath), Chapter 4 for inverse power iteration.
 */

/*
 * Limitations: Works only for SQUARE HTAs
 */

#include "npb-C.h"
#include <unistd.h>
#include <iostream>
#include <fstream>

#include "cg.h"

#define T_total   0
#define T_reduce1 1 
#define T_reduce2 2
#define T_reduce3 3
#define T_reduce4 4
#define T_reduce5 5
#define T_lmatmul1 6
#define T_lmatmul2 7
#define T_misc1    8
#define T_misc2    9
#define T_misc3    10
#define T_misc4    11
#define T_ltrans1  12
#define T_ltrans2  13
#define T_trans1   14
#define T_trans2   15 
#define T_ldot     16

typedef Tuple<2> T;
/* print and debug methods */



void print_timers ()

{
  char* name [] = { "total", "reduce1", "reduce2", "reduce3", "reduce4", "reduce5", "lmatmul1", "lmatmul2", "misc1", "misc2", "misc3", 
		    "misc4", "ltrans1", "ltrans2", "trans1", "trans2",  "ldot"};
  
  for (int i = T_total; i <= T_ldot; i++)
    IO::cout << name[i] << " : " << timer_read (i) << endl;
}


typedef HTA<double, 2, Traits::HTA_DEFAULT_TRAIT<double, 2>::TraitDense> HTA_1H_DENSE;
typedef HTA<double, 2, Traits::HTA_DEFAULT_TRAIT<double, 2>::TraitDense> HTA_0H_DENSE;
typedef HTA<double, 2, Traits::HTA_DEFAULT_TRAIT<double, 2>::TraitSparse> HTA_1H_SPARSE;
typedef HTA<double, 2, Traits::HTA_DEFAULT_TRAIT<double, 2>::TraitSparse> HTA_0H_SPARSE;
 
int NPROWS; //# of tiles per row
int NPCOLS; //# of tiles per column
const int CGITMAX = 25;
int NZ;

/* rand generator parameters */
double tran;
double amult;

HTA_1H_DENSE rho;
HTA_1H_DENSE rho0;
HTA_1H_DENSE alpha;
HTA_1H_DENSE beta;

HTA_1H_DENSE p;
HTA_1H_DENSE q;
HTA_1H_DENSE r;
HTA_1H_DENSE pt;
HTA_1H_DENSE zt;
HTA_1H_DENSE qt;
HTA_1H_DENSE qtmp2;
HTA_1H_DENSE tmp3;

/* sparse matrix components */
double** values;
int** colindex;
int** rowptr;

/**
 * initialization routine
 */
void makea(HTA_0H_SPARSE A, const Tuple<2>& index);
void makeaCRS (HTA_0H_SPARSE A, const Tuple<2>& index, double*& values, int*& rowptr, int*& colindex);

/** 
 * The main conjugate gradient routine 
 */
void  conj_grad(HTA_1H_SPARSE A, HTA_1H_DENSE x, HTA_1H_DENSE z, HTA_1H_DENSE norm);

static void sprnvc(int n, int nz, double v[], int iv[], int nzloc[],
		   int mark[]);
static int icnvrt(double x, int ipwr2);
static void vecset(int n, double v[], int iv[], int *nzv, int i, double val);

//setup the domain - the tiling is always square
//this implies we need many tiles -> one processor mapping
void setup ()
{ 
  int t = log2 (max (Traits::Default::nPlaces(), 1));  

  //find the nearest *square* grid for the given # of places
  int add = t % 2 ? 1 : 0;
  t += add;

  NPROWS = sqrt (pow((double) 2, (double)t));
  NPCOLS = NPROWS;
}


int run(int argc, char** argv) 
{

  sleep(10);

  double zeta;
  double mflops;
  char CLASS;
  double zeta_verify_value;

  if (NA == 1400 && NONZER == 7 && NITER == 15 && SHIFT == 10.0) {
    CLASS = 'S';
    zeta_verify_value = 8.5971775078648;
  } else if (NA == 7000 && NONZER == 8 && NITER == 15 && SHIFT == 12.0) {
    CLASS = 'W';
    zeta_verify_value = 10.362595087124;
  } else if (NA == 14000 && NONZER == 11 && NITER == 15 && SHIFT == 20.0) {
    CLASS = 'A';
	zeta_verify_value = 17.130235054029;
  } else if (NA == 75000 && NONZER == 13 && NITER == 75 && SHIFT == 60.0) {
    CLASS = 'B';
    zeta_verify_value = 22.712745482631;
  } else if (NA == 150000 && NONZER == 15 && NITER == 75 && SHIFT == 110.0) {
      CLASS = 'C';
      zeta_verify_value = 28.973605592845;
  } else {
    CLASS = 'U';
  }
  
  setup();

  cout << NPROWS << " " << NPCOLS << " " << CLASS << endl;
  
  int nTilesPerProc = NPROWS*NPCOLS / Traits::Default::nPlaces();
  //spare Array for each tile 
  values = new double* [nTilesPerProc];
  rowptr = new int* [nTilesPerProc];
  colindex = new int* [nTilesPerProc];

  NZ = NA*(NONZER+1)*(NONZER+1)/(NPROWS*NPCOLS)
    + NA*(NONZER+2+(NPROWS*NPCOLS)/256)/NPCOLS;  
  

  // allocate and initialize row vectors
  Sequence<Tuple<2> > tiling_row (Tuple<2>(1, NA / NPCOLS), Tuple<2>(NPROWS, NPCOLS)); 
  HTA_1H_DENSE x = HTA_1H_DENSE::alloc(tiling_row);
  HTA_1H_DENSE z = HTA_1H_DENSE::alloc(tiling_row); 
  p  = HTA_1H_DENSE::alloc(tiling_row);  
  q  = HTA_1H_DENSE::alloc(tiling_row);   
  r  = HTA_1H_DENSE::alloc(tiling_row);  

  // set-up temporary arrays .. Unncessary overhead
  Sequence<Tuple<2> > tiling_col (Tuple<2>(NA / NPCOLS, 1), Tuple<2>(NPROWS, NPCOLS)); 
  zt = HTA_1H_DENSE::alloc(tiling_col); 
  qt = HTA_1H_DENSE::alloc(tiling_col);
  pt = HTA_1H_DENSE::alloc(tiling_col); 
  qtmp2 = HTA_1H_DENSE::alloc(tiling_col);

  x = 1.0;

  Sequence<Tuple<2> > tiling_scalar (Tuple<2>(1, 1), Tuple<2>(NPROWS, NPCOLS));
  rho = HTA_1H_DENSE::alloc(tiling_scalar);
  rho0 = HTA_1H_DENSE::alloc(tiling_scalar);
  alpha = HTA_1H_DENSE::alloc(tiling_scalar);
  beta = HTA_1H_DENSE::alloc(tiling_scalar);
  HTA_1H_DENSE norm = HTA_1H_DENSE::alloc(tiling_scalar);
  tmp3  = HTA_1H_DENSE::alloc(tiling_scalar);  

  // allocate matrix A
  Sequence<Tuple<2> > tiling_a (Tuple<2>( NA / NPROWS,  NA / NPCOLS), Tuple<2> (NPROWS, NPCOLS)); 
  HTA_1H_SPARSE A = HTA_1H_SPARSE::alloc(tiling_a);  

  // CVPDEBUG
  //cout << "Allocation of A with tiling=" << tiling_a << endl;

  int idx = 0;
  // initialize A (see NAS report for more details)
  for(Shape<2>::iterator i = A.shape().begin(); i != A.shape().end(); ++i) {
    Tuple<2> t = *i;

    if (A(t).isLocal()) {
      makeaCRS (A (t), 
		t, 
		values[idx], 
		rowptr[idx], 
		colindex[idx]);
      idx++;
    }
  }

  //cout << A << endl;

  timer_start (T_total);

  for (int it = 0; it < NITER; it++) {
    conj_grad(A, x, z, norm);    
    HTA_1H_DENSE tmp1 = x * z;
    HTA_1H_DENSE tmp2 = z * z;  
  
    double norm_temp1 = tmp1.reduce(plus<double>());
    double norm_temp2 = tmp2.reduce(plus<double>());
    double temp2 = 1.0 / sqrt(norm_temp2); 
    zeta = 1.0 / norm_temp1 + SHIFT;

    if (Traits::Default::myPlace() == 0) {
      if (it==0) 
	IO::cout << "iteration\t||r|\t zeta\n";      
      IO::cout << it << "\t" << norm(T(0,0))[T(0,0)] << "\t" << zeta  << endl;   
    }

    x = temp2 * z;
  }

  timer_stop (T_total);
  
  printf(" Benchmark completed\n");
  Traits::Default::barrier();

  double max_time = timer_read (T_total);
#if (HTA_TRAIT == HTA_TRAIT_MPI)
   double thread_time = timer_read(T_total);
   MPI_Allreduce (&thread_time,
	       &max_time,
	       1,
	       MPI_DOUBLE,
	       MPI_MAX,
	       MPI_COMM_WORLD);	       
  //Just for verification...
  cout << "Thread time : " << thread_time << endl;
#endif

   double t = max_time;

  if (Traits::Default::myPlace() == 0) {
    
    bool verified = false;
    double epsilon = 1.0e-10;
    if (CLASS != 'U') {
      if (fabs(zeta - zeta_verify_value) <= epsilon) {
	verified = true;
	printf(" VERIFICATION SUCCESSFUL\n");
	printf(" Zeta is    %20.12e\n", zeta);
	printf(" Error is   %20.12e\n", zeta - zeta_verify_value);
      } else {
	verified = false;
	printf(" VERIFICATION FAILED\n");
	printf(" Zeta                %20.12e\n", zeta);
	printf(" The correct zeta is %20.12e\n", zeta_verify_value);
      }
    } else {
      verified = false;
      printf(" Problem size unknown\n");
      printf(" NO VERIFICATION PERFORMED\n");
    }
    
    if ( t != 0.0 ) {
      mflops = (2.0*NITER*NA)
	* (3.0+(NONZER*(NONZER+1)) + 25.0*(5.0+(NONZER*(NONZER+1))) + 3.0 )
	/ t / 1000000.0;
    } else {
      mflops = 0.0;
    }

    print_timers();

    c_print_results("CG", CLASS, NA, 0, 0, NITER, Traits::Default::nPlaces(), 
		    t,  mflops, "          floating point", 
		    verified, NPBVERSION, COMPILETIME,
		    CS1, CS2, CS3, CS4, CS5, CS6, CS7);
    
  }

  /* delete the sparse components */
  for (int i = 0; i < nTilesPerProc; i++)
    {
      delete [] values[i];
      delete [] colindex[i];
      delete [] rowptr[i];
    }
}

/**
 * Solve Az = x.
 * Inputs:
 * @arg1 is the matrix (A) 
 * @arg2 is the vector (x)
 * Outputs:
 * @arg3 is the output vector (z)
 * @arg4 is the output norm (rnorm) 
 */
void  conj_grad (HTA_1H_SPARSE A, 
		 HTA_1H_DENSE x, 
		 HTA_1H_DENSE z, 
		 HTA_1H_DENSE norm) 
{     
  z = 0.0;          /* z = row */
  q = 0.0;          /* q = row */

  r = x;            /* x = row */
  p = r;            /* p = row */  
    
  //cout << "point 0 " << endl;
  /* allocate this temp outside */
  HTA_1H_DENSE tmp0 = r * r;    /* r = row */
  rho = 0;
  rho = tmp0.reduce(plus<double>(), 1, 0x2); /* rho = (1,1) */

  //cout << "point 1 " << endl;

  for(int i = 0; i < CGITMAX; i++) {
    
    timer_start (T_ltrans1);
    pt = p.ltranspose();
    timer_stop (T_ltrans1);

    //cout << "point 3 " << endl;
    timer_start (T_lmatmul1);
    HTA_1H_SPARSE::lmatmul(A, pt, &qt);
    timer_stop (T_lmatmul1);

    //cout << "point 4 " << endl;
    timer_start (T_reduce1);
    qtmp2 = 0;
    qtmp2 = qt.reduce(plus<double>(), 1, 0x2); /* qt = col */ 
    timer_stop (T_reduce1);

   // cout << "point 5 " << endl;
    timer_start (T_trans1);
    q = qtmp2.transpose();
    timer_stop (T_trans1);
    
    //cout << "point 6 " << endl;
    timer_start (T_ldot);
    HTA_1H_DENSE tmp1 = HTA_1H_DENSE::ldotproduct(p, q);
    timer_stop (T_ldot);

    // cout << "point 7" << endl;
    timer_start (T_reduce2);
    tmp3 = 0;
    tmp3 =  tmp1.reduce(plus<double>(), 1, 0x2);
    timer_stop (T_reduce2);

    timer_start (T_misc1);
    alpha = rho / tmp3;
    
    //cout << "point 9" << endl;
    z = z + p * alpha;                        /* z = row */
    rho0 = rho;                               /* rho0 = scalar */

    //cout << "point 10" << endl;
    r = r - q * alpha;                        /* r = row, q = row */ 

    //cout << "point 11" << endl;
    HTA_1H_DENSE tmp2 = r * r;
    timer_stop (T_misc1);

   // cout << "point 12" << endl;
    timer_start (T_reduce3);
    rho = 0;
    rho = tmp2.reduce (plus<double>(), 1, 0x2);
    timer_stop (T_reduce3);

    //cout << "point 13" << endl;
    timer_start (T_misc2);
    beta = rho / rho0;                        /* beta = scalar */
    p = r + p * beta;                         /* p = row, r = row */
    timer_stop (T_misc2);
  }  
  
  // compute the norm
  //qt = 0; // unnecessary overhead

  //cout << "point 2 " << endl;

  timer_start (T_ltrans2);
  zt = z.ltranspose();
  timer_stop (T_ltrans2);

  timer_start (T_lmatmul2);
  HTA_1H_SPARSE::lmatmul(A, zt, &qt); /* z = row, qt = col */
  timer_stop (T_lmatmul2);
  
  timer_start (T_reduce4);
  qtmp2 = 0;
  qtmp2 = qt.reduce(plus<double>(), 1, 0x2); /* qt = col */ 
  timer_stop (T_reduce4);
  
  timer_start (T_trans2);
  r = qtmp2.transpose(); /* r = row */ 
  timer_stop (T_trans2);

  timer_start (T_misc3);
  HTA_1H_DENSE f = x - r;   /* f = row */ 
  f = f * f;
  timer_stop (T_misc3);

  timer_start (T_reduce5);
  norm = 0;
  norm = f.reduce(plus<double>(), 1, 0x2); 
  timer_stop (T_reduce5);

  timer_start (T_misc4);
  norm = norm.sqrt();
  timer_stop (T_misc4);
}


// initialize array A
void makeaCRS (HTA_0H_SPARSE A, const Tuple<2>& index,
	       double*& values, int*& rowptr, int*& colindex) 
{  
  /** init the rand generator for each tile */
  tran  = 314159265.0;
  amult = 1220703125.0;
  double zeta    =  RandUtils::randlc (&tran, amult);

  int nx = A.shape().size()[0];
  int ny = A.shape().size()[1];  
  int firstrow = index[0] * nx;
  int lastrow = (index[0] + 1) * nx - 1;
  int firstcol = index[1] * ny;
  int lastcol = (index[1] + 1) * ny - 1;
  
  int n = NA;
  double size = 1.0;
  double ratio = pow(RCOND,  (1.0 / double(n)));
  int  nnza = 0;
  
  values = new double [NZ];
  colindex = new int [NZ];
  rowptr =  new int [lastrow - firstrow + 2];

  /* temporary workspace */
  double* element = new double [NZ+1];
  int* arow = new int [NZ+1];
  int* acol = new int [NZ+1];
  int* rowidx = new int [2*n+1];
  int* iv = new int [2*NA+1];
  double* v = new double [NONZER+1];
  int iouter;

  int* numelem = new int [lastrow - firstrow + 2];
  for (int i = 0; i < lastrow - firstrow + 2; i++) {
    numelem[i] = 0;
  }
   
  int i; 
  for (i = 0; i < n; i++) {
    assert (n+i < 2*n+1);
    rowidx[n+i] = 0;
  }
 
  for (iouter = 0; iouter < n; iouter++) {
    int nzv = NONZER;
    sprnvc (n, nzv, v, iv, &(rowidx[0]), &(rowidx[n]));
    vecset (n, v, iv, &nzv, iouter, (double) 0.50);
     for (int ivelt = 0; ivelt < nzv; ivelt++) {
      int jcol = iv[ivelt];
      if (jcol >= firstcol && jcol <= lastcol) {
	double scale = size * v[ivelt];
	for (int ivelt1 = 0; ivelt1 < nzv; ivelt1++) { 
	  int irow = iv[ivelt1];
	  if (irow >= firstrow && irow <= lastrow) {
	    assert (nnza <= NZ);
	    acol[nnza] = jcol - firstcol;
	    arow[nnza] = irow - firstrow;
	    element[nnza] = v[ivelt1] * scale;
	    numelem[irow - firstrow + 1]++;
	    nnza = nnza + 1;
	  }
	}
      }
    }
    size = size * ratio;
  }
  
    
  /**
   * ... add the identity * rcond to the generated matrix to bound
   *           the smallest eigenvalue from below by rcond
   * This step leads to duplicate value for a given (acol,arow).
   */
  for (int i = firstrow; i <= lastrow; i++){
    if (i >= firstcol && i <= lastcol) {
      iouter = n + i - 1;
      assert (nnza < NZ+1);
      acol[nnza] = i - firstcol;
      arow[nnza] = i - firstrow;
      element[nnza] = RCOND - SHIFT;
      numelem[i - firstrow + 1]++;
      nnza = nnza + 1;
    }
  }
  
  /**
   * Find the global starting positions 
   * of each column in the values array
   */
  rowptr[0] = numelem[0];
  for (int i = 1; i < lastrow - firstrow + 2; i++) {
    numelem[i] += numelem[i-1];
    rowptr[i] = numelem[i];
  }
  
  /**
   * group all elements of row i together
   */
  for (int i = 0; i < nnza; i++) {
    int j = arow[i];
    int k = numelem[j];
    values[k] = element[i];
    colindex[k] = acol[i];
    numelem[j]++;
  }
  
  /**
   * Remove duplicates and generate 
   * the final output 
   */
  nnza = 0;
  double x[n];
  int nzloc[n];
  bool mark[n];

  for (int i = 0; i < n; i++) {
    x[i] = 0.0;
    nzloc[i] = 0;
    mark[i] = false;
  }

  int jajpr = rowptr[0];
  for (int j = 0; j < lastrow - firstrow + 1; j++) {
    int nzrow = 0;
    for (int k = jajpr; k < rowptr[j+1]; k++) {
      int i = colindex[k];
      x[i] = x[i] + values[k];
      if (mark[i] == false && x[i] != 0.0){	
	mark[i] = true;
	nzloc[nzrow] = i;
	nzrow = nzrow + 1;
      }
    }	

    for (int k = 0; k < nzrow; k++) {
      int i = nzloc[k];
      mark[i] = false;
      /* 
       * do not know why we need this temporary? 
       * The FORTRAN version was written this way.
       * May be something to do with branch prediction..
      */
      double xi = x[i];  
      x[i] = 0.0;
      if (xi != 0.0) {
	values[nnza] = xi;
	colindex[nnza] = i;
	nnza = nnza + 1;      
      }
    }
    jajpr = rowptr[j+1];
    rowptr[j+1] = nnza + rowptr[0];
  }

  A.initSparse (values, rowptr, colindex);

  delete [] element;
  delete [] arow;
  delete [] acol;
  delete [] rowidx;
  delete [] iv;
  delete [] v;
  delete [] numelem;
  //cout << A << endl;
  // cout << "here " << Traits::Default::myPlace() << endl;
}

// initialize array A - CCS format - not tested recently
// use at your OWN RISK
#if 0
void makeaCCS (HTA_0H_SPARSE A, const Tuple<2>& index) 
{  
  /** init the rand generator for each tile */
  tran  = 314159265.0;
  amult = 1220703125.0;
  double zeta    =  RandUtils::randlc (&tran, amult);

  int nx = A.shape().size()[0];
  int ny = A.shape().size()[1];  
  int firstrow = index[0] * nx;
  int lastrow = (index[0] + 1) * nx - 1;
  int firstcol = index[1] * ny;
  int lastcol = (index[1] + 1) * ny - 1;

  int n = NA;
  double size = 1.0;
  double ratio = pow(RCOND,  (1.0 / double(n)));
  int  nnza = 0;
  
  /* progammer should delete them */
  values = new double [NZ];
  rowindex = new int [NZ];
  colptr = new int [lastrow - firstrow + 2];

  /* temporary workspace */
  double element[NZ+1];
  int arow[NZ+1];
  int acol[NZ+1];
  int colidx[NZ+1];
  int iv[2*NA+1];
  double v[NONZER+1];
  int iouter;

  int numelem[lastcol - firstcol + 2];
  for (int i = 0; i < lastcol - firstcol + 2; i++) {
    numelem[i] = 0;
  }
    
  for (int i = 0; i < n; i++) {
    assert (n+i < NZ+1);
    colidx[n+i] = 0;
  }
  
  for (iouter = 0; iouter < n; iouter++) {
    int nzv = NONZER;
    sprnvc (n, nzv, v, iv, &(colidx[0]), &(colidx[n]));
    vecset (n, v, iv, &nzv, iouter, (double) 0.50);
     for (int ivelt = 0; ivelt < nzv; ivelt++) {
      int jcol = iv[ivelt];
      if (jcol >= firstcol && jcol <= lastcol) {
	double scale = size * v[ivelt];
	for (int ivelt1 = 0; ivelt1 < nzv; ivelt1++) { 
	  int irow = iv[ivelt1];
	  if (irow >= firstrow && irow <= lastrow) {
	    assert (nnza <= NZ);
	    acol[nnza] = jcol - firstcol;
	    arow[nnza] = irow - firstrow;
	    element[nnza] = v[ivelt1] * scale;
	    numelem[jcol - firstcol + 1]++;
	    nnza = nnza + 1;
	  }
	}
      }
    }
    size = size * ratio;
  }
  
    
  /**
   * ... add the identity * rcond to the generated matrix to bound
   *           the smallest eigenvalue from below by rcond
   * This step leads to duplicate value for a given (acol,arow).
   */
  for (int i = firstrow; i <= lastrow; i++){
    if (i >= firstcol && i <= lastcol) {
      iouter = n + i - 1;
      assert (nnza < NZ+1);
      acol[nnza] = i - firstcol;
      arow[nnza] = i - firstrow;
      element[nnza] = RCOND - SHIFT;
      numelem[i - firstcol + 1]++;
      nnza = nnza + 1;
    }
  }
  
  /**
   * Find the global starting positions 
   * of each column in the values array
   */
  colptr[0] = numelem[0];
  for (int i = 1; i < lastcol - firstcol + 2; i++) {
    numelem[i] += numelem[i-1];
    colptr[i] = numelem[i];
  }
  
  /**
   * group all elements of column i together
   */
  for (int i = 0; i < nnza; i++) {
    int j = acol[i];
    int k = numelem[j];
    values[k] = element[i];
    rowindex[k] = arow[i];
    numelem[j]++;
  }
  
  /**
   * Remove duplicates and generate 
   * the final output 
   */
  nnza = 0;
  double x[n];
  int nzloc[n];
  bool mark[n];

  for (int i = 0; i < n; i++) {
    x[i] = 0.0;
    nzloc[i] = 0;
    mark[i] = false;
  }

  int jajpr = colptr[0];
  for (int j = 0; j < lastcol - firstcol + 1; j++) {
    int nzrow = 0;
    for (int k = jajpr; k < colptr[j+1]; k++) {
      int i = rowindex[k];
      x[i] = x[i] + values[k];
      if (mark[i] == false && x[i] != 0.0){	
	mark[i] = true;
	nzloc[nzrow] = i;
	nzrow = nzrow + 1;
      }
    }	

    for (int k = 0; k < nzrow; k++) {
      int i = nzloc[k];
      mark[i] = false;
      /* 
       * do not know why we need this temporary? 
       * The FORTRAN version was written this way.
       * May be something to do with branch prediction..
      */
      double xi = x[i];  
      x[i] = 0.0;
      if (xi != 0.0) {
	values[nnza] = xi;
	rowindex[nnza] = i;
	nnza = nnza + 1;      
      }
    }
    jajpr = colptr[j+1];
    colptr[j+1] = nnza + colptr[0];
  }

  A.initSparse (values, colptr, rowindex);

  // we assume that the memory referenced by arrays values, colptr, rowind 
  // will be de-allocated by the programmer when the HTA is not used any 
  // longer ...
}
#endif

/*---------------------------------------------------------------------
  c       generate a sparse n-vector (v, iv)
  c       having nzv nonzeros
  c
  c       mark(i) is set to 1 if position i is nonzero.
  c       mark is all zero on entry and is reset to all zero before exit
  c       this corrects a performance bug found by John G. Lewis, caused by
  c       reinitialization of mark on every one of the n calls to sprnvc
  ---------------------------------------------------------------------*/
static void sprnvc(
		   int n,
		   int nz,
		   double v[],		/* v[1:*] */
		   int iv[],		/* iv[1:*] */
		   int nzloc[],	/* nzloc[1:n] */
		   int mark[] ) 	/* mark[1:n] */
{
  int nn1;
  int nzrow, nzv, ii, i;
  double vecelt, vecloc;
  
  nzv = 0;
  nzrow = 0;
  nn1 = 1;
  do {
    nn1 = 2 * nn1;
  } while (nn1 < n);
  
  /*--------------------------------------------------------------------
   *    nn1 is the smallest power of two not less than n
   *-------------------------------------------------------------------*/
  
  while (nzv < nz) {
    vecelt = RandUtils:: randlc(&tran, amult);
    
    /*--------------------------------------------------------------------
     *   generate an integer between 1 and n in a portable manner
     *-------------------------------------------------------------------*/
    vecloc =  RandUtils::randlc(&tran, amult);
    i = icnvrt(vecloc, nn1) + 1;
   

    if (i > n) continue;
    
    /*--------------------------------------------------------------------
      c  was this integer generated already?
      c-------------------------------------------------------------------*/
    if (mark[i-1] == 0) {
      nzrow = nzrow + 1;
      nzv = nzv + 1;
      mark[i-1] = 1;
      nzloc[nzrow-1] = i-1;
      v[nzv-1] = vecelt;
      iv[nzv-1] = i-1;
    }
  }
  
  for (ii = 0; ii < nzrow; ii++) {
    i = nzloc[ii];
    mark[i] = 0;
  }
}

/*---------------------------------------------------------------------
 * scale a double precision number x in (0,1) by a power of 2 and chop it
 *---------------------------------------------------------------------*/
static int icnvrt(double x, int ipwr2) {
  return ((int)(double(ipwr2) * x));
}

/*--------------------------------------------------------------------
  c       set ith element of sparse vector (v, iv) with
  c       nzv nonzeros to val
  c-------------------------------------------------------------------*/
static void vecset(
		   int n,
		   double v[],	/* v[1:*] */
		   int iv[],	/* iv[1:*] */
		   int *nzv,
		   int i,
		   double val)
{
  int k;
  bool set;
  
  set = false;
  for (k = 0; k < *nzv; k++) {
    if (iv[k] == i) {
      v[k] = val;
      set  = true;
    }
  }
  if (set == false) {
    v[*nzv] = val;
    iv[*nzv] = i;
    *nzv = *nzv + 1;
  }
}

#if (HTA_TRAIT == HTA_TRAIT_UPC)
// main is defined in the upc runtime ... (cludge code in main.o)
int upc_main(int argc, char** argv) {
  return run(argc, argv);
}
#else
int main(int argc, char** argv) {

  Traits::Default::init(argc, argv);
  int rank = Traits::Default::myPlace();
  char ranks[100];
  sprintf (ranks, "/home/bikshand/htalib/benchmarks/NAS_PRIVATE/rank%d/", rank);
  chdir (ranks);
  int result = run(argc, argv);
  Traits::Default::finalize();
  return result;
}
#endif /* HTA_TRAIT */
