#include "../common/npb-C.h"
#include "ep.h"

#define T(i) Tuple<1>(i)
#define R(l, u, s) Triplet(l, u, s)
#define isRoot() (Traits::Default::myPlace() == 0)

#define T_Total 0
#define T_Rand 1
#define T_Tabulate 2
#define T_Trans 3
#define T_Reduce 4

typedef HTA<double, 1> PROC;

namespace {
  const int mk = 16;
  const int m = M;
  const int mm = m - mk;
  const int nn = 1 << mm; /* 2^mm */
  const int nk = 1 << mk; /* 2^mk */
  const int nq = 10;
  const double epsilon = 1.0e-8;
  const double a = 1220703125.0;
  const double s = 271828183.0;

  double dum[] = {1.0, 1.0, 1.0};
  double Mops, t1, t2, an, gc;
  int np;

  void print_timers () {
    const char* const str[] = {"T_total", "T_Rand", "T_Tabulate", "T_Trans", "T_Reduce"};

    for (int i = T_Total; i <= T_Reduce; ++i)
      printf("%s %lf\n", str[i], timer_read(i));
  }

};

struct MainLoop {
  void operator() (PROC xh, PROC qh, PROC sxh, PROC syh) {
    double * x = xh.raw();
    double * q = qh.raw();
    
    double sx = 0.0;
    double sy = 0.0;
    qh = 0.0;

    const int k_offset = sxh.memMap().leafPos()[0] * np;

    for(int k = 0 ; k < np; ++k) {
      double t3, t4;
      int kk = k_offset + k;
      double t1 = s;
      double t2 = an;
      
      // c        Find starting seed t1 for this kk

      for(int i = 0 ; i < 100; ++i) {
	int ik = kk / 2;
	if( (2 * ik) != kk ) t3 = randlc(&t1, t2);
	if(!ik) break;
	t3 = randlc(&t2, t2);
	kk = ik;
      }

      // c        Compute uniform pseudorandom numbers.
      //timer_start (T_Rand);
      vranlc(2 * nk, &t1, a, x - 1);
      //timer_stop (T_Rand);

      /*
c        Compute Gaussian deviates by acceptance-rejection method and 
c        tally counts in concentric square annuli.  This loop is not 
c        vectorizable. 
       */

      //timer_start (T_Tabulate);
      for(int i = 0 ; i < nk; ++i) {
	double x1 = 2.0 * x[2*i] - 1.0;
	double x2 = 2.0 * x[2*i+1] - 1.0;
	t1 = x1 * x1 + x2 * x2;
	if(t1 < 1.0) {
	  t2 = sqrt(-2.0 * log(t1) / t1);
	  t3   = (x1 * t2);
	  t4   = (x2 * t2);
	  int l    = static_cast<int>(max(abs(t3), abs(t4)));
	  q[l] += 1.0;
	  sx   += t3;
	  sy   += t4;
	}
      }
      //timer_stop (T_Tabulate);

    }

    sxh[0] = sx;
    syh[0] = sy;
  }

};


static void run()
{
  const int no_nodes = max (Traits::Default::nPlaces(), 1);

  if(isRoot()) {
    float nvals = powf(2.0, (float)(m+1));
      puts(" NAS Parallel Benchmarks 3.2 -- EP Benchmark\n");
    printf(" Number of random numbers generated:  %.0f\n", nvals);
    printf(" Number of active processes:          %10d\n", no_nodes);
  }

  /*
c   Compute the number of "batches" of random number pairs generated 
c   per processor. Adjust if the number of processors does not evenly 
c   divide the total number
  */

  np = nn / no_nodes;

  //no_large_nodes must yield 0
  // => np_add = 0

  /*
c   Call the random number generator functions and initialize
c   the x-array to reduce the effects of paging on the timings.
c   Also, call all mathematical functions that are used. Make
c   sure these initializations cannot be eliminated as dead code.
  */

  vranlc(0, dum, dum[1], &dum[2]);
  dum[1] = randlc(&dum[1], dum[2]);

  PROC x = PROC::alloc((T(2 * nk), T(no_nodes)), NULL, ROW);
  PROC sxh = PROC::alloc((T(1), T(no_nodes)), NULL, ROW);
  PROC syh = PROC::alloc((T(1), T(no_nodes)), NULL, ROW);
  PROC qh  = PROC::alloc((T(nq), T(no_nodes)), NULL, ROW);
  
  x = -1.99;
  Mops = log(sqrt(abs(max(1.0,1.0))));

  /* IMPOSSIBLE TO SYNCHR WITH CURRENT HTA MPI IMPL AFAIK
c---------------------------------------------------------------------
c      Synchronize before placing time stamp
c---------------------------------------------------------------------
    call mpi_barrier(MPI_COMM_WORLD, ierr)
   */

  for(int i = T_Total; i <= T_Reduce; ++i)
    timer_clear( i );

  timer_start (T_Total);

  t1 = a;

  // call vranlc(0, t1, a, x) ?

  //c  Compute AN = A ^ (2 * NK) (mod 2^46).

  for(int i = 0; i <= mk; ++i)
    t2 = randlc(&t1, t1);

  an = t1;
  //double tt = s; NOT USED
  gc = 0.0;
  //sx = 0.0; //mainLoop()
  //sy = 0.0; //mainLoop()

  //q = 0.0;  //mainLoop()

  /*
c   Each instance of this loop may be performed independently. We compute
c   the k offsets separately to take into account the fact that some nodes
c   have more numbers to generate than others
  */

  x.hmap(MainLoop(), qh, sxh, syh, 0);

  double sx = sxh.reduce(plus<double>());
  double sy = syh.reduce(plus<double>());
  qh.reduce (1, plus<double>(), 0, 0, &qh);

  //for(int i = 0; i < nq; ++i)
  //  gc += q[i];

  gc = qh(T(0)).reduce(plus<double>());

  timer_stop (T_Total);

  double tm = timer_read(T_Total);

#if (HTA_TRAIT == HTA_TRAIT_MPI)
  double tm_copy = tm;
  MPI_Reduce (&tm_copy, &tm, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
#endif

  if (Traits::Default::myPlace() == 0) {
    double sx_verify_value, sy_verify_value;
    int nit = 0;
    bool verified = true;

    switch(M) {
    case 24: 
      sx_verify_value = -3.247834652034740E3;
      sy_verify_value = -6.958407078382297E3;
      break;
    case 25:
      sx_verify_value = -2.863319731645753E3;
      sy_verify_value = -6.320053679109499E3;
      break;
    case 28:
      sx_verify_value = -4.295875165629892E3;
      sy_verify_value = -1.580732573678431E4;
      break;
    case 30:
      sx_verify_value =  4.033815542441498E4;
      sy_verify_value = -2.660669192809235E4;
      break;
    case 32:
      sx_verify_value =  4.764367927995374E4;
      sy_verify_value = -8.084072988043731E4;
      break;
    default:
      verified = false;
    }

    if(verified) {
      double sx_err = fabs((sx - sx_verify_value)/sx_verify_value);
      double sy_err = fabs((sy - sy_verify_value)/sy_verify_value);
      verified = (sx_err <= epsilon) && (sy_err <= epsilon);
    }

    Mops = pow(2.0, M+1)/tm/1000000.0;
 
      puts("\nEP Benchmark Results: \n\n");
    printf("CPU Time = %6.3lf\n", tm);
    printf("N = 2^ %d\n", M);
    printf("No. Gaussian Pairs = %15.0lf\n", gc);
    printf("Sums = %25.15lf %25.15lf\n", sx, sy);
    puts("Counts:");

    for (int i = 0; i  < 10; i++)
      printf("%3d %15.0lf\n", i, qh(T(0))[i]);

    c_print_results("EP", CLASS, M+1, 0, 0, nit, no_nodes,
	 	    tm, Mops, 	
		    "Random numbers generated",
		    verified, NPBVERSION, COMPILETIME,
		    CS1, CS2, CS3, CS4, CS5, CS6, CS7);
 
    //print_timers ();
  }

}


#if (HTA_TRAIT == HTA_TRAIT_UPC)
// main is defined in the upc runtime ... (cludge code in main.o)
int upc_main(int argc, char** argv) {
  Traits::Default::init(argc, argv);
  run(argc, argv);
  Traits::Default::finalize();
  return 0;
}
#else
int main(int argc, char** argv) {
  Traits::Default::init(argc, argv);
  run();
  Traits::Default::finalize();
  return 0;
}
#endif /* HTA_TRAIT */

