/**
 * NOTE: The EP HTA version uses a different algorithm 
 * than the EP NAS version
 * EP HTA uses block-cyclic distribution of random numbers
 * EP NAS uses block distribution
 */

#include "../common/npb-C.h"
#include "ep.h"

#define T(i) Tuple<1>(i)
#define R(l, u, s) Triplet(l, u, s)
#define EPSILON		1.0e-8

#define T_Total 0
#define T_Rand 1
#define T_Tabulate 2
#define T_Trans 3
#define T_Reduce 4

void print_timers();

typedef HTA<double, 1> PROC;

template<int N>
struct Result
{
  double count[N];
  double sx;
  double sy;
};

template <typename Type, int DIM, int N>
struct Tabulate :  public Operator::MapReduce<Type, Result<N>, DIM>  
{

public:
  
  Tabulate()
  {
    result_.sx = (double) 0.0;
    result_.sy = (double) 0.0;
    for (int i = 0; i < N; i++)
      result_.count[i] = (double) 0.0;
  }
 
  void reduce (const Result<N>& d) 
  {
    for (int i = 0; i < 10; i++)
      result_.count[i] += d.count[i];

    result_.sx += d.sx;
    result_.sy += d.sy;
    
  }  

  void map_none (Type& x, Type& y)
  {
    double x1 = ((double) 2.0) * x - ((double) 1.0);
    double x2 = ((double) 2.0) * y - ((double) 1.0);
    
    assert (x1 != x2);

    double val = x1 * x1 + x2 * x2;

    //cout << "here " << val << " " << x1 << " " << x2 << endl;

    if (val < (double) 1.0)
      {
	double t0 = sqrt( ((double) -2.0) * log(val)/val);
	
	double t1 = x1 * t0;
	
	double t2 = x2 * t0;
	
	int l = max(abs(t2), abs(t1));

	result_.count[l] += (double) 1.0;

	result_.sx  += t1;

	result_.sy  += t2;
	
      }
  }

  bool doReduce () { return false; }
  
  virtual const Result<N>& result () const 
  {
    return result_;
  }
  
  Operator::MapReduceKey keyType () const { return Operator::MR_NONE; }    
 
  
private:    
  
  //intermediate results
  Result<N> result_;
  
};

int run() 
{

  double gc = (double) 0.0;
  int m = M;
  int mk = 16;
  int mm = m - mk;
  int nn = pow((double) 2, (double) mm);
  int nk = pow((double) 2, (double) mk);
  int nq = 10;
  double epsilon = 1.0e-8;
  
  double seed = (double) 271828183.0;
  double a = (double) 1220703125.0;
  
  int no_nodes = max (Traits::Default::nPlaces(), 1);

  int np = nn / no_nodes;
  
  cout << " " << np << " " << no_nodes <<  " " << nn << endl;
  
  int end = 2 * nk - 1;

  PROC x = PROC::alloc((T(2 * nk), T(no_nodes)), NULL, ROW);

  /* temporary htas */
  //  PROC x1 = PROC::alloc((T(nk), T(no_nodes)));
  //PROC x2 = PROC::alloc((T(nk), T(no_nodes)));
  PROC t =  PROC::alloc((T(nk), T(no_nodes)));

  Tabulate<double, 1, 10> tabulate; 

  timer_start (T_Total);

  for (int i = 0; i < np; i++)
    {      
      int kk = i * 2 * nk * no_nodes;
      double t1 = seed;
      double t2 = a;
      
      //find the seed for this iteration
      for (int j = 0; j < 100; j++)
	{
	  //cout << " kk " << kk << endl;
	  int ik = kk / 2;
	  if (2 * ik != kk) double t3  = RandUtils::randlc(&t1, t2);
	  if (ik == 0) break;
	  double t3 = RandUtils::randlc(&t2, t2);
	  kk = ik;
	  
	}
      
      //generate uniform random numbers (predefined in htalib)
      timer_start (T_Rand);
      RandNumGeneratorDist<double, 1> rand (a, t1);    
      x.mapReduce (rand);
      timer_stop (T_Rand);
        
      //cout << x << endl;

      //apply the box muller transformation      
      // (1) find 2 * odd-1, 2*even-1 of the original set of random numbers
      timer_start (T_Trans);
      PROC x1 =   x()(R(0, end, 2)); //odd points
      PROC x2 =   x()(R(1, end, 2)); //even points	 
      // (2) find the sum of their squares
      //t = x1 * x1 + x2 * x2;
      timer_stop (T_Trans);

      //cout << "x1 " << x1 * 2.0 - 1.0 << endl;
      //cout << "x2 " << x2 * 2.0 - 1.0 << endl;

      //tabulate the # of pair in successive square annuli            
      //TODO: rename mapReduce to map as here only map is applied
      timer_start (T_Tabulate);
      x1.mapReduce (tabulate, x2);
      timer_stop (T_Tabulate);
	
      //assert (false);
    }

  //TODO: impelement reduce(MapReduce&) in HTAimpl and DistHTAImpl
  //The key lesson is that map and reduce need to be SPLIT
  timer_start (T_Reduce);
  x.reduceL (tabulate);
  timer_stop (T_Reduce);
  
  cout << "reduction time " << timer_read(0) << endl;
  
  Result<10> result = tabulate.result();
  
  for (int i = 0; i < 10; i ++)
    gc += result.count[i];
  
  timer_stop (T_Total);

  double tm = timer_read(T_Total);

  double sx = result.sx;
  double sy = result.sy;
  bool verified = false;
  int nit = 0;
  if (M == 24) {
    if((fabs((sx- (-3.247834652034740e3))/sx) <= EPSILON) &&
       (fabs((sy- (-6.958407078382297e3))/sy) <= EPSILON)) {
      verified = TRUE;
    }
  } else if (M == 25) {
    if ((fabs((sx- (-2.863319731645753e3))/sx) <= EPSILON) &&
	(fabs((sy- (-6.320053679109499e3))/sy) <= EPSILON)) {
      verified = TRUE;
    }
  } else if (M == 28) {
    if ((fabs((sx- (-4.295875165629892e3))/sx) <= EPSILON) &&
	(fabs((sy- (-1.580732573678431e4))/sy) <= EPSILON)) {
      verified = TRUE;
    }
  } else if (M == 30) {
    if ((fabs((sx- (4.033815542441498e4))/sx) <= EPSILON) &&
	(fabs((sy- (-2.660669192809235e4))/sy) <= EPSILON)) {
      verified = TRUE;
    }
  } else if (M == 32) {
    if ((fabs((sx- (4.764367927995374e4))/sx) <= EPSILON) &&
	(fabs((sy- (-8.084072988043731e4))/sy) <= EPSILON)) {
      verified = TRUE;
    }
  }
  
  double Mops = pow(2.0, M+1)/tm/1000000.0;
  
  cout << "EP Benchmark Results: " << endl <<
    "CPU Time = " << tm << endl << 
    "N = 2^ " << M << endl <<
    "No. Gaussian Pairs = " << gc <<
    "Sums = " << sx << " " << sy << endl << 
    "Counts: " << endl;
  
  for (int i = 0; i  < 10; i++) {
    cout << i << " " << result.count[i] << endl;
  }
  
  c_print_results("EP", CLASS, M+1, 0, 0, nit, no_nodes,
		  tm, Mops, 	
		  "Random numbers generated",
		  verified, NPBVERSION, COMPILETIME,
		  CS1, CS2, CS3, CS4, CS5, CS6, CS7);
 
  print_timers ();
}

void print_timers () {
  char* str[] = {"T_total", "T_Rand", "T_Trans", "T_Tabulate", "T_Reduce"};

  for (int i = T_Total; i <= T_Reduce; ++i)
    cout << str[i] << " " << timer_read(i) << endl;
}

#if (HTA_TRAIT == HTA_TRAIT_UPC)
// main is defined in the upc runtime ... (cludge code in main.o)
int upc_main(int argc, char** argv) {
  Traits::Default::init(argc, argv);
  int ret = run(argc, argv);
  Traits::Default::finalize();
  return ret;
}
#else
int main(int argc, char** argv) {
  Traits::Default::init(argc, argv);
  int ret = run();
  Traits::Default::finalize();
  return ret;
}
#endif /* HTA_TRAIT */
