#include "util/complex"
#include "ft.h"

typedef Triplet R;
typedef Tuple<3> T;

typedef HTA<complex<double>, 3> HTA_2H;
typedef HTA<complex<double>, 3> HTA_1H;
typedef HTA<complex<double>, 3> LEAF;

typedef HTA<double, 3> HTA_2H_DOUBLE;
typedef HTA<double, 3> HTA_1H_DOUBLE;
typedef HTA<double, 3> LEAF_DOUBLE;

/* helper method */
static double ipow46(double a, int exponent) {

/*--------------------------------------------------------------------
c-------------------------------------------------------------------*/

/*--------------------------------------------------------------------
c compute a^exponent mod 2^46
c-------------------------------------------------------------------*/

    double dummy, q, r;
    int n, n2;

/*--------------------------------------------------------------------
c Use
c   a^n = a^(n/2)*a^(n/2) if n even else
c   a^n = a*a^(n-1)       if n odd
c-------------------------------------------------------------------*/
    if (exponent == 0) return 1;
    q = a;
    r = 1;
    n = exponent;

    while (n > 1) {
	n2 = n/2;
	if (n2 * 2 == n) {
            dummy = RandUtils::randlc(&q, q);
            n = n2;
	} else {
            dummy = RandUtils::randlc(&r, q);
            n = n-1;
	}
    }
    dummy = RandUtils::randlc(&r, q);
    return r;
}

/**
 * apply Fourier Transform on 
 * on the entire *raw* array
 * stored in each of the tiles.
 * @arg1 = dimension of transform (0..DIM-1).
 * @arg2 = direction (forward = 1, backward = -1).
 * @arg3 = dest. array (NULL by default) => *this* is overwritten
 */
struct FT
{
  
  FT (int dim, int dir) : 
    dim_(dim), dir_ (dir) {}
  
  void operator () (HTA_2H h) const 
  {
    assert (dim_ >=0 && dim_ < 3);
        
    // cout << "Here " << h.memMap().leafPos() << endl;
    //    Shape<3>::iterator i = h.shape().begin();
    //for (; i!=h.shape().end(); i++)
    // cout << h(*i).memMap().leafPos() << " " ;
    //cout << endl;

    // stride between two consecutive elements along 'dim_' 
    int stride[2] = {0, h.memMap().linearStep()[dim_]}; 
    
    T tile_size = h.memMap().leafSize();
    
    int size = tile_size[dim_];  // size of the transform 
    
    tile_size[dim_] = 1;
    
    NasFT <complex<double> > mkl(size, stride);    
    
    int size_0 = tile_size[0];
    int size_1 = tile_size[1];
    int size_2 = tile_size[2];
    
    if (dir_ == 1)  //forward FOURIER transfrom (time --> freqency domain)
      
      for (int i = 0; i < size_0; i++)
	for (int j = 0; j < size_1; j++)
	  for (int k = 0; k < size_2; k++)
	    	    
	    mkl.dftForward(&(h[T(i, j, k)]));
    
    else // inverse FOURIER transform (frequency --> time domain)
      
      for (int i = 0; i < size_0; i++)
	for (int j = 0; j < size_1; j++)
	  for (int k = 0; k < size_2; k++)
	
	    mkl.dftBackward(&(h[T(i,j,k)]));
  }
private : 
  int dim_;
  int dir_;
};


struct Checksum 
{
  void operator() (HTA_2H u0, 
	    HTA_2H sum) {
    
    int d1 = NZ;
    int d2 = NY;
    int d3 = NX;
    
    int s0 = u0.memMap().leafSize()[0];
    int s1 = u0.memMap().leafSize()[1];
    int s2 = u0.memMap().leafSize()[2];
    
    int startZ = u0.memMap().leafPos()[2];
    int startY = u0.memMap().leafPos()[1];
    int startX = u0.memMap().leafPos()[0];
    
    int endZ = u0.memMap().leafPos()[2] + s2 - 1;
    int endY = u0.memMap().leafPos()[1] + s1 - 1;
    int endX = u0.memMap().leafPos()[0] + s0 - 1;
    
    complex<double> csum(0.0, 0.0); 
    for  (int i = 1; i <= 1024; i++) {
      int ii = 5 * i % d1;
      if( ii >= startX && ii <= endX) {
	int ji = (3 * i) % d2;
	if (ji >= startY && ji <= endY) {
	  int ki =  i % d3;
	  if (ki >= startZ && ki <= endZ) {
	    csum = csum + u0 [T(ii-startX, ji-startY, ki-startZ)];
	  }
	}
      }  
    }
    
    sum [0] = csum * (1/(double)(d1*d2*d3));
  }
};

struct IndexMap {

  IndexMap (Tuple<3> offset): t(offset) {}

  void operator() (HTA_1H_DOUBLE twiddle)
  {
    int d1 = NZ;
    int d2 = NY;
    int d3 = NX;
    
    int s1 = twiddle.memMap().leafSize()[0];
    int s2 = twiddle.memMap().leafSize()[1];
    int s3 = twiddle.memMap().leafSize()[2];
    
    int startZ = twiddle.memMap().leafPos()[2];
    int startY = twiddle.memMap().leafPos()[1];
    int startX = twiddle.memMap().leafPos()[0] + t[2] * s3; //this is required for CLASS A (see the call site)
    
    double pi = 3.141592653589793238;
    double ap = -4.0 * 1.0e-6 * pi * pi;
    
    for (int i = 0; i < s1; i++) {
      for (int j = 0; j < s2; j++) {
	for (int k = 0; k < s3; k++){
	int ii = (i + startX + d3/2) % d3 - d3/2;    
	int jj = (j + startY + d2/2) % d2 - d2/2;
	int kk = (k + startZ + d1/2) % d1 - d1/2;
	twiddle [T(i, j, k)] = exp(ap*double(kk*kk + jj*jj + ii*ii));
	}         
      }
    }
  }

private :
  Tuple<3> t;
};


struct Initial {

  void operator() (HTA_1H u0) 
  {
    assert(u0.isLocal());
    
    /* Sequence splitting */
    /* rand number generator inputs */
    double seed = 314159265.00;
    double a = 1220703125.00;
    
    int d1 = NZ;
    int d2 = NY;
    int d3 = NX;
    
    int s1 = u0.memMap().leafSize()[0];
    int s2 = u0.memMap().leafSize()[1];
    int s3 = u0.memMap().leafSize()[2];
    
    int startZ = u0.memMap().leafPos()[2];
    int startY = u0.memMap().leafPos()[1];
    int startX = u0.memMap().leafPos()[0];
       
    double start = seed;
    double an = ipow46(a, 2*startX*d2*d3);
    double dummy = RandUtils::randlc(&start, an);   
    an = ipow46(a, 2*d2*d3);
    
    /* Go through z planes filling in one square at a time */  
    double x0 = start;
    for (int i = 0; i < s1; i++) {
      // double x0 = start;
      for (int j = 0; j < s2; j++) {
	for (int k = 0; k < s3; k++) {
	  double real = RandUtils::randlc(&x0, a);
	double imag = RandUtils::randlc(&x0, a);
	u0 [T(i, j, k)] = complex<double>(real, imag);	
	}
      }
      // if (i != s1-1) dummy = RandUtils::randlc(&start, an);
    }
  }
};
