/*
 * Version: $Id: mg.orig.cpp,v 1.1 2007/04/14 19:37:46 bikshand Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

/** 
 * HTA++ implementation of NAS MG benchmark 
 */

#include "../common/npb-C.h"
#include "mg.h"

// #include <unistd.h>
#define epsilon  1.0e-8
#include "Interp.h"
#include "Rprj3.h"
#include "Resid.h"
#include "Psinv.h"

// ordering TILE is necessary to use the operator lconv 
// which is a built-in library function. lconv requires 
// contiguous memory of its operand and destination.
#define order TILE

int nx;
int ny; 
int nz;

/**
 * The main Multi-Grid V-Cycle routine
 */
void mg3P    (Grid* r,  Grid& v, Grid* u, double a[4], double c[4], int k);

Grid conv (Grid s, double w[4]);

Grid conv2 (Grid s, Grid r, double w[4], int, Tuple<3>);

/**
 * the interpolation function 
 * interpolates the grid z to grid u
 */
void interp  (Grid& z, Grid& u, int k);
void interp_normal (Grid z, Grid u );
void interp_special (Grid z, Grid u );

/**
 * the projection function
 * projects (restricts) the grid r to grid s 
 */
void rprj3   (Grid& r, Grid& s, int k);

/**
 * the residual function
 * find the residual r = u - A*v 
 */
void resid   (Grid& u, Grid& v, Grid&r, double a[4], int k);

/**
 * matrix inversion 
 * the jacobi iterative scheme
 */
void psinv   (Grid& r, Grid& s, double c[4], int k);

/**
 * initialization routine
 * fills z with 10 +1 and 10 -1 values
 * at random positions determined as 
 * the positions of top 10 and bottom 10 values
 * of array z filled with random numbers.
 */
void zran3   (Grid& z, int gx, int gy, int gz);

/**
 * norm2u3. 
 * finds the 2-norm 
 */
void norm2u3 (Grid& r, double *rnm2, double *rnmu, int nx = NX_DEFAULT, int ny = NY_DEFAULT, int nz = NZ_DEFAULT);
void rep_nrm (Grid &x, char* callee, int k );

/**
 * the communication routine 
 * perfoms communication across all three borders
 */
void comm3    (Grid& u);
void comm3_ex (Grid& u, int s1, int s2, int s3);

void zero3   (Grid& z);
void print_timers(void);

/**
 * tile leve initialization function
 */
void zran3_inner (Leaf z, Tuple<3> tileIndex,  int gx, int gy, int gz);


/*
 * These should be get rid of soon...
 */
void comm_ex_ew (Grid& u);
void comm_ex_ns (Grid& u);
void comm_ex_fb (Grid& u);

/**
 * Helper functions
 * See the function implementation for more details
 */
//double power (double a, int n);
void setup (Grid* r, Grid* u, int nprocs);




/**
 * map-reduce operator to find 10 min-max values
 */
template<typename T, int N, int DIM>
struct XchgData 
{
  Tuple<DIM> maxKey[N];
  T maxVal[N];
  Tuple<DIM> minKey[N];
  T minVal[N];
};

template<typename T, int N, int DIM>
struct FilterN : public Operator::MapReduce<T, XchgData<T, N, DIM>, DIM>  
{
  FilterN() : minIdx_(0), maxIdx_(0) { 
    for (int i = 0; i < N; i++) {
      data_.minVal[i] = 1.0;
      data_.maxVal[i] = 0.0;
    }
  }
  
  void reduce (const XchgData<T, N, DIM>& d) 
  { 
    for (int i = 0; i < N; i++) { 
      map_tuple(d.maxKey[i], d.maxVal[i]);
      map_tuple(d.minKey[i], d.minVal[i]);
    }
  }
  
  void map_tuple (const Tuple<DIM>& key, const T& val) 
  { 
    if (val > data_.maxVal[minIdx_]) {
      // new maximum
      data_.maxVal[minIdx_] = val;
      data_.maxKey[minIdx_] = key;
      T min = val;
      for (int i = 0; i < N; i++) {
	if (data_.maxVal[i] < min) {
	  min = data_.maxVal[i];
	  minIdx_ = i;
	}
      }
    } 
    
    if (val > 0.0 && val < data_.minVal[maxIdx_]) {
      // new minimum
      data_.minVal[maxIdx_] = val;
      data_.minKey[maxIdx_] = key;
      T max = val;
      for (int i = 0; i < N; i++) {
	if (data_.minVal[i] > max) {
	  max = data_.minVal[i];
	  maxIdx_ = i;
	}
      }
    }
  } // map
  

  virtual bool doReduce () { return true; }
  virtual const XchgData<T, N, DIM>& result () const 
  {
    return data_;
  }

  Operator::MapReduceKey keyType () const { return Operator::MR_TUPLE; }

  XchgData<T, N, DIM> data_;
  int minIdx_;
  int maxIdx_;
}; // struct FilterN

/**
 * map-reduce operator to compute the square sum and find maxiumum
 */
template<typename T, int DIM>
struct SquareSum : public Operator::MapReduce<T, SquareSum<T, DIM>, DIM>  
{
  SquareSum() : sqsum(0), max (0) {}

  void reduce (const SquareSum<T, DIM>& d) 
  { 
    sqsum += d.sqsum;
    Operator::setIfLarger<double>()(max, d.max);
  }
  
  void map_none (const T& val) {
    sqsum += std::pow(val, 2.0);
    Operator::setIfLarger<double>()(max, std::fabs(val));
  }

  virtual const SquareSum<T, DIM>& result () const 
  {
    return *this;
  }
  
  Operator::MapReduceKey keyType () const { return Operator::MR_NONE; }
  
  T sqsum;
  T max;
}; // struct SquareSum


/**
 * main routine
 */
int run(int ntiles = -1)
{ 

  sleep (10);

   double rnm2;
   double rnmu;
     
   int nPlaces = Traits::Default::nPlaces();
   
   if (ntiles <= 0)
     ntiles = nPlaces;
   
   //if ntiles < 8, set ntiles = 8
   ntiles = max (ntiles, 1);
   
   IO::cout << "tiles  = " << ntiles << endl;
   IO::cout << "places = " << nPlaces << endl;

   //assert (ntiles == 1 || ntiles % 8 == 0);
     
   lt = LT_DEFAULT;
   
   Grid r[lt];
   Grid u[lt];  
   setup (r, u, ntiles);
   int sx = r[lt-1].shape().size()[0];
   int sy = r[lt-1].shape().size()[1];
   int sz = r[lt-1].shape().size()[2];
  
   nx = NX_DEFAULT / sx;
   ny = NY_DEFAULT / sy;
   nz = NZ_DEFAULT / sz;
   int nit = NIT_DEFAULT;

   Sequence<Tuple<3> > tiling(Tuple<3> (nx+2, ny+2, nz+2), Tuple<3> (sx, sy, sz));
   Grid v = HTA<double,3>::alloc(1, tiling, NULL, order);
   
   zran3 (v, NX_DEFAULT, NY_DEFAULT, NZ_DEFAULT);
   zero3 (u[lt-1]);

   // enable or disable debugging mode
   debug_vec[0] = 0;
   
   if ( (NX_DEFAULT != NY_DEFAULT) || (NX_DEFAULT != NY_DEFAULT) ) Class = 'U';
   else if( NX_DEFAULT == 32 && nit == 4 )   Class = 'S';
   else if( NX_DEFAULT == 64 && nit == 40 )  Class = 'W';
   else if( NX_DEFAULT == 256 && nit == 20 ) Class = 'B';
   else if( NX_DEFAULT == 512 && nit == 20 ) Class = 'C';
   else if( NX_DEFAULT == 256 && nit == 4 )  Class = 'A';
   else 	Class = 'U';
   
   a[0] = -8.0/3.0;
   a[1] =  0.0;
   a[2] =  1.0/6.0;
   a[3] =  1.0/12.0;

   /*
    *     Coefficients for the S(a) smoother
    */

   if (Class == 'A' || Class == 'S' || Class =='W') {
     /*
      * Coefficients for the S(a) smoother
      */        
     c[0] =  -3.0/8.0;
     c[1] =  1.0/32.0;
     c[2] =  -1.0/64.0;
     c[3] =   0.0;
   } else {
     /*
      * Coefficients for the S(b) smoother
      */       
     c[0] =  -3.0/17.0;
     c[1] =  1.0/33.0;
     c[2] =  -1.0/61.0;
     c[3] =   0.0;
   }
   
   //one iteration for startup
   mg3P (r, v, u, a, c, lt-1);     
   resid (u[lt-1], v, r[lt-1], a, lt-1);
   setup (r, u, ntiles);
   zero3 (u[lt-1]);
   zran3 (v, NX_DEFAULT, NY_DEFAULT, NZ_DEFAULT);
     
   timer_start (T_total);
   resid (u[lt-1], v, r[lt-1], a, lt-1);
   norm2u3 (r[lt-1], &rnm2, &rnmu);

   for (int it = 0; it < nit; it++) {
     IO::cout << "Iteration " << it << endl;
     mg3P (r, v, u, a, c, lt-1);
     resid (u[lt-1], v, r[lt-1], a, lt-1);
   }

   norm2u3 (r[lt-1], &rnm2, &rnmu);

   timer_stop (T_total);

   IO::cout << endl << "Benchmark completed. " << endl;
   Traits::Default::barrier();

   double max_time = timer_read (T_total);
#if (HTA_TRAIT == HTA_TRAIT_MPI)
   double thread_time = timer_read(T_total);
   MPI_Allreduce (&thread_time,
	       &max_time,
	       1,
	       MPI_DOUBLE,
	       MPI_MAX,
	       MPI_COMM_WORLD);
  //Just for verification...
  cout << "Thread time : " << thread_time << endl;	       
#endif

   if (Traits::Default::myPlace() == 0) {
   
     bool verified = false;
     double verify_value = 0.0;
     
     if (Class != 'U') {
       if (Class == 'S')  verify_value = 0.530770700573e-04;
       else if (Class == 'W')  verify_value = 0.250391406439e-17;  /* 40 iterations*/
       /*				0.183103168997d-044 iterations*/
       else if (Class == 'A')   verify_value = 0.2433365309e-5;
       else if (Class == 'B')   verify_value = 0.180056440132e-5;
       else if (Class == 'C') verify_value = 0.570674826298e-06;
       
       
       if ( fabs( rnm2 - verify_value ) <= epsilon ) {
	 verified = true;
	 IO::cout << " VERIFICATION SUCCESSFUL\n";
	 IO::cout << " L2 Norm is " <<  rnm2 << endl;
	 IO::cout << " Error is  " <<  rnm2 - verify_value << endl;
       } else {
	 verified = false;
	 IO::cout << " VERIFICATION FAILED\n";
	 IO::cout << " L2 Norm is              " << rnm2;
	 IO::cout << " The correct L2 Norm is  " <<  verify_value << endl;
       }
     } else {
       verified = false;
       IO::cout << " Problem size unknown\n";
       IO::cout << " NO VERIFICATION PERFORMED\n";
     }
     
     int nn = NX_DEFAULT * NY_DEFAULT * NZ_DEFAULT;
     double t = max_time;
     double mflops;
     
     if (t != 0)
       mflops = 58. * nit * nn * 1.0E-6 / t;
     else
       mflops = 0.0;
     
     c_print_results("MG", // char *name, 
		     Class, // char cls, 
		     NX_DEFAULT, NY_DEFAULT, NZ_DEFAULT, // int n1, int n2, int n3, 
		     nit, //int niter, 
		     nPlaces , // int tiles, 
		     t, // double t,
		     mflops, // double mops, 
		     "floating point", // char *optype, 
		     verified, // int passed_verification,
		     NPBVERSION, // char *npbversion, 
		     COMPILETIME, // char *compiletime, 
		     CS1, // char *cc,
		     CS2, // char *clink, 
		     CS3, // char *c_lib, 
		     CS4, // char *c_inc,
		     CS5, // char *cflags, 
		     CS6, // char *clinkflags, 
		     CS7); // char *rand);
     
     IO::cout << endl << endl;
     print_timers ();
     IO::cout << endl;
   }
   return 0;
}

void mg3P (Grid* r, Grid& v, Grid* u, double a[4], double c[4], int k)
{

  if (k > 0) {
    rprj3 (r[k], r[k-1], k);
    mg3P (r, v, u, a, c, k-1);      
    if (k != lt-1) zero3 (u[k]);
    interp (u[k-1], u[k], k);
    k == lt-1 ? resid (u[k], v, r[k], a, k) : resid (u[k], r[k], r[k], a, k) ;
    psinv (r[k], u[k], c, k);
  } else  {
    zero3 (u[k]);
    psinv (r[k], u[k], c, k); 
  }  
}

void resid (Grid& u, Grid& v, Grid& r, double a[4], int k)
{     
  timer_start (T_resid);
    
   r = v - conv(u, a);
   //r = v - u.lconv(a);
  
  timer_stop (T_resid); 
  
  comm3 (r);
  if (debug_vec[0]) 
    rep_nrm (r, "resid", k);
  
} 


void psinv (Grid& r, Grid& u, double c[4], int k)
{  
  timer_start (T_psinv);
  u += conv(r, c);

  timer_stop (T_psinv);
  
  comm3 (u);
  
  if (debug_vec[0]) rep_nrm (u, "psinv", k);
} 

void rprj3 (Grid& h, Grid& s, int k)
{   
  timer_start (T_rprj3);
  
  Tuple<3> hSize = h.shape().size();
  Tuple<3> sSize = s.shape().size();

  int d1 = hSize[0] == sSize[0] ? 0 : 1;
  int d2 = hSize[1] == sSize[1] ? 0 : 1;
  int d3 = hSize[2] == sSize[2] ? 0 : 1;
  
  Grid r = h((Triplet (d1, hSize[0]-1, d1+1), 
	      Triplet (d2, hSize[1]-1, d2+1),
	      Triplet (d3, hSize[2]-1, d3+1)));
    
  /*
   * TODO: this kind of statments (r(0,0,0)) may be a bottle neck
   * Try to use some other function. 
   */
  int iSize = s (T(0,0,0)).shape().size()[0];
  int jSize = s (T(0,0,0)).shape().size()[1];
  int kSize = s (T(0,0,0)).shape().size()[2];
  
  double a[4];
  a[0] = 0.5;
  a[1] = 0.25;
  a[2] = 0.125;
  a[3] = 0.0625;
  
/* 
  Grid tmp = r.lconv(a);
  s()((R(1, iSize-2), R(1, jSize-2), R(1, kSize-2))) = tmp()((R(2-d1, 2 * (iSize-2) - d1, 2), R(2-d2, 2 * (jSize-2) - d2, 2), R(2-d3, 2 * (kSize-2) - d3, 2)));
*/

  Tuple<3> ratio = hSize / sSize + 1; 
  Grid tmp = s()((R(1, iSize-2), R(1, jSize-2), R(1, kSize-2)));
  conv2 (r, tmp, a, 2, ratio); 

  timer_stop (T_rprj3);
  
  comm3 (s);
  
  if (debug_vec[0]) rep_nrm (s, "rprj3", k-1);
 } 

Grid conv (Grid s, double w[4])
{
  //size of the leaf HTA (assuming HTAs are regular)
  Tuple<3> size = s(T(0,0,0)).shape().size();  
  
  /* GB: temporaries -- allocate it outside the function */
  Sequence<Tuple<3> > tiling(size, s.shape().size());  
  Grid u1 = HTA<double, 3>::alloc(1, tiling, s.dist().clone(), order);
  Grid u2 = HTA<double, 3>::alloc(1, tiling, s.dist().clone(), order); 
  
  /* GB: result -- allocate it outside the function */
  Grid r = HTA<double, 3>::alloc(1, tiling, s.dist().clone(), order);   

  Triplet I(1, size[0]-2);
  Triplet J(1,size[1]-2);
  Triplet K(0, size[2]-1);

  Triplet RI(1, size[0] - 2);
  Triplet RJ(1, size[1] - 2);
  Triplet RK(0, size[2] - 1);
  
  /* TODO: Check if the compiler unrolls all the loops and contracts the temporary arrays -- I doubt it
   * TODO: See if Expression Templates are useful
   */
  
  u1()((RI, RJ, RK)) =  
    (Grid) s()((I, J-1, K)) + 
    (Grid) s()((I, J+1, K)) +
    (Grid) s()((I-1, J, K)) +
    (Grid) s()((I+1, J, K));
  
  u2()((RI, RJ, RK)) =  
    (Grid) s()((I-1, J-1, K)) +
    (Grid) s()((I-1, J+1, K)) +
    (Grid) s()((I+1, J-1, K)) +
    (Grid) s()((I+1, J+1, K));
  
  K = Triplet(1,  size[2] - 2);
  RK = Triplet(1, size[2] - 2);
  
  r()((RI, RJ, RK)) = 
    w[0] *  (Grid) s()((I, J, K)) + 
    w[1] * ((Grid) u1()((RI, RJ, RK))  +   (Grid) s()((I, J, K-1)) +  (Grid) s()((I, J, K+1)))  + 
    w[2] * ((Grid) u2()((RI, RJ, RK))  +  (Grid) u1()((RI, RJ, RK-1)) + (Grid) u1()((RI, RJ, RK+1))) + 
    w[3] * ((Grid) u2()((RI, RJ, RK-1)) +  (Grid) u2()((RI, RJ, RK+1))); 
  
  return r;
}


Grid conv2 (Grid s, Grid r, double w[4], int step, Tuple<3> d)
{
  //size of the leaf HTA (assuming HTAs are regular)
  Tuple<3> size = s(T(0,0,0)).shape().size();  

  Tuple<3> rsize = Tuple<3>::max((size - 2) / step, 1) + 2;  
  
  /* GB: temporaries -- allocate it outside the function */
  Sequence<Tuple<3> > tiling(rsize, s.shape().size());  
  Grid u1 = HTA<double, 3>::alloc(1,tiling, s.dist().clone(), order);
  Grid u2 = HTA<double, 3>::alloc(1,tiling, s.dist().clone(), order); 
  Grid u3 = HTA<double, 3>::alloc(1,tiling, s.dist().clone(), order);
  Grid u4 = HTA<double, 3>::alloc(1,tiling, s.dist().clone(), order); 

  u1 = 0.0;
  
  u2 = 0.0;
  
  u3 = 0.0;

  u4  = 0.0;
  
  Triplet I(2*2-d[0], 2*(rsize[0]-1)-d[0], step);
  Triplet J(2*2-d[1], 2*(rsize[1]-1)-d[1], step);
  Triplet K(2*2-d[2], 2*rsize[2]-d[2], step);

  Triplet RI(1, rsize[0] - 2);
  Triplet RJ(1, rsize[1] - 2);
  Triplet RK(1, rsize[2] - 1);
    
  /*
   * TODO: Check if the compiler unrolls all the loops and contracts the temporary arrays 
   * I doubt it will do it. 
   */
  
  u1()((RI, RJ, RK)) =  
    (Grid) s()((I, J-1, K-1)) + 
    (Grid) s()((I, J+1, K-1)) +
    (Grid) s()((I-1, J, K-1)) +
    (Grid) s()((I+1, J, K-1));

  u2()((RI, RJ, RK)) =  
    (Grid) s()((I-1, J-1, K-1)) +
    (Grid) s()((I-1, J+1, K-1)) +
    (Grid) s()((I+1, J-1, K-1)) +
    (Grid) s()((I+1, J+1, K-1));
  
  K = Triplet(2*2-d[2], 2*(rsize[2]-1)-d[2], step);
  RK = Triplet(1, rsize[2] - 2);
  
  u3()((RI, RJ, RK)) =  
    (Grid) s()((I, J-1, K)) + 
    (Grid) s()((I, J+1, K)) +
    (Grid) s()((I-1, J, K)) +
    (Grid) s()((I+1, J, K));

  u4()((RI, RJ, RK)) =  
    (Grid) s()((I-1, J-1, K)) +
    (Grid) s()((I-1, J+1, K)) +
    (Grid) s()((I+1, J-1, K)) +
    (Grid) s()((I+1, J+1, K));

  //   r()((RI, RJ, RK)) = 
  r = 
    w[0] *  (Grid) s()((I, J, K)) + 
    w[1] * ((Grid) u3()((RI, RJ, RK))  +   (Grid) s()((I, J, K-1)) +  (Grid) s()((I, J, K+1)))  + 
    w[2] * ((Grid) u4()((RI, RJ, RK))  +  (Grid) u1()((RI, RJ, RK)) + (Grid) u1()((RI, RJ, RK+1))) + 
    w[3] * ((Grid) u2()((RI, RJ, RK+1)) +  (Grid) u2()((RI, RJ, RK)));   
  
   return r;
}

void interp (Grid& z, Grid& u, int k)
{
  timer_start (T_interp); 
  
  if (z.shape() == u.shape()) {
      interp_normal (z, u);
  } else {
    int s1 = z.shape().size()[0] == u.shape().size()[0] ? 1 : 2;
    int s2 = z.shape().size()[1] == u.shape().size()[1] ? 1 : 2;
    int s3 = z.shape().size()[2] == u.shape().size()[2] ? 1 : 2;
    
    Sequence<Triplet> t_args(Triplet (s1-1, u.shape().size()[0]-1, s1), 
			     Triplet (s2-1, u.shape().size()[1]-1, s2), 
			     Triplet (s3-1, u.shape().size()[2]-1, s3));
    interp_special (z, u(t_args));     
    comm3_ex (u, s1, s2, s3);
  }  

  timer_stop (T_interp);
  
  if (debug_vec[0]) rep_nrm (z, "interp", k-1);
  if (debug_vec[0]) rep_nrm (u, "interp", k);
}

void interp_normal (Grid z, Grid u )
{  

  timer_start (T_interp_inner);

  
   int iSize = z (T(0,0,0)).shape().size()[0];
   int jSize = z (T(0,0,0)).shape().size()[1];
   int kSize = z (T(0,0,0)).shape().size()[2];

   int nx = z.shape().size()[0];
   int ny = z.shape().size()[1];
   int nz = z.shape().size()[2];
     
   Sequence<Tuple<3> > tiling(Tuple<3> (iSize, jSize, kSize), 
 			     Tuple<3>(nx, ny, nz));

   /* GB: allocate it outside the loops */
   Grid tmpz1 = HTA<double, 3>::alloc(1,tiling, z.dist().clone(), order);
   Grid tmpz2 = HTA<double, 3>::alloc(1,tiling, z.dist().clone(), order);
   Grid tmpz3 = HTA<double, 3>::alloc(1,tiling, z.dist().clone(), order);
  
   Triplet I(0, iSize-2);
   Triplet J(0, jSize-2);
   Triplet K(0, kSize-1);

   tmpz1()((I, J, K)) =  z()((I,     J + 1, K)) + z()((I,     J, K));
   tmpz2()((I, J, K)) =  z()((I + 1, J,     K)) + z()((I,     J, K));
   tmpz3()((I, J, K)) =  z()((I + 1, J + 1, K)) + z()((I + 1, J, K));  
   tmpz3()((I, J, K)) =   tmpz3()((I, J, K)) + tmpz1()((I, J, K));    
  
   Triplet X(0, iSize-2);
   Triplet Y(0, jSize-2);
   Triplet Z(0, kSize-2);

   Triplet A(0, 2 * (iSize-2), 2);
   Triplet B(0, 2 * (jSize-2), 2);
   Triplet C(0, 2 * (kSize-2), 2);

   u()((A,     B,     C + 1)) =  u()((A,     B,     C + 1)) + 0.5 *  ( z()(((X, Y,    Z + 1))) +  z()((X, Y, Z)));
   u()((A + 1, B,     C + 1)) =  u()((A + 1, B,     C + 1)) + 0.25 * ( tmpz2()((X, Y, Z)) +  tmpz2()((X, Y, Z + 1) ));
   u()((A,     B + 1, C + 1)) =  u()((A,     B + 1, C + 1)) + 0.25 * ( tmpz1()((X, Y, Z)) +  tmpz1()((X, Y, Z + 1)));
   u()((A + 1, B + 1, C + 1)) =  u()((A + 1, B + 1, C + 1)) + 0.125 *( tmpz3()((X, Y, Z)) +  tmpz3()((X, Y, Z + 1)));
  
   u()((A,     B,     C)) =       u()((A,     B,     C)) +            z()((X, Y, Z));  
   u()((A + 1, B,     C)) +=      0.5 * tmpz2()((X, Y, Z));
   u()((A,     B + 1, C)) +=      0.5 * tmpz1()((X, Y, Z));
   u()((A + 1, B + 1, C)) +=      0.25 * tmpz3()((X, Y, Z));    
   
  timer_stop (T_interp_inner);  
}

void interp_special (Grid z, Grid u)
{

   int n1 = u(T(0,0,0)).shape().size()[0]; 
   int n2 = u(T(0,0,0)).shape().size()[1]; 
   int n3 = u(T(0,0,0)).shape().size()[2]; 
  
   int d1 = n1 == 3 ? 1 : 0;
   int t1 = n1 == 3 ? 0 : -1;
   int d2 = n2 == 3 ? 1 : 0;
   int t2 = n2 == 3 ? 0 : -1;
   int d3 = n3 == 3 ? 1 : 0;
   int t3 = n3 == 3 ? 0 : -1;

   int mm1 = z(T(0,0,0)).shape().size()[0];
   int mm2 = z(T(0,0,0)).shape().size()[1];
   int mm3 = z(T(0,0,0)).shape().size()[2];
 
  
   cout << "Here " << d1 <<  " " << mm1 << endl;
 
   {
     Triplet X(d1, mm1-2);
     Triplet Y(d2, mm2-2);
     Triplet Z(d3, mm3-2);
     Triplet A1(d1, 2 * (mm1-2)-d1, 2);
     Triplet B1(d2, 2 * (mm2-2)-d2, 2);
     Triplet C1(d3, 2 * (mm3-2)-d3, 2);
     Triplet A2(-t1, 2 * (mm1-2)-t1, 2);
     Triplet X2(0, mm1-2);
    
     u()((A1,     B1,     C1)) =   u()((A1, B1, C1)) + z()((X, Y, Z));  
     u()((A2,     B1,     C1)) +=   0.5 * (z()((X2 + 1, Y, Z)) +  z()((X2, Y, Z)));
   }

   {          
     Triplet I1(d1, mm1-2);
     Triplet I2(0, mm2-2);
     Triplet I3(d3, mm3-2);

     Triplet J1(d1, 2*(mm1-2)-d1, 2);
     Triplet J2(-t2, 2*(mm2-2)-t2, 2);
     Triplet J3(d3, 2*(mm3-2)-d3, 2);

     u()((J1, J2, J3)) += 0.5 * (z()((I1, I2+1, I3)) + z()((I1, I2, I3)));
   }
  

   { 
     Triplet I1(0, mm1-2);
     Triplet I2(0, mm2-2);
     Triplet I3(d3, mm3-2);
    
     Triplet J1(-t1, 2*(mm1-2)-t1, 2);
     Triplet J2(-t2, 2*(mm2-2)-t2, 2);
     Triplet J3(d3,  2*(mm3-2)-d3, 2);
    
         cout << "Here " << endl;
     u()((J1, J2, J3)) +=  0.25 * (z()((I1+1, I2+1, I3)) + z()((I1+1, I2, I3))
 				  + z()((I1, I2, I3)) + z()((I1, I2+1, I3)));        
   }
  
   {
     Triplet I1(d1, mm1-2);
     Triplet I2(d2, mm2-2);
     Triplet I3(0, mm3-2);
    
     Triplet J1(d1, 2*(mm1-2)-d1, 2);
     Triplet J2(d2, 2*(mm2-2)-d2, 2);
     Triplet J3(-t3, 2*(mm3-2)-t3, 2);
     u()((J1, J2, J3)) = 0.5 * (z()((I1, I2, I3+1)) + z()((I1, I2, I3)));    
   }
  
   {
     Triplet I1(0, mm1-2);
     Triplet I2(d2, mm2-2);
     Triplet I3(0, mm3-2);
    
     Triplet J1(-t1, 2*(mm1-2)-t1, 2);
     Triplet J2(d2, 2*(mm2-2)-d2, 2);
     Triplet J3(-t3, 2*(mm3-2)-t3, 2);
    
     u()((J1, J2, J3)) = 0.25 * (z()((I1+1, I2, I3+1)) + z()((I1, I2, I3+1)) + z()((I1+1, I2, I3)) + z()((I1, I2, I3)));
   }
  
   {
       Triplet I1(d1, mm1-2);
       Triplet I2(0, mm2-2);
       Triplet I3(0, mm3-2);
      
       Triplet J1(d1, 2*(mm1-2)-d1, 2);
       Triplet J2(-t2, 2*(mm2-2)-t2, 2);
       Triplet J3(-t3, 2*(mm3-2)-t3, 2);
      
       u()((J1, J2, J3)) = 0.25 * (z()((I1, I2+1, I3)) + z()((I1, I2, I3+1)) + z()((I1, I2+1, I3)) + z()((I1, I2, I3)));
     }
  
   {
     Triplet I1(0, mm2-2);
     Triplet I2(0, mm2-2);
     Triplet I3(0, mm3-2);
    
     Triplet J1(-t1, 2*(mm1-2)-t1, 2);
     Triplet J2(-t2, 2*(mm2-2)-t2, 2);
     Triplet J3(-t3, 2*(mm3-2)-t3, 2);
    
     u()((J1, J2, J3)) = 0.125 * (z()((I1+1, I2+1, I3+1)) + 
 				 z()((I1+1, I2, I3+1)) + 
 				 z()((I1,   I2+1, I3+1)) + 
 				 z()((I1,   I2,   I3+1)) + 
 				 z()((I1+1, I2+1, I3)) + 
 				 z()((I1+1, I2, I3)) + 
 				 z()((I1, I2+1, I3)) + 
 				 z()((I1, I2, I3)));
   }
  
}


void zero3 (Grid& z)
{
  timer_start (T_zero3);
  z = 0.0;
  timer_stop (T_zero3);
}

void comm3 (Grid& u)
{
  timer_start (T_comm3);
  
  int NX = u.shape().size()[0] - 1;
  int NY = u.shape().size()[1] - 1;
  int NZ = u.shape().size()[2] - 1;
  
  int nx = u(T(0,0,0)).shape().size()[0] - 1;
  int ny = u(T(0,0,0)).shape().size()[1] - 1;
  int nz = u(T(0,0,0)).shape().size()[2] - 1;
  
  /* 
   * TODO: avoid the extra statement for wraparound
   * communication. Instead, Wrap the Range (Triplet)
   * using some operator (e.g. operator <<). For e.g.,
   * u(R(0, NX), R(0, NY), R(0, NZ)](R(nx, nx), R(1, ny-1), R(1, nz-1)] = u(R(0, NX) << 1, R(0, NY), R(0, NZ)](R(1,1), R(1, ny-1), R(1,nz-1)];

   * Another possibility is to have the concept of  AllRange (A for short or '_'). Then we would have,
   * u[_, _, _](R(nx, nx), R(1, ny-1), R(1, nz-1)] = u[_ << 1, _, _](R(1,1), R(1, ny-1), R(1,nz-1)];
   */

  //north-south

  Traits::Default::async();

  //  cout << "Final " << ((Grid)  u((R(0, NX-1), R(0, NY), R(0, NZ)))).dist() << ((Grid)  u((R(1, NX), R(0, NY), R(0, NZ)))).dist() << endl;

  if (NX > 0) 
    u((R(0, NX-1), R(0, NY), R(0, NZ)))((R(nx, nx), R(1, ny-1), R(1, nz-1))) = u((R(1, NX), R(0, NY), R(0, NZ)))((R(1,1), R(1, ny-1), R(1,nz-1)));

  //  cout << "after " << endl;
  u((R(NX, NX), R(0, NY), R(0, NZ)))((R(nx, nx), R(1, ny-1), R(1, nz-1))) = u((R(0,0), R(0, NY), R(0, NZ)))((R(1,1), R(1, ny-1), R(1,nz-1)));
  
  if (NX > 0) 
    u((R(1, NX), R(0, NY), R(0, NZ)))((R(0,0), R(1, ny-1), R(1, nz-1))) = u((R(0, NX-1), R(0, NY), R(0, NZ)))((R(nx-1,nx-1), R(1, ny-1), R(1,nz-1)));
  u((R(0, 0), R(0, NY), R(0, NZ)))((R(0,0), R(1, ny-1), R(1, nz-1))) = u((R(NX, NX), R(0, NY), R(0, NZ)))((R(nx-1,nx-1), R(1, ny-1), R(1,nz-1)));

  Traits::Default::sync();
  Traits::Default::async();

  //east-west
  if (NY > 0) 
    u((R(0, NX), R(0, NY-1), R(0, NZ)))((R(0, nx), R(ny, ny), R(1, nz-1))) = u((R(0, NX), R(1, NY), R(0, NZ)))((R(0, nx), R(1, 1), R(1, nz-1)));
  u((R(0, NX), R(NY, NY), R(0, NZ)))((R(0, nx), R(ny, ny), R(1, nz-1))) = u((R(0, NX), R(0,0), R(0, NZ)))((R(0,nx), R(1, 1), R(1, nz-1)));
  if (NY > 0) 
    u((R(0, NX), R(1, NY), R(0, NZ)))((R(0, nx), R(0, 0), R(1, nz-1))) = u((R(0, NX), R(0, NY-1), R(0, NZ)))((R(0, nx), R(ny-1, ny-1), R(1, nz-1)));
  u((R(0, NX), R(0, 0), R(0, NZ)))((R(0, nx), R(0, 0), R(1, nz-1))) = u((R(0, NX), R(NY, NY), R(0, NZ)))((R(0,nx), R(ny-1, ny-1), R(1, nz-1)));

  Traits::Default::sync();
  Traits::Default::async();

  //front-back
  
  if (NZ > 0) 
    u((R(0, NX), R(0, NY), R(0, NZ-1)))((R(0, nx), R(0, ny), R(nz, nz))) = u((R(0, NX), R(0, NY), R(1, NZ)))((R(0, nx), R(0, ny), R(1, 1)));
  u((R(0, NX), R(0, NY), R(NZ, NZ)))((R(0, nx), R(0, ny), R(nz, nz))) = u((R(0, NX), R(0, NY), R(0, 0)))((R(0, nx), R(0, ny), R(1, 1)));

  if (NZ > 0) 
    u((R(0, NX), R(0, NY), R(1, NZ)))((R(0, nx), R(0, ny), R(0, 0))) = u((R(0, NX), R(0, NY), R(0, NZ-1)))((R(0, nx), R(0, ny), R(nz-1, nz-1)));
  u((R(0, NX), R(0, NY), R(0, 0)))((R(0, nx), R(0, ny), R(0, 0))) = u((R(0, NX), R(0, NY), R(NZ, NZ)))((R(0, nx), R(0, ny), R(nz-1, nz-1)));   
  Traits::Default::sync();

  timer_stop (T_comm3);
}

#if 0
void comm3 (Grid& u)
{
  int NX = u.shape().size()[0] - 1;
  int NY = u.shape().size()[1] - 1;
  int NZ = u.shape().size()[2] - 1;
  
  int nx = u(T(0,0,0)).shape().size()[0] - 1;
  int ny = u(T(0,0,0)).shape().size()[1] - 1;
  int nz = u(T(0,0,0)).shape().size()[2] - 1;
  
  /* 
   * TODO: avoid the extra statement for wraparound
   * communication. Instead, Wrap the Range (Triplet)
   * using some operator (e.g. operator <<). For e.g.,
   * u[R(0, NX), R(0, NY), R(0, NZ)][R(nx, nx), R(1, ny-1), R(1, nz-1)] = u[R(0, NX) << 1, R(0, NY), R(0, NZ)][R(1,1), R(1, ny-1), R(1,nz-1)];
   
   * Another possibility is to have the concept of  AllRange ('_' for short). Then we would have,
   * u[_, _, _][R(nx, nx), R(1, ny-1), R(1, nz-1)] = u[_ << 1, _, _][R(1,1), R(1, ny-1), R(1,nz-1)];
   
   * Thus, we can reduce the *length* of a statement!
   */
  
  //north-south
  u ((R(0, NX),     R(0, NY), R(0, NZ))) ((R(nx, nx), R(1, ny-1), R(1, nz-1))) = u ((R(0, NX) % 1, R(0, NY), R(0, NZ))) ((R(1,1), R(1, ny-1),       R(1,nz-1)));
  u ((R(0, NX) % 1, R(0, NY), R(0, NZ))) ((R(0,0),    R(1, ny-1), R(1, nz-1))) = u ((R(0, NX),     R(0, NY), R(0, NZ))) ((R(nx-1,nx-1), R(1, ny-1), R(1,nz-1)));
    
  //east-west
  u ((R(0, NX), R(0, NY),     R(0, NZ))) ((R(0, nx), R(ny, ny), R(1, nz-1))) = u ((R(0, NX), R(0, NY) % 1, R(0, NZ))) ((R(0, nx), R(1, 1),       R(1, nz-1)));
  u ((R(0, NX), R(0, NY) % 1, R(0, NZ))) ((R(0, nx), R(0, 0),   R(1, nz-1))) = u ((R(0, NX), R(0, NY),     R(0, NZ))) ((R(0, nx), R(ny-1, ny-1), R(1, nz-1)));
  
  //front-back
  u ((R(0, NX), R(0, NY), R(0, NZ)))     ((R(0, nx), R(0, ny), R(nz, nz))) = u ((R(0, NX), R(0, NY), R(0, NZ) % 1)) ((R(0, nx), R(0, ny), R(1, 1)));  
  u ((R(0, NX), R(0, NY), R(0, NZ) % 1)) ((R(0, nx), R(0, ny), R(0, 0)))   = u ((R(0, NX), R(0, NY), R(0, NZ)))     ((R(0, nx), R(0, ny), R(nz-1, nz-1)));
  
}
#endif

 void comm3_ex (Grid& u, int s1, int s2, int s3)
 {
   // GBDEBUG 
   // cout << "here ... " << endl; 

   if (s1 == 2) comm_ex_ns (u);
   if (s2 == 2) comm_ex_ew (u);
   if (s3 == 2) comm_ex_fb (u);
 }


 void comm_ex_ns (Grid& u)
 {

   int NX = u.shape().size()[0] - 1;
   int NY = u.shape().size()[1] - 1;
   int NZ = u.shape().size()[2] - 1;

   int ny = u (T(0,0,0)).shape().size()[1] - 1; 
   int nz = u (T(0,0,0)).shape().size()[2] - 1; 

   u((SR(0, NX, 2), R(0, NY), R(0, NZ)))((R(2, 2), R(0, ny), R(0, nz))) = u((SR(1, NX, 2), R(0, NY), R(0, NZ)))((R(1, 1), R(0, ny), R(0, nz)));
   u((SR(2, NX-1, 2), R(0, NY), R(0, NZ)))((R(0, 1), R(0, ny), R(0, nz))) = u((SR(1, NX-1, 2), R(0, NY), R(0, NZ)))((R(1, 2), R(0, ny), R(0, nz)));
   u((R(0, 0), R(0, NY), R(0, NZ)))((R(0, 1), R(0, ny), R(0, nz))) = u((R(NX, NX), R(0, NY), R(0, NZ)))((R(1, 2), R(0, ny), R(0, nz)));
   
}

 void comm_ex_ew (Grid& u)
 {

   int NX = u.shape().size()[0] - 1;
   int NY = u.shape().size()[1] - 1;
   int NZ = u.shape().size()[2] - 1;

   int nx = u (T(0,0,0)).shape().size()[0] - 1; 
   int nz = u (T(0,0,0)).shape().size()[2] - 1; 

   u((R(0, NX), SR(0, NY, 2), R(0, NZ)))((R(0, nx), R(2, 2), R(0, nz))) = u((R(0, NX), SR(0, NY, 2), R(0, NZ)))((R(0, nx), R(1, 1), R(0, nz)));
   u((R(0, NX), SR(2, NY-1, 2), R(0, NZ)))((R(0, nx), R(0, 1), R(0, nz))) = u((R(0, NX), SR(1, NY-1, 2), R(0, NZ)))((R(0, nx), R(1, 2), R(0, nz)));
   u((R(0, NX), R(0, 0), R(0, NZ)))((R(0, nx), R(0, 1), R(0, nz))) = u((R(0, NX), R(NY, NY), R(0, NZ)))((R(0, nx), R(1, 2), R(0, nz)));
   
 }

void comm_ex_fb (Grid& u)
 {

   int NX = u.shape().size()[0] - 1;
   int NY = u.shape().size()[1] - 1;
   int NZ = u.shape().size()[2] - 1;

   int nx = u (T(0,0,0)).shape().size()[0] - 1; 
   int ny = u (T(0,0,0)).shape().size()[1] - 1;       
   
   u((R(0, NX), R(0, NY), SR(0, NZ, 2)))((R(0, nx), R(0, ny), R(2, 2))) = u((R(0, NX), R(0, NY), SR(1, NZ, 2)))((R(0, nx), R(0, ny), R(1, 1)));
   u((R(0, NX), R(0, NY), SR(1, NZ-1, 2)))((R(0, nx), R(0, ny), R(0, 1))) = u((R(0, NX), R(0, NY), SR(1, NZ-1, 2)))((R(0, nx), R(0, ny), R(1, 2)));
   u((R(0, NX), R(0, NY), R(0, 0)))((R(0, nx), R(0, ny), R(0, 1))) = u((R(0, NX), R(0, NY), R(NZ, NZ)))((R(0, nx), R(0, ny), R(1, 2)));
   
}


/**
 * GB: Norm is a very basic vector/matrix operation
 * Hence this should be defined in htalib...
 */
 void norm2u3 (Grid& r, double *rnm2, double *rnmu, int nx, int ny, int nz)
 {
   timer_start (T_norm);
  
   double n =(double) nx * ny * nz;
   int tile_x = r (T(0,0,0)).shape().size()[0]; 
   int tile_y = r (T(0,0,0)).shape().size()[1]; 
   int tile_z = r (T(0,0,0)).shape().size()[2]; 
   SquareSum <double, 3> mr = SquareSum <double, 3>();
   Grid tmp = r()((R(1, tile_x-2), R(1, tile_y-2), R(1, tile_z-2)));
   tmp.mapReduce(mr);
   *rnm2 = sqrt (mr.sqsum / (double) n);
   *rnmu = mr.max;

   timer_stop (T_norm);
 }

 void rep_nrm (Grid &x, char* callee, int k )
 {
   double rnm2;
   double rnmu;

   int div = (int) pow(2.0,lt-k-1);
   int nx = NX_DEFAULT / div;
   int ny = NY_DEFAULT / div;
   int nz = NZ_DEFAULT / div;

   norm2u3 (x, &rnm2, &rnmu, nx, ny, nz);
   IO::cout << callee << " rep_nrm:: nx=" << nx << " ny=" << ny << " nz=" << nz << " rnm2=" << rnm2 << " rnmu=" << rnmu << "k " << k <<  endl;
 }

 void zran3 (Grid & z, int gx, int gy, int gz)
 {

   zero3 (z);   
   Shape<3>::iterator iz = z.shape().begin();   
   Shape<3>::iterator iz_end = z.shape().end();   
   for (; iz != iz_end; iz++) {
     if (z(*iz).isLocal()) 
       zran3_inner(z(*iz), *iz, gx, gy, gz);
   }

   // create and apply map-reduce
   FilterN <double, 10, 3> mr = FilterN <double, 10, 3>();
   
   z.mapReduce(mr);
   zero3 (z);

   //sort the result in some order
   XchgData<double, 10, 3> r = mr.result();
   double maxVal[10];
   double minVal[10];
   Tuple<3> maxKey[10];
   Tuple<3> minKey[10];
   for (int i = 0; i < 10; i++)
     {
       maxVal[i] = r.maxVal[i];
       maxKey[i] = r.maxKey[i];
       minVal[i] = r.minVal[i];
       minKey[i] = r.minKey[i];
     }
   
   for (int i = 0; i < 10; i++)
     for (int j = i+1; j < 10; j++)
       {
	 if (maxVal[i] > maxVal[j])
	   {
	     double tmp = maxVal[i];
	     maxVal[i] = maxVal[j];
	     maxVal[j] = tmp;
	     
	     Tuple<3> t = maxKey[i];
	     maxKey[i] = maxKey[j];
	     maxKey[j] = t;
	   }
	 
	 if (minVal[i] > minVal[j])
	   {
	     double tmp = minVal[i];
	     minVal[i] = minVal[j];
	     minVal[j] = tmp;
	     
	     Tuple<3> t = minKey[i];
	     minKey[i] = minKey[j];
	     minKey[j] = t;
	   }
       }
   
   // set minimas and maximas
   for (int i = 0; i < 10; ++i) {
     z [minKey[i]] = -1;
     z [maxKey[i]] = 1;
   }

 }


/* helper method: initialize grid z with random values */
void zran3_inner (Leaf z, Tuple<3> tileIndex, int gx, int gy, int gz)
{
  double a = pow(5.0, 13);
  int nx = z.shape().size()[0] - 2;
  int ny = z.shape().size()[1] - 2;
  int nz = z.shape().size()[2] - 2;

  Tuple<3> localIndexStart = tileIndex * Tuple<3> (nx, ny, nz);
  int pi = localIndexStart[0]  + localIndexStart[1] * gx + localIndexStart[2] * gx * gy;

  double ai = RandUtils::power (a, pi); 
  double a1 = RandUtils::power (a, gx);
  double a2 = RandUtils::power (a, gx*gy);

  double x0 = 314159265.0;
  double rdummy = RandUtils::randlc (&x0, ai);
  for (int k = 1; k < z.shape().size()[2]-1; k++) {
    double x1 = x0;
    for (int j = 1; j < z.shape().size()[1]-1; j++) {
	  double x2 = x1;
	  for (int i = 1; i < z.shape().size()[0]-1; i++) 
	    z [T(i,j,k)] = RandUtils::randlc (&x2, a);
	  rdummy = RandUtils::randlc (&x1, a1);
    }
    rdummy = RandUtils::randlc (&x0, a2);
  }
}	      


void setup (Grid* r, Grid* u, int nprocs)
{    
  int pi[3];
  int ng[3];
  int di[3];
  int si[3];

  double log_p  = log(double(nprocs)+0.0001)/log(2.0);
  int dx     = int(log_p/3);
  pi[0] = (int) pow(2.0, dx);
  
  int dy     = int((log_p-dx)/2);
  pi[1]  = (int) pow (2.0, dy);
  
  pi[2]  = nprocs/(pi[0]*pi[1]);
  
  ng[0] = NX_DEFAULT;
  ng[1] = NY_DEFAULT;
  ng[2] = NZ_DEFAULT;

  for (int i = 0; i < 3; i++)
    di[i] =   pi[i]; 
  
  for (int k = lt-1; k >=0; k--)
    {
      for (int i = 0; i < 3; i++)
	{  
	  di[i] = di[i] - (ng[i] % di[i] );
	  int totalSize = ng[i] + di[i] * 2; //size + 2 * shadow region for each dimension (di)
	  si[i] = totalSize / di[i];
	  ng[i] /= 2; 
	}
                   
      //set a different distribution ((1:di[i]-1:pi[i]/di[i])) for end cases
      if (di[0] != pi[0] || di[1] != pi[1] || di[2] != pi[2])
	{
	  cout << "Here 1" << endl;
	  int s0 = di[0] != pi[0] ? 1 : 0;
	  int s1 = di[1] != pi[1] ? 1 : 0;
	  int s2 = di[2] != pi[2] ? 1 : 0;
	  
	  Grid tmp = r[k+1];
	  Tuple<3> sz = tmp.shape().size();

	  Grid tmp2 =  tmp((R(s0, sz[0]-1, s0+1), 
			    R(s1, sz[1]-1, s1+1), 
			    R(s2, sz[2]-1, s2+1)));
	
	  cout << "After 1 " << endl;
	  r[k] = HTA<double,3>::alloc (1,(Tuple<3>(si[0],si[1],si[2]), Tuple<3>(di[0],di[1],di[2])), tmp2.dist().clone(), order);
	  u[k] = HTA<double,3>::alloc (1,(Tuple<3>(si[0],si[1],si[2]), Tuple<3>(di[0],di[1],di[2])), tmp2.dist().clone(), order);
	  cout <<"After 2" << endl;					 
	} else
	  {
	    r[k] = HTA<double,3>::alloc (1,(Tuple<3>(si[0],si[1],si[2]), Tuple<3>(di[0],di[1],di[2])), NULL, order);
	    u[k] = HTA<double,3>::alloc (1,(Tuple<3>(si[0],si[1],si[2]), Tuple<3>(di[0],di[1],di[2])), NULL, order);
	  }
           
      zero3 (r[k]);
      zero3 (u[k]);
    }
  
  MPI_Barrier (MPI_COMM_WORLD);
}


/* print and debug methods */

void print_timers ()

{
  char* name [] = { "total", "resid", "resid_inner", "psinv", "psinv_inner", "rprj3", "rprj3_inner", "interp", "interp_inner",
		    "norm", "comm3", "zero3"};
  
  for (int i = T_total; i <= T_zero3; i++)
    IO::cout << name[i] << " : " << timer_read (i) << endl;
}

/* main method */

#if (HTA_TRAIT == HTA_TRAIT_UPC)
// main is defined in the upc runtime ... (cludge code in main.o)
int upc_main( int argc, char** argv) {
  Traits::Default::init(argc, argv);
  int ret = run(argc, argv);
  Traits::Default::finalize();
  return ret;
  
}
#else
// main defined for serial and mpi
int main( int argc, char** argv) {
  
  Counters::disable();
  Tracing::dlevel(-1);

  Traits::Default::init(argc, argv);
  
  /* determine number of tiles to use */
  int ret;
  int ntiles;
  if (argc != 2) 
    ntiles = -1;
  else 
    ntiles = atoi(argv[1]);
  ret = run(ntiles);
  Traits::Default::finalize();
  return ret;
}

#endif /* HTA_TRAIT */
