/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: AbstractArray.h,v 1.10.4.1 2007/08/13 16:47:14 brodman2 Exp $
 * Authors: Christoph von Praun
 */

#ifndef  __ABSTRACT_ARRAY_H__
#define  __ABSTRACT_ARRAY_H__

#include <cassert>
#include "MemMapping.h"
#include "Shape.h"
#include "util/VectorOps.h"

using namespace std;

/** 
 * An abstract array has a logical shape and a physical mapping.
 */

template<typename T, int DIM>
class AbstractArray 
{

public:

  virtual const Shape<DIM>& shape () const = 0;

  virtual const MemMapping<DIM>& memMap () const = 0;

  virtual AbstractArray<T, DIM>& accessTile (const Tuple<DIM>& t) const = 0;

  virtual AbstractArray<T, DIM>& accessTile (int idx) const = 0;

  virtual int ref () const = 0;
  
  virtual T& accessScalar (const Tuple<DIM>& t) const = 0;

  virtual size_t level () const = 0;

  virtual bool isLocal () const = 0;

  virtual bool isDistributed () const = 0;

  virtual int home () const = 0;

  const typename Tuple<DIM>::Seq& tiling () const 
  {
    return memMap().tiling();
  }
  
  inline int hash () const 
  {
    if (hash_ == -1) 
      computeHash_();
    return hash_;
  }

  /**
   * only instances tha are an immediate result of clone or alloc shall 
   * be cachable. Other instance are 'views' / aliases of cacheable arrays 
   * and may not be cached themselves.
   */
  inline bool isCachable () const
  {
    return cacheable_;
  }

  inline void setCachable () 
  {
    cacheable_ = true;
  }

  static bool shapeConformable (const AbstractArray<T, DIM>& arg1, const AbstractArray<T, DIM>& arg2) 
  {
    size_t level = arg1.level();
    bool ret = level == arg2.level();
    if (ret) {
      const AbstractArray<T, DIM>* arg1_tmp = &arg1;
      const AbstractArray<T, DIM>* arg2_tmp = &arg2;
      for (int i = level; ret && i >= 0; --i) {
	if (arg1_tmp->shape() != arg2_tmp->shape())
	  ret = false;
	else {
	  if (i > 0) {
	    const Tuple<DIM>& t_low = arg1_tmp->shape().low();
	    arg1_tmp = &(arg1_tmp->accessTile(t_low)); 
	    arg2_tmp = &(arg2_tmp->accessTile(t_low));
	  }
	}
      }
    }
    return ret;
  }

  static bool memConformable (const AbstractArray<T, DIM>& arg1, const AbstractArray<T, DIM>& arg2) 
  {
    return arg1.memMap().tiling() == arg2.memMap().tiling();
  }
   
  
  /** 
   * This virtual destructor is essential to have the 
   * destructors of subclasses called, e.g. if instances of 
   * AbstractArray are removed form the HTA Pool.
   */
  virtual ~AbstractArray() { }

protected:

  AbstractArray() : hash_(-1), cacheable_(false) {}
   
  AbstractArray (const AbstractArray<T, DIM>& other) :
    hash_ (other.hash_),
    cacheable_ (other.cacheable_) { cout << "Copy " << endl;}
  
public:
  
  void computeHash_() const
  {
    AbstractArray<T, DIM>* this_non_const = const_cast<AbstractArray<T, DIM>*>(this);

    // encode the level
    int h = (level() << 28);
    // encode the memory-mapping
    for (size_t i = 0; i < level(); ++i) {
      h += (VectorOps<Tuple<DIM>, DIM>::selfProduct(tiling()[0])) << 10;
    }

    // encode the logical tiling
    const AbstractArray<T, DIM>* tmp = this;
    for (int i = level(); i >= 0; --i) {
      h += VectorOps<Tuple<DIM>, DIM>::selfProduct(tmp->shape().size());
      if (i > 0)
	tmp = &(tmp->accessTile(tmp->shape().low()));//JG: this is a tileAt in destruction
      
    }
    
    this_non_const->hash_ = h;
  }

  int hash_;
  bool cacheable_;
};

#endif /* __ABSTRACT_ARRAY_H__ */
