/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: BinExpr.h,v 1.10 2007/02/14 16:09:58 bikshand Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef  __BIN_EXPR_H__
#define  __BIN_EXPR_H__

#include <cassert>
#include <functional>
#include <iostream>

using namespace std;

/** 
 * A BinExpr serves to defer evaluation of a binary expression until 
 * an assignment statement. This avaoids temporaries in operations such 
 * as A = B + C. This corresponds to a 'poor mans expression templates'.
 * we do this here, because it covers the common case and is not as 
 * complex as the expression templates.
 */

template<typename OP, typename TRAIT>
class BinExpr {

  enum Kind_ { HH, HS, SH };

public:
    
  BinExpr (const typename TRAIT::Type& arg1, const typename TRAIT::Type& arg2, typename TRAIT::TypeImpl* dst = NULL) :
    kind_(HH),
    arg1_(arg1), 
    arg1s_(),
    arg2_(arg2),
    arg2s_(), 
    dst_(dst)
  {
    if (dst != NULL)
      dst->refCount_(1);
  }

  BinExpr (const typename TRAIT::Type& arg1, const typename TRAIT::TypeScalar& arg2, typename TRAIT::TypeImpl* dst = NULL) :
    kind_(HS),
    arg1_(arg1), 
    arg1s_(),
    arg2_(),
    arg2s_(arg2), 
    dst_(dst) 
  {
    if (dst != NULL)
      dst->refCount_(1);
  }

  BinExpr (const typename TRAIT::TypeScalar& arg1, const typename TRAIT::Type& arg2, typename TRAIT::TypeImpl* dst = NULL) :
    kind_(SH),
    arg1_(), 
    arg1s_(arg1),
    arg2_(arg2),
    arg2s_(), 
    dst_(dst)
  {
    if (dst != NULL)
      dst->refCount_(1);
  }

  virtual ~BinExpr () 
  {
    if (dst_ && dst_->refCount_(-1) == 0) 
      delete dst_;
  }

  void eval (typename TRAIT::Type& dst) const
  {
    eval_(*dst.wrapped_);
  }

  template<typename OP2>
  void evalAndAccum (typename TRAIT::Type& dst,
		     OP2 acc) const
  {
    evalAndAccum_(*dst.wrapped_, acc);
  }
  
  template<typename OP2>
  void evalAndAccum (typename TRAIT::TypeImpl& dst,
		     OP2 acc) const
  {
    evalAndAccum_(dst, acc);
  }

  typename TRAIT::Type eval () const
  {
    if (dst_) {
      eval_(*dst_);
      return typename TRAIT::Type(*dst_);
    } else {
      return static_cast<typename TRAIT::Type> (*this);
    }
  }

  /**
   * Implicit conversion of a binary-op to an HTA
   */
  operator typename TRAIT::Type () const 
  {   
    switch (kind_) {
    case HH: 
    case HS: {
      typename TRAIT::Type dst = arg1_.clone();
      eval (dst);
      return dst;
    }
    case SH: {
      typename TRAIT::Type dst = arg2_.clone();
      eval (dst);
      return dst;
    }
    default:
      assert (false);
      return typename TRAIT::Type();
    }
    }

  // operator+ --------------------------------------------------------------------

  typename TRAIT::Type operator+ (const typename TRAIT::TypeScalar& rhs) const
  {
    typename TRAIT::Type result = static_cast<typename TRAIT::Type>(*this);
    result += rhs;
    return result;
  }

  BinExpr<plus<typename TRAIT::TypeScalar>, TRAIT> operator+ (const typename TRAIT::Type& rhs) const
  {
     typename TRAIT::Type result = static_cast<typename TRAIT::Type>(*this);
     return BinExpr<plus<typename TRAIT::TypeScalar>, TRAIT> (result, rhs);
  }

  // operator- --------------------------------------------------------------------

  typename TRAIT::Type operator- (const typename TRAIT::TypeScalar& rhs) const
  {
    typename TRAIT::Type result = static_cast<typename TRAIT::Type>(*this);
    result -= rhs;
    return result;
  }

  BinExpr<minus<typename TRAIT::TypeScalar>, TRAIT> operator- (const typename TRAIT::Type& rhs) const
  {
     typename TRAIT::Type result = static_cast<typename TRAIT::Type>(*this);
     return BinExpr<minus<typename TRAIT::TypeScalar>, TRAIT> (result, rhs);
  }
  

  // operator* --------------------------------------------------------------------
  
  typename TRAIT::Type operator* (const typename TRAIT::TypeScalar& rhs) const
  {
    // materialize the expression, then multiply in place ...
    typename TRAIT::Type result = static_cast<typename TRAIT::Type>(*this);
    result *= rhs;
    return result;
  }

  BinExpr<multiplies<typename TRAIT::TypeScalar>, TRAIT> operator* (const typename TRAIT::Type& rhs) const
  {
     typename TRAIT::Type result = static_cast<typename TRAIT::Type>(*this);
     return BinExpr<multiplies<typename TRAIT::TypeScalar>, TRAIT> (result, rhs);
  }
  
  // operator/ --------------------------------------------------------------------
  
  typename TRAIT::Type operator/ (const typename TRAIT::TypeScalar& rhs) const
  {
    typename TRAIT::Type result = static_cast<typename TRAIT::Type>(*this);
    result /= rhs;
    return result;
  }

  BinExpr<divides<typename TRAIT::TypeScalar>, TRAIT> operator/ (const typename TRAIT::Type& rhs) const
  {
     typename TRAIT::Type result = static_cast<typename TRAIT::Type>(*this);
     return BinExpr<divides<typename TRAIT::TypeScalar>, TRAIT> (result, rhs);
  }

private:

  void eval_ (typename TRAIT::TypeImpl& dst) const
  {
    switch (kind_) {
    case HH:
      arg2_.wrapped_->communicate_ (arg1_.wrapped_->dist());
      TRAIT::TypeImpl::map(OP(), *arg1_.wrapped_, *arg2_.wrapped_, dst);
      break;
    case HS:
      TRAIT::TypeImpl::map(binder2nd<OP>(OP(), arg2s_), *arg1_.wrapped_, dst); 
      break;
    case SH:
      TRAIT::TypeImpl::map(binder1st<OP>(OP(), arg1s_), *arg2_.wrapped_, dst);
      break;
    }
  }
  
  template<typename OP2>
  void evalAndAccum_ (typename TRAIT::TypeImpl& dst,
		      const OP2 & acc) const
  {
    switch (kind_) {
    case HH:
      assert (false);
      break;
    case HS:
      TRAIT::TypeImpl::mapAccum (binder1st<OP>(OP(), arg2s_), acc, *arg1_.wrapped_, dst);
      break;
    case SH:
      TRAIT::TypeImpl::mapAccum (binder1st<OP>(OP(), arg1s_), acc, *arg2_.wrapped_, dst);
      break;
    }
  }
  
  const Kind_ kind_;
  const typename TRAIT::Type arg1_;
  const typename TRAIT::TypeScalar arg1s_;
  const typename TRAIT::Type arg2_;
  const typename TRAIT::TypeScalar arg2s_;
  typename TRAIT::TypeImpl* dst_;
};  // class BinExpr


template<class OP, typename TRAIT> 
BinExpr<plus<typename TRAIT::TypeScalar>, TRAIT> operator+ (const typename TRAIT::TypeScalar& val, 
							     const BinExpr<OP, TRAIT>& binexpr)
{
  typename TRAIT::Type result = static_cast<typename TRAIT::Type>(binexpr);
  return BinExpr<plus<typename TRAIT::TypeScalar>, TRAIT> (val, result);
}

template<class OP, typename TRAIT> 
BinExpr<minus<typename TRAIT::TypeScalar>, TRAIT> operator- (const typename TRAIT::TypeScalar& val, 
							      const BinExpr<OP, TRAIT>& binexpr)
{
  typename TRAIT::Type result = static_cast<typename TRAIT::Type>(binexpr);
  return BinExpr<minus<typename TRAIT::TypeScalar>, TRAIT> (val, result);
}

template<class OP, typename TRAIT> 
BinExpr<multiplies<typename TRAIT::TypeScalar>, TRAIT> operator* (const typename TRAIT::TypeScalar& val, 
								   const BinExpr<OP, TRAIT>& binexpr)
{
  typename TRAIT::Type result = static_cast<typename TRAIT::Type>(binexpr);
  return BinExpr<multiplies<typename TRAIT::TypeScalar>, TRAIT> (val, result);
}

template<class OP, typename TRAIT> 
BinExpr<divides<typename TRAIT::TypeScalar>, TRAIT> operator/ (const typename TRAIT::TypeScalar& val, 
								const BinExpr<OP, TRAIT>& binexpr)
{
  typename TRAIT::Type result = static_cast<typename TRAIT::Type>(binexpr);
  return BinExpr<divides<typename TRAIT::TypeScalar>, TRAIT> (val, result);
}

#endif /* __BIN_EXPR_H__ */
