/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: CyclicDistribution.h,v 1.13 2007/02/14 16:09:58 bikshand Exp $
 * Authors: Christoph von Praun
 */


#ifndef __CYCLIC_DISTRIBUTION_H__
#define __CYCLIC_DISTRIBUTION_H__

using namespace std;

#include "Distribution.h"
#include "Traits.h"
#include "Tuple.h"
#include "util/VectorOps.h"

template <int DIM>
class CyclicDistribution : public Distribution<DIM>
{
public:

  CyclicDistribution () 
  {
    procGrid_[0] = Traits::Default::nPlaces();
    for (int i = 1; i < DIM; ++i) 
      procGrid_[i] = 1;
  }
  
  CyclicDistribution (int nprocs) 
  {
    assert (nprocs <= Traits::Default::nPlaces());

    procGrid_[0] = nprocs;
    for (int i = 1; i < DIM; ++i) 
      procGrid_[i] = 1;
  }

  CyclicDistribution (const Tuple<DIM>& proc_grid)
  { 
    // grid must correpsond to number of processors
    int tmp1 = Traits::Default::nPlaces();
    int tmp2 = VectorOps<Tuple<DIM>, DIM>::selfProduct(proc_grid);
    assert (tmp1 >= tmp2);

    procGrid_[DIM-1] = proc_grid[DIM-1];
    for (int i = DIM-2; i >= 0; --i) 
      procGrid_[i] = proc_grid[i] * procGrid_[i-1];
  }

private:

  /* forbid copy constructor */
  CyclicDistribution (const CyclicDistribution& rhs) : 
    procGrid_(rhs.procGrid_)
  {}

public:

  virtual ~CyclicDistribution () {}

  int operator[] (const Tuple<DIM>& t) const 
  { 
    // compute the ordinal number of t in the processor grid
    Tuple<DIM> tmp = t % procGrid_;
    int idx = tmp[DIM-1];
    for (int i = DIM-2; i >= 0; --i) 
      idx += tmp[i] * procGrid_[i+1];
    return this->operator[](idx);
  }

  int operator[] (int idx) const 
  { 
    return idx % Traits::Default::nPlaces();
  }

  virtual int course (int idx) const
  {
    int nnodes = procGrid_[0];
    return idx / nnodes;
  }

  bool operator== (const Distribution<DIM>& rhs) const
  {
    bool ret = false;
    const CyclicDistribution<DIM>* tmp = static_cast<const CyclicDistribution<DIM>*>(&rhs);
    if (tmp != NULL) 
      ret = (tmp->procGrid_ == this->procGrid_);
    return ret;
  }
  
  virtual int nPlaces () const 
  {
    return procGrid_[0];
  }


 virtual Distribution<DIM>* clone() const 
  {
    return new CyclicDistribution (*this);
  }


private:
  
  /* cumulative processor grid */
  Tuple<DIM> procGrid_;

  template <int D> 
  friend ostream& operator << (ostream& ost, const CyclicDistribution<D>& d);

  void print_(ostream& ost, int indent = 0) const 
  {
    ost << "CyclicDistribution(procGrid=" << procGrid_  << ")" << endl;
  }
};

template<int DIM> 
ostream& operator << (ostream& ost, const CyclicDistribution<DIM>& d) {  
  d.print_(ost);
  return ost;
}

#endif /* __CYCLIC_DISTRIBUTION_H__ */
