/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: GenericDistribution.h,v 1.1 2007/02/14 16:09:58 bikshand Exp $
 * Authors: Christoph von Praun
 */


#ifndef __GENERIC_DISTRIBUTION_H__
#define __GENERIC_DISTRIBUTION_H__

using namespace std;

#include <cassert>
#include "AbstractArray.h"
#include "Distribution.h"
#include "Tuple.h"


template <int DIM>
class GenericDistribution : public Distribution<DIM>
{
public:

  GenericDistribution (const Shape<DIM>& shape) : shape_(shape)
  {
  }

  virtual ~GenericDistribution () {}

  /** 
   * @return the node correspnoding to an index 
   */
  virtual int operator[] (int idx) const 
  {     
    int proc;
    
    Tuple<DIM> coordGlobal (shape_.coord(idx) * shape_.step() + shape_.low()); 
    proc = VectorOps<Tuple<DIM>, DIM>::dotProduct (coordGlobal, shape_.linearStep());
    
    //cout << "Here " << proc << shape_.linearStep() << shape_.high() << shape_.low() << shape_.coord(idx) << coordGlobal << endl;
    return proc % Traits::Default::nPlaces();
  }
  
  virtual int operator[] (const Tuple<DIM>& t) const 
  { 
    return shape_.ord(t) % Traits::Default::nPlaces() ;
  }  
  
  virtual Distribution<DIM>* clone() const 
  {
    return new GenericDistribution<DIM> (this->shape_);
  }
  
  virtual Distribution<DIM>* subDist (const Shape<DIM>& shape) const
  {
    Shape<DIM> s (shape);
    s.relocate (shape.low() * shape_.step() + shape_.low());
    s.step() = shape.step() * shape_.step();
    s.linearStep() = shape_.linearStep();
      
    return new GenericDistribution<DIM> (s); 
   }

  /** 
   * @return the 'course' within the node, i.e. the block / tile number 
   * relative to the node.
   */
  virtual int course (int idx) const
  {
    int proc = (*this)[idx];
    int nPlaces = Traits::Default::nPlaces();
    return proc / nPlaces;
  }

  virtual bool operator== (const Distribution<DIM>& rhs) const
  {
    bool ret = false;
    const GenericDistribution<DIM>* tmp = dynamic_cast<const GenericDistribution<DIM>*>(&rhs);
    if (tmp != NULL) 
      ret = (tmp->shape_ == this->shape_);
    return ret;
  }

  virtual int nPlaces () const 
  {
    assert (false);
    return 0;
  }


  const Shape<DIM>& shape () const 
  {
    return shape_;
  }

private:

  const Shape<DIM> shape_;

  template <int DIM2> 
  friend ostream& operator << (ostream& ost, const GenericDistribution<DIM2>& d);

  void print_(ostream& ost, int indent = 0) const 
  {
    ost << "GenericDistribution()" << shape_ <<  endl;
  }

};

template<int DIM> 
ostream& operator << (ostream& ost, const GenericDistribution<DIM>& d) 
{  
  d.print_(ost);
  return ost;
}

#endif /* __GENERIC_DISTRIBUTION_H__ */
