/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: HTA.h,v 1.276.2.7 2007/08/23 17:04:21 fraguela Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef  __HTA_H__
#define  __HTA_H__

#include <cassert>

#include "BinExpr.h"
#include "Distribution.h"
#include "HelperIndx.h"
#include "HTACache.h"
#include "MemMapping.h"
#include "Operators.h"
#include "Shape.h"
#include "Tuple.h"
#include "Order.h"
#include "Router.h"
#include "debug/Tracing.h"
#include "util/RandUtils.h"
#include "maps/RandNumGenerator.h"
#include "maps/AccumArray.h"
#include "ReduceExpr.h"
#include "TransExpr.h"
#include "Overlap.h"
//#include "mpi/HTAODistImpl.h"
#include "serial/HTAOImpl.h"

#ifndef HTA_DEFAULT_TRAIT
#error "HTA_DEFAULT_TRAIT not defined!"
#endif

using namespace std;

/** 
 * Handle class to HTA - this class implements the public API of arrays in htalib.
 * operations are forwarded to the implementation classes.
 */
template<class T, int DIM, typename TRAIT = Traits::HTA_DEFAULT_TRAIT<T, DIM> >  
class HTA {
  
  public:
  
  /** 
   * Default constructor  should only be used for the creation of HTA []
   */
  HTA () : 
  wrapped_ (NULL) {}
  
  /**
   * @param a the wrapped array
   */
  HTA (typename TRAIT::TypeImpl& a) : 
    wrapped_(&a)
  { 
    wrapped_->refCount_(1);
  }

  /**
   * @param a the overlapped HTA
   *        ol the new overlapped size
   */
  HTA (HTA<T, DIM, TRAIT>& rhs, Overlap<DIM> * ol)
  {

    wrapped_ = TRAIT::TypeImpl::alloc(rhs.wrapped_, ol);
    wrapped_->refCount_(1);
  }


  HTA(Router<T,DIM,TRAIT> & rhs)
  {
    wrapped_ = rhs.value();
    wrapped_->refCount_(1);
  }
  /** 
   * Copy constructor - adjusts ref count of wrapped object 
   */
  HTA (const HTA<T,DIM, TRAIT>& rhs) : 
    wrapped_(rhs.wrapped_)
  {
    assert(wrapped_);
    wrapped_->refCount_(1);
  }



  virtual ~HTA() 
  {
    if (wrapped_ != NULL) {
      if (wrapped_->refCount_(-1) == 0) {
	if (! HTACache<T, DIM>::put (wrapped_)) {
	  delete wrapped_;
	}
      }
      
    }
  }
  
  /** 
   * "Simulates" the MATLAB equivalent of h{}*().
   * (chain of responsibility). User must ensure that sizeof(sv) == n.
   * This operator is subject to change -- for now replaced by the
   * 'Router-mechanism'
   * operator[][][] 
   */
  typename TRAIT::Type tileAt(int n, const Shape<DIM>* sv) const 
  {
    assert (wrapped_);    
    HTA_DEPRECATED("HTA::tileAt(n, const Shape*)");
    typename TRAIT::TypeImpl* ws = wrapped_->tileAt(n, sv);
    //JG: add hooks to the OL-HTA if needed.
    ws->addParent(wrapped_, n, sv);
    return typename TRAIT::Type(*ws);
  }

  /**
   * tileAt operator
   */
  typename TRAIT::TypeTile operator() (const Tuple<DIM>& t)  const
  {
    assert (wrapped_);
    assert (this->level() > 0);
    typename TRAIT::TypeTileImpl* ws = wrapped_->tileAt(t);
    assert (ws != NULL);
    //JG: add hooks to the OL-HTA if needed.
    ws->addParent(wrapped_, t);
    return typename TRAIT::TypeTile (*ws);
  }

  //Simple syntax
  typename TRAIT::TypeTile operator() (int i) const
  {
    return this->operator()(Tuple<DIM>(i));
  }
  
  typename TRAIT::TypeTile operator() (int i, int j) const
  {
    return this->operator()(Tuple<DIM>(i, j));
  }

  typename TRAIT::TypeTile operator() (int i, int j, int k) const
  {
    return this->operator()(Tuple<DIM>(i, j, k));
  }

  typename TRAIT::TypeTile operator() (int i, int j, int k, int l) const
  {
    return this->operator()(Tuple<DIM>(i, j, k, l));
  }


  typename TRAIT::TypeTile operator() (const bool* indices)  const
  {
    assert (wrapped_);
    assert (this->level() > 0);
    typename TRAIT::TypeTileImpl* ws = wrapped_->tileAt(indices);
    assert (ws != NULL);
    //JG: add hooks to the OL-HTA if needed.
    ws->addParent(wrapped_);
    return typename TRAIT::TypeTile (*ws);
  }
  
  void initDistribution (const Distribution<DIM>& dist)
  {
    wrapped_->initDistribution (dist);
    wrapped_->initLocalTiles();
  }

  void initTile (const Tuple<DIM>& t, typename TRAIT::TypeTile& tile)
  {
    assert (wrapped_);
    assert (tile.wrapped_);
    assert (this->level() > 0);
    wrapped_->initTile(t, tile.wrapped_);
  }

  void initRemoteTile (const Tuple<DIM>& t, typename TRAIT::TypeTile& tile)
  {
    assert (wrapped_);
    assert (tile.wrapped_);
    assert (this->level() > 0);
    wrapped_->initRemoteTile(t, tile.wrapped_);
  }
    
  Router<T, DIM, TRAIT> operator() () const
  {
    
    typename Triplet::Seq ts;
    
    for (int i = 0; i < DIM; ++i) 
      ts[i] = this->shape()[i];
    
    return Router<T, DIM, TRAIT>(*wrapped_, ts);
  }
 
  Router<T, DIM, TRAIT> operator() (const typename Triplet::Seq& s) const 
  {
    assert (wrapped_);
    assert (this->level() > 0);
    return Router<T, DIM, TRAIT>(*wrapped_, s);
  }

  //Simple syntax
  Router<T, DIM, TRAIT> operator() (const Triplet &t) const
  {
    assert (wrapped_);
    assert (this->level() > 0);
    return Router<T, DIM, TRAIT>(*wrapped_, typename Triplet::Seq(t));
  }

  Router<T, DIM, TRAIT> operator() (const Triplet &t1, const Triplet &t2) const
  {
    assert (wrapped_);
    assert (this->level() > 0);
    return Router<T, DIM, TRAIT>(*wrapped_, typename Triplet::Seq(t1, t2));
  }

  Router<T, DIM, TRAIT> operator() (const Triplet &t1, const Triplet &t2, const Triplet &t3) const
  {
    assert (wrapped_);
    assert (this->level() > 0);
    return Router<T, DIM, TRAIT>(*wrapped_, typename Triplet::Seq(t1, t2, t3));
  }

  Router<T, DIM, TRAIT> operator() (const Triplet &t1, const Triplet &t2, const Triplet &t3, const Triplet &t4) const
  {
    assert (wrapped_);
    assert (this->level() > 0);
    return Router<T, DIM, TRAIT>(*wrapped_, typename Triplet::Seq(t1, t2, t3, t4));
  }

  /** @brief scalarAt operator 
   *
   *  @bug breaks on leaf HTADistImpl
   */
  inline T& operator[] (const Tuple<DIM>& t) const 
    //   __attribute__((always_inline))
    //__attribute__((const))    
  {
    assert (wrapped_);    

    /** BBF: The code below is not correct because if this HTA is for example
      *      a HTADistImpl leaf, HTAImpl<T, DIM, TRAIT>::scalarAt_leaf_(t)
      *      wil throw !this->isAvailable_() in all the processors but the home.
      *      If an HTADistImpl::scalarAt_leaf_ that broadcasts the scalar to
      *      all the processors is written the code for this function could be
      *
      * return (wrapped_->actual_().level_ == 0) ? wrapped_->scalarAt_leaf_(t) : wrapped_->scalarAt(t);
      *
      * But the problem then is that scalarAt_leaf_ is invoked on leaf HTADistImpl
      * in situations in which no broadcast is required, thus leading to stall,
      * so the bug is still here : <<One cannot use h[t] on a leaf HTADistImpl h>>
      */
    
       //GBDEBUG: I find this to be faster than having
       // the same branch inside scalarAt (for reasons unknown).
      
       //if(wrapped_->actual_().is_leaf_)
       if(wrapped_->actual_().level_ == 0)
         return (static_cast<const HTAImpl<T, DIM, TRAIT>*>(wrapped_))->scalarAt_leaf_(t);
       else 
         return wrapped_->scalarAt (t);
  }

  /**
   * scalarAt operator 
   * GB: To avoid calls to scalarAt 
   * This function makes sense as it should be possible
   * to index any n-dimensional array using an integer
   * (MATLAB allows this)
   */
   
  /** BBF: This syntax is VERBOTEN with the new one
      i.e: this method is incompatible with having the
      syntax A[i][j] instead of A[Tuple<2>(i,j)]
  inline T& operator[] (const int& i) const
  {
    //assert (DIM==1);
    assert (this->level() == 0);
    assert (wrapped_);
    return wrapped_->scalarAt (i);
  }
 */

  typename HelperIndx<T, DIM, TRAIT, 1>::RType operator[] (int i) {
    assert (wrapped_);
    return HelperIndx<T, DIM, TRAIT, 1>(*this, i);
  }

  typename HelperIndxShape<T, DIM, TRAIT, 1>::RType operator[] (const Triplet &t) {
    assert (wrapped_);
    return HelperIndxShape<T, DIM, TRAIT, 1>(*this, t);
  }

  typename TRAIT::Type operator[] (const Shape<DIM>& s) const
  {
    assert (wrapped_);

    if( !this->level() )
      return typename TRAIT::Type (*wrapped_->scalarAt(s));
    else {
      /* BBF TODO: This can be optimized A LOT modifying
         ShapeH interface so that a copy of the shapes is not required in
	 advance of the intersection. Also notice that the result of the
	 intersection needs to be assigned to a ShapeH with copy-in-depth
       */
      ShapeH<DIM> hierarch_shape(this->shape());
      const int ncard = this->shape().card();
      for(int i = 0; i < ncard; ++i)
	hierarch_shape.shapeAt(i) = wrapped_->tileAt(i)->shape();
      hierarch_shape = hierarch_shape.intersect(s); //deep copy of tmp
      return ((Router<T, DIM, TRAIT>(*wrapped_, hierarch_shape.myshape()))(hierarch_shape));
    }
  }

  /** pointwise from binary expression */
  template <typename OP>
  typename TRAIT::Type& operator= (const BinExpr<OP, TRAIT>& expr) 
  {
    assert(wrapped_);
    expr.eval(*this);
    return *this;
  }
 
  /** result of TransExpr */
  template<int TYPE>
  typename TRAIT::Type& operator= (const TransExpr<TRAIT, TYPE>& expr) 
  {
    assert(wrapped_);
    expr.eval(*this);
    return *this;
  }

  /** result of ReduceExpr */
  template <typename OP>
  typename TRAIT::Type& operator= (const ReduceExpr<OP, TRAIT>& expr) 
  {
    assert(wrapped_);
    expr.eval(*this);
    return *this;
  }
 
  /** 
   * Assignment operator - this is not subsumed by the template version! 
   */
  typename TRAIT::Type& operator= (const HTA<T, DIM, TRAIT>& rhs) {
    assert (rhs.wrapped_);
    if (this != &rhs) {
      if (wrapped_ == NULL) { 
        // initialization
        wrapped_ = rhs.wrapped_; 
        wrapped_->refCount_(1);
      } else  {
	assert (Shape<DIM>::conformable(this->wrapped_->shape(), rhs.wrapped_->shape()));

	// communicate 
	rhs.wrapped_->actual_().communicate_(this->wrapped_->dist(), this->wrapped_);
	
	// pointwise copy 
        // *wrapped_ = *rhs.wrapped_;
	
	// invalidate
	// rhs.wrapped_->invalidate_();
      }
    }    
    return *this;
  }

  /** 
   * Intialization or pointwise assignment 
   */
  /*
  // used by the leaf; not valid anymore; add if in the default operator=
  typename TRAIT::Type& operator= (const HTA<T, DIM, TRAIT>& rhs) 
  { 
    assert (this->level() != rhs.level());
    assert (rhs.wrapped_);
    assert (0); // not yet implemented
    return *this;
  }
  */

  /** pointwise assignment */
  typename TRAIT::Type& operator= (const T& rhs) 
  {
    assert (wrapped_);
    *wrapped_ = rhs;
    return *this;
  }


  // dotproduct -----------------------------------------------------------------

  static typename TRAIT::TypeScalar dotproduct (const typename TRAIT::Type& larg, 
						const typename TRAIT::Type& rarg)
  {
    assert (larg.shape() == rarg.shape());
    typename TRAIT::TypeScalar ret = TRAIT::TypeImpl::dotproduct (*larg.wrapped_, 
								  *rarg.wrapped_);
    //    TRAIT::barrier();
    return ret;
  }

  static typename TRAIT::Type ldotproduct (const typename TRAIT::Type& larg, 
						const typename TRAIT::Type& rarg)
  {
    assert (TRAIT::dim == 2);

    Tuple<2>::Seq tiling;
    tiling[0] = Tuple<2>(1,1);
    tiling[1] = larg.wrapped_->tiling()[1];
   
    typename TRAIT::TypeImpl* dst = TRAIT::TypeImpl::alloc (larg.level(), tiling, larg.wrapped_->memMap().order());

    assert (larg.shape() == rarg.shape());
    TRAIT::TypeImpl::ldotproduct (*larg.wrapped_, 
				  *rarg.wrapped_,
				  *dst);
    //    TRAIT::barrier();
    return typename TRAIT::Type (*dst); 
  }
  
  // matmul ----------------------------------------------------------------------

  static  typename TRAIT::Type matmul (const typename TRAIT::Type& larg, 
				       const typename TRAIT::Type& rarg, 
				       typename TRAIT::Type* res = NULL) 
  {
    assert (DIM == 2);
    assert (larg.shape().size()[1] == rarg.shape().size()[0]);
    
    // allocate the result array if necessary
    typename TRAIT::TypeImpl* dst;
    if (res == NULL) {
      typename Tuple<DIM>::Seq res_tiling;
      const typename Tuple<DIM>::Seq& larg_tiling = larg.wrapped_->tiling();
      const typename Tuple<DIM>::Seq& rarg_tiling = rarg.wrapped_->tiling();
      for (size_t i = 0; i <= larg.level(); ++i) {
	res_tiling[i] = Tuple<DIM>(larg_tiling[i][0], rarg_tiling[i][1]);
      }
      Order order = larg.wrapped_->memMap().order();
      dst = TRAIT::TypeImpl::alloc(larg.level(), res_tiling, order);
    } else {
      dst = res->wrapped_;
      assert (dst->shape().size()[0] == larg.shape().size()[0] && 
	      dst->shape().size()[1] == rarg.shape().size()[1]);
    } 

    TRAIT::TypeImpl::matmul(*larg.wrapped_, *rarg.wrapped_, *dst);
    //    TRAIT::barrier();
    return typename TRAIT::Type (*dst); 
  }
  
  
  // leaf-convolution -----------------------------------------------------------------

  /**
   * @param weights   array of szie DIM + 1 that specifes the coefficients
   *                  for elements of distance i (0... DIM+1) in the grid
   * @param dst       the destination array
   * @return          teh destination or a temporary HTA.
   */
  HTA<T, DIM, TRAIT> lconv (const T* weights,
			    HTA<T, DIM, TRAIT>* dst = NULL) const
  {
    assert (weights != NULL);
    assert (this->wrapped_);
    HTA<T, DIM, TRAIT> result;

    // limitation of the current implementation
    if (dst == NULL) {
      result = clone(false);
      dst = &result;
    } else 
      assert (dst->wrapped_);
    
    wrapped_->lconv(weights, *dst->wrapped_);
    //TRAIT::barrier();
    return *dst;
  }

  /** 
   *  overlapped shadow region update (manually)
   */
  void update(bool * idx=NULL) const{
    assert(this->wrapped_);
    wrapped_->update(idx);
  }
  
  void setUpdate(bool update) const{
    wrapped_->setUpdate(update);
  }

  void sendUpdate(bool *idx=NULL) const {
    assert(this->wrapped_);
    wrapped_->sendUpdate(idx);
  }

  void recvUpdate(bool *idx =NULL) const {
    assert(this->wrapped_);
    wrapped_->recvUpdate(idx);
  }
  /**
   * overlapped shadow region set clean 
   */
  void setClean(int val =0) const 
  {
    assert(wrapped_);
    wrapped_->setClean(val);
  }

  // leaf-matmul -------------------------------------------------------------------

  static typename TRAIT::TraitDense::Type lmatmul (const typename TRAIT::Trait::Type& larg,
						   const typename TRAIT::TraitDense::Type& rarg,
						   typename TRAIT::TraitDense::Type* res = NULL) 
  {        
    // allocate the result array if necessary
    typename TRAIT::TraitDense::TypeImpl* dst;
    if (res == NULL) {
      const typename Tuple<DIM>::Seq& larg_tiling = larg.wrapped_->tiling();
      const typename Tuple<DIM>::Seq& rarg_tiling = rarg.wrapped_->tiling();
      typename Tuple<DIM>::Seq res_tiling = rarg.wrapped_->tiling();
     
      /* only the leaf tiling changes */
      Tuple<DIM> t = larg_tiling[0];
      t[0] = larg_tiling[0][0];
      t[1] = rarg_tiling[0][1];
      res_tiling[0] = t;

      Order order = larg.wrapped_->memMap().order();
      /* GBDEBUG: Another problem due to order */
      dst = TRAIT::TraitDense::TypeImpl::alloc(rarg.level(), res_tiling, TILE, &larg.dist());
    
    } else {
      dst = res->wrapped_;
      assert (dst->shape().size()[0] == larg.shape().size()[0] && 
	      dst->shape().size()[1] == rarg.shape().size()[1]);
    } 
    
    TRAIT::TypeImpl::lmatmul (*larg.wrapped_, *rarg.wrapped_, *dst);
    //    TRAIT::barrier();
    return typename TRAIT::TraitDense::Type (*dst); 
  }

  // operator+ --------------------------------------------------------------------

  template<typename T2, int DIM2, typename TRAIT2> 
  friend HTA<T2, DIM2, TRAIT2> operator+ (const T2& val, 
					  const HTA<T2, DIM2, TRAIT2>& arr);

  BinExpr<plus<T>, TRAIT> operator+ (const T& rhs) const
  {
    return BinExpr<plus<T>, TRAIT>(*this, rhs);
  }

  BinExpr<plus<T>, TRAIT> operator+ (const HTA<T, DIM, TRAIT>& rhs) const
  {
    return BinExpr<plus<T>, TRAIT>(*this, rhs);
  }

  template<typename OP>
  BinExpr<plus<T>, TRAIT> operator+ (const BinExpr<OP, TRAIT>& rhs) const
  {
    HTA<T, DIM, TRAIT> rhs_eval = rhs.eval();
    // re-use the temporary that comes with the bin-expr arg
    return BinExpr<plus<T>, TRAIT>(*this, rhs_eval, rhs_eval.wrapped_);
  }

  template <typename T2>
  HTA<T, DIM, TRAIT> operator+ (const T2& rhs) const
  {
    // allocate the result array
    HTA<T, DIM, TRAIT> result = clone();
    TRAIT::TypeImpl::map(binder2nd<plus<T2> >(plus<T2>(), rhs), *wrapped_, *result.wrapped_);
    return result;
  }

  template <typename T2, typename TRAIT2>
  HTA<T, DIM, TRAIT> operator+ (const HTA<T2, DIM, TRAIT2>& rhs) const
  {
    // allocate the result array
    assert (rhs.level() <= this->level());                    // limitation of the current implementation
    HTA<T, DIM, TRAIT> result = clone();  // allocate the result array
    TRAIT::TypeImpl::map(plus<T>(), *wrapped_, *rhs.wrapped_, *result.wrapped_);
    return result;
  }

  /** pointwise from binary expression */
  template <typename OP>
  typename TRAIT::Type& operator+= (const BinExpr<OP, TRAIT>& expr) 
  {
    assert(wrapped_);
    expr.evalAndAccum(*this, plus<T>());
    return *this;
  }
 
  // this non-template version is necessary to help the compiler to 
  // auto-convert BinExpr and Router instances to the rhs HTA argument.
  HTA<T, DIM, TRAIT>& operator+= (const HTA<T, DIM, TRAIT>& rhs)
  {
    assert (wrapped_);
    TRAIT::TypeImpl::map (plus<T>(), *wrapped_, *rhs.wrapped_, *wrapped_);
    return *this;
  }

  template <typename TRAIT2>
  HTA<T, DIM, TRAIT>& operator+= (const HTA<T, DIM, TRAIT2>& rhs)
  {
    assert (wrapped_);
    assert (this->level() >= rhs.level());
    TRAIT::TypeImpl::map(plus<T>(), *wrapped_, *rhs.wrapped_, *wrapped_);
    return *this;
  }    

  HTA<T, DIM, TRAIT>& operator+= (const T& rhs)
  {
    assert(wrapped_);
    TRAIT::TypeImpl::map(binder2nd<plus<T> >(plus<T>(), rhs), *wrapped_, *wrapped_);
    return *this;
  }

  // operator- --------------------------------------------------------------------

  template<typename T2, int DIM2, typename TRAIT2> 
  friend HTA<T2, DIM2, TRAIT2> operator- (const T2& val, 
					  const HTA<T2, DIM2, TRAIT2>& arr);

  BinExpr<minus<T>, TRAIT> operator- (const T& rhs) const
  {
    return BinExpr<minus<T>, TRAIT>(*this, rhs);
  }

  BinExpr<minus<T>, TRAIT> operator- (const HTA<T, DIM, TRAIT>& rhs) const
  {
    return BinExpr<minus<T>, TRAIT>(*this, rhs);
  }

  template<typename OP>
  BinExpr<minus<T>, TRAIT> operator- (const BinExpr<OP, TRAIT>& rhs) const
  {
    HTA<T, DIM, TRAIT> rhs_eval = rhs.eval();
    // re-use the temporary that comes with the bin-expr arg
    return BinExpr<minus<T>, TRAIT>(*this, rhs_eval, rhs_eval.wrapped_);
  }

  HTA<T, DIM, TRAIT> operator- () const
  {
    // allocate the result array
    //    HTA<T, DIM, TRAIT> result = clone();
    TRAIT::TypeImpl::map (negate<T>(), *wrapped_, *wrapped_);
    return *this;
  }
  
  template <typename T2>
  HTA<T, DIM, TRAIT> operator- (const T2& rhs) const
  {
    // allocate the result array
    HTA<T, DIM, TRAIT> result = clone();
    TRAIT::TypeImpl::map (binder2nd<minus<T2> >(minus<T2>(), rhs), *wrapped_, *result.wrapped_);
    return result;
  }
  
  template <typename T2, typename TRAIT2>
  HTA<T, DIM, TRAIT> operator- (const HTA<T2, DIM, TRAIT2>& rhs) const
  {
    // allocate the result array
    assert (rhs.level() <= this->level());                    // limitation of the current implementation
    HTA<T, DIM, TRAIT> result = clone(); // allocate the result array
    TRAIT::TypeImpl::map (minus<T>(), *wrapped_, *rhs.wrapped_, *result.wrapped_);
    return result;
  }

  // this non-template version is necessary to help the compiler to 
  // auto-convert BinExpr and Router instances to the rhs HTA argument.
  HTA<T, DIM, TRAIT>& operator-= (const HTA<T, DIM, TRAIT>& rhs)
  {
    assert (wrapped_);
    TRAIT::TypeImpl::map (minus<T>(), *wrapped_, *rhs.wrapped_, *wrapped_);
    return *this;
  }

  template <typename TRAIT2>
  HTA<T, DIM, TRAIT>& operator-= (const HTA<T, DIM, TRAIT2>& rhs)
  {
    assert (wrapped_);
    assert (this->level() >= rhs.level());
    TRAIT::TypeImpl::map (minus<T>(), *wrapped_, *rhs.wrapped_, *wrapped_);
    return *this;
  }    

  HTA<T, DIM, TRAIT>& operator-= (const T& rhs)
  {
    assert (wrapped_);
    TRAIT::TypeImpl::map (binder2nd<minus<T> >(minus<T>(), rhs), *wrapped_, *wrapped_);
    return *this;
  }

  // operator* --------------------------------------------------------------------

  template<typename T2, int DIM2, typename TRAIT2> 
  friend HTA<T2, DIM2, TRAIT2> operator* (const T2& val, 
					  const HTA<T2, DIM2, TRAIT2>& arr);

  BinExpr<multiplies<T>, TRAIT> operator* (const T& rhs) const
  {
    return BinExpr<multiplies<T>, TRAIT>(*this, rhs);
  }

  BinExpr<multiplies<T>, TRAIT> operator* (const HTA<T, DIM, TRAIT>& rhs) const
  {
    return BinExpr<multiplies<T>, TRAIT>(*this, rhs);
  }

  template<typename OP>
  BinExpr<plus<T>, TRAIT> operator* (const BinExpr<OP, TRAIT>& rhs) const
  {
    HTA<T, DIM, TRAIT> rhs_eval = rhs.eval();
    // re-use the temporary that comes with the bin-expr arg
    return BinExpr<multiplies<T>, TRAIT>(*this, rhs_eval, rhs_eval.wrapped_);
  }

  template <typename T2>
  HTA<T2, DIM, TRAIT> operator* (const T2& rhs) const
  {
    // allocate the result array
    HTA<T2, DIM, TRAIT> result = clone();
    TRAIT::TypeImpl::map (binder2nd<multiplies<T2> >(multiplies<T2>(), rhs), *wrapped_, *result.wrapped_);
    return result;
  }
  
  template <typename T2, typename TRAIT2>
  HTA<T, DIM, TRAIT> operator* (const HTA<T2, DIM, TRAIT2>& rhs) const
  {
    // allocate the result array
    assert (rhs.level() <= this->level());                    // limitation of the current implementation
    HTA<T, DIM, TRAIT> result = clone();  // allocate the result array
    
    TRAIT::TypeImpl::map(multiplies<T>(), *wrapped_, *rhs.wrapped_, *result.wrapped_);
    
    return result;
  }

  // this non-template version is necessary to help the compiler to 
  // auto-convert BinExpr and Router instances to the rhs HTA argument.
  HTA<T, DIM, TRAIT>& operator*= (const HTA<T, DIM, TRAIT>& rhs)
  {
    assert (wrapped_);
    TRAIT::TypeImpl::map (multiplies<T>(), *wrapped_, *rhs.wrapped_, *wrapped_);
    return *this;
  }

  template <typename T2, typename TRAIT2>
  HTA<T, DIM, TRAIT>& operator*= (const HTA<T2, DIM, TRAIT2>& rhs)
  {
    assert (wrapped_);
    TRAIT::TypeImpl::map (multiplies<T>(), *wrapped_, *rhs.wrapped_, *wrapped_);
    return *this;
  }

  template <typename TRAIT2>
  HTA<T, DIM,TRAIT>& operator*= (const HTA<T, DIM, TRAIT2>& rhs)
  {
    assert (wrapped_);
    assert (this->level() >= rhs.level());
    TRAIT::TypeImpl::map (multiplies<T>(), *wrapped_, *rhs.wrapped_, *wrapped_);
    return *this;
  }    

  HTA<T, DIM,  TRAIT>& operator*= (const T& rhs)
  {
    assert (wrapped_);
    TRAIT::TypeImpl::map (binder2nd<multiplies<T> >(multiplies<T>(), rhs), *wrapped_, *wrapped_);
    return *this;
  }

  // operator/ --------------------------------------------------------------------

  template<typename T2, int DIM2, typename TRAIT2> 
  friend HTA<T2, DIM2, TRAIT2> operator/ (const T2& val, 
					  const HTA<T2, DIM2, TRAIT2>& arr);

  BinExpr<divides<T>, TRAIT> operator/ (const T& rhs) const
  {
    return BinExpr<divides<T>, TRAIT>(*this, rhs);
  }

  BinExpr<divides<T>, TRAIT> operator/ (const HTA<T, DIM, TRAIT>& rhs) const
  {
    return BinExpr<divides<T>, TRAIT>(*this, rhs);
  }

  template<typename OP>
  BinExpr<plus<T>, TRAIT> operator/ (const BinExpr<OP, TRAIT>& rhs) const
  {
    HTA<T, DIM, TRAIT> rhs_eval = rhs.eval();
    // re-use the temporary that comes with the bin-expr arg
    return BinExpr<divides<T>, TRAIT>(*this, rhs_eval, rhs_eval.wrapped_);
  }

  template <typename T2>
  HTA<T, DIM, TRAIT> operator/ (const T2& rhs) const
  {
    assert (rhs != 0);
    // allocate the result array
    HTA<T, DIM, TRAIT> result = clone();
    TRAIT::TypeImpl::map (binder2nd<divides<T2> >(divides<T2>(), rhs), *wrapped_, *result.wrapped_);
    return result;
  }
  
  template <typename T2, typename TRAIT2>
  HTA<T, DIM, TRAIT> operator/ (const HTA<T2, DIM, TRAIT2>& rhs) const
  {
    // allocate the result array
    assert (rhs.level() <= this->level());                    // limitation of the current implementation
    HTA<T, DIM, TRAIT> result = clone();  // allocate the result array
    TRAIT::TypeImpl::map (divides<T>(), *wrapped_, *rhs.wrapped_, *result.wrapped_);
    return result;
  }

  // this non-template version is necessary to help the compiler to 
  // auto-convert BinExpr and Router instances to the rhs HTA argument.
  HTA<T, DIM, TRAIT>& operator/= (const HTA<T, DIM, TRAIT>& rhs)
  {
    assert (wrapped_);
    TRAIT::TypeImpl::map (divides<T>(), *wrapped_, *rhs.wrapped_, *wrapped_);
    return *this;
  }

  template <typename TRAIT2>
  HTA<T, DIM, TRAIT>& operator/= (const HTA<T, DIM, TRAIT2>& rhs)
  {
    assert (wrapped_);
    assert (this->level() >= rhs.level());
    TRAIT::TypeImpl::map (divides<T>(), *wrapped_, *rhs.wrapped_, *wrapped_);
    return *this;
  }    

  HTA<T, DIM, TRAIT>& operator/= (const T& rhs)
  {
    assert (wrapped_);
    TRAIT::TypeImpl::map (binder2nd<divides<T> >(divides<T>(), rhs), *wrapped_, *wrapped_);
    return *this;
  }

  // operator== --------------------------------------------------------------------
  
  bool operator== (const HTA<T, DIM, TRAIT>& rhs) const
  {
    bool ret;
    if (this == &rhs || rhs.wrapped_ == wrapped_) {
      ret = true;
    } else {
      assert (rhs.wrapped_ && wrapped_);
      ret = (*wrapped_ == *rhs.wrapped_);
    }
    return ret;
  }
  
  bool operator!= (const HTA<T, DIM, TRAIT>& rhs) const
  {
    return ! (this->operator==(rhs)); 
  }

  /**
   * pointwise application of sqrt operator.
   */
  HTA<T, DIM, TRAIT> sqrt (HTA<T, DIM, TRAIT>* dst = NULL) const 
  {
    assert (wrapped_);
    HTA<T, DIM, TRAIT> result;
    if (dst == NULL) {
      result = clone();
      dst = &result;
    } else 
      assert (dst->wrapped_);
    
    TRAIT::TypeImpl::map (Operator::sqrt<T>(), *wrapped_, *dst->wrapped_);
    return *dst;
  }
  
#if 0
  /**
   * pointwise application of sqrt operator.
   */
  HTA<T, DIM, TRAIT> pow2 (HTA<T, DIM, TRAIT>* dst = NULL) const 
  {
    assert (wrapped_);
    HTA<T, DIM, TRAIT> result;
    if (dst == NULL) {
      result = clone();
      dst = &result;
    } else 
      assert (dst->wrapped_);
    
    TRAIT::TypeImpl::map (Operator::pow2<T>(), *wrapped_, *dst->wrapped_);
    return *dst;
  }
#endif
    
  /**
   * pointwise application of floor operator.
   */
  template <typename TRAIT2>
  HTA<int, DIM, TRAIT2> floor (HTA<int, DIM, TRAIT2>* dst = NULL) const 
  {

    assert (wrapped_);
    HTA<int, DIM, TRAIT2> result;

    assert (dst); //current limitation - i am unable to use the original TRAIT 
    if (dst == NULL) {
      //result = HTA<int, DIM, TRAIT>::alloc(wrapped_->tiling(), wrapped_->order_());
      //dst = &result;
      assert (false);
    } else 
      assert (dst->wrapped_);
    
    TRAIT::TypeImpl::map (Operator::floor<T>(), *wrapped_, *dst->wrapped_);
    return *dst;
  }
  
   // collectives 
  
  /** 
   * all-to-all reduction over all dimensions in the array
   */
  template <typename OP> 
  T reduce (const OP& binary_op) const
  {
    assert (wrapped_);
    return wrapped_->reduce(binary_op);
  }

  
  /* HTA scan operation
   * oplevel decide *AT* which level the scan is applied
   * For a 1-LEVEL HTA,
   * scan(1) => scan of the top level tiles
   * scan(0) => scan inside each of the level 0 tiles.
   */
  HTA<T, DIM, TRAIT> scan(int oplevel,
			  HTA<T, DIM, TRAIT>* dst = NULL) const
  {
    typename TRAIT::TypeImpl* dst_actual = NULL;
    
    if (dst)
      dst_actual = dst->wrapped_;
    else	
      {
	HTA<T, DIM, TRAIT> result = clone();
	dst_actual = result.wrapped_;
      }
    
    wrapped_->scan(oplevel, *dst_actual);
    return HTA<T, DIM, TRAIT> (*dst_actual);
  }

  //===TILE BY TILE operations =====================

 // fourier transform -----------------------------------------------------------------
  void ft (int dim, int dir) const
  {
    wrapped_->ft(dim, dir);
  }
  
  void sort (int level = 0) 
  {
    wrapped_->HTAsort (level);    
  }
  
 
  void accumArray (HTA<T, DIM, TRAIT> indices, int rshift = 0, int lshift = 0, int level = 0)
  {
    wrapped_->accumArray (indices.wrapped_, rshift, lshift, level);
  }
  
  /* ================ DEFUNCT ======================== USE accumArray instead ========*/
  /* GBDEBUG: I keeping it for few more days */
  template <typename TRAIT2>
  void accumarray(HTA<int, DIM, TRAIT2> countHTA) const 
  {
    assert (false);
    AccumArray<T, DIM, TRAIT2> counter(countHTA);
    this->mapReduce(counter);   
  }
  
  /**
   * redcution over one dimension which is specified by argument dim.
   * communication occurs if reduction is all_to_all and the array is 
   * distributed.
   *
   * @param all_to_all_mask mask that specifies at which level to 
   *                        reduce in all-to-all manner
   */
  /*
  template <typename OP> 
  HTA<T, DIM, TRAIT> reduce (const OP& binary_op, 
				    int dim, 
				    unsigned all_to_all_mask = 0x0,
				    HTA<T, DIM, TRAIT>* dst = NULL) const
  {
    assert (wrapped_);
    assert (dst == NULL); // limitation of the current implementation, result is allocated here
    assert (dim < DIM);
    
   
    // (1) allocate the result and initialize it 
    typename TRAIT::TypeImpl* dst_actual;
    // (1a) compute tiling according to all_to_all_mask
    typename Tuple<DIM>::Seq tiling_red = wrapped_->tiling();
    for (size_t i = 0; i <= this->level(); ++i) {
      if (!(all_to_all_mask & (1 << i)))
	// no all-to-all along this tiling level
	tiling_red[i] = tiling_red[i].mask(dim, 1);
    }
    
    Order order = wrapped_->memMap().order();
    dst_actual = TRAIT::TypeImpl::alloc (this->level(), tiling_red, order);
    
    T bottom = TypeHelper<T>::identityVal(binary_op);
    *dst_actual = bottom;
    
    // (2) then compute the reduction
    wrapped_->actual_().reduce_ (binary_op, dim, *dst_actual);
    
    return HTA<T, DIM, TRAIT>(*dst_actual);
  } */

  template <typename OP> 
  ReduceExpr<OP, TRAIT> reduce (const OP& binary_op, 
			     int dim, 
			     unsigned all_to_all_mask = 0x0,
			     HTA<T, DIM, TRAIT>* dst = NULL) const
  {
    assert (wrapped_);
    assert (dst == NULL); // limitation of the current implementation, result is allocated here
    assert (dim < DIM);
    return ReduceExpr<OP, TRAIT> (binary_op, *this, dim, all_to_all_mask);
  }
   

  /* *oplevel* decides *FROM* which level the reduction should start
   * levels *BELOW* oplevel will do no reduction, but just assignment
   * The meaning of OPLEVEL is quite different from that of SCAN.
   * We should soon adopt a consistent description.
   */
  template <typename OP> 
  HTA<T, DIM, TRAIT> reduce (int oplevel,
			     const OP& binary_op, 
			     int dim, 
			     unsigned all_to_all_mask = 0x0,
			     HTA<T, DIM, TRAIT>* dst = NULL) const
  {
    assert (wrapped_);
    typename TRAIT::TypeImpl* dst_actual;
    if (dst)
      dst_actual = dst->wrapped_;
    else {
      // (1) allocate the result and initialize it 
      // (1a) compute tiling according to all_to_all_mask
      typename Tuple<DIM>::Seq tiling_red = wrapped_->tiling();
      for (int i = oplevel; i <= this->level(); ++i) {
	if (!(all_to_all_mask & (1 << i)))
	  // no all-to-all along this tiling level
	  tiling_red[i] = tiling_red[i].mask(dim, 1);
      }

      Order order = wrapped_->memMap().order();
      dst_actual = TRAIT::TypeImpl::alloc(this->level(), tiling_red, order);
    }

    // (2) then compute the reduction
    wrapped_->actual_().reduce_ (oplevel, binary_op, dim, *dst_actual);
    
    return HTA<T, DIM, TRAIT>(*dst_actual);
  }


  /**
   * map (one argument)
   * I have to rename hmap to avoid name collision with another 'map' 
   */

  template <typename OP>
  void hmap (OP op, int level = 0, bool* mask = NULL)
  {
    assert (wrapped_);
    wrapped_->hmap(op, level, mask);
  }

  template <typename OP, class T1, int DIM1, typename TRAIT1>
  void hmap (OP op, HTA<T1, DIM1, TRAIT1> arg1, int level = 0, bool* mask = NULL)
  {
    assert (wrapped_);
    //    cout << "Hello " << endl;
    wrapped_->hmap(op, level,arg1.wrapped_, mask);
  }

  /**
   * map (two argument)
   */
  template <typename OP, class T1, int DIM1, typename TRAIT1, class T2, int DIM2, typename TRAIT2>
  void hmap (OP op, HTA<T1, DIM1, TRAIT1> arg1, 
	    HTA<T2, DIM2, TRAIT2> arg2, int level = 0, bool* mask = NULL)
  {
    assert (wrapped_);
    wrapped_->hmap(op, level,arg1.wrapped_, arg2.wrapped_, mask);
  }

  /**
   * map (three argument)
   */
  template <typename OP, class T1, int DIM1, typename TRAIT1, class T2, int DIM2, typename TRAIT2, class T3, int DIM3, typename TRAIT3>
  void hmap (OP op, HTA<T1, DIM1, TRAIT1> arg1, 
	    HTA<T2, DIM2, TRAIT2> arg2, 
	    HTA<T3, DIM3, TRAIT3> arg3, 
	    int level = 0,
	    bool* mask = NULL) 
  {
    assert (wrapped_);
    wrapped_->hmap(op, level, arg1.wrapped_, arg2.wrapped_, arg3.wrapped_, mask);
  }
  
  /**
   * map (four argument)
   */
  template <typename OP>
  void hmap (OP op, HTA<T, DIM, TRAIT> arg1, 
	     HTA<T, DIM, TRAIT> arg2, 
	     HTA<T, DIM, TRAIT> arg3, 
	     HTA<T, DIM,TRAIT> arg4,
	     int level = 0,
	     bool* mask = NULL) 
  {
    assert (wrapped_);
    wrapped_->hmap(op, level, arg1.wrapped_, arg2.wrapped_, arg3.wrapped_, arg4.wrapped_, mask);
  }
  
  /**
   * map (five argument)
   */
  template <typename OP>
  void hmap (OP op, HTA<T, DIM, TRAIT> arg1, 
	     HTA<T, DIM, TRAIT> arg2, 
	     HTA<T, DIM, TRAIT> arg3, 
	     HTA<T, DIM,TRAIT> arg4,
	     HTA<T, DIM,TRAIT> arg5,
	     int level = 0) 
  {
    assert (wrapped_);
    wrapped_->hmap(op, level, arg1.wrapped_, arg2.wrapped_, arg3.wrapped_, arg4.wrapped_, arg5.wrapped_);
  }

 template <typename OP>
  void hmap (OP op, HTA<T, DIM, TRAIT> arg1, 
	     HTA<T, DIM, TRAIT> arg2, 
	     HTA<T, DIM, TRAIT> arg3, 
	     HTA<T, DIM,TRAIT> arg4,
	     HTA<T, DIM,TRAIT> arg5,
	     HTA<T, DIM,TRAIT> arg6,
	     HTA<T, DIM,TRAIT> arg7,	     
	     int level = 0,
	     bool* mask = NULL) 
  {
    assert (wrapped_);
    wrapped_->hmap(op, level, 
		   arg1.wrapped_, 
		   arg2.wrapped_, 
		   arg3.wrapped_, 
		   arg4.wrapped_, 
		   arg5.wrapped_,
		   arg6.wrapped_, 
		   arg7.wrapped_,
		   mask);
  }
  

  template <typename OP>
  void mapReduce (OP& mr_op) const
  { 
    assert (wrapped_);
    wrapped_->mapReduce(mr_op);
  }


  template <typename OP>
  void mapReduce (OP& mr_op,
		  HTA<T, DIM, TRAIT> arg1) const
  { 
    assert (wrapped_);
    wrapped_->mapReduce(mr_op, *arg1.wrapped_);
  }

  /** 
   * MapReduceHTA - similar to MapReduce, with level specialization
   * This is called MapReduceHTA, just to not break existing code
   * This should be soon merged and renamed as mapReduce
   */
  template <typename OP>
  void mapReduceHTA (OP& mr_op) const
  { 
    assert (wrapped_);
    wrapped_->mapReduceHTA (mr_op);
  }
  
  template <typename OP>
  void mapReduceHTA (OP& mr_op, HTA<T, DIM, TRAIT> arg1) const
  { 
    assert (wrapped_);
    wrapped_->mapReduceHTA (mr_op, *arg1.wrapped_);
  }
  
  template <typename OP>
  void mapReduceHTA (OP& mr_op, HTA<T, DIM, TRAIT> arg1, HTA<T, DIM, TRAIT> arg2) const
  { 
    assert (wrapped_);
    wrapped_->mapReduceHTA (mr_op, *arg1.wrapped_, *arg2.wrapped_);
  }
   
  template <typename OP>
  void mapReduceHTA (OP& mr_op, HTA<T, DIM, TRAIT> arg1, HTA<T, DIM, TRAIT> arg2,  HTA<T, DIM, TRAIT> arg3,
		      HTA<T, DIM, TRAIT> arg4,  HTA<T, DIM, TRAIT> arg5,  HTA<T, DIM, TRAIT> arg6, 
		      HTA<T, DIM, TRAIT> arg7) const
  { 
    assert (wrapped_);
    wrapped_->mapReduceHTA (mr_op, *arg1.wrapped_, *arg2.wrapped_, *arg3.wrapped_, *arg4.wrapped_, 
			    *arg5.wrapped_, *arg6.wrapped_,
			    *arg7.wrapped_);
  }
 
 
  /** MapReduce with REVERSE iteration
   * GB: Technically, this is wrong. I am using it in LU
   * Very soon I will remove this and use an alternative way */
  template <typename OP>
  void mapReduceHTAR (OP& mr_op, HTA<T, DIM, TRAIT> arg1, HTA<T, DIM, TRAIT> arg2,  HTA<T, DIM, TRAIT> arg3,
		      HTA<T, DIM, TRAIT> arg4,  HTA<T, DIM, TRAIT> arg5,  HTA<T, DIM, TRAIT> arg6, 
		      HTA<T, DIM, TRAIT> arg7) const
  { 
    assert (wrapped_);
    wrapped_->mapReduceHTAR(mr_op, *arg1.wrapped_, *arg2.wrapped_, *arg3.wrapped_, *arg4.wrapped_, 
			    *arg5.wrapped_, *arg6.wrapped_,
			    *arg7.wrapped_);
  }
  

  template <typename OP>
  void reduceL (OP& mr_op) const
  { 
    assert (wrapped_);
    wrapped_->reduceL (mr_op);
  }

  template <typename OP>
  static HTA<T, DIM, TRAIT>& map(const OP& nullary_op, 
				 HTA<T, DIM, TRAIT>& dst)
  { 
    assert (dst.wrapped_);
    TRAIT::TypeImpl::map(nullary_op, *dst.wrapped_);
    return dst;
  } 

  /* 
   * applies op to all elements in the HTA
   * dst can be null if the output of the result of the 
   * unary operation is void
   */
  template <typename OP>
  static HTA<T, DIM, TRAIT> map (const OP& unary_op, 
				 const HTA<T, DIM, TRAIT>& arg, 
				 HTA<T, DIM, TRAIT>* dst = NULL)
  { 
    assert (arg.wrapped_);
    HTA<T, DIM, TRAIT> result;

    if (dst == NULL) {
      result = &arg.clone();
      dst = &result;
    }
    TRAIT::TypeImpl::map(unary_op, *arg.wrapped_, *dst->wrapped_);
    return *dst;
  }
  
  template <typename OP> 
  static HTA<T, DIM, TRAIT> map (const OP& binary_op, 
					const HTA<T, DIM, TRAIT>& larg, 
					const HTA<T, DIM, TRAIT>& rarg, 
					HTA<T, DIM, TRAIT>* dst = NULL) 
  {
    assert (larg.wrapped_);
    assert (rarg.wrapped_);
    HTA<T, DIM, TRAIT> result;

    // allocate the result array
    if (dst == NULL) {
      result = larg.clone();
      dst = &result;
    } 
    TRAIT::TypeImpl::map(binary_op, *larg.wrapped_, *rarg.wrapped_, *dst->wrapped_);
    return *dst;
  }

  /* 
   * Shape with tileAt coordinates - subclasses must override 
   * this method (even if they do not specify an implementation)
  */
  inline const Shape<DIM>& shape () const 
  { 
    assert (wrapped_);
    return wrapped_->shape();
  }

  inline const MemMapping<DIM>& memMap () const 
  {
    assert (wrapped_);
    return wrapped_->memMap();
  }
  
  inline const Tuple<DIM>& leafSize () const 
  {
    assert (wrapped_);
    return wrapped_->leafSize();
  }

  inline const Distribution<DIM>& dist () const 
  {
    assert (wrapped_);
    return wrapped_->dist();
  }

  /** rewrite of the above */
  HTA<T, DIM, TRAIT> clone (bool copy_values = false) const
  {
    assert (wrapped_);
    typename TRAIT::TypeImpl* ret = NULL;
    bool is_distributed = wrapped_->isDistributed();

    if (wrapped_->isHomogeneous_()) {
      // try to find an appropriate HTA in the cache
      AbstractArray<T, DIM>* cache_val = HTACache<T, DIM>::take(*wrapped_);
      if (cache_val != NULL) 
	ret = dynamic_cast<typename TRAIT::TypeImpl*>(cache_val);
      
      if (ret == NULL) {
	// collect tiling info according to logical shapes
	typename Tuple<DIM>::Seq tiling;
	const AbstractArray<T, DIM>* tmp = wrapped_;
	for (int i = this->level(); i >= 0; --i) {
	  tiling[i] = tmp->shape().size();
	  if (i > 0) 
	    tmp = &(tmp->accessTile (tmp->shape().low())); 
	}
	
	// then allocate a new HTA based on this tiling info
	
	// BBF Use always &wrapped_->dist() instead of const Distribution<DIM>* dist = is_distributed ? &wrapped_->dist() : NULL;

	// GBDEBUG -- changed the second argument to TILE for LU

	// GBDEUG -- not sure if the fix is correct
	// BBF Fixed that "wrapped_->tileAt(0)->memMap().order()" breaks for level 0 HTAs,
	//     thus replaced by "this->level() ? wrapped_->tileAt(0)->memMap().order() : this->memMap().order()"
	//Modified by Jia
	if(wrapped_->ol()){
	  ret = TRAIT::TypeImpl::alloc(this->level(), tiling, this->level() ? wrapped_->tileAt(0)->memMap().order() : this->memMap().order(), wrapped_->ol(), &wrapped_->dist());
	}else{
	  ret = TRAIT::TypeImpl::alloc(this->level(), tiling, this->level() ? wrapped_->tileAt(0)->memMap().order() : this->memMap().order(), &wrapped_->dist());
	}

	if (copy_values)
	  *ret = *this->wrapped_;
	
      } 
    } else {	//GBDEBUG: this is my new addition for non-homogenous cases
      ret = wrapped_->clone (copy_values);
    } 
    
    //    if (is_distributed) 
    // TRAIT::barrier();
    return HTA<T, DIM, TRAIT>(*ret);
  }
  

  HTA<T, DIM, TRAIT> permute (const Tuple<DIM>& dims, 
				     HTA<T, DIM, TRAIT>* dst = NULL,
				     unsigned mask = 0xFF) const 
  {
    assert (wrapped_);
    typename TRAIT::TypeImpl* ret = wrapped_->permute(dims, (dst == NULL) ? NULL : dst->wrapped_, mask);
    //    TRAIT::barrier();
    return HTA<T, DIM, TRAIT>(*ret);
  }

  TransExpr<TRAIT, DTRANSPOSE> dpermute (const Tuple<2>& dims) const 
  {
    assert (wrapped_);
    return TransExpr<TRAIT, DTRANSPOSE> (*this, dims, 0xFF);
  }
  
  void htranspose (HTA<T, DIM, TRAIT> dst) const
  {
    assert (wrapped_);
    wrapped_->htranspose (*dst.wrapped_);
    //    TRAIT::barrier();
  }
  
  TransExpr<TRAIT, TRANSPOSE> transpose () const
  {
    assert (DIM == 2);
    assert (wrapped_); 
    Tuple<2> dims(1,0);   
    return TransExpr<TRAIT, TRANSPOSE> (*this, dims, 0xFF);
  }
  
  TransExpr<TRAIT, TRANSPOSE> ltranspose () const
  {
    assert (wrapped_);    
    Tuple<2> dims(1,0);   
    return TransExpr<TRAIT, TRANSPOSE> (*this, dims, 0x01);
  }

  /* HTA<T, DIM, TRAIT> transpose (HTA<T, DIM, TRAIT>* dst = NULL) const 
  {
    assert (DIM == 2);
    typename TRAIT::TypeImpl* ret = wrapped_->permute(Tuple<DIM>(1, 0), (dst == NULL) ? NULL : dst->wrapped_);
    //    TRAIT::barrier();
    return HTA<T, DIM, TRAIT>(*ret);
    }*/
  
  
  /* HTA<T, DIM, TRAIT> ltranspose (HTA<T, DIM, TRAIT>* dst = NULL) const 
  {
    assert (DIM == 2);
    assert (wrapped_);
    typename TRAIT::TypeImpl* ret = wrapped_->permute(Tuple<DIM>(1,0), (dst == NULL) ? NULL : dst->wrapped_, 0x01);
    //TRAIT::barrier();
    return HTA<T, DIM, TRAIT>(*ret);
    } */

  /**
   * check if the current hta is local(true) or private (true) or remote (false)
   * proxy leaves should override this method to return false
   */
  inline bool isLocal () const 
  { 
    assert (wrapped_);
    return wrapped_->isLocal(); 
  }

  inline int home () const
  {
    assert(wrapped_);
    return wrapped_->home();
  }

  T* raw () const
  {
    assert (wrapped_);
    return wrapped_->raw();
  }

  void initSparse (T* values, const int* start, const int* idx) 
  {
    assert (values);
    assert (start);
    assert (idx);
    wrapped_->initSparse(values, start, idx);
  }


  template<typename TRAIT2>
  static inline bool conformable (const HTA<T, DIM, TRAIT>& larg, 
				  const HTA<T, DIM, TRAIT2>& rarg) 
  {
    assert (larg.wrapped_);
    assert (rarg.wrapped_);
    // this is only correct if all tiles in larg resp. rarg have the same shape ..

    if (larg.level() > rarg.level()) 
      return TRAIT::TypeImpl::conformable(*larg->tileAt(Tuple<DIM>::zero), *rarg.wrapped_);
    else 
      return TRAIT::TypeImpl::conformable(*rarg->tileAt(Tuple<DIM>::zero), *larg.wrapped_);
  }

  static inline bool conformable (const HTA<T, DIM, TRAIT>& larg, 
				  const HTA<T, DIM, TRAIT>& rarg)
  {
    return TRAIT::TypeImpl::conformable(*larg->wrapped_, *rarg->wrapped_);
  }

  static inline bool conformable (const HTA<T, DIM, TRAIT>& larg, const T& rarg) 
  {
    return true;
  }

  static inline bool conformable (const T& rarg, const HTA<T, DIM, TRAIT>& larg) 
  {
    return true;
  }

  /**
   * Factory method for top-down creation of HTA implementations
   * 
   * @param  tiling    complete tiling information (index i-1 is tiling at level i)
   * @param  data      flat data to be managed by that HTA (only ehwn allocating locally)
   * @param  ordering  row, column, tiled
   * @return           implementation of HTA boxed in an HTA
   */
  static HTA<T, DIM, TRAIT> alloc (size_t levels, 
				   const typename Tuple<DIM>::Seq& tiling, 
				   const Distribution<DIM>* dist = NULL, 
				   const Order ordering = TILE)
  {
    typename TRAIT::TypeImpl* ret = TRAIT::TypeImpl::alloc(levels, tiling, ordering, dist);
    ret->setCachable();
    //TRAIT::barrier();
    return HTA<T, DIM, TRAIT> (*ret);
  }
  
  // for overlapped tiling
  static HTA<T, DIM, TRAIT> alloc (size_t levels, 
				   const typename Tuple<DIM>::Seq& tiling, 
				   Overlap<DIM>* ol,
				   const Distribution<DIM>* dist = NULL, 
				   const Order ordering = TILE)
  {
    typename TRAIT::TypeImpl* ret = TRAIT::TypeImpl::alloc(levels, tiling, ordering, ol, dist);
    ret->setCachable();
    //TRAIT::barrier();
    return HTA<T, DIM, TRAIT> (*ret);
  }


  static HTA<T, DIM, TRAIT> alloc (const typename Tuple<DIM>::Seq& tiling, 
				   const Distribution<DIM>* dist = NULL, 
				   const Order ordering = TILE)
  {
    return alloc(tiling.length()-1, tiling, dist, ordering);
  }

  /**
   * Local allocation 
   */
  static HTA<T, DIM, TRAIT> allocLocal (size_t levels,
					const typename Tuple<DIM>::Seq& tiling,
					T* init = NULL,
					const Order ordering = TILE)
  {
    typename TRAIT::TypeImpl* ret = TRAIT::TypeImpl::allocLocal(levels, tiling, init, ordering);
    ret->setCachable();
    return HTA<T, DIM, TRAIT> (*ret);
  }

  /**
   * Returns an HTA at a partuclar level for which the tiles are not intialized.
   * Such HTA can be understood as a 'shell', where tile are intialized subsequently. 
   * This HTA constructor is in support of the bottom up creation of HTAs.
   */
  static HTA<T, DIM, TRAIT> allocShell (const size_t levels,
					const Tuple<DIM>& tiling, 
					const Distribution<DIM>* dist = NULL)
  {
    typename TRAIT::TypeImpl* ret = TRAIT::TypeImpl::allocShell(levels, tiling, dist);
    return HTA<T, DIM, TRAIT> (*ret);
  }

  static HTA<T, DIM, TRAIT> allocShell (const size_t levels,
					const typename Tuple<DIM>::Seq & tiling, 
					const Distribution<DIM>* dist = NULL)
  {
    typename TRAIT::TypeImpl* ret = TRAIT::TypeImpl::allocShell(levels, tiling, dist);
    return HTA<T, DIM, TRAIT> (*ret);
  }


  /**
   * Factory method for top-down creation of RANDOM HTAs
   * i.e HTAs filled with uniform random numbers -- inspired from MATLAB rand
   * Just a syntactic short-cut.
   * @param  tiling    complete tiling information (index i-1 is tiling at level i)
   * @param  data      flat data to be managed by that HTA (only ehwn allocating locally)
   * @param  ordering  row, column, tiled
   * @return           implementation of HTA boxed in an HTA
   */
  static HTA<T, DIM, TRAIT> allocRand (size_t level,
				       const typename Tuple<DIM>::Seq& tiling, 
				       const Order ordering,
				       double seed = 314159265.00,
				       double a = 1220703125)
  {
    assert (DIM <=2);
    typename TRAIT::TypeImpl* ret = TRAIT::TypeImpl::alloc (level, tiling, ordering);
    ret->setCachable();
    //TRAIT::barrier();
    RandNumGeneratorDist<double, DIM> rand (a, seed);    
    ret->mapReduce(rand);    
    return HTA<T, DIM, TRAIT> (*ret);
  }

  void initMemMapping() 
  {
    assert (wrapped_);
    wrapped_->initMemMapping();
  }

  void updateShape()
  {
    assert(wrapped_);
    wrapped_->updateShape();
  }

  bool isLeaf() const
  {
    return wrapped_->isLeaf();
  }

  size_t level() const 
  {
    return wrapped_->level();
  }

  // for debugging
  typename TRAIT::TypeImpl* impl() const
  {
      return wrapped_; 
  }

private:

  // all other template instance of this class should be friends (required when accessing wrapped_)
  template<class T2, int DIM2, typename TRAIT2>
  friend class HTA;
  
  template<class OP, typename TRAIT2>
  friend class BinExpr;

  template<class OP, typename TRAIT2>
  friend class ReduceExpr;

  template<typename TRAIT2, int TYPE>
  friend class TransExpr;

  /** pointer to the implementation of the HTA - this internal to the class */
  typename TRAIT::TypeImpl* wrapped_;

  template<class T2, int DIM2, typename TRAIT2>
  friend ostream& operator<< (ostream& ost, const HTA<T2, DIM2, TRAIT2>& a);
  
  template<class T2, int DIM2, typename TRAIT2>
  friend class Router;  
}; // class HTA

template<class T, int DIM, typename TRAIT> 
HTA<T, DIM, TRAIT> operator+ (const T& val, const HTA<T, DIM, TRAIT>& arr) 
{
  // allocate the result array
  HTA<T, DIM, TRAIT> result = arr.clone();
  TRAIT::TypeImpl::map(binder1st<plus<T> >(plus<T>(), val), *arr.wrapped_, *result.wrapped_);
  return result;
}

template<class T, int DIM, typename TRAIT> 
HTA<T, DIM, TRAIT> operator- (const T& val, const HTA<T, DIM, TRAIT>& arr) 
{
  // allocate the result array
  HTA<T, DIM, TRAIT> result = arr.clone();
  TRAIT::TypeImpl::map(binder1st<minus<T> >(minus<T>(), val), *arr.wrapped_, *result.wrapped_);
  return result;
}

template<class T, int DIM, typename TRAIT> 
HTA<T, DIM, TRAIT> operator* (const T& val, const HTA<T, DIM, TRAIT>& arr) 
{
  // allocate the result array
  HTA<T, DIM, TRAIT> result = arr.clone();
  TRAIT::TypeImpl::map(binder1st<multiplies<T> >(multiplies<T>(), val), *arr.wrapped_, *result.wrapped_);
  return result;
}

template<class T, int DIM, typename TRAIT> 
HTA<T, DIM, TRAIT> operator/ (const T& val, const HTA<T, DIM, TRAIT>& arr) 
{
  // allocate the result array
  HTA<T, DIM, TRAIT> result = arr.clone();
  TRAIT::TypeImpl::map(binder1st<divides<T> >(divides<T>(), val), *arr.wrapped_, *result.wrapped_);
  return result;
}

template<typename T, int DIM, typename TRAIT>
ostream& operator<< (ostream& ost, const HTA<T, DIM, TRAIT>& a) 
{
  // let all threads advance to this point-otherwise stale data might be printed 
  TRAIT::Default::barrier();
  
  // communicate to place 0 - that is where printing occurs

  
  /* 
   * GBDEBUG -- there is some bug ; several non-zero entries are missing in the output 
   * As a temporary fix I am letting each processor print its local  tile
   */

  //a.wrapped_->communicate_(ConstDistribution<DIM>::get(0));  
  //if (Traits::Default::myPlace() == 0) {
  if (a.wrapped_ == NULL) 
    ost << "[]" << endl;
  else 
    ost << *a.wrapped_;
  //}
  
  return ost;
}  

#endif /* __HTA_H__ */
