/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: HTA_0.h,v 1.44 2006/05/30 06:52:28 bikshand Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef  __HTA_0_H__
#define  __HTA_0_H__

#include <cassert>
#include <iostream>

#include "BinExpr.h"
#include "Operators.h"
#include "Order.h"
#include "Router.h"
#include "Shape.h"
#include "Tuple.h"
#include "util/Iterator.h"
#include "debug/Tracing.h"

#ifndef HTA_DEFAULT_TRAIT
#error "HTA_DEFAULT_TRAIT not defined!"
#endif

using namespace std;

/** 
 * This is a specialization of class HTA that just bends the semantics of operatror[]
 * to the usual array access semantics -- it returns a scalar value - not an HTA.
 */
template<class T, int DIM, typename TRAIT>
class HTA<T, DIM, 0, TRAIT> {

public:
 
  /** 
   * Default constructor  should only be used for the creation of HTA []
   */
  HTA () : 
    wrapped_ (NULL) {}

  /**
   * @param a   the wrapped array
   */
  HTA (typename TRAIT::TypeImpl& a) : 
    wrapped_(&a)
  { 
    wrapped_->refCount_(1);
  }

  /** 
   * copy constructor - adjusts ref count of wrapped object 
   */
  HTA (const HTA<T,DIM, 0, TRAIT>& rhs) : 
    wrapped_(rhs.wrapped_)
  {
    assert(wrapped_);
    wrapped_->refCount_(1);
  }

  virtual ~HTA() 
  {
    if (wrapped_ != NULL && wrapped_->refCount_(-1) == 0) 
       delete wrapped_;
  }
  
  /** 
   * scalarAt operator -- only implemented for arrays at level 0
   */
  T& operator[] (const Tuple<DIM>& t) const
  {
    assert (wrapped_);
    return wrapped_->scalarAt (t);
  }

  typename TRAIT::Type operator[] (const Shape<DIM>& t) const
  {
    assert (wrapped_);
    return typename TRAIT::Type (*wrapped_->scalarAt(t));
  }
  
  /**
   * Assignment operator - assigning a sparse array to LeafHTA
   */
  typename TRAIT::Type& operator= (const typename TRAIT::TypeDenseInternal& rhs) 
  {
    assert (false); // not yet implemented
    return *this;
  }

  typename TRAIT::Type& operator= (const typename TRAIT::TypeSparseInternal& rhs) 
  {
    *wrapped_ = rhs;
    return *this;
  }
     
  /** 
   * Assignment operator - this is not subsumbed by the template version! 
   */
  typename TRAIT::Type& operator= (const HTA<T, DIM, 0>& rhs) {
    if (this != &rhs) {
      if (wrapped_ == NULL) { 
        // initialization
        wrapped_ = rhs.wrapped_->clone();
        wrapped_->refCount_(1);
      } else  {
        // communication / pointwise copy 
        assert (Shape<DIM>::conformable(this->shape(), rhs.shape()));
        rhs.wrapped_->communicate_(this->wrapped_->dist(), this->wrapped_);
      }
    }    
    return *this;
  }

  /** 
   * Intialization or pointwise assignment 
   */
  template <int LEVEL2>
  typename TRAIT::Type& operator= (const HTA<T, DIM, LEVEL2>& rhs) 
  { 
    assert (0 != LEVEL2);
    assert (rhs.wrapped_);
    assert (0); // not yet implemented
    return *this;
  }
  
  /** pointwise assignment */
  typename TRAIT::Type& operator= (const T& rhs) 
  {
    assert (wrapped_);
    *wrapped_ = rhs;
    return *this;
  }

  // dotproduct ----------------------------------------------------------------------

  static typename TRAIT::TypeScalar dotproduct (const typename TRAIT::Type& larg, 
						 const typename TRAIT::Type& rarg)
  {
    assert (larg.shape() == rarg.shape());
    return TRAIT::TypeImpl::dotproduct (*larg.wrapped_, *rarg.wrapped_);
  }

  // matmul ----------------------------------------------------------------------
  
  static  typename TRAIT::Type matmul (const typename TRAIT::Type& larg, 
				       const typename TRAIT::Type& rarg, 
				       typename TRAIT::Type* res = NULL) 
  {
    assert (DIM == 2);
    assert (larg.shape().size()[1] == rarg.shape().size()[0]);
    
    // allocate the result array if necessary
    typename TRAIT::TypeImpl* dst;
    if (res == NULL) {
      typename Tuple<DIM>::Seq res_tiling;
      const typename Tuple<DIM>::Seq& larg_tiling = larg.wrapped_->tiling();
      const typename Tuple<DIM>::Seq& rarg_tiling = rarg.wrapped_->tiling();
      for (int i = 0; i <= 0; ++i) {
	res_tiling[i] = Tuple<DIM>(larg_tiling[i][0], rarg_tiling[i][1]);
      }
      Order order = larg.memMap().order();
      dst = TRAIT::TypeImpl::alloc(res_tiling, order);
    } else {
      dst = res->wrapped_;
      assert (dst->shape().size()[0] == larg.shape().size()[0] && 
	      dst->shape().size()[1] == rarg.shape().size()[1]);
    }

    TRAIT::TypeImpl::matmul (*larg.wrapped_, *rarg.wrapped_, *dst);
    return typename TRAIT::Type (*dst); 
  }

  template <typename TRAIT2>
  static typename TRAIT2::Type matvecmul (const typename TRAIT::Type& larg,
					  const typename TRAIT2::Type& rarg,
					  typename TRAIT2::Type* res = NULL) 
  {    
    assert (DIM == 2);
    assert (larg.shape().size()[1] == rarg.shape().size()[0]);
    const int LEVEL = 0;

    // allocate the result array if necessary
    typename TRAIT::TypeImpl* dst;
    if (res == NULL) {
      typename Tuple<DIM>::Seq res_tiling;
      const typename Tuple<DIM>::Seq& larg_tiling = larg.wrapped_->tiling_();
      const typename Tuple<DIM>::Seq& rarg_tiling = rarg.wrapped_->tiling_();

      res_tiling[LEVEL] = Tuple<DIM>(larg_tiling[LEVEL][0], rarg_tiling[LEVEL][1]);
      Order order = larg.memMap().order();
      dst = TRAIT2::TypeImpl::alloc(res_tiling, order);
    } else {
      dst = res->wrapped_;
      assert (dst->shape().size()[0] == larg.shape().size()[0] && 
	      dst->shape().size()[1] == rarg.shape().size()[1]);
    } 

    TRAIT::TypeImpl::matvecmul (*larg.wrapped_, *rarg.wrapped_, *dst);
    TRAIT::barrier();
    return typename TRAIT2::Type (*dst);    
  }
  
  // leaf-convolution -----------------------------------------------------------------

  HTA<T, DIM, 0, TRAIT> lconv (const T* weights,
			       HTA<T, DIM, 0, TRAIT>* dst = NULL) const
  {
    assert (weights != NULL);
    assert (this->wrapped_);
    HTA<T, DIM, 0, TRAIT> result;

    // limitation of the current implementation
    if (dst == NULL) {
      result = clone(false);
      dst = &result;
    } else 
      assert (dst->wrapped_);
    
    wrapped_->lconv(weights, *dst->wrapped_);
    TRAIT::barrier();
    return *dst;
  }


  // operator+ --------------------------------------------------------------------

  template<class T2, int DIM2, int LEVEL2, typename TRAIT2> 
  friend HTA<T2, DIM2, LEVEL2, TRAIT2> operator+ (const T2& val, 
						  const HTA<T2, DIM2, LEVEL2, TRAIT2>& arr);

  BinExpr<plus<T>, TRAIT> operator+ (const T& rhs) const
  {
    return BinExpr<plus<T>, TRAIT>(*this, rhs);
  }

  BinExpr<plus<T>, TRAIT> operator+ (const HTA<T, DIM, 0, TRAIT>& rhs) const
  {
    return BinExpr<plus<T>, TRAIT>(*this, rhs);
  }

  template<typename OP>
  BinExpr<plus<T>, TRAIT> operator+ (const BinExpr<OP, TRAIT>& rhs) const
  {
    HTA<T, DIM, 0, TRAIT> rhs_eval = rhs.eval();
    // re-use the temporary that comes with the bin-expr arg
    return BinExpr<plus<T>, TRAIT>(*this, rhs_eval, rhs_eval.wrapped_);
  }

  template<typename T2>
  HTA<T, DIM, 0, TRAIT> operator+ (const T2& rhs) const
  {
    HTA<T, DIM, 0, TRAIT> result = clone();
    TRAIT::TypeImpl::map (binder2nd<plus<T2> >(plus<T2>(), rhs), *wrapped_, *result.wrapped_);
    return result;
  }

  template <int LEVEL2, typename TRAIT2>
  HTA<T, DIM, 0, TRAIT> operator+ (const HTA<T, DIM, LEVEL2, TRAIT2>& rhs) const
  {
    assert (LEVEL2 <= 0);                    // limitation of the current implementation
    HTA<T, DIM, 0, TRAIT> result = clone();  // allocate the result array
    TRAIT::TypeImpl::map (plus<T>(), *wrapped_, *rhs.wrapped_, *result.wrapped_);
    return result;
  }

  // this non-template version is necessary to help the compiler to 
  // auto-convert BinExpr and Router instances to the rhs HTA argument.
  HTA<T, DIM, 0>& operator+= (const HTA<T, DIM, 0, TRAIT>& rhs)
  {
    assert (wrapped_);
    TRAIT::TypeImpl::map (plus<T>(), *wrapped_, *rhs.wrapped_, *wrapped_);
    return *this;
  }

  template <int LEVEL2, typename TRAIT2>
  HTA<T, DIM, 0, TRAIT>& operator+= (const HTA<T, DIM, LEVEL2, TRAIT2>& rhs)
  {
    assert (wrapped_);
    assert (0 >= LEVEL2);
    TRAIT::TypeImpl::map (plus<T>(), *wrapped_, *rhs.wrapped_, *wrapped_);
    return *this;
  }    

  HTA<T, DIM, 0, TRAIT>& operator+= (const T& rhs)
  {
    assert(wrapped_);
    TRAIT::TypeImpl::map (binder2nd<plus<T> >(plus<T>(), rhs), *wrapped_, *wrapped_);
    return *this;
  }

  // operator- --------------------------------------------------------------------

  template<class T2, int DIM2, int LEVEL2, typename TRAIT2> 
  friend HTA<T2, DIM2, LEVEL2, TRAIT2> operator- (const T2& val, 
						  const HTA<T2, DIM2, LEVEL2, TRAIT2>& arr);

  BinExpr<minus<T>, TRAIT> operator- (const T& rhs) const
  {
    return BinExpr<minus<T>, TRAIT>(*this, rhs);
  }

  BinExpr<minus<T>, TRAIT> operator- (const HTA<T, DIM, 0, TRAIT>& rhs) const
  {
    return BinExpr<minus<T>, TRAIT>(*this, rhs);
  }

  template<typename OP>
  BinExpr<minus<T>, TRAIT> operator- (const BinExpr<OP, TRAIT>& rhs) const
  {
    HTA<T, DIM, 0, TRAIT> rhs_eval = rhs.eval();
    // re-use the temporary that comes with the bin-expr arg
    return BinExpr<minus<T>, TRAIT>(*this, rhs_eval, rhs_eval.wrapped_);
  }

  template<typename T2>
  HTA<T, DIM, 0, TRAIT> operator- (const T2& rhs) const
  {
    // allocate the result array
    HTA<T, DIM, 0, TRAIT> result = clone();
    TRAIT::TypeImpl::map (binder2nd<minus<T2> >(minus<T2>(), rhs), *wrapped_, *result.wrapped_);
    return result;
  }
  
  template <int LEVEL2, typename TRAIT2>
  HTA<T, DIM, 0, TRAIT> operator- (const HTA<T, DIM, LEVEL2, TRAIT2>& rhs) const
  {
    // allocate the result array
    assert (LEVEL2 <= 0);                    // limitation of the current implementation
    HTA<T, DIM, 0, TRAIT> result = clone(); // allocate the result array
    TRAIT::TypeImpl::map (minus<T>(), *wrapped_, *rhs.wrapped_, *result.wrapped_);
    return result;
  }

  // this non-template version is necessary to help the compiler to 
  // auto-convert BinExpr and Router instances to the rhs HTA argument.
  HTA<T, DIM, 0>& operator-= (const HTA<T, DIM, 0, TRAIT>& rhs)
  {
    assert (wrapped_);
    TRAIT::TypeImpl::map (minus<T>(), *wrapped_, *rhs.wrapped_, *wrapped_);
    return *this;
  }    

  template <int LEVEL2, typename TRAIT2>
  HTA<T, DIM, 0>& operator-= (const HTA<T, DIM, LEVEL2, TRAIT2>& rhs)
  {
    assert (wrapped_);
    assert (0 >= LEVEL2);
    TRAIT::TypeImpl::map (minus<T>(), *wrapped_, *rhs.wrapped_, *wrapped_);
    return *this;
  }    

  HTA<T, DIM, 0, TRAIT>& operator-= (const T& rhs)
  {
    assert (wrapped_);
    TRAIT::TypeImpl::map (*binder2nd<minus<T> >(minus<T>(), rhs), *wrapped_, *wrapped_);
    return *this;
  }

  // operator* --------------------------------------------------------------------

  template<class T2, int DIM2, int LEVEL2, typename TRAIT2> 
  friend HTA<T2, DIM2, LEVEL2, TRAIT2> operator* (const T2& val, 
						  const HTA<T2, DIM2, LEVEL2, TRAIT2>& arr);

  BinExpr<multiplies<T>, TRAIT> operator* (const T& rhs) const
  {
    return BinExpr<multiplies<T>, TRAIT>(*this, rhs);
  }

  BinExpr<multiplies<T>, TRAIT> operator* (const HTA<T, DIM, 0, TRAIT>& rhs) const
  {
    return BinExpr<multiplies<T>, TRAIT>(*this, rhs);
  }

  template<typename OP>
  BinExpr<multiplies<T>, TRAIT> operator* (const BinExpr<OP, TRAIT>& rhs) const
  {
    HTA<T, DIM, 0, TRAIT> rhs_eval = rhs.eval();
    // re-use the temporary that comes with the bin-expr arg
    return BinExpr<multiplies<T>, TRAIT>(*this, rhs_eval, rhs_eval.wrapped_);
  }

  template<typename T2>
  HTA<T, DIM, 0, TRAIT> operator* (const T2& rhs) const
  {
    // allocate the result array
    HTA<T, DIM, 0, TRAIT> result = clone();
    TRAIT::TypeImpl::map (binder2nd<multiplies<T2> >(multiplies<T2>(), rhs), *wrapped_, *result.wrapped_);
    return result;
  }
  
  template <int LEVEL2, typename TRAIT2>
  HTA<T, DIM, 0, TRAIT> operator* (const HTA<T, DIM, LEVEL2, TRAIT2>& rhs) const
  {
    // allocate the result array
    assert (LEVEL2 <= 0);                    // limitation of the current implementation
    HTA<T, DIM, 0, TRAIT> result = clone();  // allocate the result array
    TRAIT::TypeImpl::map (multiplies<T>(), *wrapped_, *rhs.wrapped_, *result.wrapped_);
    return result;
  }

  // this non-template version is necessary to help the compiler to 
  // auto-convert BinExpr and Router instances to the rhs HTA argument.
  HTA<T, DIM, 0>& operator*= (const HTA<T, DIM, 0, TRAIT>& rhs)
  {
    assert (wrapped_);
    TRAIT::TypeImpl::map (multiplies<T>(), *wrapped_, *rhs.wrapped_, *wrapped_);
    return *this;
  }

  template <int LEVEL2, typename TRAIT2>
  HTA<T, DIM, 0, TRAIT>& operator*= (const HTA<T, DIM, LEVEL2, TRAIT2>& rhs)
  {
    assert (wrapped_);
    assert (0 >= LEVEL2);
    TRAIT::TypeImpl::map (multiplies<T>(), *wrapped_, *rhs.wrapped_, *wrapped_);
    return *this;
  }    

  HTA<T, DIM, 0, TRAIT>& operator*= (const T& rhs)
  {
    assert (wrapped_);
    TRAIT::TypeImpl::map (binder2nd<multiplies<T> >(multiplies<T>(), rhs), *wrapped_, *wrapped_);
    return *this;
  }

  template <int OPLEVEL>
  HTA<T, DIM, 0, TRAIT> scan() const
  {
    HTA<T, DIM, 0, TRAIT> result = clone();
    wrapped_->scan<OPLEVEL>(*result.wrapped_);
    return result;
  }

  // operator/ --------------------------------------------------------------------

  template<class T2, int DIM2, int LEVEL2, typename TRAIT2> 
  friend HTA<T2, DIM2, LEVEL2, TRAIT2> operator/ (const T2& val, 
						  const HTA<T2, DIM2, LEVEL2, TRAIT2>& arr);

  BinExpr<divides<T>, TRAIT> operator/ (const T& rhs) const
  {
    return BinExpr<divides<T>, TRAIT>(*this, rhs);
  }

  BinExpr<divides<T>, TRAIT> operator/ (const HTA<T, DIM, 0, TRAIT>& rhs) const
  {
    return BinExpr<divides<T>, TRAIT>(*this, rhs);
  }
  
  template<typename OP>
  BinExpr<divides<T>, TRAIT> operator/ (const BinExpr<OP, TRAIT>& rhs) const
  {
    HTA<T, DIM, 0, TRAIT> rhs_eval = rhs.eval();
    // re-use the temporary that comes with the bin-expr arg
    return BinExpr<divides<T>, TRAIT>(*this, rhs_eval, rhs_eval.wrapped_);
  }

  template<typename T2>
  HTA<T, DIM, 0, TRAIT> operator/ (const T2& rhs) const
  {
    assert (rhs != 0);
    // allocate the result array
    HTA<T, DIM, 0, TRAIT> result = clone();
    TRAIT::TypeImpl::map (binder2nd<divides<T2> >(divides<T2>(), rhs), *wrapped_, *result.wrapped_);
    return result;
  }
  
  template <int LEVEL2, typename TRAIT2>
  HTA<T, DIM, 0, TRAIT> operator/ (const HTA<T, DIM, LEVEL2, TRAIT2>& rhs) const
  {
    // allocate the result array
    assert (LEVEL2 <= 0);                    // limitation of the current implementation
    HTA<T, DIM, 0, TRAIT> result = clone();  // allocate the result array
    TRAIT::TypeImpl::map (divides<T>(), *wrapped_, *rhs.wrapped_, *result.wrapped_);
    return result;
  }

  // this non-template version is necessary to help the compiler to 
  // auto-convert BinExpr and Router instances to the rhs HTA argument.
  HTA<T, DIM, 0, TRAIT>& operator/= (const HTA<T, DIM, 0, TRAIT>& rhs)
  {
    assert (wrapped_);
    TRAIT::TypeImpl::map (divides<T>(), *wrapped_, *rhs.wrapped_, *wrapped_);
    return *this;
  }

  template <int LEVEL2, typename TRAIT2>
  HTA<T, DIM, 0, TRAIT>& operator/= (const HTA<T, DIM, LEVEL2, TRAIT2>& rhs)
  {
    assert (wrapped_);
    assert (0 >= LEVEL2);
    TRAIT::TypeImpl::map (divides<T>(), *wrapped_, *rhs.wrapped_, *wrapped_);
    return *this;
  }    

  HTA<T, DIM, 0, TRAIT>& operator/= (const T& rhs)
  {
    assert (wrapped_);
    TRAIT::TypeImpl::map (binder2nd<divides<T> >(divides<T>(), rhs), *wrapped_, *wrapped_);
    return *this;
  }

  
  // operator== --------------------------------------------------------------------
  
  bool operator== (const HTA<T, DIM, 0, TRAIT>& rhs) const
  {
    bool ret;
    if (this == &rhs || rhs.wrapped_ == this->wrapped_) {
      ret = true;
    } else {
      assert (rhs.wrapped_ && wrapped_);
      ret = (wrapped_ == *rhs.wrapped_);
    }
    return ret;
  }
  
  bool operator!= (const HTA<T, DIM, 0, TRAIT>& rhs) const
  {
    return ! (this->operator==(rhs)); 
  }

  // collectives 

  template <typename OP> 
  T reduce (const OP& binary_op) const
  {
    assert (wrapped_);
    return wrapped_->reduce(binary_op);
  }

  template <typename OP> 
  HTA<T, DIM, 0, TRAIT> reduce (const OP& binary_op,
				int dim, 
				unsigned all_to_all_mask = 0x0,
				HTA<T, DIM, 0, TRAIT>* dst = NULL) const
  { 
    assert (wrapped_);
    assert (dst == NULL); // limitation of the current implementation, result is allocated in the callee.

     // (1) allocate the result and initialize it 
    typename TRAIT::TypeImpl* dst_actual;
    // (1a) compute tiling according to all_to_all_mask
    typename Tuple<DIM>::Seq tiling_red = wrapped_->tiling();
    for (int i = 0; i <= 0; ++i) {
      if (!(all_to_all_mask & (1 << i)))
	// no all-to-all along this tiling level
	tiling_red[i] = tiling_red[i].mask(dim, 1);
    }
    Order order = this->memMap().order();
    dst_actual = TRAIT::TypeImpl::alloc(tiling_red, order);
    
    // (2) then do the reduction
    wrapped_->reduce (binary_op, dim, *dst_actual);
    return HTA<T, DIM, 0, TRAIT>(*dst_actual);
  }

  template <typename OP>
  void mapReduce (OP& mr_op) 
  { 
    assert (wrapped_);
    wrapped_->TRAIT::TypeImpl::mapReduce(mr_op);
  }

  template <typename OP>
  static HTA<T, DIM, 0, TRAIT>& map (const OP& nullary_op, 
				     HTA<T, DIM, 0, TRAIT>& dst)
  { 
    assert (dst.wrapped_);
    TRAIT::TypeImpl::map(nullary_op, *dst.wrapped_);
    return dst;
  } 

  /* 
   * applies op to all elements in the HTA
   * dst can be null if the output of the result of the 
   * unary operation is void
   */
  template <typename OP>
  static HTA<T, DIM, 0, TRAIT> map (const OP& unary_op, 
				        const HTA<T, DIM, 0, TRAIT>& arg, 
				        HTA<T, DIM, 0, TRAIT>* dst = NULL)
  { 
    assert (arg.wrapped_);
    HTA<T, DIM, 0, TRAIT> result;

    if (dst == NULL) {
      result = &arg.clone();
      dst = &result;
    }
    TRAIT::TypeImpl::map(unary_op, *arg, *dst);
    return *dst;
  }
  
  template <typename OP> 
  static HTA<T, DIM, 0, TRAIT> map (const OP& binary_op, 
					const HTA<T, DIM, 0, TRAIT>& larg, 
					const HTA<T, DIM, 0, TRAIT>& rarg, 
					    HTA<T, DIM, 0, TRAIT>* dst = NULL) 
  {
    assert (larg.wrapped_);
    assert (rarg.wrapped_);
    HTA<T, DIM, 0, TRAIT> result;

    // allocate the result array
    if (dst == NULL) {
      result = larg.clone();
      dst = &result;
    } 
    TRAIT::TypeImpl::map(binary_op, *larg.wrapped_, *rarg.wrapped_, *dst->wrapped_);
    return *dst;
  }

  /* 
   * Shape with tileAt coordinates - subclasses must override 
   * this method (even if they do not specify an implementation)
  */
  const Shape<DIM>& shape() const { 
    assert (wrapped_);
    return wrapped_->shape();
  }

  /**
   * clones the structure of the HTA, but does not copy the values. 
   * the result HTA is initialized with zeros.
   */
  HTA<T, DIM, 0, TRAIT> clone(bool copy_values = false) const
  {
    assert (wrapped_);
    typename TRAIT::TypeImpl* ret = wrapped_->clone(copy_values);
    return HTA<T, DIM, 0, TRAIT>(*ret);
  }
  
  HTA<T, DIM, 0, TRAIT> permute (const Tuple<DIM>& dims, HTA<T, DIM, 0>* dst = NULL) const 
  {
    assert (wrapped_);
    typename TRAIT::TypeImpl* ret = wrapped_->permute(dims, (dst == NULL) ? NULL : dst->wrapped_);
    return HTA<T, DIM, 0, TRAIT>(*ret);
  }
  
  HTA<T, DIM, 0, TRAIT> transpose(HTA<T, DIM, 0>* dst = NULL) const 
  {
    assert (DIM == 2);
    return permute (Tuple<DIM>(1, 0), dst);
  }

  /*
   * check if the current hta is local(true) or private (true) or remote (false)
   * proxy leaves should override this method to return false
   */
  bool isLocal () const 
  { 
    assert (wrapped_);
    return wrapped_->isLocal(); 
  }

  int home() const
  {
    assert(wrapped_);
    return wrapped_->home();
  }

  const MemMapping<DIM>& memMap () const 
  {
    return wrapped_->memMap();
  }

  T* raw() const
  {
    assert (wrapped_);
    return wrapped_->raw();
  }

  void initSparse (T* values, const int* start, const int* idx) 
  {
    assert (values);
    assert (start);
    assert (idx);
    wrapped_->initSparse(values, start, idx);
  }

  /**
   * Factory method for top-down creation of HTA implementations
   * 
   * @param  tiling    complete tiling information (index i-1 is tiling at level i)
   * @param  data      flat data to be managed by that HTA (only ehwn allocating locally)
   * @param  ordering  row, column, tiled
   * @return           implementation of HTA boxed in an HTA
   */
  static HTA<T, DIM, 0, TRAIT> alloc (const typename Tuple<DIM>::Seq& tiling) 
  {
    typename TRAIT::TypeImpl* ret = TRAIT::TypeImpl::alloc(tiling, TILE);
    return HTA<T, DIM, 0, TRAIT> (*ret);
  }

  /**
   * Local allocation 
   */
  static HTA<T, DIM, 0, TRAIT> allocLocal (const typename Tuple<DIM>::Seq& tiling,
					   T* init = NULL)
  {
    typename TRAIT::TypeImpl* ret = TRAIT::TypeImpl::allocLocal(tiling, init, TILE);
    return HTA<T, DIM, 0, TRAIT> (*ret);
  }
  
private:

  // all other template instance of this class should be friends (required when accessing wrapped_)
  template<class T2, int DIM2, int LEVEL2, typename TRAIT2>
  friend class HTA;

  template<class OP, typename TRAIT2>
  friend class BinExpr;

  /** pointer to the implementation of the HTA - this internal to the class */
  typename TRAIT::TypeImpl* wrapped_;

  template<class T2, int DIM2, int LEVEL2, typename TRAIT2>
  friend ostream& operator<< (ostream& ost, const HTA<T2, DIM2, LEVEL2, TRAIT2>& a);
  
}; // class HTA

#endif /* __HTA_0_H__ */
