/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: IrregularMemMapping.h,v 1.5 2006/10/10 18:03:05 bikshand Exp $
 * Authors: Christoph von Praun
 */


#ifndef __IRREGULAR_MEM_MAPPING_H__
#define __IRREGULAR_MEM_MAPPING_H__

#include <cassert>

#include "MemMapping.h"
#include "LeafMemMapping.h"
#include "Shape.h"
#include "Traits.h"
#include "Tuple.h"

using namespace std;

template <int DIM>
class IrregularMemMapping : public MemMapping<DIM>
{

public:
     
  IrregularMemMapping (const MemMapping<DIM>** children,
		       int n_children,
		       int level, 
		       int offset,
		       int nlevels,
		       Tuple<DIM> leaf_size,
		       const Tuple<DIM>& leafPos, 
		       const Tuple<DIM>* linear_step = NULL) :
    MemMapping<DIM> (TILE, level, offset, nlevels, leaf_size, linear_step), 
    children_(children),
    n_children_(n_children),
    leafPos_(leafPos)
  { 
    children_ = new const MemMapping<DIM>* [n_children];	
    memcpy (children_, children, sizeof(MemMapping<DIM>*) * n_children);
  }
  
public:
 
  virtual ~IrregularMemMapping() 
  {
    for (int i = 0; i < n_children_; ++i) 
      {
	const MemMapping<DIM>* tmp = children_[i];
       	
	if (tmp && tmp->refCount_(-1) == 0) 
	  {	    
	    delete tmp;
	    children_[i] = NULL;
	  }
      }
    
    delete[] children_;
    children_ = NULL;
  }
  
  virtual const MemMapping<DIM>* select (const MemMapping<DIM>** children, 
					 const Tuple<DIM>& new_tiling) const
  {
    assert (false);
    return NULL;
  }


  const MemMapping<DIM>* clone () const
  {
    int offset = 0;
    return clone_helper(offset);
  }

  virtual const MemMapping<DIM>* clone_helper (int offset) const
  {         
    const MemMapping<DIM>** children = new const MemMapping<DIM>* [n_children_];	
    
    int my_offset = offset;

    //initialize children - they follow the same order as the parent
    for (int i = 0; i < n_children_; i++) { 
      if (i > 0) 
	offset += children_[i-1]->linearCard();
      children[i] = children_[i]->clone_helper(offset);
      children[i] ->refCount_(1);                 
    }
    
    MemMapping<DIM>* ret = new IrregularMemMapping<DIM> (children, n_children_, this->level(), my_offset, this->level(), 
							 this->leafSize(), this->leafPos_, NULL);
    delete [] children;
    return ret;
  }
    
  virtual const MemMapping<DIM>& operator[] (int idx) const
  {
    assert (this->level() > 0);
    assert (idx >= 0);
    assert (idx < n_children_);

    return *children_[idx];
  } 
  
  virtual const MemMapping<DIM>* create (typename Tuple<DIM>::Seq& tiling, 
					 const Distribution<DIM>& dist) const 
  {
    //responsibility of Array 
    assert (false);
    return NULL;
  }
  
  /**
   * virtual constructor, creates a new distribution according to a specified tiling
   */
  virtual const MemMapping<DIM>* create (typename Tuple<DIM>::Seq& tiling) const 
  {
    //responsibility of Array 
    assert (false);
    return NULL;
  }
  
  /**
   * @return the position that this array has within the 
   *         flattened overall array.
   */
  virtual const Tuple<DIM>& leafPos() const 
  {
    return leafPos_;
  }
 
private:

  virtual void print_(ostream& ost, int indent = 0) const 
  {
    ost << "IrregularMemMapping<" << DIM << ">(" 
	<< "this=" << this
	<< " level=" << this->level() 
	<< " order=" << this->order() 
	<< " offset=" << this->offset() << ")";
  }

  /**
   * pointers to children_ in row-major order.
   */
  const MemMapping<DIM>** children_;

  /* no. of children */
  int n_children_;

  const Tuple<DIM> leafPos_;
};

#endif /* __LOCAL_MEM_MAPPING_H__ */
