/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: LocalMemMapping.h,v 1.20.4.2 2007/08/21 17:47:03 brodman2 Exp $
 * Authors: Christoph von Praun
 */


#ifndef __LOCAL_MEM_MAPPING_H__
#define __LOCAL_MEM_MAPPING_H__

#include <cassert>

#include "MemMapping.h"
#include "LeafMemMapping.h"
#include "Shape.h"
#include "Traits.h"
#include "Tuple.h"

using namespace std;

template <int DIM>
class LocalMemMapping : public MemMapping<DIM>
{

public:
  
  /** 
   * @param level    hierarchy level of this tile
   * @param offset   offset in the local memory 
   * @param nlevels  the total number of levels of the HTA
   */
  LocalMemMapping(int level, 
		  int offset,
		  int nlevels,
		  const typename Tuple<DIM>::Seq& tiling,
		  Order order, 
		  const Tuple<DIM>& position, 
		  const Tuple<DIM>* linear_step = NULL) :
    MemMapping<DIM> (order, level, offset, nlevels, tiling, linear_step), 
    leafPos_(position * this->leafSize())
  { 
    assert (level > 0);
    int n_children = VectorOps<Tuple<DIM>, DIM>::selfProduct(tiling[level]);
    assert (n_children > 0);
    children_ = new const MemMapping<DIM>* [n_children];
    
    HTA_DEBUG (2, "LocalMemMapping::<init> = " << *this);
    
    switch (order) {
    case TILE:
      initTiled_(n_children, offset, position);
      break;
    case ROW:
      initRowMajor_(position);
      break;
    case COL:
      assert (false); // not implemented
      break;
    default:
      assert(0);
    }
  }

  /*
   * Add this constructor for overlapped tiling where the level-1 leaf size 
   * should be modified. leaf_size is added as a new parameter.
   */
 LocalMemMapping(int level, 
		 int offset,
		 int nlevels,
		 const typename Tuple<DIM>::Seq& tiling,
		 Tuple<DIM> leaf_size, //add this parameter
		 Order order, 
		 const Tuple<DIM>& position, 
		 const Tuple<DIM>* linear_step = NULL) :
    MemMapping<DIM> (order, level, offset, nlevels, tiling, leaf_size, linear_step), 
    leafPos_(position * this->leafSize())
  { 
    assert (level > 0);
    int n_children = VectorOps<Tuple<DIM>, DIM>::selfProduct(tiling[level]);
    assert (n_children > 0);
    children_ = new const MemMapping<DIM>* [n_children];
    
    HTA_DEBUG (2, "LocalMemMapping::<init> = " << *this);
    
    switch (order) {
    case TILE:
      initTiled_(n_children, offset, position);
      break;
    case ROW:
      initRowMajor_(position);
      break;
    case COL:
      assert (false); // not implemented
      break;
    default:
      assert(0);
    }
  }


private:
  
  /**
   * private constructor which is only used by the virtual 
   * factory method select().
   */
  LocalMemMapping (const LocalMemMapping& ld, 
		   int offset,
		   const MemMapping<DIM>** children, 
		   const typename Tuple<DIM>::Seq tiling, 
		   const Tuple<DIM>& linear_step) :
    MemMapping<DIM>(ld.order(), ld.level(), offset, ld.nlevels_, tiling, &linear_step),
    children_(children), leafPos_(ld.leafPos())
  {
    assert (children != NULL);


    // increment reference count for children
    int n_children = VectorOps<Tuple<DIM>, DIM>::selfProduct(tiling[this->level()]);
    for (int i = 0; i < n_children; ++i) {
#ifndef NDEBUG
      assert (children_[i] != NULL);
      for (int j = 0; j < ld.level(); ++ j) {
	assert (children_[i]->tiling()[j] == tiling[j]);
      }
#endif
      children_[i]->refCount_(1);
    }
  }
  

public:
  
  /**
   * "Virtual Constructor"
   * 
   * @param  children  distribution of the children - responsibility to de-allocate this
   *                   memory is transferred to the newly created instance of LocalMemMapping
   * @param  tiling    new tiling information at the current level
   * @return           new instance of LocalMemMapping
   */
  virtual const MemMapping<DIM>* select (const MemMapping<DIM>** children, 
					 const Tuple<DIM>& new_tiling) const
  {
    typename Tuple<DIM>::Seq tmp_tiling;
    
    // adjust the tiling at the level of the children
    for (int i = 0; i < this->level(); ++i)
      tmp_tiling[i] = children[0]->tiling()[i];
    // adjust the tiling at the current level
    tmp_tiling[this->level()] = new_tiling;

    // compute offset
    int offset = children[0]->offset();

    return new LocalMemMapping(*this, offset, children, tmp_tiling, this->linearStep_);
  }
  
  virtual const MemMapping<DIM>* clone () const
  {
    return new LocalMemMapping(this->level(), this->offset(), this->level(), this->tiling(), 
			       this->order(), this->leafPos_ ) ; /// this->leafSize());	
  }
 
  virtual const MemMapping<DIM>* clone_helper (int offset) const
  {
    return new LocalMemMapping(this->level(), offset, this->level(), this->tiling(), 
			       this->order(), this->leafPos_);// / this->leafSize());	
  }

  virtual ~LocalMemMapping() 
  {
    int n_children = VectorOps<Tuple<DIM>, DIM>::selfProduct(this->tiling_[this->level()]);

    for (int i = 0; i < n_children; ++i) 
      {
	const MemMapping<DIM>* tmp = children_[i];
       	
	if (tmp && tmp->refCount_(-1) == 0) 
	  {	    
	    delete tmp;
	    children_[i] = NULL;
	  }

      }

    delete[] children_;
    children_ = NULL;
  }
  
  virtual const MemMapping<DIM>& operator[] (int idx) const
  {
    assert (this->level() > 0);
    assert (idx >= 0);
    assert (idx < (VectorOps<Tuple<DIM>, DIM>::selfProduct(this->tiling_[this->level()])));
    return *children_[idx];
  } 
  
  virtual const MemMapping<DIM>* create (typename Tuple<DIM>::Seq& tiling, 
					 const Distribution<DIM>& dist) const 
  {
    return create (tiling);
  }
  
  /**
   * virtual constructor, creates a new distribution according to a specified tiling
   */
  virtual const MemMapping<DIM>* create (typename Tuple<DIM>::Seq& tiling) const 
  {
    return new LocalMemMapping<DIM>(this->level(),
				    this->offset(), this->nlevels_,
				    tiling, this->order(), Tuple<DIM>::zero);
  }
  
  /**
   * @return the position that this array has within the 
   *         flattened overall array.
   */
  virtual const Tuple<DIM>& leafPos() const 
  {
    return leafPos_;
  }

private:

  void initTiled_(int n_children, int offset, const Tuple<DIM>& position) 
  {
    // calculate size of one tile at leaf level
    Shape<DIM> s (this->tiling_[this->level()]); 
    Tuple<DIM> leaf_size_t = this->leafSize();
    int leaf_size = VectorOps<Tuple<DIM>, DIM>::selfProduct(leaf_size_t);
    int tile_size = leaf_size / n_children;
    assert (leaf_size % n_children == 0);
    int j = 0;
    typename Shape<DIM>::iterator end = s.end();
    
    // then initialize children - they follow the same order as the parent
    for (typename Shape<DIM>::iterator i = s.begin(); i != end; i++, j++) {
      Tuple<DIM> pos_cur_level = (*i);
      int child_offset = offset + j * tile_size;
      MemMapping<DIM>* tmp = (this->level() > 1) ? 
	(MemMapping<DIM>*) new LocalMemMapping<DIM> (this->level() - 1,  child_offset, this->nlevels_, 
						     this->tiling_, this->order_,  pos_cur_level) :
	(MemMapping<DIM>*) new LeafMemMapping<DIM> (child_offset, this->nlevels_, this->tiling_, pos_cur_level); 
      tmp->refCount_(1);
      children_[j] = tmp;
    }
  }

  void initRowMajor_(const Tuple<DIM>& position) {

    Shape<DIM> s (this->tiling_[this->level()]); 
    int j = 0; 
    int child_offset;
    int level = this->level();
    Tuple<DIM> linear_step = this->linearStep_;
    typename Shape<DIM>::iterator end = s.end();

    if (level > 1) {
      linear_step = linear_step * this->computeLeafSize_(level - 1);
      for (typename Shape<DIM>::iterator i = s.begin(); i != end; i++, j++) {
	Tuple<DIM> pos_cur_level = position * this->tiling_[this->level()] + (*i);
	MemMapping<DIM>* tmp;	
	child_offset = VectorOps<Tuple<DIM>, DIM>::dotProduct(pos_cur_level, linear_step);
	//child_offset = VectorOps<Tuple<DIM>, DIM>::dotProduct(pos_cur_level - leafPos_, linear_step);
	tmp = new LocalMemMapping<DIM> (this->level() - 1, child_offset, 
					this->nlevels_, this->tiling_, this->order_,  pos_cur_level, &this->linearStep_);
	tmp->refCount_(1);
	children_[j] = tmp;

	//cout << "Offset " << child_offset << endl;
      }
    } else { // level == 1
      for (typename Shape<DIM>::iterator i = s.begin(); i != end; i++, j++) {	

	Tuple<DIM> pos_cur_level = position * this->tiling_[this->level()] + (*i);
	MemMapping<DIM>* tmp;
	pos_cur_level = pos_cur_level * this->tiling_[0];
	child_offset = this->offset_ + VectorOps<Tuple<DIM>, DIM>::dotProduct(pos_cur_level - leafPos_, linear_step); 
	tmp = new LeafMemMapping<DIM> (child_offset, this->nlevels_, this->tiling_, pos_cur_level, &this->linearStep_);
	tmp->refCount_(1);
	children_[j] = tmp;

	//cout << "Offset 2" << child_offset << endl;
      }
    }
  }

private:

  virtual void print_(ostream& ost, int indent = 0) const 
  {
    ost << "LocalMemMapping<" << DIM << ">(" 
	<< "this=" << this
	<< " level=" << this->level() 
	<< " order=" << this->order() 
	<< " offset=" << this->offset() << ")";
  }

  /**
   * pointers to children_ in row-major order.
   */
  const MemMapping<DIM>** children_;
  
  const Tuple<DIM> leafPos_;
};

#endif /* __LOCAL_MEM_MAPPING_H__ */
