/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 */

/*
 * Version: $Id: MapReduceHTA.h,v 1.3 2006/10/10 18:03:05 bikshand Exp $
 * Authors:Christoph von Praun
 */

#ifndef __MAP_REDUCE_HTA_H__
#define __MAP_REDUCE_HTA_H__

#include <cassert>
#include <Tuple.h>
#include <MapContext.h>

using namespace std;

template <typename T, int DIM, typename TRAIT> class HTA;
template <typename T, int DIM, typename TRAIT> class HTAImpl;
template <typename T, int DIM, typename TRAIT> class HTADistImpl;

namespace OperatorHTA {

  /* TODO: remove this */
  enum MapReduceKey { 
    MR_NONE,
    MR_PROC_OFF,  // processor and local offset in array
    MR_TUPLE      // coordinates at level 0
  }; 

  
  /**
   * base class for map-reduce operators 
   * subclasses should override the required methods
   */
  template<typename T, typename DATA, int DIM, typename TRAIT>
  class MapReduceHTA
  {
    
  public:
    
    /* data type to be exchanged during the reduction */
    typedef DATA XData;

    MapReduceHTA() : level_ (0) {}

    MapReduceHTA(size_t level) : level_ (level) {}

    virtual ~MapReduceHTA() {}

    void reduce (const DATA& data) { } 
        
    /** 
     * Init method 
     */
    void init (const MapContext<T, DIM>& context) {}
    
    /** 
     * map functions - one of those is called by the runtime system, 
     * depending on the key type      
     */
    void map_none (const T& val) { assert (false); }
    void map_proc_off (int proc, int idx, const T& val) { assert (false); }
    void map_tuple (const Tuple<DIM>& t, const T& val) { assert (false); }
    void map_init (const Tuple<DIM>& t) { assert (false); }
  
    /**
     * map functions that take an HTA as its input
     */
    virtual void map_hta_tuple (const Tuple<DIM>& pos, HTA<T, DIM, TRAIT> h) { assert (false); }
    virtual void map_hta_none (HTA<T, DIM, TRAIT> h) { assert (false); }
    virtual void map_hta_none (HTA<T, DIM, TRAIT> h, HTA<T, DIM, TRAIT> r) { assert (false); }
    virtual void map_hta_none (HTA<T, DIM, TRAIT> h, HTA<T, DIM, TRAIT> r, HTA<T, DIM, TRAIT> s) { assert (false); }
    virtual void map_hta_none (HTA<T, DIM, TRAIT> h, HTA<T, DIM, TRAIT> r, 
				HTA<T, DIM, TRAIT> s, 	HTA<T, DIM, TRAIT> s1,
				HTA<T, DIM, TRAIT> s2, 	HTA<T, DIM, TRAIT> s3,
				HTA<T, DIM, TRAIT> s4, HTA<T, DIM, TRAIT> s5) { assert (false); }
    /**
     * Wrapper classes around map_hta_none_intl
     * so that HTAImpl need not know about HTA
     * GDDEBUG: Any suggestions for an alternative design?
     */
    void map_hta_tuple_intl (const Tuple<DIM>& pos, HTAImpl<T, DIM, TRAIT>& h) 
    { 
      this->map_hta_tuple (pos, HTA<T, DIM, TRAIT>(h)); 
    }
    
    void map_hta_none_intl (HTAImpl<T, DIM, TRAIT>& h) 
    { 
      this->map_hta_none (HTA<T, DIM, TRAIT>(h)); 
    }
    void map_hta_none_intl (HTAImpl<T, DIM, TRAIT>& h, HTAImpl<T, DIM, TRAIT>& r) 
    { 
      this->map_hta_none (HTA<T, DIM, TRAIT>(h), HTA<T, DIM, TRAIT>(r)); 
    } 
 
    void map_hta_none_intl (HTAImpl<T, DIM, TRAIT>& h, HTAImpl<T, DIM, TRAIT>& r, HTAImpl<T, DIM, TRAIT>& s) 
    { 
      this->map_hta_none (HTA<T, DIM, TRAIT>(h), HTA<T, DIM, TRAIT>(r),  HTA<T, DIM, TRAIT>(s)); 
    }
    
    void map_hta_tuple_intl (const Tuple<DIM>& pos, HTADistImpl<T, DIM, TRAIT>& h) 
    { 
      this->map_hta_tuple (pos, HTA<T, DIM, TRAIT>(h)); 
    }
    
    void map_hta_none_intl (HTADistImpl<T, DIM, TRAIT>& h) 
    { 
      this->map_hta_none (HTA<T, DIM, TRAIT>(h)); 
    }
    
    void map_hta_none_intl (HTADistImpl<T, DIM, TRAIT>& h, HTADistImpl<T, DIM, TRAIT>& r) 
    { 
      this->map_hta_none (HTA<T, DIM, TRAIT>(h), HTA<T, DIM, TRAIT>(r)); 
    } 
 
    void map_hta_none_intl (HTADistImpl<T, DIM, TRAIT>& h, HTADistImpl<T, DIM, TRAIT>& r, HTADistImpl<T, DIM, TRAIT>& s) 
    { 
      this->map_hta_none (HTA<T, DIM, TRAIT>(h), HTA<T, DIM, TRAIT>(r),  HTA<T, DIM, TRAIT>(s)); 
    }
    
    void map_hta_none_intl (HTADistImpl<T, DIM, TRAIT>& h, HTADistImpl<T, DIM, TRAIT>& r,
			     HTADistImpl<T, DIM, TRAIT>& s, HTADistImpl<T, DIM, TRAIT>& s1,
			     HTADistImpl<T, DIM, TRAIT>& s2, HTADistImpl<T, DIM, TRAIT>& s3,
			     HTADistImpl<T, DIM, TRAIT>& s4, HTADistImpl<T, DIM, TRAIT>& s5) 
    { 
      this->map_hta_none (HTA<T, DIM, TRAIT>(h), HTA<T, DIM, TRAIT>(r),  HTA<T, DIM, TRAIT>(s),
			   HTA<T, DIM, TRAIT>(s1),  HTA<T, DIM, TRAIT>(s2),
			   HTA<T, DIM, TRAIT>(s3),  HTA<T, DIM, TRAIT>(s4),
			   HTA<T, DIM, TRAIT>(s5)); 
    }


    void map_hta_none_intl (HTAImpl<T, DIM, TRAIT>& h, HTAImpl<T, DIM, TRAIT>& r,
			     HTAImpl<T, DIM, TRAIT>& s, HTAImpl<T, DIM, TRAIT>& s1,
			     HTAImpl<T, DIM, TRAIT>& s2, HTAImpl<T, DIM, TRAIT>& s3,
			     HTAImpl<T, DIM, TRAIT>& s4, HTAImpl<T, DIM, TRAIT>& s5) 
    { 
      this->map_hta_none (HTA<T, DIM, TRAIT>(h), HTA<T, DIM, TRAIT>(r),  HTA<T, DIM, TRAIT>(s),
			   HTA<T, DIM, TRAIT>(s1),  HTA<T, DIM, TRAIT>(s2),
			   HTA<T, DIM, TRAIT>(s3),  HTA<T, DIM, TRAIT>(s4),
			   HTA<T, DIM, TRAIT>(s5)); 
    }

    /**
     * The keytype determines the map function that is invoked 
     * by the runtime system.
     */
    virtual MapReduceKey keyType () const = 0;

    /**
     * @return reference to data that participates in the exchange
     */
    virtual const DATA& result () const = 0;
    
    /** 
     * Information that the runtime system might exploit 
     * for optimization.
     */
    bool isThreadSafe () { return false; }

    /** 
     * Information that the runtime system might exploit 
     * for optimization.
     */
    bool doReduce () { return true; }

    size_t level () { return level_; }

  public:

    size_t level_; 
    
    /* it is the responsiblity of the derived classes to initialize this using *init* */
    MapContext<T, DIM> context_;     
  };


}
  
#endif /* __MAP_REDUCE_H__ */
