/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: MemMapping.h,v 1.26.4.2 2007/08/21 17:47:03 brodman2 Exp $
 * Authors: Christoph von Praun
 */


#ifndef __MEM_MAPPING_H__
#define __MEM_MAPPING_H__

#include <cassert>

#include "Distribution.h"
#include "Order.h"
#include "Tuple.h"
#include "util/RefCounted.h"
#include "util/VectorOps.h"

using namespace std;


/**
 * MemMapping objects give information about the physical memory location of an arry / tile.
 * This information consists of the place and the offset address in the global memory block
 * provided by an allocator. This information is necessary when an HTA is allocated / cloned 
 * / permuted.
 */
template <int DIM>
class MemMapping : public RefCounted
{

public:

  /**
   * @param offset in the physical representation
   */
  MemMapping(Order order, 
	     int level, int offset, int nlevels, 
	     const typename Tuple<DIM>::Seq& tiling, 
	     const Tuple<DIM>* linear_step = NULL) : 
    RefCounted(),
    nlevels_(nlevels),
    tiling_(tiling),
    level_(level),
    offset_(offset), 
    order_(order),
    leafSize_(computeLeafSize_()),
    linearStep_( (linear_step == NULL) ? computeLinearStep_() : *linear_step ), 
    linearCard_(computeLinearCard_()),
    isContiguous_(computeIsContiguous_())
  {}
  
  /**
   * @param offset in the physical representation
   */
  MemMapping(Order order, 
	     int level, int offset, int nlevels, 
	     const typename Tuple<DIM>::Seq& tiling, 
	     Tuple<DIM> leaf_size,
	     const Tuple<DIM>* linear_step = NULL) : 
    RefCounted(),
    nlevels_(nlevels),
    level_(level),
    offset_(offset), 
    order_(order),
    tiling_(tiling),  //JG: add for overlapped tiling
    leafSize_(leaf_size),
    linearStep_( (linear_step == NULL) ? computeLinearStep_() : *linear_step ), 
    linearCard_(computeLinearCard_()),
    isContiguous_(computeIsContiguous_())
  {
  }

  virtual ~MemMapping () 
  {}

  /**
   * "Virtual constructor" is used to create a new distribution, e.g., in the 
   * context of a subref operation that returns an HTA of the same level (MATLAB style).
   */
  virtual const MemMapping<DIM>* select (const MemMapping<DIM>** children, const Tuple<DIM>& tiling) const = 0;

  //virtual const MemMapping<DIM>* select (const MemMapping<DIM>** children, const Tuple<DIM>& tiling,
  //				 Order& o) const = 0;

  virtual const MemMapping<DIM>* clone () const  = 0;

  //GBDEBUG: Useful for recursive creation of self 
  virtual const MemMapping<DIM>* clone_helper (int offset) const { assert (false); return NULL; }

  
  /**
   * @return  the distribution of the idx's tile in the 
   *          enumeration of the tiles according to the ordering
   */
  virtual const MemMapping<DIM>& operator[] (int idx) const = 0;
  
  /** 
   * the way the physical memory is partitioned among subordinate 
   * tiles. This is different from the information carried in Shape, 
   * which describes only how the memory of the array is represented 
   * at the current level which is *always* row major.
   */
  Order order() const 
  {
    return order_;
  }

  virtual const MemMapping<DIM>* create (typename Tuple<DIM>::Seq& tiling, 
					 const Distribution<DIM>& dist) const 
  {
    assert (false); // subclasses should override this
    return NULL;
  };


 virtual const MemMapping<DIM>* create (typename Tuple<DIM>::Seq& tiling, 
					 const Distribution<DIM>& dist,
					 Order order) const 
  {
    assert (false);
    return NULL;
  }
  /**
   * @return the position that this array has within the 
   *         flattened overall array.
   */
  virtual const Tuple<DIM>& leafPos () const 
  {
    // this method should be overridden -- the only use of this method is 
    // currently in LeafMemMapping
    assert (false);
    return Tuple<DIM>::zero; // avoid compile warning
  }

  inline const Tuple<DIM>& leafSize () const 
  {
    return leafSize_;
  }

  inline const Tuple<DIM>& linearStep () const 
  {
    //assert (order_ != DIST);
    return linearStep_;
  }

  /**
   * note that this is not an offset in bytes, but in scalar 
   * elements (size of scalar type)
   */
  virtual int offset () const
  {
    assert (order_ != DIST);
    return offset_;
  }

  const typename Tuple<DIM>::Seq& tiling () const 
  {
    return tiling_;
  }

  /**
   * the number of tiling levels below this tile.
   */
  inline int level () const 
  {
    return level_;
  } 
 
  inline int linearCard () const 
  {
    return linearCard_;
  }
    
  inline bool isContiguous () const 
  {
    return isContiguous_;
  }
  
  /**
   * Ordinal value of Tuple. The ordinal value expresses the
   * linear distance (number of elements) of a tuple from the 
   * start of this memory block. Tuple<DIM>::zero  = 0.
   * in must be a coordinate at the leaf level.
   */
  inline int ord (const Tuple<DIM>& in) const
  {
    // assert (this->level() == 0);
    // assert (!this->isDistributed()) ;
    assert (order_ != DIST);
    return VectorOps<Tuple<DIM>, DIM>::dotProduct (in, linearStep_); 
  }

protected:
  
  int computeLinearCard_ () const 
  {
    return leafSize_[0] * linearStep_[0];	
  }

  bool computeIsContiguous_ () const 
  {
    bool ret = false;
    if (order_ != DIST) {
      // least significant dimension is DIM-1
      int acc = leafSize_[DIM-1];
      if (linearStep_[DIM-1] == 1) {
	ret = true;
	for (int i = DIM-2; i >= 0; --i) {
	  if (acc != linearStep_[i]) {
	    ret = false;
	    break;
	  }
	  acc *= leafSize_[i];
	}
      }
    }

    return ret;
  }

  /**
   * @return step in each dimension in the linearized representation at 
   *         level 0
   */
  Tuple<DIM> computeLinearStep_ () const
  {
    // compute linear_step
    Tuple<DIM> linear_step;
    
    // dimension 0 is most significant
    linear_step[DIM-1] = 1;
    for (int i = DIM-1 ; i > 0; i--)
      linear_step[i-1] = linear_step[i] * leafSize_[i];

    //    cout << "hello " << linear_step << endl;
    return linear_step;
  }

  /**
   * number of variables of this tile at level 0.
   */
  Tuple<DIM> computeLeafSize_ (int l = -1) const
  {
    assert (l >= -1);
    const typename Tuple<DIM>::Seq& t = this->tiling_;
    if (l == -1)
      l = level_;
    assert (l >= 0);

    // compute size of all dimensions at the leaf-level
    Tuple<DIM> leaf_size = t[0];
    for (int i = 1; i <= l; i++) 
      leaf_size = leaf_size * t[i];


    return leaf_size;
  }

  /**
   * the total number of levels in the tiling above the 
   * contiguous physical memory 
   */
  const int nlevels_;

  /**
   * top-level tiling specified as tiling_[LEVEL], 
   * leaf-size is tiling_[0]
   */
  const typename Tuple<DIM>::Seq tiling_;

  /* the level to which this tile belongs to */
  const int level_;

  /* the offset of this tile in the physical representation */
  const int offset_;
  
  /** 
   * the way the physical memory is partitioned among subordinate 
   * tiles.
   */
  const Order order_;

  const Tuple<DIM> leafSize_;

  const Tuple<DIM> linearStep_;

  const int linearCard_;
  
  const bool isContiguous_;
  
private:

  template <int D> 
  friend ostream& operator << (ostream& ost, const MemMapping<D>& d);
  virtual void print_(ostream& ost, int indent = 0) const = 0;

};

template<int DIM> 
ostream& operator << (ostream& ost, const MemMapping<DIM>& d) 
{  
  d.print_(ost, 0);
  return ost;
}

#endif /* __MEM_MAPPING_H__ */
