#ifndef __NEIGHBOR_LIST_H__
#define  __NEIGHBOR_LIST_H__

#include "Tuple.h"
#include "util/RefCounted.h"
#include "Overlap.h"

struct NT_record{
  int NT_index;
  int position;
  int send_flag;
  int recv_flag;
  

};

struct NT_list{
  NT_record NT_node[6]; // assert overlapped dim <=3 
  int num_NT;
};
  
template <int DIM>
class NeighborList: public RefCounted 
{
  
public:
  NeighborList(): RefCounted(),
		  NT_list_(NULL),
		  NT_list_len_(0),
		  send_all_(false)
  {}
  
  NeighborList(int len): RefCounted(),
			 NT_list_len_(len)
  {
    NT_list_ = new NT_list[len];
  }

  ~NeighborList(){
    if(NT_list_){
      delete [] NT_list_;
    }
    NT_list_=NULL;
  }
  
    /*
   * the encoding algorithm wants to transfer the 
   * coordinate (a,b,c) into digit abc. To deal with -1, 
   * it adds 1 to all coordinate
   */
  inline int ord(const Tuple<DIM>& coord) const
  {
    int ret=0;
    for(int i=0; i<DIM; i++){
      ret = ret*10 + (coord[i]+1); // add 1 to all coordinates
    }
    return ret;
  }
  
  inline Tuple<DIM> coord(const int position) const
  {
    int remain = position;
    Tuple<DIM> ret;
    for(int i=DIM-1; i>=0; i--){
      ret[i]= remain%10-1;  // minus 1 to get all coordinates
      remain = remain/10;
    }
    return ret;
  }
  
  inline NT_list * get_list(const int i=0) const
  {
    assert(i>=0);
    return &NT_list_[i];
  }
  
  inline bool send_all() const
  {
    return send_all_;
  }

  inline void set_send_all(bool flag)
  {
    send_all_ = flag;
  }
  inline void insert_NT(const int iList, const int iNode, const int nt_idx, const Tuple<DIM>& coord)
  {
    NT_record *curr = &NT_list_[iList].NT_node[iNode];
    curr->NT_index = nt_idx;
    curr->position = ord(coord);
    curr->send_flag = 0;
    curr->recv_flag = 0;
  }

  inline void set_all_send_flag(const int iList, const int value)
  {
    assert(iList >= 0);
    assert(value ==1 || value ==0);
    NT_record *curr =  NT_list_[iList].NT_node;
    int len = NT_list_[iList].num_NT;
    for(int i=0; i<len; i++){
      curr[i].send_flag = value;
    }
  }
  
  inline void set_all_recv_flag(const int iList, const int value)
  {
    assert(iList >= 0);
    assert(value ==1 || value ==0);
    NT_record *curr = NT_list_[iList].NT_node;
    int len = NT_list_[iList].num_NT;
    for(int i=0; i<len; i++){
      curr[i].recv_flag = value;
    }
  }

  inline void set_one_send_flag(const int iList, const int iNode, const int value)
  {
    assert(iList >= 0);
    assert(value ==1 || value ==0);
    NT_list_[iList].NT_node[iNode].send_flag=value;
  }

  inline void set_one_recv_flag(const int iList, const int iNode, const int value)
  {
    assert(iList >= 0);
    assert(value ==1 || value ==0);
    NT_list_[iList].NT_node[iNode].recv_flag=value;
  }

  inline int get_send_flag(const int iList, const int iNode) const
  {
    assert(iList >= 0);
    return NT_list_[iList].NT_node[iNode].send_flag;
  }
  inline int get_recv_flag(const int iList, const int iNode) const
 {
    assert(iList >= 0);
    return NT_list_[iList].NT_node[iNode].recv_flag;
  }

 /*
   * use shape to calculate the linear distance of (1,1,0) with respect to
   * the current liear_idx.
   * Arrange the neighbor list according to the lexicalgraphic order of the 
   * offset index. This is important to guarantee the correctness of sending
   * and receiving data. 
   */
  inline void create_NT_list(const int iList, const int linear_idx, const Shape<DIM>& shape, const Overlap<DIM>& ol)
  {
  
    Tuple<DIM> current = shape.coord(linear_idx);
    int iter =0;
    for(int i=0; i<DIM; i++){
      // for negative dir
      Tuple<DIM> offset = Tuple<DIM>::zero;
      if(ol.neg()[i]!=0){ // has the neighbor in that direction and dimension
	if (current[i]!=0){
	  offset[i]=-1;
	  int linear_off = VectorOps<Tuple<DIM>, DIM>::dotProduct(offset, shape.linearStep());
	  int nt_idx = linear_off + linear_idx;
	  insert_NT(iList, iter++, nt_idx, offset);
	  
	}else if(ol.mode()==PERIODIC){
	  offset[i]=-1+ shape.size()[i];
	  int linear_off = VectorOps<Tuple<DIM>, DIM>::dotProduct(offset, shape.linearStep());
	  offset[i]=-1;
	  int nt_idx = linear_off + linear_idx;
	  insert_NT(iList, iter++, nt_idx, offset);
	}
      }
    }
    for(int i=DIM-1; i>=0; i--){
      Tuple<DIM> offset = Tuple<DIM>::zero;
      
      if(ol.pos()[i]!=0){ // has the neighbor in that direction and dimension
	if (current[i]!=shape.size()[i]-1){
	  offset[i]=1;
	  int linear_off =  VectorOps<Tuple<DIM>, DIM>::dotProduct(offset, shape.linearStep());
	  int nt_idx = linear_off + linear_idx;
	  insert_NT(iList, iter++, nt_idx, offset);
	}else if(ol.mode()==PERIODIC){
	  offset[i]=0;
	  int linear_off =  VectorOps<Tuple<DIM>, DIM>::dotProduct(offset, shape.linearStep());
	  offset[i]=1;
	  int nt_idx = linear_off + linear_idx;
	  insert_NT(iList, iter++, nt_idx, offset);
	}
      }
    }
    NT_list_[iList].num_NT = iter;
  }

  void clear_NT_list(int val = 0)
  {
    for (int i=0; i<NT_list_len_; i++){
      for(int j=0; j<NT_list_[i].num_NT; j++){
	NT_record * p = & (NT_list_[i].NT_node[j]);
	p->send_flag =val;
	p->recv_flag = val;
      }
    }
    send_all_ = (val==0)?false:true;
  }

  void print_NT_list(ostream& ost) const
  {
    for (int i=0; i<NT_list_len_; i++){
      ost<<i<<"th tile, len: "<<NT_list_[i].num_NT<<endl;
      for(int j=0; j<NT_list_[i].num_NT; j++){
	NT_record * p = & (NT_list_[i].NT_node[j]);
	ost<<"NT_idx: "<<p->NT_index<<" pos: "<<p->position<<" "<<coord(p->position);
	ost<<" send: "<<p->send_flag<<" recv: "<<p->recv_flag<<endl;
      }
      ost<<endl;
    }
  }

private:
  
  NT_list * NT_list_;
  int NT_list_len_;
  bool send_all_;
};
  template <int DIM>
ostream& operator << (ostream& ost, const NeighborList<DIM> & nl) 
{
  return nl.print_NT_list(ost);
}

#endif // __NEIGHBOR_LIST_H__
