/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Operators.h,v 1.21 2006/07/11 04:24:06 bikshand Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef __OPERATORS_H__
#define __OPERATORS_H__

#include <cassert>
#include <cmath>
#include <cstdlib>

using namespace std;

/**
 * Namespace for auxiliary operators specific to htalib.
 * Most operators are defined in the C++ std library
 * in the header <functional>
 */
namespace Operator {

  // binary operators ---------------------------------------

  template<typename T>
  struct max 
  {
    T operator() (const T x, const T y) const 
    { 
      return x > y ? x : y;
    }
  };

  template<typename T> 
  struct min 
  {
    T operator() (const T x, const T y) const 
    { 
      return x <= y ? x : y;
    }
  };

  // unary operators -----------------------------------------

  template<typename T>
  struct idem 
  {
    T operator() (const T x) const 
    { 
      return x;
    }
  };
  
  // unary operators -----------------------------------------
  
  template<typename T>
  struct sqrt
  {
    T operator() (const T& x) const { assert (false); return 0; }
  };

  // specialization for double
  template<>
  struct sqrt<double>
  {
    double operator() (const double& x) const { return std::sqrt(x); }
  };
  

  template<typename T>
  struct pow2
  {
    T operator() (const T& x) const { assert (false); return 0; }
  };
  
  // specialization for double
  template<>
  struct pow2<double>
  {
    double operator() (const double& x) const { return x*x; }
  };
  
  template<typename T>
  struct floor
  {
    int operator() (const T& x) const { assert (false); return 0; }
  };

  // specialization for double
  template<>
  struct floor<double>
  {
    double operator() (const double& x) const { return std::floor(x); }
  };
  

  // nullary operators ----------------------------------------

  template<typename T>
  struct cnst 
  {
    T val_;
    cnst(T val) : val_(val) { }
    T operator() () const 
    { 
      return val_;
    }
  };

  template<typename T>
  struct rand 
  {
    rand(T max, T seed) { assert (false); }
    T operator() () const { assert (false); return 0; }
  };

  // specialization for int
  template<>
  struct rand<int>
  {
    const int max_;
    rand(int max, int seed = 0) : max_(max)
    { 
      assert (max > 0);
      std::srand(seed); 
    }
    int operator() () const { return std::rand() % max_; }
  };

  // other conditional operators  ----------------------------------------
  
  template<typename T>
  struct setIfLarger
  {
    T& operator() (T& x, const T& y) const 
    { 
      if (y > x)
	x = y;
      return x;
    }
  };

};



#endif /* __OPERATORS_H__ */
