/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: PermutedDistribution.h,v 1.2 2006/06/14 00:17:34 vonpraun Exp $
 * Authors: Christoph von Praun
 */


#ifndef __PERMUTED_DISTRIBUTION_H__
#define __PERMUTED_DISTRIBUTION_H__

using namespace std;

#include <cassert>
#include "Distribution.h"
#include "Shape.h"
#include "Tuple.h"


/* this is proxy to another distribution */
template <int DIM>
class PermutedDistribution : public Distribution<DIM>
{

public:

  /**
   * @param dist   the distribution to be permuted
   * @param tiles  the tile space to be permuted
   * @param perm   the permutation
   */
  PermutedDistribution (const Distribution<DIM>& dist, const Tuple<DIM>& tiles, const Tuple<DIM>& perm) : 
    toPermute_(dist), 
    toPermuteTiles_(tiles),
    tiles_(tiles.permute(perm)),
    perm_(perm)
  { 
    dist.refCount_(1);
  }

  virtual ~PermutedDistribution () 
  {
    if (toPermute_.refCount_(-1) == 0)
      delete &toPermute_;
  }

  /** 
   * @return the home node corresponding to an index 
   */
  int operator[] (int idx) const { 
    Tuple<DIM> t = tiles_.coord(idx);
    Tuple<DIM> t_perm = t.permute(perm_);
    return toPermute_[t_perm];
  }

  /** 
   * @return the home node corresponding to a coordinate
   */
  int operator[] (const Tuple<DIM>& t) const 
  { 
    assert (tiles_.contains(t));
    Tuple<DIM> t_perm = t.permute(perm_);
    return toPermute_[t_perm];
  }


  /** 
   * @return the 'course' within the node, i.e. the block / tile number 
   * relative to the node.
   */
  virtual int course (int idx) const
  {
    Tuple<DIM> t = tiles_.coord(idx);
    Tuple<DIM> t_perm = t.permute(perm_);
    int idx_tile = toPermuteTiles_.ord(t_perm);
    return toPermute_.course (idx_tile);
  }

  virtual bool operator== (const Distribution<DIM>& rhs) const
  {
    // not yet implemented
    return false;
  }

  virtual int nPlaces () const 
  {
    return toPermute_.nPlaces();
  }

private:

  const Distribution<DIM>& toPermute_;
  const Shape<DIM> toPermuteTiles_;
  const Shape<DIM> tiles_;
  const Tuple<DIM> perm_;

  template <int DIM2> 
  friend ostream& operator << (ostream& ost, const PermutedDistribution<DIM2>& d);

  void print_(ostream& ost, int indent = 0) const 
  {
    ost << "PermutedDistribution()" << endl;
  }

};

template<int DIM> 
ostream& operator << (ostream& ost, const PermutedDistribution<DIM>& d) 
{  
  d.print_(ost);
  return ost;
}

#endif /* __PERMUTED_DISTRIBUTION_H__ */
