/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.
 */

/*
 * Version: $Id: Router.h,v 1.35.4.5 2007/08/18 19:01:31 fraguela Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef  __ROUTER_H__
#define  __ROUTER_H__

#include <cassert>
#include <iostream>

#include "Triplet.h"
#include "HTA.h"
#include "ShapeH.h"
#include "TransExpr.h"

using namespace std;

/** 
 * A router serves to emulate the {}{}.. {} MATLAB operator.
 * It is a temporary object that collectes the 'selection' 
 * information at each level, and then. when used as an lhs 
 * or rhs value - behaves like an array.
 */
template<typename T, int DIM, typename TRAIT>
class Router {

public:
    
  Router (typename TRAIT::TypeImpl& arr, const typename Triplet::Seq& t) : 
    last_(0), 
    arr_(arr), 
    wrapped__(NULL),
    isHomogeneous_(true)
  {
    arr_.refCount_(1);
    val_[last_++] = new Shape<DIM>(t);
  }

  ~Router () 
  {    
    if (this->wrapped__ && this->wrapped__->refCount_(-1) == 0)
    {
      delete wrapped__;
    }
    
    if (arr_.refCount_(-1) == 0) 
      delete &arr_;
    
    //cout << "Me followed by " << endl;
    for (int i = 0; i < last_; i++) 
      if (!static_cast<const ShapeH<DIM>*> (val_[i]))
	delete val_[i];     
  }
  
  /** 
   * Copy constructor - adjusts ref counts
   */
  Router<T, DIM, TRAIT> (const Router<T, DIM, TRAIT>& rhs) : 
    last_(rhs.last_), 
    arr_(rhs.arr_), 
    wrapped__(rhs.wrapped__),
    isHomogeneous_(rhs.isHomogeneous_)
  {
    arr_.refCount_(1);
    if (this->wrapped__)
      this->wrapped__->refCount_(1);

    for (int i = 0; i < last_; i++) 
      val_[i] = rhs.val_[i];//->clone();     
  }

  Router<T, DIM, TRAIT>& operator() ()
  {
    Triplet::Seq tmp;
  
    //JG: add the case to index the whole region including the shadow regions.
    if(arr_.level()-last_ ==0 && arr_.ol()!=NULL){
      Tuple<DIM> tmp = arr_.data_[0]->shape().size();
      Tuple<DIM> neg = arr_.ol()->neg();
      Tuple<DIM> pos = arr_.ol()->pos();
      tmp = tmp+neg+pos;
      Shape<DIM> s(tmp);
      val_[last_] = new Shape<DIM>(s-neg);
      //      val_[last_]->relocate(-neg);
    }else{
      const typename Tuple<DIM>::Seq& tiling = arr_.tiling();
      const Tuple<DIM> tiling_cur_level = tiling[arr_.level() - last_];
      
      for (int i = 0; i < DIM; ++i)
	tmp[i] = Triplet(0, tiling_cur_level[i] - 1);
      
      val_[last_] = new Shape<DIM>(tmp);
    }
    last_++;
    return *this;
  }
  
  Router<T, DIM, TRAIT>& operator() (const typename Triplet::Seq& t)
  {
    val_[last_++] = new Shape<DIM>(t);
    return *this;
  }

  //Simple syntax
  Router<T, DIM, TRAIT>& operator() (const Triplet &t)
  {
    val_[last_++] = new Shape<DIM>(typename Triplet::Seq(t));
    return *this;
  }

  Router<T, DIM, TRAIT>& operator() (const Triplet &t1, const Triplet &t2)
  {
    val_[last_++] = new Shape<DIM>(typename Triplet::Seq(t1, t2));
    return *this;
  }

  Router<T, DIM, TRAIT>& operator() (const Triplet &t1, const Triplet &t2, const Triplet &t3)
  {
    val_[last_++] = new Shape<DIM>(typename Triplet::Seq(t1, t2, t3));
    return *this;
  }

  Router<T, DIM, TRAIT>& operator() (const Triplet &t1, const Triplet &t2, const Triplet &t3, const Triplet &t4)
  {
    val_[last_++] = new Shape<DIM>(typename Triplet::Seq(t1, t2, t3, t4));
    return *this;
  }

  Router<T, DIM, TRAIT>& operator() (const Shape<DIM>& s)
  {
    //if any one Shape is hierarchical, the result of wrapped_ is assumed to be non-homogenous
    if (isHomogeneous_)
      {
	if (static_cast<const ShapeH<DIM>*> (&s))
	  isHomogeneous_ = false;
      }

    val_[last_++] = s.clone();
    return *this;
  }

  BinExpr<divides<T>, TRAIT> operator/ (Router<T, DIM, TRAIT>& t)
  {
    typename TRAIT::TypeImpl* lhs = this->wrapped_();
    typename TRAIT::TypeImpl* rhs = t.wrapped_();
    
    
    return BinExpr<divides<T>, TRAIT>(*lhs, *rhs);    
  }

  template<typename OP>
  BinExpr<divides<T>, TRAIT> operator/ (const BinExpr<OP, TRAIT>& rhs) const
  {
    HTA<T, DIM, TRAIT> rhs_eval = rhs.eval();
    typename TRAIT::TypeImpl* lhs = this->wrapped_();
    
    return BinExpr<divides<T>, TRAIT>(*lhs, rhs_eval, rhs_eval.wrapped_);
  }

  BinExpr<multiplies<T>, TRAIT> operator* (Router<T, DIM, TRAIT>& t)
  {
    typename TRAIT::TypeImpl* lhs = this->wrapped_();
    typename TRAIT::TypeImpl* rhs = t.wrapped_();
    
    
    return BinExpr<multiplies<T>, TRAIT>(*lhs, *rhs);    
  }

  BinExpr<multiplies<T>, TRAIT> operator* (const HTA<T, DIM, TRAIT>& rhs) const
  {
    typename TRAIT::TypeImpl* lhs = this->wrapped_();
    
    return BinExpr<multiplies<T>, TRAIT>(*lhs, *rhs.wrapped_);    
  }

  template<typename OP>
  BinExpr<multiplies<T>, TRAIT> operator* (const BinExpr<OP, TRAIT>& rhs) const
  {
    HTA<T, DIM, TRAIT> rhs_eval = rhs.eval();
    typename TRAIT::TypeImpl* lhs = this->wrapped_();
    
    return BinExpr<multiplies<T>, TRAIT>(*lhs, rhs_eval, rhs_eval.wrapped_);
  }

  /** pointwise from binary expression */
  template <typename OP>
  Router<T, DIM, TRAIT>& operator+= (const BinExpr<OP, TRAIT>& expr) 
  {
    expr.evalAndAccum(*(this->wrapped_()), plus<T>());
    return *this;
  } 

  BinExpr<plus<T>, TRAIT> operator+ (Router<T, DIM, TRAIT>& t)
  {
    typename TRAIT::TypeImpl* lhs = this->wrapped_();
    typename TRAIT::TypeImpl* rhs = t.wrapped_();
    
    
    return BinExpr<plus<T>, TRAIT>(*lhs, *rhs);    
  }

  template <int TYPE>
  Router<T, DIM, TRAIT>& operator= (const TransExpr<TRAIT, TYPE>& expr) 
  {
    expr.eval(*this->wrapped_());
    return *this;
  }

  BinExpr<plus<T>, TRAIT> operator+ (const HTA<T, DIM, TRAIT>& rhs) const
  {
    typename TRAIT::TypeImpl* lhs = this->wrapped_();
    
    return BinExpr<plus<T>, TRAIT>(*lhs, *rhs.wrapped_);    
  }
  
  template<typename OP>
  BinExpr<plus<T>, TRAIT> operator+ (const BinExpr<OP, TRAIT>& rhs) const
  {
    HTA<T, DIM, TRAIT> rhs_eval = rhs.eval();
    typename TRAIT::TypeImpl* lhs = this->wrapped_();
    
    return BinExpr<plus<T>, TRAIT>(*lhs, rhs_eval, rhs_eval.wrapped_);
  }

  BinExpr<minus<T>, TRAIT> operator- (const HTA<T, DIM, TRAIT>& rhs) const
  {
    typename TRAIT::TypeImpl* lhs = this->wrapped_();
    
    return BinExpr<minus<T>, TRAIT>(*lhs, *rhs.wrapped_);    
  }


  BinExpr<minus<T>, TRAIT> operator- (Router<T, DIM, TRAIT>& t)
  {
    typename TRAIT::TypeImpl* lhs = this->wrapped_();
    typename TRAIT::TypeImpl* rhs = t.wrapped_();
    
    
    return BinExpr<minus<T>, TRAIT>(*lhs, *rhs);    
  }

  template<typename OP>
  BinExpr<minus<T>, TRAIT> operator- (const BinExpr<OP, TRAIT>& rhs) const
  {
    HTA<T, DIM, TRAIT> rhs_eval = rhs.eval();
    typename TRAIT::TypeImpl* lhs = this->wrapped_();
    
    return BinExpr<minus<T>, TRAIT>(*lhs, rhs_eval, rhs_eval.wrapped_);
  }

  Router<T,DIM,TRAIT> operator-() const
  {
    typename TRAIT::TypeImpl* lhs = this->wrapped_();
    TRAIT::TypeImpl::map(negate<T>(), *lhs, *lhs);
    return *this;
  }
  /**
   * map (one argument)
   * I have to rename hmap to avoid name collision with another 'map' 
   */
  template <typename OP>
  void hmap (OP op, 
	     Router<T, DIM, TRAIT>& arg1, 
	     int level, 
	     bool* mask = NULL)
  {
    arr_.hmap (last_, 
	      val_, 
	      arg1.val_,
	      op, 
	      level, 
	      &arg1.arr_, 
	      mask);
  }

  template <typename OP>
  void hmap (OP op, Router<T, DIM, TRAIT>& arg1, 
	     Router<T, DIM, TRAIT>& arg2,
	     int level, bool* mask = NULL)
  {
    arr_.hmap (last_, 
	       val_, 
	       arg1.val_,
	       arg2.val_,
	       op, 
	       level, 
	       (&(arg1.arr_)), 
	       (&(arg2.arr_)), 
	       mask);
  }

  template <typename OP>
  void hmap (OP op, Router<T, DIM, TRAIT>& arg1, 
	     Router<T, DIM, TRAIT>& arg2,
	     Router<T, DIM, TRAIT>& arg3,
	     int level, bool* mask = NULL)
  {
    arr_.hmap (last_, 
	      val_, 
	      arg1.val_,
	      arg2.val_,
	      arg3.val_,
	      op, 
	      level, 
	      (&(arg1.arr_)), 
	      (&(arg2.arr_)), 
	      (&(arg3.arr_)), 
	      mask);
  }

  template <typename OP>
  void hmap (OP op, Router<T, DIM, TRAIT>& arg1, 
	     Router<T, DIM, TRAIT>& arg2,
	     Router<T, DIM, TRAIT>& arg3,
	     Router<T, DIM, TRAIT>& arg4,
	     int level, bool* mask = NULL)
  {
    arr_.hmap (last_, 
	      val_, 
	      arg1.val_,
	      arg2.val_,
	      arg3.val_,
	      arg4.val_,
	      op, 
	      level, 
	      (&(arg1.arr_)), 
	      (&(arg2.arr_)), 
	      (&(arg3.arr_)), 
	      (&(arg4.arr_)), 
	      mask);
  }
  
  template <typename OP>
  void hmap (OP op, Router<T, DIM, TRAIT>& arg1, 
	     Router<T, DIM, TRAIT>& arg2,
	     Router<T, DIM, TRAIT>& arg3,
	     Router<T, DIM, TRAIT>& arg4,
	     Router<T, DIM, TRAIT>& arg5,
	     int level, bool* mask = NULL)
  {
    arr_.hmap (last_, 
	       val_, 
	       arg1.val_,
	       arg2.val_,
	       arg3.val_,
	       arg4.val_,
	       arg5.val_,
	       op, 
	       level, 
	       (&(arg1.arr_)), 
	       (&(arg2.arr_)), 
	       (&(arg3.arr_)), 
	       (&(arg4.arr_)), 
	       (&(arg5.arr_)), 
	       mask);
  }

 template <typename OP>
  void hmap (OP op, Router<T, DIM, TRAIT>& arg1, 
	     Router<T, DIM, TRAIT>& arg2,
	     Router<T, DIM, TRAIT>& arg3,
	     Router<T, DIM, TRAIT>& arg4,
	     Router<T, DIM, TRAIT>& arg5,
	     Router<T, DIM, TRAIT>& arg6,
	     Router<T, DIM, TRAIT>& arg7,
	     int level, bool* mask = NULL)
  {
    arr_.hmap (last_, 
	       val_, 
	       arg1.val_,
	       arg2.val_,
	       arg3.val_,
	       arg4.val_,
	       arg5.val_,
	       arg6.val_,
	       arg7.val_,
	       op, 
	       level, 
	       (&(arg1.arr_)), 
	       (&(arg2.arr_)), 
	       (&(arg3.arr_)), 
	       (&(arg4.arr_)), 
	       (&(arg5.arr_)), 
	       (&(arg6.arr_)), 
	       (&(arg7.arr_)), 
	       mask);
  }
  
#if 0
  HTA<T, DIM, TRAIT> pow2 () const
  {
    // allocate the result array
    typename TRAIT::TypeImpl* result = this->wrapped_()->clone();
    TRAIT::TypeImpl::map(Operator::pow2<T>(), *this->wrapped_(), *result);
    return *result;
  }
  
#endif
  
  Router<T, DIM, TRAIT>& operator= (const T& t)
  {
    *wrapped_() = t;    // forward assignment to array
    return *this;
  }

  Router<T, DIM, TRAIT>& operator= (const Router<T, DIM, TRAIT>& t)
  {
    typename TRAIT::TypeImpl* lhs = this->wrapped_();
    typename TRAIT::TypeImpl* rhs = t.wrapped_();
    
    
    assign_(*lhs, *rhs);
    return *this;
  }

  Router<T, DIM, TRAIT>& operator= (const HTA<T, DIM, TRAIT>& t)
  {
    typename TRAIT::TypeImpl* lhs = this->wrapped_();
    typename TRAIT::TypeImpl* rhs = t.wrapped_;
    assign_(*lhs, *rhs);
    return *this;
  }

  /**
   * Conversion of a router to an array.
   */ 
    
  operator typename TRAIT::Type () const 
    {
      typename TRAIT::TypeImpl* lhs = this->wrapped_();
      
      return typename TRAIT::Type(*lhs);
    }
  

  typename TRAIT::TypeImpl * value() const
  {
    return this->wrapped_();
  }
private:
  
  void assign_ (typename TRAIT::TypeImpl& lhs, const typename TRAIT::TypeImpl& rhs) 
  {
    // communicate 
    rhs.communicate_(lhs.dist(), &lhs); 

    // forward assignment to array
    // lhs = rhs;
    
    // invalidate
    // rhs.invalidate_();
  }
  
  template<class T2, int DIM2, typename TRAIT2>
  friend class HTA;

  template<typename T1, int DIM1, typename TRAIT1>
  friend ostream& operator<< (ostream& ost, const Router<T1, DIM1, TRAIT1>& a) ;

  
  template<class T1, int DIM1, class TRAIT1>
  friend BinExpr<multiplies<T1>, TRAIT1> operator* (const T1& val, const Router<T1, DIM1, TRAIT1>& arr);
    template<class T1, int DIM1, typename TRAIT1> 
    friend HTA<T1, DIM1, TRAIT1> operator+ (const T1& val, const Router<T1, DIM1, TRAIT1>& arr) ;
  template<class T1, int DIM1, typename TRAIT1> 
  friend HTA<T1, DIM1, TRAIT1> operator- (const T1& val, const Router<T1, DIM1, TRAIT1>& arr) ;
  template<class T1, int DIM1, typename TRAIT1> 
  friend HTA<T1, DIM1, TRAIT1> operator/ (const T1& val, const Router<T1, DIM1, TRAIT1>& arr) ;
    
  /** forbid heap allocation */
  void* operator new(size_t size) throw () { assert(0); return NULL; }

  /**  forbid heap allocation */
  void operator delete(void* to_delete) { assert (0); }
  
  typename TRAIT::TypeImpl* wrapped_() const 
  {
    Router<T, DIM, TRAIT>* this_non_const = const_cast<Router<T, DIM, TRAIT>*>(this);
    assert (last_ >= 0);
    if (wrapped__ == NULL) {
      if (last_ == 1) 
	{
	  this_non_const->wrapped__ = arr_.tileAt(*val_[0]);
	}
      else 
	{ 
	  if (isHomogeneous_) //result is a homogenous hta
	    {
	      this_non_const->wrapped__ = arr_.tileAt(last_, val_);	
	    } else { //result is a non-homogenous hta
	      this_non_const->wrapped__ = arr_.tileAt(last_, val_, Tuple<DIM>::zero);	
	      //this_non_const->wrapped__->initMemMapping ();
	    }
	}
      
      this_non_const->wrapped__->refCount_(1);
      
    }
    //JG: add hook to OL-HTA if needed
    wrapped__->addParent(&arr_, last_, val_);
    return wrapped__;
  }
  
  int last_;
  /* val_[0] is shape at top-level, val_[1] one level below, etc .. */
  //gti define a max level instead of 5 if we want to allocate on stack; 
  //gti else use dynamic allocation; ; check speed issues
  const Shape<DIM>* val_[5];//gti ALLOCATE SPACE FOR THIS IN THE CONSTRUCTOR this->level() + 1
  typename TRAIT::TypeImpl& arr_;
  typename TRAIT::TypeImpl* wrapped__;
  bool isHomogeneous_;
};

template<typename T, int DIM, typename TRAIT>
ostream& operator<< (ostream& ost, const Router<T, DIM, TRAIT>& a) 
{
  typename TRAIT::Type tmp (*a.wrapped_());
  ost << tmp; // forward message to array
  return ost;
}  


template<class T, int DIM, typename TRAIT> 
BinExpr<multiplies<T>, TRAIT> operator* (const T& val, const Router<T, DIM, TRAIT>& arr) 
{
  // allocate the result array
  return BinExpr<multiplies<T>, TRAIT>(*arr.wrapped_(), val);
}


template<class T, int DIM, typename TRAIT> 
HTA<T, DIM, TRAIT> operator+ (const T& val, const Router<T, DIM, TRAIT>& arr) 
{
  // allocate the result array
  typename TRAIT::TypeImpl* lhs = arr.wrapped_();
  typename TRAIT::TypeImpl* result = lhs->clone();
  TRAIT::TypeImpl::map(binder1st<plus<T> >(plus<T>(), val), *lhs, *result);
  return *result;
}


template<class T, int DIM, typename TRAIT> 
HTA<T, DIM, TRAIT> operator- (const T& val, const Router<T, DIM, TRAIT>& arr) 
{
  // allocate the result array
  typename TRAIT::TypeImpl* lhs = arr.wrapped_();
  typename TRAIT::TypeImpl* result = lhs->clone();
  TRAIT::TypeImpl::map(binder1st<minus<T> >(minus<T>(), val), *lhs, *result);
  return *result;
}


template<class T, int DIM, typename TRAIT> 
HTA<T, DIM, TRAIT> operator/ (const T& val, const Router<T, DIM, TRAIT>& arr) 
{
  // allocate the result array
  typename TRAIT::TypeImpl* lhs = arr.wrapped_();
  typename TRAIT::TypeImpl* result = lhs->clone();
  TRAIT::TypeImpl::map(binder1st<divides<T> >(divides<T>(), val), *lhs, *result);
  return *result;
}




#endif /* __ROUTER_H__ */
