/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Shape.h,v 1.159.4.3 2007/08/21 17:47:03 brodman2 Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef  __SHAPE_H__
#define  __SHAPE_H__

#include <cassert>
#include <iostream>

#include "Triplet.h"
#include "Tuple.h"
#include "debug/Tracing.h"
#include "util/Iterator.h"
#include "util/Sequence.h"
#include "util/VectorOps.h"


using namespace std;


/**
 * Shape class that defines a multi-diimensional coordinate system and 
 * its mapping to an underlying dense, linearized representation.
 * The mapping to ordinals assumes that 0 is the least signifcant 
 * dimension.
 *
 * Instances of Shape are immutable.
 */
template <int DIM> class Shape {
    
public:

  // support for ,-syntax and sequences 
  typedef Sequence<Shape<DIM> > Seq;
  Seq operator,(const Shape<DIM>& s) 
  {
    return Seq(*this, s);
  }

private:
  
  class ShapeIterator_spec {
  public:
    typedef int                          iterator_type;
    typedef Shape<DIM>                   collection_type;
    typedef random_access_iterator_tag   iterator_category;  
    typedef Tuple<DIM>                   value;
    typedef value*                       pointer;
    typedef const value*                 const_pointer;
    typedef value&                       reference;
    typedef const value&                 const_reference;
  };

  class ShapeIterator : public Iterator<ShapeIterator_spec> {
  private:

    const typename ShapeIterator_spec::collection_type& coll_;
    bool valid_;
    int low_;
    Tuple<DIM> currentTuple_;

  public:

    ShapeIterator (const typename ShapeIterator_spec::collection_type& coll, int cur) : 
      Iterator<ShapeIterator_spec>(cur), 
      coll_(coll), 
      valid_(true),
      currentTuple_()
    { 
      assert (cur >= 0 || cur == -1); 
      Tuple<DIM> tmp_tuple;
      int tmp = cur;

      // compute the current tuple from cur, which  is the index 
      // within the iteration space, 0 if currentTuple_ == coll_.low_.

      for(int i = 0; i <  DIM - 1 ; i++) {
      	tmp_tuple[i] = tmp / coll_.linearStep_[i];  
	tmp -= tmp_tuple[i] * coll_.linearStep_[i]; 
      }
      tmp_tuple[DIM-1] = tmp;

      currentTuple_ = tmp_tuple + coll.low_;
      low_ = VectorOps<Tuple<DIM>, DIM>::dotProduct (coll_.low_, coll_.linearStep_);
    }

    /**
     * Prefix increment 
     */
    ShapeIterator& operator++() 
    {     
      advance_(1);
      return *this;
    }
    
    /**
     * Postfix increment 
     */
    ShapeIterator operator++(int) 
    {
      assert (valid_);
      ShapeIterator tmp(*this);
      advance_(1);
      return tmp;
    }

 
    /**
     * Postfix increment 
     */
    ShapeIterator operator--(int) 
    {
      assert (valid_);
      ShapeIterator tmp(*this);
      stepback_(1);
      return tmp;
    }

    ShapeIterator& operator+=(int n) {
      advance_(n);
      return *this;
    }
    
    typename ShapeIterator_spec::value operator*() const 
    {
      assert (valid_);
      return currentTuple_;      
    }
    
    typename ShapeIterator_spec::value operator[](int n) const 
    { 
      assert (valid_);
      ShapeIterator tmp = *this;
      tmp.advance_(n);
      return *tmp;
    }
    
    int ord() const 
    {
      assert (valid_);
      return this->current_;
    }
    
    /* This is for debugging only. Not doen in an overload of operator<<
     * because it seems that gcc 3.4 does not manage inner class declarations of 
     * template classes correctly */
    char print_(ostream& ost) const {
      ost << "Shape<" << DIM << ">::iterator("
	  << "current=" << this->current_ 
	  << " currentTuple=" << this->currentTuple_ 
	  << " valid=" << this->valid_ << ")";
      return '\0';
    }

  private:
    
    void advance_(int n) 
    {
      int acc = 0;
      bool step_done = false;

      // least significant dimension is at DIM-1
      for (int i = DIM-1; i >= 0; --i) {
	if (!step_done) {
	  if (currentTuple_[i] + (n * coll_.step_[i]) > coll_.high_[i]) {
	    currentTuple_[i] = coll_.low_[i];
	  } else {
	    currentTuple_[i] += n * coll_.step_[i];
	    step_done = true;
	  }
	}
	acc += currentTuple_[i] * coll_.linearStep_[i];
      }

      if (!step_done) {
	// iterator hit the end - make it look like "end()"
	this->current_ = -1;
	valid_ = false;
      } else {
	assert (acc > this->current_);
	this->current_ = acc - low_;
      }
      HTA_DEBUG(3, "Shape::Iterator::advance this=" << this << 
		" acc=" << acc << 
		" current=" << this->current_ << 
		" currentTuple=" << this->currentTuple_  << 
		" step_done=" << step_done);
    }

    void stepback_(int n) 
    {
      int acc = 0;
      bool step_done = false;
      
      // least significant dimension is at DIM-1
      for (int i = DIM-1; i >= 0; --i) {
	if (!step_done) {
	  if (currentTuple_[i] - (n * coll_.step_[i]) < coll_.low_[i]) {
	    currentTuple_[i] = coll_.high_[i];
	  } else {
	    currentTuple_[i] -= n * coll_.step_[i];
	    step_done = true;
	  }
	}
	acc += currentTuple_[i] * coll_.linearStep_[i];
      }
      
      if (!step_done) {
	// iterator hit the end - make it look like "end()"
	this->current_ = -1;
	valid_ = false;
      } else {
	//cout << acc << " " << this->current_ << endl;
	assert (acc < this->current_);
	this->current_ = acc - low_;
      }
      HTA_DEBUG(3, "Shape::Iterator::advance this=" << this << 
		" acc=" << acc << 
		" current=" << this->current_ << 
		" currentTuple=" << this->currentTuple_  << 
		" step_done=" << step_done);
    }
  };
    
public:

  typedef ShapeIterator iterator;

  /** 
   * Default constructor
   */
  Shape() :
    size_(),
    high_(),
    low_(),
    step_(),
    linearStep_(),
    localStep_(),
    mod_()
  { }

  /** 
   * Constructor
   * 
   * TODO: case like 8:-1:4 is not yet handled!
   * 
   * @param t  array of Triplets of length DIM-1 that specifies the dimensions of 
   *           this Shape. The most significant dimension should be found at index 0.
   */
  Shape (const typename Triplet::Seq& t) :
    size_(), 
    high_(),
    low_(), 
    step_(), 
    mod_()
  {
    for (int i = 0; i < DIM; ++ i)
      dims_[i] = t[i];
    init_(t.raw());
  }

  /** 
   * Convenience constructor.
   *
   * @param  t upper coordinates of the shape - not included
   * @return shape with lower bound (0,0,..), step (1,1,..).
   */
  Shape (const Tuple<DIM>& t)
  {
    for (int i = 0; i < DIM; i++) {
      dims_[i] = Triplet(0, t[i] - 1); 
    }
    init_(dims_);
  }

   /** 
   * Convenience constructor.
   *
   * @param  begin lower coordinates of shape
   * @param  end   upper coordinates of the shape - not included
   * @return       shape with step (1,1,..).
   */
  Shape (const Tuple<DIM>& begin, const Tuple<DIM> &end)
  {
    for (int i = 0; i < DIM; i++)
      dims_[i] = Triplet(begin[i], end[i] - 1);
    init_(dims_);
  }

  /**
   * Copy constructor 
   */
  Shape (const Shape<DIM>& s) : 
    size_ (s.size_), 
    high_ (s.high_), 
    low_ (s.low_), 
    step_ (s.step_), 
    linearStep_ (s.linearStep_),
    localStep_ (s.localStep_),
    mod_ (s.mod_) //,
    //dims_ (s.dims_) JCB added explicit copy
  { 
    for (int i = 0; i < DIM; i++) {
      dims_[i] = s.dims_[i];
    }
    HTA_DEBUG(2, "Shape::<init> this=" << this << " linearStep_=" << linearStep_ << " high:" << high_);
  }

  /**
   * @return a new shape that masks out one dimension
   */
  Shape<DIM> mask (int dim) const 
  {
    assert (dim >=0 && dim < DIM);

    typename Triplet::Seq tmp;
    for (int i = 0; i < DIM; ++i) 
      tmp[i] = (i == dim) ? Triplet(0, 0, 1) : this->dims_[i];
    return Shape<DIM>(tmp);
  }

  virtual ~Shape() {}
private:

  /**
   * Initializer method
   * @param t  t[0] is the most significant dimension.
   */
  void init_(const Triplet* t) 
  {
    int highs[DIM];
    int sizes[DIM];
    int lows [DIM];
    int steps[DIM];
    int linear_sizes[DIM];
    
    for (int i = 0; i < DIM; ++i) {
      highs[i]        = t[i].high();
      linear_sizes[i] = t[i].high() - t[i].low() + 1; 
      sizes[i]        = t[i].size();
      lows[i]         = t[i].low();
      steps[i]        = t[i].step();
      mod_[i]         = t[i].mod();
    }
    
    size_ = Tuple<DIM>(sizes);
    high_ = Tuple<DIM>(highs);
    low_ =  Tuple<DIM>(lows);
    step_ = Tuple<DIM>(steps);

    // memorize values that aid the linearization function
    // DIM-1 is the *least* significant dimension in row-major 
    // (row is in dimension 0)

    // compute cumulative linearStep
    linearStep_[DIM-1] = 1;
    for (int i = DIM-1; i > 0; --i) 
      linearStep_[i-1] = linear_sizes[i] * linearStep_[i];
    
    localStep_[DIM-1] = 1;
    for (int i = DIM-1; i > 0; --i) 
      localStep_[i-1] = sizes[i] * localStep_[i];

    HTA_DEBUG(2, "Shape::init_ this=" << this << " linearStep_=" << linearStep_ << " high:" << high_);
  }

public:

  Shape<DIM> setLinearStep (const Shape<DIM>& s) const
  {
    Shape<DIM> ret(*this);
    ret.linearStep_ = s.linearStep_;
    return ret;
  }

  virtual const Shape<DIM>* clone () const
  {
    assert (false);
    return NULL;
  }

  virtual const Shape<DIM>& operator [] (const Tuple<DIM>& i) const
  {
    return *this;
  }


  virtual const Shape<DIM>& shapeAt (int i) const
  {
    return *this;
  }

  /**
   * Accessor to dimension 
   *
   * @param    i dimension index (0 .. DIM-1)
   * @return   triplet that characterizes dimension i of this shape.
   *           most significant dimenation in row-major layout is found 
   *           at index 0.
   */
  const Triplet& operator[] (int d) const 
  {
    assert(d >= 0 && d < DIM);
    return dims_[d]; 
  }
  
  /**
   * @param  s1    shape
   * @param  s2    shape
   * @return       true if both shapes are conformable.
   */
  static inline bool conformable (const Shape<DIM>& s1, const Shape<DIM>& s2) 
  {
     if (&s1 == &s2)
      return true;
     else 
      return (s1.size_ == s2.size_);
  }

  /**
   * @param  s1    shape
   * @param  s2    shape
   * @return       true if both shapes can be accessed with the same non-linear iterator.
   *               the property is reflexive and transitive.
   */
  static inline bool iteratorConformable_ (const Shape<DIM>& s1, const Shape<DIM>& s2) 
  {
     if (&s1 == &s2)
      return true;
    else 
      return (s1.linearStep_ == s2.linearStep_);
  }

  inline bool contains (const Tuple<DIM>& t) const 
  {
    bool ret = (t >= low_ && t <= high_);
    if (ret && step_ != Tuple<DIM>::one) {
      for (int i=0; ret && i < DIM; ++i) 
	ret &= (t[i] - low_[i]) % step_[i] == 0;
    }
    return ret;
  }

  inline bool contains (const Shape<DIM>& s) const 
  {
    // current limitation 
    assert (s.step_ == Tuple<DIM>::one && step_ == Tuple<DIM>::one);
    return s.low_ >= low_ && s.high_ <= high_;
  }


  /**
   * Equality operator
   */
  inline bool operator== (const Shape<DIM>& other) const 
  {  
    if (&other == this)
      return true;
    else 
      return (other.low_ == low_ &&
	      other.high_ == high_ && 
	      other.step_ == step_);
  }

  /**
   * Inequality operator
   */
  inline bool operator!= (const Shape<DIM>& other) const 
  {  
    return ! (this->operator==(other));
  }
  
  /**
   * Accessor
   */
  inline const Tuple<DIM>& size() const // __attribute__((always_inline))
  {
    return size_;
  }

  inline const Tuple<DIM>& lows() const // __attribute__((always_inline))
  {
    return low_;
  }
  
  inline const Tuple<DIM>& highs() const // __attribute__((always_inline))
  {
    return high_;
  }
  
  /**
   * Accessor
   *
   * @return  the number of elements returned by an iteration over this shape
   */
  inline int card () const
  {
    return VectorOps<Tuple<DIM>, DIM>::selfProduct(size_);
  }

  /**
   * Accessor
   */
  inline Tuple<DIM>& linearStep () // __attribute__((always_inline))
  {
    return linearStep_;
  }

  inline Tuple<DIM>& step () // __attribute__((always_inline))
  {
    return step_;
  }

  inline const Tuple<DIM>& step () const // __attribute__((always_inline))
  {
    return step_;
  }

  inline const Tuple<DIM>& linearStep () const // __attribute__((always_inline))
  {
    return linearStep_;
  }

  /**
   * Accessor
   */
  inline const Tuple<DIM>& high () const // __attribute__((always_inline))
  {
    return high_;
  }

  /**
   * Accessor
   */
  inline const Tuple<DIM>& low () const // __attribute__((always_inline))
  {
    return low_;
  }

  /**
   * Ordinal value of Tuple. Ordinal value does not include offset
   * and does not consider step. If the shape starts at (5, 3), 
   * then (5, 4) has ord 1 on row-major. 
   * The ordinal value of low() is always 0.
   * The ordinal value of high() is always <= linearCard().
   */
  inline int ord (const Tuple<DIM>& in) const
  {
    //
    // TODO: optimize for mod_ == 0 (common case)
    // 
    // assert (contains(in)) could be too strong because 
    // a call to ord_(5,5) should be allowed in a shape 
    // (0..5:2,0..5:2), where (5,5) is not actually part 
    // of the shape.
    assert (in >= low_ && in <= high_);

    // assert (mod_ == Tuple<DIM>::zero); // current limitation

    Tuple<DIM> tmp = (in - low_) % (high_ - low_ + Tuple<DIM>::one);
    int ret =  VectorOps<Tuple<DIM>, DIM>::dotProduct (tmp, linearStep_); 
    return ret;
  }
  
  /* GB: coord is local coordinate */
  inline Tuple<DIM> coord (int idx) const 
  {
    assert (idx >= 0 && idx < this->card());
    Tuple<DIM> ret;

    /* start with most significant dimension 0 */
    for (int d = 0; idx > 0 && d < DIM; ++d) {
      ret[d] = idx / localStep_[d];
      idx -= ret[d] * localStep_[d];
    }

    assert (idx == 0);
    return ret;
  }
   
  /**
   * Iterator factory
   */
  inline iterator begin () const 
  { 
    return ShapeIterator(*this, 0);
  }
  
  /**
   * Iterator factory
   */
  inline iterator end () const 
  {
    return ShapeIterator(*this, -1);
  }

  inline bool isContiguous () const 
  {
    return step_ == Tuple<DIM>::one;
  }

  inline iterator rend () const 
  {
    return ShapeIterator(*this, -1);
  }

  inline iterator rbegin () const 
  { 
    return ShapeIterator (*this, this->card() - 1);
  }
  
public:
  
  /**
   * Relocates the coordinate system represented by 
   * this shape. The underlying mapping to linearized 
   * representation remains unchanged.
   *
   * @param origin    new low coordinate of the shape
   */
  inline void relocate (const Tuple<DIM>& origin) 
  {
    Tuple<DIM> dir(origin - low_);
    low_ = origin;
    high_ += dir;
  }
  
  /**
   * Relocates and remaps.
   */
  inline Shape<DIM> operator- (const Tuple<DIM>& dir) const
  {
    return (*this) + (-dir);
  }

    
  /**
   * Relocates and remaps.
   */
  inline Shape<DIM> operator+ (const Tuple<DIM>& dir) const
  {
    Shape<DIM> ret(*this);    
    Tuple<DIM> new_origin = low_ + dir;
    ret.relocate(new_origin);
    return ret;
  }
  
  /**
   * Assignment-move operator 
   */
  Shape& operator+= (const Tuple<DIM>& dir)
  {
    Tuple<DIM> new_origin = low_ + dir;
    relocate(new_origin);
    return *this;
  }
  
  /**
   * Assignment-move operator 
   */
  Shape& operator-= (const Tuple<DIM>& dir)
  {
    Tuple<DIM> new_origin = low_ - dir;
    relocate(new_origin); 
    return *this;
  }

  template <int D> friend ostream& operator << (ostream& ost, const Shape<D>& s);

private:
  
  /* size of each dimension - considering step */
  Tuple<DIM> size_;

  /* highest index in each dimension (included) */
  Tuple<DIM> high_;

  /* lowest index in each dimension (included) */
  Tuple<DIM> low_;

  /* step in each dimension - usually 1 */
  Tuple<DIM> step_; 

  /* the number of elements to skip in the linearized 
   * array until the respective dimension increments (unlike step, this is cumulative) */
  Tuple<DIM> linearStep_;

  /* the number of elements to be skipped in the local co-ordinate system */
  Tuple<DIM> localStep_;

  /* the circshift parameter for each dimension */
  Tuple<DIM> mod_;

  /* the complete dimension information stored as a sequence of triplets */
  Triplet dims_[DIM];

  

};

template<int DIM>  ostream& operator << (ostream& ost, const Shape<DIM>& s) {  
  ost << "Shape<" << DIM << ">(" 
      << " @=" << &s
      << " low=" << s.low_ 
      << " high=" << s.high_
      << " step=" << s.step_ 
      << " size=" << s.size_ 
      << " linearStep=" << s.linearStep_  << ")";
  return ost;
}

#endif /* __SHAPE_H__ */
