/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: ShapeH.h,v 1.4.4.3 2007/08/18 19:01:31 fraguela Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef  __SHAPE_ARRAY_H__
#define  __SHAPE_ARRAY_H__

#include <cassert>
#include <iostream>

#include "Triplet.h"
#include "Tuple.h"
#include "debug/Tracing.h"
#include "util/Iterator.h"
#include "util/Sequence.h"
#include "util/VectorOps.h"
#include "Shape.h"

using namespace std;

template <int DIM> 
class ShapeH : public Shape<DIM>
{
  template <int D> 
  friend ostream& operator << (ostream& ost, const ShapeH<D>& s);

public:

  ShapeH ()
   : Shape<DIM>(), children_(0)
  { }
  
  ShapeH (const Tuple<DIM>& size) 
    : Shape<DIM>(size), children_(new Shape<DIM> [this->card()])
  { }
  
  ShapeH (Shape<DIM>* children, const Shape<DIM>& shape)
   : Shape<DIM>(shape), children_(children)
  { }
  
  ShapeH (const Shape<DIM>& shape)
   : Shape<DIM>(shape), children_(new Shape<DIM> [shape.card()])
  { }

  ShapeH (const ShapeH<DIM>& s)
  : Shape<DIM>(s), children_ (s.children_)
  { }

  virtual ~ShapeH()
  {
    //cout << "This ... " << endl;
    if(children_)
      delete [] children_;
  }

  virtual const Shape<DIM>& operator [] (const Tuple<DIM>& i) const
  {
    return children_[ord(i)];
  }

  virtual const Shape<DIM>& shapeAt (int i) const
  {
    return children_[i];
  }
  
  Shape<DIM>& shapeAt (int i)
  {
    return children_[i];
  }

  virtual ShapeH<DIM> operator [] (const Shape<DIM>& s) const
  {
    return ShapeH<DIM> (children_, Shape<DIM>(s.size()));
  }

  virtual const ShapeH<DIM>& operator= (const ShapeH<DIM>& s) 
  {
    if(this != &s) {
      
      Shape<DIM>::operator=(s);
      
      if(children_)
	delete [] children_;
      
      const int ncard = this->card();
      children_ = new Shape<DIM> [ncard];
      for (int i = 0; i < ncard; i++)
	children_[i] = s.children_[i];
    }
    
    return *this;
  }

  virtual const ShapeH<DIM>& operator= (const Shape<DIM>& s)
  {
    const int ncard = this->card();
    for(int j = 0; j < ncard; ++j)    
      children_[j] = s;
    return *this;
  }

  void  assign (const Shape<DIM>& s, const Tuple<DIM>& pos)
  {
    children_[ord(pos)] = s;
  }

  typename Triplet::Seq myshape() const 
  {
    typename Triplet::Seq range;
    for(int i = 0; i < DIM; ++i) {
      range[i] = Shape<DIM>::operator[](i);
    }
    return range;
  }
  
  //BBF: MMmmm... not actually cloning anything... (?)
  virtual const ShapeH<DIM>* clone () const
  {
    //    Shape<DIM>* children = new Shape<DIM> [shape_.card()];
    //for (int i = 0; i < shape_.card(); i++)
      //children[i] = children_[i];

    //    return new ShapeH<DIM> (children, shape_);
    return this;
  }


  /** 
   *GBDEBUG: Does not work for all cases
   */
  inline ShapeH<DIM> intersect (const Shape<DIM>& s) const
  {
    const int ncard = Shape<DIM>::card();
    if(!ncard)
      return ShapeH<DIM>();

    //TO create the children field. It may be larger than actual 
    // because some of the tiles may not be indexed.
    Shape<DIM>* children = new Shape<DIM> [ncard];
    typename Shape<DIM>::iterator i = Shape<DIM>::begin();
    typename Shape<DIM>::iterator i_end = Shape<DIM>::end();
    Tuple<DIM> ii;
    int k = 0;
    int real_k = 0;
    Tuple<DIM> start = Tuple<DIM>::zero;
    Tuple<DIM> end; // = children_[0].highs();
    const Tuple<DIM> lows = s.lows();
    const Tuple<DIM> highs = s.highs();
    Tuple<DIM> new_shape[2];

    for (; i != i_end; ++i, k++) {

      if(k != 0) {
	const Tuple<DIM>& size = children_[k - 1].size();
	Tuple<DIM> delta = *i - ii;
	for(int d = 0; d < DIM; d++)
	  if(delta[d] < 0) {
	    start[d] = 0;
	  } else {
	    start[d] += delta[d] * size[d];
	  }
	  /* Another implementation would be instead of the previous for loop:
	   Tuple<DIM> filter = Tuple<DIM>::one;
	   for(int d = 0; d < DIM; d++) {
	     if( delta[d] < 0) {
	       filter[d]=0;
	     }
	   }
	   start = filter * (start + delta * size);
	  */
      }
      end = start + children_[k].size() - 1;

      if(start<=highs && end >=lows) {
        typename Triplet::Seq range;
	if(real_k == 0) {
	  new_shape[0] = *i;
	}
	new_shape[1] = *i; //  to catch the last valid tile
	for (int j = 0; j < DIM; j++) {
	  
	  int begin = max((lows[j] - start[j]), 0);
	  //int begin = children_[k].lows()[j];
	  //begin += start[j] >= lows[j] ? 0 : lows[j] - start[j];
	  
	  int last = min(highs[j], end[j]) - start[j];
	  //int last =  children_[k].highs()[j];
	  //last -= end[j] <= highs[j] ? 0 : end[j] - highs[j];
	  
	  range[j] = Triplet(begin, last);
	}  
	children[real_k] = Shape<DIM> (range);
	real_k++; //only the valid tiles will be put into children
      }
      
      ii = *i;
      
    }
    
    typename Triplet::Seq range0;
    for(int j = 0; j < DIM; j++) {
      range0[j] = Triplet(new_shape[0][j], new_shape[1][j]);
    }
    Shape<DIM> ns = Shape<DIM>(range0); 
    return ShapeH<DIM>(children, ns);
  }

  inline ShapeH<DIM> operator+ (const Tuple<DIM>& dir) const
  {
    const int ncard = Shape<DIM>::card();
    Shape<DIM>* children = new Shape<DIM> [ncard];
    for (int i = 0; i < ncard; i++)
      children[i] = children_[i] + dir;

    return ShapeH<DIM> (children, *this);
  }
  
  inline ShapeH<DIM> operator- (const Tuple<DIM>& dir) const
  {
    const int ncard = Shape<DIM>::card();
    Shape<DIM>* children = new Shape<DIM> [ncard];
    for (int i = 0; i < ncard; i++)
      children[i] = children_[i] - dir;

    return ShapeH<DIM> (children, *this);
  }

private:

  Shape<DIM>* children_;
  
};


template <int D> 
ostream& operator << (ostream& ost, const ShapeH<D>& s)
{
  ost << static_cast<Shape<D> >(s) << endl;
  for (int i = 0; i < s.card(); ++i)
    ost << s.children_[i] << endl;
  return ost;
}

#endif
