/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Triplet.h,v 1.30.4.1 2007/08/16 00:33:37 fraguela Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef __TRIPLET_H__
#define __TRIPLET_H__

#include <iostream>
#include <cassert>
#include <cmath>

#include "util/Sequence.h"

using namespace std;


class Triplet {

public:

  typedef Sequence<Triplet> Seq;
  Seq operator,(const Triplet& t) 
  {
    return Seq(*this, t);
  }

  // constructors

  Triplet() : low_(0), high_(0), step_(0), mod_(0) {}

  Triplet(const Triplet& t) : low_(t.low_), high_(t.high_), step_(t.step_), mod_(t.mod_) { } 
  
  Triplet(int h) : low_(h), high_(h), step_(1), mod_(0) 
  {
    assert (h >= 0);
  }

  Triplet(int l, int h, int s=1, int mod = 0) : low_(l), high_(h), step_(s), mod_(mod)
  {
    assert ((l <= h && s > 0 ) || ( l > h && s < 0 ));
    assert (s != 0);
  }

  Triplet& operator=(const Triplet& rhs) 
  {
    if (&rhs != this) {
      low_ = rhs.low_;
      high_ = rhs.high_;
      step_ = rhs.step_;
      mod_ = rhs.mod_;
    }
    return *this;
  }

  bool operator==(const Triplet& rhs) const {
    return (&rhs == this) || (rhs.low_ == low_ && 
			      rhs.high_ == high_ && 
			      rhs.step_ == step_);
  }
  
  bool operator!=(const Triplet& rhs) const {
    return !this->operator==(rhs);
  }

  /**
   * @return the number of elements covered by this triplet
   */
  inline int size() const 
  { 
    return max ( (int) ceil( (double) ( abs(high_ - low_) + 1 ) / (double) abs(step_) ) , 0 ); 
  }

  inline int low() const 
  {
    return low_;
  }

  inline int high() const 
  {
    return high_;
  }
  
  inline int step() const 
  {
    return step_;
  }

  inline Triplet operator % (int r) 
  {
    return Triplet(low_, high_, step_, mod_ + r);
  }
  
  inline Triplet operator + (int r) 
  {
    return Triplet(low_ + r, high_ + r, step_, mod_);
  }
  
  inline Triplet operator - (int r) 
  {
    return Triplet(low_ - r, high_ - r, step_, mod_);
  }

  inline int mod() const
  {
    return mod_;
  }
  
private:

  /** forbid heap allocation */
  void* operator new(size_t size) throw () { assert(0); return NULL; }

  /**  forbid heap allocation */
  void operator delete(void* to_delete) { assert (0); }

  int low_;
  int high_;
  int step_;
 
  /** the circhsift parameter */
  int mod_;
};


ostream& operator << (ostream& ost, const Triplet& t) 
{
  return ost << t.low() << ":" << t.step() << ":" << t.high() << ":" << t.mod();
}

typedef Triplet Range;

#endif /* __TRIPLET_H__ */
