/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Tuple.h,v 1.65.4.1 2007/08/21 17:47:03 brodman2 Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef __TUPLE_H__
#define __TUPLE_H__

#include <cassert>
#include <iostream>

#include "util/Sequence.h"
#include "util/VectorOps.h"

using namespace std;

template <int DIM> 
class Tuple {

  template <int D> friend ostream& operator<< (ostream& ost, const Tuple<D>& t);

public:

  /* support for ,-syntax and sequences */
  typedef Sequence<Tuple<DIM> > Seq;
  Seq operator,(const Tuple<DIM>& t) 
  {
    return Seq(*this, t);
  }

  inline explicit Tuple() 
  {
    for (int i = 0; i < DIM; i++)
      values_[i] = 0;
  }

  inline explicit Tuple(const int* values) 
  {
    //    VectorOps<int[DIM], DIM>::copy(values_, values);
    for (int i = 0; i < DIM; i++)
      values_[i] = values[i];
  }
  
  inline explicit Tuple(int i, int j, int k, int l)
  {
    values_[0] = i;
    values_[1] = j;
    values_[2] = k;
    values_[3] = l;
  }
  
  
  /** 
   * Copy constructor 
   */
  inline Tuple(const Tuple<DIM>& t) 
  {
    VectorOps<int[DIM], DIM>::copy(values_, t.values_);
  }

  inline Tuple<DIM> operator+ (const Tuple<DIM>& other) const 
  {    
    Tuple<DIM> ret;
    VectorOps<int[DIM], DIM>::add(ret.values_, values_, other.values_);
    return ret;
  }

  inline void operator+= (const Tuple<DIM>& other)  
  {   
    VectorOps<int[DIM], DIM>::addIn(values_, other.values_);
  }

  inline void operator-= (const Tuple<DIM>& other)  
  {   
    VectorOps<int[DIM], DIM>::subIn(values_, other.values_);
  }

  inline Tuple<DIM> operator+ (int d) const 
  {    
    return Tuple<DIM> (values_[0] + d, values_[1] + d, values_[2] + d, values_[3] + d);
  }

  inline Tuple<DIM> operator- (const Tuple<DIM>& other) const 
  {    
    Tuple<DIM> ret;
    VectorOps<int[DIM], DIM>::sub(ret.values_, values_, other.values_);
    return ret;
  }

  inline Tuple<DIM> operator- (int d) const 
  {    
    return Tuple<DIM> (values_[0] - d, values_[1] - d, values_[2] - d, values_[3] - d);
  }

  inline Tuple<DIM> operator- () const
  {
    Tuple<DIM> ret;
    VectorOps<int[DIM], DIM>::uminus(ret.values_, values_);
    return ret;
  }

  inline Tuple<DIM> operator* (const Tuple<DIM>& other) const 
  {    
    Tuple<DIM> ret;
    VectorOps<int[DIM], DIM>::mult(ret.values_, values_, other.values_);
    return ret;
  }

  inline Tuple<DIM> operator/ (const Tuple<DIM>& other) const 
  {    
    Tuple<DIM> ret;
    VectorOps<int[DIM], DIM>::div(ret.values_, values_, other.values_);
    return ret;
  }

  inline Tuple<DIM> operator% (const Tuple<DIM>& other) const 
  {
    Tuple<DIM> ret;
    VectorOps<int[DIM], DIM>::mod(ret.values_, values_, other.values_);
    return ret;
  }

  inline bool operator< (const Tuple<DIM>& other) const 
  {    
    for(int i=0; i < DIM; i++)
      if((*this)[i] >= other[i])
	return false;    
    return true;
  }
  
  inline bool operator<= (const Tuple<DIM>& other) const 
  {    
    for(int i=0; i < DIM; i++)
      if((*this)[i] > other[i])
	return false;    
    return true;
  }

  inline bool operator> (const Tuple<DIM>& other) const 
  {    
    for (int i=0; i < DIM; i++)
      if ((*this)[i] <= other[i])
	return false;    
    return true;
  }


  inline bool operator>= (const Tuple<DIM>& other) const 
  {    
    for (int i=0; i < DIM; i++)
      if ((*this)[i] < other[i])
	return false;    
    return true;
  }


  inline  bool operator== (const Tuple<DIM>& other) const 
  {
    return (&other == this) || VectorOps<int[DIM], DIM>::equal(values_, other.values_);
  }

  inline  bool operator!= (const Tuple<DIM>& other) const 
  {
    return VectorOps<int[DIM], DIM>::different(values_, other.values_);
  }

  inline Tuple<DIM>& operator= (const Tuple<DIM> & other) 
  { 
    VectorOps<int[DIM], DIM>::copy(values_,other.values_);
    return *this;
  }
  
  // JCB - removed const in return type as it is meaningless.
  inline int operator[] (int i) const 
  { 
    return values_[i]; 
  }
  
  inline int& operator[] (int i) 
  { 
    return values_[i]; 
  }

  /**
   * mask a value in a tuple with the given value (default 0).
   *
   * @param dim    the dimension to mask
   * @param value  the value to put into the dimension
   */
  Tuple<DIM> mask(int dim, int value = 0) const 
  {
    Tuple<DIM> out = *this;
    out[dim] = value;
    return out;
  }
  
  static Seq mask (const Seq& seq, int dim, int value = 0) 
  {
    Seq ret;
    for (int i = 0; i < seq.length(); ++i) 
      ret[i] = seq[i].mask(dim, value);
    return ret;
  }

  /**
   * permute a tuple.
   *
   * @param permutation - must be exactly of length DIM
   */
  Tuple<DIM> permute (const Tuple<DIM>& perm) const
  {
    Tuple<DIM> out;
    for (int i = 0; i < DIM; i++) {
      assert (perm[i] >= 0 && perm[i] < DIM);
      out[i] = values_[perm[i]];
    }
    return out;
  }

  static Seq permute (const Seq& seq, const Tuple<DIM>& perm) 
  {
    Seq ret;
    for (int i = 0; i < seq.length(); ++i) 
      ret[i] = seq[i].permute(perm);
    return ret;
  }

 
private:

  int values_ [DIM]; 

  /** forbid heap allocation */
  void* operator new(size_t size) throw () 
  { 
    assert (false); 
    return NULL; 
  }

  /**  forbid heap allocation */
  void operator delete(void* to_delete) 
  { 
    assert (false); 
  }

public:

  static const Tuple<DIM> zero;
  static const Tuple<DIM> one;
};

template <int DIM> 
ostream& operator<< (ostream& ost, const Tuple<DIM>& t) 
{
  ost << "(";
  for (int i=0; i < DIM - 1; i++)
    ost << t.values_[i] << ", ";
  ost << t.values_[DIM-1] << ")";
  return ost;
};

template <int DIM>
const Tuple<DIM> Tuple<DIM>::zero;

// include specializations
#include "Tuple_1.h"
#include "Tuple_2.h"
#include "Tuple_3.h"
#include "Tuple_4.h"


#endif /* end __TUPLE_H__*/
