/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Tuple_1.h,v 1.7.4.2 2007/08/21 17:47:03 brodman2 Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef __TUPLE_1_H__
#define __TUPLE_1_H__

#include <cassert>
#include <iostream>

#include "Tuple.h"
#include "util/Sequence.h"

using namespace std;

/**
 * This is a template specialization of Tuple for DIM=1 
 */
template <>
class Tuple<1> {

  static const int  DIM = 1;
  
  template<int D>
  friend ostream& operator<< (ostream& ost, const Tuple<D>& t);

public:

  // support for ,-syntax and sequences 
  typedef Sequence<Tuple<DIM> > Seq;
  Seq operator,(const Tuple<DIM>& t) 
  {
    return Seq(*this, t);
  }

  inline explicit Tuple() : x(0) { }

  inline explicit Tuple(const int* values) : x(values[0]) {} 

  inline explicit Tuple (const int i) : x(i) { } 
  
  /** 
   * Copy constructor 
   */
  inline Tuple(const Tuple<DIM>& t) : x(t.x) {}

  inline Tuple<DIM> operator+ (const Tuple<DIM>& other) const 
  {    
    return Tuple<DIM>(x + other.x);
  }

  inline void operator+= (const Tuple<DIM>& other)  
  {   
    x += other.x;
  }

  inline void operator-= (const Tuple<DIM>& other)  
  {   
    x -= other.x;
  }

  inline Tuple<DIM> operator- (const Tuple<DIM>& other) const 
  {    
    return Tuple<DIM>(x - other.x);
  }

  inline Tuple<DIM> operator- () const
  {
    return Tuple<DIM> (-x);
  } 

  inline Tuple<DIM> operator- (int d) const {
    return Tuple<DIM> (x - d);
  }

  inline Tuple<DIM> operator+ (int d) const {
    return Tuple<DIM> (x + d);
  }

  inline Tuple<DIM> operator* (int d) const {
    return Tuple<DIM> (x * d);
  }

  inline Tuple<DIM> operator/ (int d) const {
    return Tuple<DIM> (x / d);
  }

  inline Tuple<DIM> operator* (const Tuple<DIM>& other) const 
  {    
    return Tuple<DIM> (x * other.x);
  }

  inline Tuple<DIM> operator/ (const Tuple<DIM>& other) const 
  {    
    return Tuple<DIM> (x / other.x);
  }

  inline bool operator< (const Tuple<DIM>& other) const 
  {    
    return (x < other.x);
  }
  
  inline bool operator<= (const Tuple<DIM>& other) const 
  {    
    return (x <= other.x);
  }
  
  inline bool operator> (const Tuple<DIM>& other) const 
  {    
    return (x > other.x);
  }

  inline bool operator>= (const Tuple<DIM>& other) const 
  {    
    return (x >= other.x);
  }


  inline  bool operator== (const Tuple<DIM>& other) const 
  {
    return &other == this || x == other.x;
  }

  inline  bool operator!= (const Tuple<DIM>& other) const 
  {
    return x != other.x;
  }

  inline Tuple<DIM>& operator= (const Tuple<DIM> & other) 
  { 
    x = other.x;
    return *this;
  }

  // JCB - removed const from return type as it is meaningless
  inline int operator[] (int i) const 
  { 
    assert (i >= 0 && i < DIM);
    return x;
  }
  
  inline int& operator[] (int i) 
  { 
    assert (i >= 0 && i < DIM);
    return x;
  }

  inline Tuple<DIM> operator% (const Tuple<DIM>& other) const 
  {
     return Tuple<DIM>(x % other.x);
  }

  Tuple<DIM> mask(int dim, int value = 0) const 
  {
    assert (dim == 0);
    return Tuple<DIM>(value);
  }

  static Seq mask (const Seq& seq, int dim, int value = 0) 
  {
    Seq ret;
    for (int i = 0; i < seq.length(); ++i) 
      ret[i] = seq[i].mask(dim, value);
    return ret;
  }
  
  Tuple<DIM> permute(const Tuple<DIM>& perm) const
  {
    assert (perm == Tuple<DIM>::zero);
    return Tuple<DIM>(*this);
  }

  static Seq permute (const Seq& seq, const Tuple<DIM>& perm) 
  {
    Seq ret;
    for (int i = 0; i < seq.length(); ++i) 
      ret[i] = seq[i].permute(perm);
    return ret;
  }

private:
  
  int x;

  /** forbid heap allocation */
  void* operator new(size_t size) throw () 
  { 
    assert(false); return NULL; 
  }

  /**  forbid heap allocation */
  void operator delete(void* to_delete) 
  { 
    assert (false); 
  }

public:

  static const Tuple<DIM> zero;
  static const Tuple<DIM> one;
};

template <>
ostream& operator<< (ostream& ost, const Tuple<1>& t) 
{
  ost << "(" << t.x << ")";
  return ost;
};

const Tuple<1> Tuple<1>::one(1);

const Tuple<1> Tuple<1>::zero;

#endif /* end __TUPLE_1_H__*/
