/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Tuple_3.h,v 1.16.4.1 2007/08/21 17:47:03 brodman2 Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef __TUPLE_3_H__
#define __TUPLE_3_H__

#include <cassert>
#include <iostream>

#include "Tuple.h"
#include "util/Sequence.h"

using namespace std;

/**
 * This is a template specialization of Tuple for DIM=3, 
 * The performance benefit over Tuple<DIM> is minimal in 
 * with g++.
 */
template <>
class Tuple<3> {

  static const int  DIM = 3;
  
  template<int D>
  friend ostream& operator<< (ostream& ost, const Tuple<D>& t);

public:

  // support for ,-syntax and sequences 
  typedef Sequence<Tuple<DIM> > Seq;
  Seq operator,(const Tuple<DIM>& t) 
  {
    return Seq(*this, t);
  }

  inline explicit Tuple() : x(0), y(0), z(0) { }

  inline explicit Tuple(const int* values) : x(values[0]), y(values[1]), z(values[2]) { }

  inline explicit Tuple (const int i, const int j, const int k) : x(i), y(j), z(k) { } 
  
  /** 
   * Copy constructor 
   */
  inline Tuple(const Tuple<DIM>& t) : x(t.x), y(t.y), z(t.z) {}

  inline Tuple<DIM> operator+ (const Tuple<DIM>& other) const 
  {    
    return Tuple<DIM>(x + other.x, y + other.y, z + other.z);
  }

  inline void operator+= (const Tuple<DIM>& other)  
  {   
    x += other.x;
    y += other.y;
    z += other.z;
  }

  inline void operator-= (const Tuple<DIM>& other)  
  {   
    x -= other.x;
    y -= other.y;
    z -= other.z;
  }

  inline Tuple<DIM> operator- (const Tuple<DIM>& other) const 
  {    
    return Tuple<DIM>(x - other.x, y - other.y, z - other.z);
  }

  inline Tuple<DIM> operator- (int d) const 
  {    
    return Tuple<DIM> (x - d, y - d, z - d);
  }

  inline Tuple<DIM> operator+ (int d) const 
  {    
    return Tuple<DIM> (x  + d, y + d, z + d);
  }

  static inline Tuple<DIM> max (Tuple<DIM>t, int d) 
  {    
    return Tuple<DIM> (t.x > d ? t.x : d,  t.y > d ? t.y : d, t.z > d ? t.z : d);
  }

  inline Tuple<DIM> operator- () const
  {
    return Tuple<DIM> (-x, -y, -z);
  } 

  inline Tuple<DIM> operator* (const Tuple<DIM>& other) const 
  {    
    return Tuple<DIM> (x * other.x, y * other.y, z * other.z);
  }
  
  inline Tuple<DIM> operator/ (const Tuple<DIM>& other) const 
  {    
    return Tuple<DIM> (x / other.x, y / other.y, z / other.z);
  }

  inline Tuple<DIM> operator/ (int d) const 
  {    
    return Tuple<DIM> (x / d, y / d, z / d);
  }
  
  inline bool operator< (const Tuple<DIM>& other) const 
  {    
    if (x >= other.x)
      return false;
    if (y >= other.y)
      return false;
    return (z < other.z);
  }
  
  inline bool operator<= (const Tuple<DIM>& other) const 
  {    
    if (x > other.x)
      return false;
    if (y > other.y)
      return false;
    return (z <= other.z);
  }

  inline bool operator> (const Tuple<DIM>& other) const 
  {    
    if (x <= other.x)
      return false;
    if (y <= other.y)
      return false;
    return (z > other.z);
  }

  inline bool operator>= (const Tuple<DIM>& other) const 
  {    
    if (x < other.x)
      return false;
    if (y < other.y)
      return false;
    return (z >= other.z);
  }


  inline  bool operator== (const Tuple<DIM>& other) const 
  {
    return (&other == this) || (x == other.x && y == other.y && z == other.z);
  }

  inline  bool operator!= (const Tuple<DIM>& other) const 
  {
    return (x != other.x || y != other.y || z != other.z );
  }

  inline Tuple<DIM>& operator= (const Tuple<DIM> & other) 
  { 
    x = other.x;
    y = other.y;
    z = other.z;
    return *this;
  }

  // JCB - removed const from return type as it is meaningless
  inline int operator[] (int i) const // __attribute__((always_inline))
  { 
    assert (i >= 0 && i < DIM);
    return (i == 0) ? x : (i == 1) ? y : z;
  }
  
  inline int& operator[] (int i) // __attribute__((always_inline))
  { 
    assert (i >= 0 && i < DIM);
    return (i == 0) ? x : (i == 1) ? y : z;
  }

  inline Tuple<DIM> operator% (const Tuple<DIM>& other) const 
  {
     return Tuple<DIM>(x % other.x, y % other.y, z % other.z);
  }

  Tuple<DIM> mask (int dim, int value = 0) const 
  {
    assert (dim >= 0 && dim < 3);
    return Tuple<DIM> (dim == 0 ? value : x , 
		       dim == 1 ? value : y, 
		       dim == 2 ? value : z);
  }
     
  static Seq mask (const Seq& seq, int dim, int value = 0) 
  {
    Seq ret;
    for (int i = 0; i < seq.length(); ++i) 
      ret[i] = seq[i].mask(dim, value);
    return ret;
  }

  Tuple<DIM> permute(const Tuple<DIM>& perm) const
  {
    return Tuple<DIM>(operator[](perm[0]), 
		      operator[](perm[1]), 
		      operator[](perm[2]));
  }

  Tuple<DIM> swap (int i, int j) const
  {
    Tuple<DIM> ret = *this;
    
    //swap
    int tmp = ret[i];
    ret[i] = ret[j];
    ret[j] = tmp;
    return ret;
  }
  
  static Seq permute (const Seq& seq, const Tuple<DIM>& perm) 
  {
    Seq ret;
    for (int i = 0; i < seq.length(); ++i) 
      ret[i] = seq[i].permute(perm);
    return ret;
  }

  static Seq permute (const Seq& seq, const Tuple<DIM>& perm, unsigned permute_mask = 0xFF) 
  {
    Seq ret;
    for (int i = 0; i < seq.length(); ++i) { 
      ret[i] = permute_mask & (1 << i) ?  seq[i].permute(perm) : seq[i];
    }
    return ret;
  }
public:
  
  int x;
  int y;
  int z;

  /** forbid heap allocation */
  void* operator new(size_t size) throw () 
  { 
    assert(false); return NULL; 
  }

  /**  forbid heap allocation */
  void operator delete(void* to_delete) 
  { 
    assert (false); 
  }

public:

  static const Tuple<DIM> zero;
  static const Tuple<DIM> one;
};

template <>
ostream& operator<< (ostream& ost, const Tuple<3>& t) 
{
  ost << "(" << t.x << ", " << t.y << ", " << t.z << ")";
  return ost;
};

const Tuple<3> Tuple<3>::one(1,1,1);

const Tuple<3> Tuple<3>::zero;

#endif /* end __TUPLE_3_H__*/
