/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Alloc.h,v 1.11 2006/07/08 22:43:41 vonpraun Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */


#ifndef __ALLOC_H__
#define __ALLOC_H__

#include <cassert>
#include <cstdlib>
#include <iostream>

#include "Traits.h"

using namespace std;

/**
 * This file contains overloaded new and delete
 * that keeps track of allocated, deallocated and
 * undeleted heap objects.
 * Check for memcount = 0 at the end of 'main'. 
 * Also invoke undeleted(), that will
 * print out the list of undeleted objects.
 */

namespace Alloc {

  /**
   * a link list to keep track of total size of allocated objects
   * I am unable to use stl maps, as my overloaded version of operator new
   * makes the calls to map[] infinitely recursive.
   */
  struct memsize
  {
    void* mem;
    int size;
    bool arr;
    struct memsize* ptr;
  };

  static struct memsize* head = NULL;

  static struct memsize* cur = NULL;

  /**
   * tracks the number of heap allocated objects
   */
  static int memcount = 0;

  /**
   * call this function to check if there are 
   * any undeleted objects.
   * This function also frees the memory occupied
   * by the link list. Hence, call this function only once.
   */
  void undeleted ()
  {
#if (HTA_DEBUG_MEM == 1)
    for (struct memsize* i = head; i != NULL; i = i->ptr) {
      if ((*i).size > 0) 
	cout << "ALLOC[" << Traits::Default::myPlace() << "] undeleted " << (*i).mem <<  " size=" << (*i).size << " arr=" << (*i).arr << endl;
    }
    cout << "ALLOC[" << Traits::Default::myPlace() << "] memcount=" << memcount << endl;
#endif
  }

  /**
   * The internal link list to keep track of 'size' of 
   * allocated objects
   */

  struct memsize* next()
  {
    struct memsize* tmp = (struct memsize*) malloc (sizeof(struct memsize));
    assert (tmp);

    tmp->size = -1;
    tmp->mem = NULL;
    tmp->ptr = NULL;

    if (!head) {
      // first entry
      head = tmp;
      cur = tmp;
    } else {
      // append at the end
      cur->ptr = tmp;
      cur = cur->ptr;
    }
    
    return cur;
  }

  struct memsize* find (void* node)
  {
    for(struct memsize* i = head; i != NULL; i = i->ptr) {
      if (i->mem == node)
	return i;
    }
    return NULL;
  }

  void insert (void* mem, int size, bool arr)
  {
    struct memsize* node = find (mem);

    if (node != NULL) {
      node->size = size;
      node->arr = arr;
    } else {
      struct memsize* tmp = next();
      tmp->mem = mem;
      tmp->size = size;
      tmp->arr = arr;
    }
  }
} // namespace Alloc


#if (HTA_DEBUG_MEM == 1)
/**
 * overloaded new operator 
 */
void* operator new (size_t size)
{
  assert (size >= 0);
  void* mem = malloc (size);
  cout << "ALLOC[" << Traits::Default::myPlace() << "] new mem="  << mem << " size=" << size << endl;
  Alloc::memcount++;
  Alloc::insert (mem, size, false);
  assert (Alloc::find (mem));
  return mem;
}

void* operator new [] (size_t size)
{
  assert (size >= 0);
  void* mem = malloc (size);
  cout << "ALLOC[" << Traits::Default::myPlace() << "] new[] mem="  << mem << " size=" << size << endl;
  Alloc::memcount++;
  Alloc::insert (mem, size, true);
  assert (Alloc::find (mem));
  return mem;
}


/**
 * overloaded delete operator
 */
void operator delete (void* mem)
{

  Alloc::memcount--;
  struct  Alloc::memsize* tmp =  Alloc::find (mem);

  if (tmp == NULL) {
    cout << "ALLOC[" << Traits::Default::myPlace() << "] delete INVALID mem=" << mem << endl;
    assert (false);
  } else if (tmp->size == 0) {
    cout << "ALLOC[" << Traits::Default::myPlace() << "] delete DUPLICATE mem=" << mem << endl;
    assert (false);
  } else if (tmp->arr) {
    cout << "ALLOC[" << Traits::Default::myPlace() << "] delete MISMATCHED mem=" << mem << endl;
    assert (false);
  } else  {
    cout << "ALLOC[" << Traits::Default::myPlace() << "] delete mem="  << mem << " size=" << tmp->size << endl;
    Alloc::insert (mem, 0, false);  
  }
  // de-alloc at the end because this might crash
  free (mem);
}

void operator delete [] (void* mem)
{
  Alloc::memcount--; 
  struct Alloc::memsize* tmp = Alloc::find(mem);

  if (tmp == NULL) {
    cout << "ALLOC[" << Traits::Default::myPlace() << "] delete[] INVALID mem=" << mem << endl;
    int i;
    cin >> i;
  } else if (tmp->size == -1) {
    cout << "ALLOC[" << Traits::Default::myPlace() << "] delete[] DUPLICATE mem=" << mem << endl;
    int i;
    cin >> i;
  } else if (!tmp->arr) {
    cout << "ALLOC[" << Traits::Default::myPlace() << "] delete[] MISMATCHED mem=" << mem << endl;
    int i;
    cin >> i;
  } else {
    cout << "ALLOC[" << Traits::Default::myPlace() << "] delete[] "  << mem << " size=" << tmp->size <<endl;
    Alloc::insert (mem, 0, true);  
  }
  // de-alloc at the end because this might crash
  free (mem);
}

#endif // HTA_DEBUG_MEM

#endif /* __ALLOC_H__ */
