/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: RandNumGenerator.h,v 1.9 2006/06/06 16:11:57 bikshand Exp $
 * Authors: Ganesh Bikshandi
 */

#ifndef  __RAND_NUM_GENERATOR_H__
#define  __RAND_NUM_GENERATOR_H__

#include <cassert>
#include "MapReduce.h"
#include "util/RandUtils.h"

using namespace std;

/* State of the random number generator */
struct RandNumGeneratorState_ 
{
  double seed;
  double a;
};

/**
 * map-reduce operator to fill an HTA with uniformly distributed pseudo-random numbers
 */
template<typename T, int DIM>
struct RandNumGenerator : public Operator::MapReduce<T, RandNumGeneratorState_, DIM>  
{
 
  RandNumGenerator (double a, double seed) {
    state_.seed = seed;
    state_.a = a;
  }
  
  void map_none (T& val) 
  {
    val = RandUtils::randlc(&state_.seed, state_.a);
  }
  
  /** Why don't we always declare the member to be 'result'
   *  and move this function to super class (MapReduce) */
  virtual const RandNumGeneratorState_& result () const 
  {
    return state_;
  }
  
  Operator::MapReduceKey keyType () const { return Operator::MR_NONE; }
  
  RandNumGeneratorState_ state_; /* name this a 'result' always; see comment above in result() */
  
}; // struct RandNumGenerator


/**
 * map-reduce operator to fill a distributed HTA with uniformly distributed pseudo-random numbers

 * TODO: works only for 1D and 2D (1 X N) HTAs.
 */
template<typename T, int DIM>
struct RandNumGeneratorDist : public Operator::MapReduce<T, RandNumGeneratorState_, DIM>  
{

  RandNumGeneratorDist (double a, double seed) : a_(a), seed_(seed) { }

  void init (MapContext<T, DIM>& context) 
  {
    this->context_ = context;   
    
    state_.seed = seed_;
    state_.a = a_;

    //global starting position
    Tuple<DIM> localIndexStart = this->context_.position_ * this->context_.child_->shape().size();
    
    int pi = 0;
    for (int i = 1; i < DIM; i++) {
      pi = pi + localIndexStart[i] * this->context_.parent_->memMap().leafSize()[i-1];
    }
    pi = pi + localIndexStart[0];

    double ai = RandUtils::power(state_.a, pi);
    double rdummy = RandUtils::randlc(&state_.seed, ai);
    
    //cout << " seed " << state_.seed << " " << state_.a << endl;
  }
    
  void map_none (T& val) 
  {
    val = RandUtils::randlc(&state_.seed, state_.a);
  }
  
  bool doReduce () {return false;}

  /** Why don't we always declare the member to be 'result'
   *  and move this function to super class (MapReduce) */
  virtual const RandNumGeneratorState_& result () const 
  {
    return state_;
  }
  
  Operator::MapReduceKey keyType () const { return Operator::MR_NONE; }
  
  RandNumGeneratorState_ state_; /* name this a 'result' always; see comment above in result() */
  
private:
  double seed_;
  double a_;
}; // struct RandNumGenerator

#endif /* __RAND_NUM_GENERATOR_H__ */
