/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: HTAImpl.h,v 1.158.2.3 2007/08/13 21:19:03 brodman2 Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef __HTA_IMPL_H__
#define __HTA_IMPL_H__

#include <cassert>
#include <functional>
#include <iostream>

//#include "Array.h"
#include "Maps.h"
#include "MemMapping.h"
#include "Shape.h"
#include "Traits.h"
#include "util/TypeHelper.h"
#include "util/VectorOps.h"
#include "MapReduce.h"

#include "LeafMemMapping.h"
#include "Operators.h"
#include "tbb/blocked_range.h"
#include "tbb/parallel_for.h"
#include <vector>


using namespace std;

/**
 * Core implementation class for HTA instance with LEVEL >= 0
 */
template<class T, int DIM, typename TRAIT>
class HTAImpl : public Array<T, DIM, TRAIT>
{

  typedef Array<T, DIM, TRAIT> base;

public:
  
  /** 
   * Constructor.
   *
   * @param data   must be unaliased and heap allocated - the HTA object will take care of 
   *               de-allocating the memory
   */
  HTAImpl (size_t levels, 
	   typename TRAIT::TypeTileImpl** data, 
	   const Shape<DIM>& shape, 
	   const typename TRAIT::TypeAllocator& alloc,
	   const MemMapping<DIM>* mmap, 
	   int home) :
    Array<T, DIM, TRAIT>(levels, shape, alloc, mmap, home),
    data_(data)
  {
    //assert (this->level() > 0);
    //assert(data);
    if(! this->isLeaf()){
      //if not leaf
      int len = shape.card();
      for(int i = 0; i < len; i++) {
	typename TRAIT::TypeTileImpl* tileAt = data[i];
	if (tileAt != NULL){
	  tileAt->refCount_(1);
	}
	// else this is a shell HTA
      } 
      HTA_DEBUG(1, "HTAImpl<LEVEL=" << this->level() << ">::<init> this=" << this << " data_[0]=" << data_[0]);
    }
    else{
      //leaf HTA
      // check if the memory provided by the allocator tolerates access 
      // to the whole range of indices of the shape

      if (home == Traits::Default::myPlace()) {
	assert (mmap != NULL);
	HTA_DEBUG(3, "HTAImpl<Leaf>::init mmap=" << mmap << " mmap.place=" << home << " mmap.offset=" << mmap->offset());
	
	int offset = mmap->offset();
	data_leaf_ = alloc.get(offset);
#ifndef NDEBUG
	int ord_low = mmap->ord(shape.low());
	int ord_high = mmap->ord(shape.high());
	//	cout<<"offset: "<<offset<<" low: "<<ord_low<<" high: "<<ord_high<<endl;
	//	cout<<"shape: "<<shape<<endl;
	assert (this->alloc_.checkAddress_(offset + ord_low));
	assert (this->alloc_.checkAddress_(offset + ord_high));
#endif
      } else {
	data_leaf_ = NULL;
      }
    }


  }
    
 HTAImpl (size_t levels, 
	   typename TRAIT::TypeTileImpl** data, 
	   const Shape<DIM>& shape, 
	   const typename TRAIT::TypeAllocator& alloc,
	   const MemMapping<DIM>* mmap, 
	   int home, 
	   bool sparse) :
    Array<T, DIM, TRAIT>(levels, shape, alloc, mmap, home),
    data_(data)
  {
    //for sparse storage don't allocate data like for dense
    if(! this->isLeaf()){
      //if not leaf
      int len = shape.card();
      for(int i = 0; i < len; i++) {
	typename TRAIT::TypeTileImpl* tileAt = data[i];
	if (tileAt != NULL){
	  tileAt->refCount_(1);
	}
	// else this is a shell HTA
      } 
      HTA_DEBUG(1, "HTAImpl<LEVEL=" << this->level() << ">::<init> this=" << this << " data_[0]=" << data_[0]);
    }
  }

  virtual ~HTAImpl() 
  {
    //used for leaf and not leaf HTA
    if(!this->isLeaf()) assert(data_);
    //gti add code for leaves

    if(!this->isLeaf()){

      int len = this->shape_.card();

      for(int i = 0; i < len; i++) {

	typename TRAIT::TypeTileImpl* tile = data_[i];

	if (tile && tile->refCount_(-1) == 0) 
	  {
	    delete tile;
	  }
      }
      delete[] data_;
    }
    else{
      this->alloc_.recycle();
    }

  }
  
  typename TRAIT::TypeImpl& operator= (const typename TRAIT::TypeImpl& rhs) 
  {
    // compiler forces to define this method due to member data_
    // that member remains unchanged and we just revert to the implementation 
    // in the superclass.

    //GB: I perform the memcpy as soon as I find that the HTAs have
    //contiguous memory (irrespective of the tiling).
    //GB: THis fix is required for HTA with 2 level tiling, for performance (e.g. FT)
   // This may lead to problems in other benchmarks.
   //Should do a general implementation. 

    if(!this->isLeaf() && (this->memMapping__ && this->memMap().order() != ROW))
      //if(!this->isLeaf())
    {
      Array<T, DIM, TRAIT>::operator= (rhs);
      return this->actual_();
    }
    else
    {
      //leaf
      if (this != &rhs && this->memMapping__) {
	assert (this->isLocal());
	if (this->memMap().isContiguous() && rhs.memMap().isContiguous() &&
	    this->shape().isContiguous() && rhs.shape().isContiguous() && 
	    rhs.ol()==NULL && this->ol()==NULL) { // add the last two condition to exclude overlapped case. TODO: make one case for OL efficiently
          assert (this->shape().card() == rhs.shape().card());
	  memcpy(this->raw(), rhs.raw(), this->memMap().linearCard() * sizeof(T));

	} 
	else
	{ 
          if (this->level() == 0){
	    map_leaf_ (Operator::idem<T>(), rhs, *this);	    
	  }
          else
	    Array<T, DIM, TRAIT>::operator= (rhs);
	}
      }
      return this->actual_();
    }
  }

  typename TRAIT::TypeImpl& operator= (const T& rhs) 
  {
    Array<T, DIM, TRAIT>::map(Operator::cnst<T>(rhs), *this);
    return this->actual_();
  }

  /** 
   * access methods for non-leaf htas
   * if called on leaf htas, should raise exception
   */
  typename TRAIT::TypeTileImpl* tileAt (const Tuple<DIM>& t) const 
  {
    assert (this->shape().contains(t)); // bounds check;
    int idx = this->shape().ord(t);
    assert (idx >= 0 && idx < this->shape().card());
    return data_[idx];
  }
  
  /**
   * efficient tile access through index inside the library
   */
  inline typename TRAIT::TypeTileImpl* tileAt (int tile_idx) const
  { 
    assert (data_); //BBF: level 0 HTAs have data==0
    assert (tile_idx >= 0 && tile_idx < this->shape_.card());
    return data_[tile_idx];
  }

  void initTile (const Tuple<DIM>& t, typename TRAIT::TypeTileImpl* tile)
  {
    assert(!this->isLeaf());
    assert (this->shape().contains(t)); // bounds check;
    int idx = this->shape().ord(t);
    assert (idx >= 0 && idx < this->shape_.card());
    assert (data_[idx] == NULL); // not initialized twice!
    const_cast <typename TRAIT::TypeTileImpl**> (data_) [idx] = tile;
    tile->refCount_(1);
  }

  void initRemoteTile (const Tuple<DIM>& t, typename TRAIT::TypeTileImpl* tile)
  {
    initTile (t, tile);
  }

  /**
   * access a sub-region of HTAImpl
   * the convex hull of shape s should
   * lie within the shape of the HTAImpl
   */
  typename TRAIT::TypeImpl* tileAt_leaf_ (const Shape<DIM>& s) const 
  {
    return scalarAt(s);
  }


  typename TRAIT::TypeImpl* tileAt (const Shape<DIM>& s) const
  {
    //    cout << "begin tileAt(S)" << endl;
    //cout << s << endl;
    //JG: comment this off for overlapped tiling.
    //    assert (this->shape_.contains(s.low()));  // bounds check
    //    assert (this->shape_.contains(s.high())); // bounds check
    if(this->isLeaf()) {
      return scalarAt (s);
    }
    else { 
      typename TRAIT::TypeImpl* ret;
      // compute the shape of the result
      Shape<DIM> shape(s.size());
      int len = shape.card();
      
      // (1) compute the tiles of the result - iterate over shape, but access this
      typename TRAIT::TypeTileImpl** tiles = new typename TRAIT::TypeTileImpl* [len];
      const MemMapping<DIM>** child_dist = new const MemMapping<DIM>* [len];    
      typename Shape<DIM>::iterator i = s.begin();
      typename Shape<DIM>::iterator end = s.end();
      for (int j = 0; i != end; i++, j++) {
	assert (j < len);
	tiles[j] = this->tileAt(*i); 
	child_dist[j] = tiles[j]->memMapping__;
      } 
      
      // (2) allocate and initialize the result
      // create new HTAImpl
      Tuple<DIM> linear_size = shape.high() - shape.low() + Tuple<DIM>::one;
      const MemMapping<DIM>* mmap = NULL;

      mmap = this->memMapping__->select(child_dist,  linear_size);

      // same allocator, i.e., the return HTA is a view on the current HTA
      ret = new typename TRAIT::TypeImpl (this->level(), tiles, shape, this->alloc_, mmap, this->home());
      // initialize the distribution

      // current limitation - we do not have a distribution in closed form for a 
      // selection of tiles of a distributed array      
      //cout << "GBDEBUG: Here " << s << endl;
      if (ret->isDistributed())
	{
	  // current limitation - we do not have a distribution in closed form for a 
	  // selection of tiles on a distributed array
	  const GenericDistribution<DIM>* old_dist = dynamic_cast<const GenericDistribution<DIM>*>(this->dist__);    
	  
	  //if old dist is a Cyclic distribution.
	  if (!old_dist)
	    old_dist = new GenericDistribution<DIM> (this->shape_);
	  
	  //the new distribution is a Sub-distribution of old distribution
	  ret->initDistribution (old_dist->subDist (s));
	  ret->initLocalTiles(); 
	}     

      return ret;
    }

  }
 
  /**
   * LOGICAL INDEXING of HTAs
   * @arg1: indices, a boolean array
   * @output: the resulting HTA, with 'false' tiles dummy
   
   * NOTES: Now I use the HTADistImpl to implement this. 
   * This should rather be
   * a different sub-class or a parent of HTADistImpl
   */
  typename TRAIT::TypeImpl* tileAt (const bool* indices) const
  {
    Shape<DIM> s = this->shape();
    
    //current limitation
    assert (this->isHomogeneous_());
    //assert (this->level() == 1);
    
    typename TRAIT::TypeImpl* ret;
    // compute the shape of the result
    Shape<DIM> shape(s.size());
    int len = shape.card();
    
    // (1) compute the tiles of the result - iterate over shape, but access this
    typename TRAIT::TypeTileImpl** tiles = new typename TRAIT::TypeTileImpl* [len];
    const MemMapping<DIM>** child_dist = new const MemMapping<DIM>* [len];    
    for (int j = 0; j < len; j++) {
      assert (j < len);

      //if this tile position is true set the tile
      if (indices[j]) 
	{
	  tiles[j] = this->tileAt(j); 
	}
      else //else set dummy tile
	{
	  typename TRAIT::TypeAllocator alloc;
	  
	  // collect tiling info according to logical shapes
	  typename Tuple<DIM>::Seq tiling;
	  const AbstractArray<T, DIM>* tmp = this->tileAt(j);
	  for (int i = this->level()-1; i >= 0; --i) {
	    tiling[i] = tmp->shape().size();
	    if (i > 0) 
	      tmp = &(tmp->accessTile (tmp->shape().low())); 
	  }
	  
	  typename TRAIT::TypeTileImpl* tile = TRAIT::TypeImpl::allocDummy (this->level() - 1, tiling, TILE);

	  tiles[j] = tile;
	}
      

      //JG:  Change it so that the new MemMapping keeps 
      // all the memMapping info of the old ones.
      // This is useful in overlapped tiling to convert 
      // between parent tiles and child tiles.

      //child_dist[j] = &tiles[j]->memMap();
      child_dist[j]= &this->tileAt(j)->memMap();
      assert (child_dist[j]);
    } 
    
    // (2) allocate and initialize the result
    // create new HTAImpl
    Tuple<DIM> linear_size = shape.high() - shape.low() + Tuple<DIM>::one;
    const MemMapping<DIM>* mmap = this->memMap().select(child_dist,  linear_size);
    //const MemMapping<DIM>* mmap = NULL;
    // same allocator, i.e., the return HTA is a view on the current HTA
    ret = new typename TRAIT::TypeImpl (this->level(), tiles, shape, this->alloc_, mmap, this->home());
    // initialize the distribution
    
    if (ret->isDistributed())
      {
	// current limitation - we do not have a distribution in closed form for a 
	// selection of tiles of a distributed array
	//      ret->initDistribution(NULL);
	ret->initDistribution (new GenericDistribution<DIM> (this->shape()));
	ret->initLocalTiles(); //JG: add this so that it has the localTiles
      }
    return ret;
  }
  
  /**
   * accessor method MATLAB style
   */
  typename TRAIT::TypeImpl* tileAt (int n, const Shape<DIM>* const* sv) const 
  {
    assert (this->level() > 0);
    typename TRAIT::TypeImpl* ret;
    Shape<DIM> shape((*sv[0]).size());
    int len = shape.card();

    //cout << "GBDEBUG tileAt " << *sv[0] << " " << *sv[1] << endl;
    // (1) allocate / compute the children
    // responsibility to de-allocate is taken by the newly created 
    // MemMapping and HTA respectively
    typename TRAIT::TypeTileImpl** tiles = new typename TRAIT::TypeTileImpl* [len];
    const MemMapping<DIM>** child_dist = new const MemMapping<DIM>* [len];
    typename Shape<DIM>::iterator i = (*sv[0]).begin();
    typename Shape<DIM>::iterator end = (*sv[0]).end();
    for (int j = 0; i != end; i++, j++) {
      typename TRAIT::TypeTileImpl* tmp = tileAt(*i);
      if (n > 2) 
        // if the child is not a leaf hta, "forward" the next tileAt request to it 
        tiles[j] = tmp->tileAt(n-1, sv+1);   
      else 
	tiles[j] = tmp->tileAt(*sv[1]); 
      child_dist[j] = tiles[j]->memMapping__;
    }

    // (2) allocate and initialize the result
    // compute mem-mapping
    Tuple<DIM> linear_size = shape.high() - shape.low() + Tuple<DIM>::one;

    const MemMapping<DIM>* mmap = NULL;

    //    if (this->memMapping__)
      mmap = this->memMapping__->select(child_dist, linear_size);
      //else
      //delete [] child_dist;

      // same allocator, i.e., the return HTA is a view on the current HTA
    ret = new typename TRAIT::TypeImpl (this->level(), tiles, shape, this->alloc_, mmap, this->home());

    // initialize the distribution
    if (ret->isDistributed())
      {
	// current limitation - we do not have a distribution in closed form for a 
	// selection of tiles on a distributed array
	const GenericDistribution<DIM>* old_dist = dynamic_cast<const GenericDistribution<DIM>*>(this->dist__);    

	//if old dist is a Cyclic distribution.
	if (!old_dist)
	  old_dist = new GenericDistribution<DIM> (this->shape_);

	//the new distribution is a Sub-distribution of old distribution
	ret->initDistribution (old_dist->subDist (*sv[0]));
	ret->initLocalTiles(); 
	
      }

    return ret;
  }

  /** 
   * scalarAt operation using a Tuple 
   * @return reference to a scalar
   */
  inline T& scalarAt_leaf_ (const Tuple<DIM>& t) const
    //__attribute__((always_inline))
    //__attribute__((const))
    {
    assert (this->isAvailable_());
    //    assert (this->shape_.contains(t)); // bounds check;
    //Tuple<DIM> linearStep = (static_cast<const MemMapping<DIM>*>(this->memMapping__))->linearStep();
    //int ord =  t.x * linearStep.x + t.y * linearStep.y + t.z * linearStep.z + t.t * linearStep.t;
    //int ord = VectorOps<Tuple<DIM>, DIM>::dotProduct (t, linearStep);
    int ord = this->memMapping__->ord(t);

    assert (!this->isLocal() || // bounds check with allocator only for local tiles
	    this->alloc_.checkAddress_(this->memMap().offset() + ord));
    return data_leaf_[ord]; 
  }

  inline T& scalarAt (const Tuple<DIM>& t) const
  {
    if(this->isLeaf()) {
      return scalarAt_leaf_(t);
    }
    else {
      assert (this->level() > 0);
      assert (this->shape().low() == Tuple<DIM>::zero);
      assert (this->shape().step() == Tuple<DIM>::one);
      Order ord = this->memMap().order();
      
      switch (ord) {
      case TILE: {
	// determine the tile where t falls in
	
	Tuple<DIM> linear_size_scalarAt = this->memMap().leafSize();
	
	Tuple<DIM> tile_idx, rest_idx;
	for (int i = 0; i < DIM; ++i) {	  
	  int size = linear_size_scalarAt [i]/this->shape().size()[i];
	  
	  tile_idx[i]= t[i] / size;
	  rest_idx[i]= t[i] % size;
	} 
	
	// then call scalarAt recursively on that tile
	HTA_DEBUG (2, "HTAImpl::scalarAt t=" << t << "falls into tile " << tile_idx);
	typename TRAIT::TypeTileImpl* tile = tileAt(tile_idx);
	return tile->scalarAt(rest_idx);
      }
      case ROW: 
      case COL: 
      default: {
	assert (this->isHomogeneous_());
	assert (ord == ROW); // current limitation
#ifndef NDEBUG
	Tuple<DIM> linear_size_scalarAt = this->memMap().leafSize();
	assert (t >= Tuple<DIM>::zero && t < linear_size_scalarAt); // bounds check
#endif
	typename TRAIT::TypeInternal data_leaf_linear = this->raw();
	Tuple<DIM> linear_step_scalarAt = this->memMap().linearStep();
	int off =	VectorOps<Tuple<DIM>, DIM>::dotProduct (t, linear_step_scalarAt);
	return data_leaf_linear[off];
      }
      }
    }
  }
  
  template <int DIM2>
  inline T& scalarAt (const Tuple<DIM2>& t) const
  {
    assert (false);
  }
    
  inline T& scalarAt (const int& i) const
  {
    return scalarAt_ (i);
  }

  inline T& scalarAt (const int& i, const int& j) const
  {
    return scalarAt_ (i, j);
  }

protected:

  /** 
   * This accessor allows other leaf-instances to efficiently access the current 
   * instance. 
   *
   * @see map
   */
  T& scalarAt_ (int idx) const 
  {
    assert (this->isAvailable_());

    
#ifndef NDEBUG
    if (this->isLocal()) {
      // if this is a local access, then the address must be within bounds....
      //      cout << idx << " " << this->memMap().linearCard() << " " << this->memMap().offset() << endl;
      if(!this->alloc_.checkAddress_(this->memMap().offset() + idx)){
	cout<<"offset: "<<this->memMap().offset()<<" idx: "<<idx<<endl;
      }
      assert (this->alloc_.checkAddress_(this->memMap().offset() + idx)); 
    }
    else 
      // else - access to cache
      assert (idx < this->memMap().linearCard()); 
    //    assert (!this->memMap().isContiguous() || (idx >= 0 && idx < this->memMap().linearCard())); //JG: comment it for overlapped tiling.
#endif
    return data_leaf_[idx]; 
  }
  

  T& scalarAt_ (const typename Shape<DIM>::iterator& it) const 
  {
    assert (this->isAvailable_());

    // compute address
    int idx;
    if (this->memMap().isContiguous()) {
      // relative address corresponds to the ordinal number returned by th iterator
      // this is fast
      idx = it.ord();
    } else {
      Tuple<DIM> t = *it;
      idx = this->memMap().ord(t);
    }
    // if this is a local access, then the address must be within bounds....
    assert (!this->isLocal() || this->alloc_.checkAddress_(this->memMap().offset() + idx)); 
    
    // else - access to cache
    assert (this->isLocal() || idx < this->memMap().linearCard()); 
    
    // the following assertion may be violated if the shape covers only 
    // a part of the underlying array. It would still be OK to do the 
    // access
    // assert (idx < this->shape_.linearCard()); 
    
    HTA_DEBUG (3, "LeafHTAImpl::scalarAt_ = idx=" << idx);
    return data_leaf_[idx];
  }

public:

  void copy (const Tuple<DIM>* perm,
	      typename TRAIT::TypeImpl* dst,
	      unsigned permute_mask = 0xFF) const
  {
    copy_ (perm, dst, permute_mask);
  }
  /** 
   * scalarAt operation using a Shape
   * 
   * @return   HTA of the same level as this HTA which is a 'view' 
   *           restricted to the specified variables
   */
  typename TRAIT::TypeImpl* scalarAt (const Shape<DIM>& s) const
  {   
    //JG: comment this off because of overlapped tiling.
    //assert (this->shape_.contains(s.low()));  // bounds check;
    //assert (this->shape_.contains(s.high())); // bounds check;
    
    // compute the shape of the result
    int offset = this->memMap().offset() + this->memMap().ord(s.low());
    Shape<DIM> shape(s);
    shape.relocate(Tuple<DIM>::zero);

    Tuple<DIM> linear_size = s.high() - s.low() + Tuple<DIM>::one;
    MemMapping<DIM>* mmap = new LeafMemMapping<DIM> (offset, 0, linear_size,
						     this->memMap().leafPos() + s.low(), 
						     &this->memMap().linearStep()); 
    // same allocator, i.e., the return HTA is a view on the current HTA
    return new typename TRAIT::TypeImpl (this->level(), NULL, shape, this->alloc_, mmap, this->home());
  }

  //gti check these two

  //typename TRAIT::TypeImpl* tileAt (const Shape<DIM>& s) const 
  //{
  //return scalarAt(s);
  //}
  
  //inline typename TRAIT::TypeTileImpl* tileAt (int tile_idx) const
  //{
  //assert (false);
  //return NULL;
  //}




  // dotproduct ----------------------------------------------------------------------

  /** 
   * Dot product for leaf HTA
   */
  static void ldotproduct_leaf_ (const Array<T, DIM, TRAIT>& larg, 
				 const Array<T, DIM, TRAIT>& rarg,
				 Array<T, DIM, TRAIT>& dst)
  {
    assert (DIM == 2); // current limitation
    dst.scalarAt_(0) = dotproduct2D_ (larg, rarg);
  }


  /**
   * dot product on non leaf HTA
   */
  static void ldotproduct (const Array<T, DIM, TRAIT>& larg, 
			   const Array<T, DIM, TRAIT>& rarg,
			   Array<T, DIM, TRAIT>& dst)
  {
    assert (larg.shape() == rarg.shape());
    if(larg.isLeaf()){
      ldotproduct_leaf_(larg,rarg,dst);
    }
    else{
      for (int i=0 ; i < larg.numTiles_; i++) 
       {
          int j = larg.localTiles_[i];
	  TRAIT::TypeTileImpl::ldotproduct (*larg.tileAt(j), *rarg.tileAt(j), *dst.tileAt(j));
       }
    }
  }


  /**
   * dot product on non leaf HTA
   */
  static T dotproduct_non_leaf_ (const Array<T, DIM, TRAIT>& larg, 
		       const Array<T, DIM, TRAIT>& rarg)
  {
    assert (larg.shape() == rarg.shape());
    T ret = TypeHelper<T>::nullVal();
    
    int end = larg.shape().card();
    for (int i=0 ; i != end; i++) 
      ret += TRAIT::TypeTileImpl::dotproduct (*larg.tileAt(i), *rarg.tileAt(i));
    return ret;
  }

  
  static typename TRAIT::TypeScalar dotproduct_leaf_ (const Array<T, DIM, TRAIT>& larg, 
						const Array<T, DIM, TRAIT>& rarg)
  {
    assert (larg.isLocal());
    assert (rarg.isLocal());
    
    T ret;
    if (DIM <= 2) {
      switch (DIM) {
      case 1:
	ret = dotproduct1D_ (larg, rarg);
	break;
      case 2:
	ret = dotproduct2D_ (larg, rarg);
	break;
      default:
	ret = TypeHelper<T>::nullVal();
	break;
      }
    } 
    HTA_DEBUG(5, "LeafHTAImpl::dotproduct larg=" << larg << endl << " rarg= " << rarg << endl << " ret=" << ret);
    return ret;
  }


  /**
   * dot product on two HTAs
   */
  static typename TRAIT::TypeScalar dotproduct (const Array<T, DIM, TRAIT>& larg, 
						const Array<T, DIM, TRAIT>& rarg) 
  {
    if(larg.leaf())
    {
      return dotproduct_leaf_(larg, rarg);
    }
    else
    {
      return dotproduct_non_leaf_(larg, rarg);
    }
  }


  /** 
   * Dot product for leaf HTA; internal; protected 
   */
  static T dotproduct1D_ (const Array<T, DIM, TRAIT>& larg, 
			  const Array<T, DIM, TRAIT>& rarg)
  {
    assert (larg.shape() == rarg.shape());
    assert (DIM == 1 || larg.shape().size()[0] == 1 || larg.shape().size()[1] == 1);

    T ret = TypeHelper<T>::nullVal();
    const Tuple<DIM>& size = larg.shape().size();
    const Tuple<DIM> step_larg = larg.memMap().linearStep() * larg.shape().step();
    const Tuple<DIM> step_rarg = rarg.memMap().linearStep() * rarg.shape().step();
 
    int size_0 = size[0];
    int step_larg_0 = step_larg[0];
    int step_rarg_0 = step_rarg[0];
    
    for (int i = 0; i < size_0; ++ i) 
      ret += larg.scalarAt_(i * step_larg_0) * rarg.scalarAt_(i * step_rarg_0);

    return ret;
  }

  /** 
   * Dot product for leaf HTA
   */
  static T dotproduct2D_ (const Array<T, DIM, TRAIT>& larg, 
			  const Array<T, DIM, TRAIT>& rarg)
  {
    assert (larg.shape() == rarg.shape());
    assert (DIM == 2);

    T ret = TypeHelper<T>::nullVal();
    const Tuple<DIM>& size = larg.shape().size();
    const Tuple<DIM> step_larg = larg.memMap().linearStep() * larg.shape().step();
    const Tuple<DIM> step_rarg = rarg.memMap().linearStep() * rarg.shape().step();

    int size_0 = size[0];
    int step_larg_0 = step_larg[0];
    int step_rarg_0 = step_rarg[0];

    int size_1 = size[1];
    int step_larg_1 = step_larg[1];
    int step_rarg_1 = step_rarg[1];
    
    for (int j = 0; j < size_0; ++ j) 
      for (int i = 0; i < size_1; ++ i) 
	ret += larg.scalarAt_(i * step_larg_1 + j * step_larg_0) *
	  rarg.scalarAt_(i * step_rarg_1 + j * step_rarg_0);

    return ret;
  }


  
  // leaf-matmul ----------------------------------------------------------------------

  static void lmatmul (const Array<T, DIM, TRAIT>& larg, 
		       const Array<T, DIM, TRAIT>& rarg, 
		       Array<T, DIM, TRAIT>& dst)
  {
     if(larg.isLeaf()) {
      TRAIT::TypeTileImpl::matmul_leaf_(larg, rarg, dst);
     }else {
     
       assert (larg.shape() == rarg.shape());
       assert (larg.shape() == dst.shape());
       
       int end = larg.shape().card();
       for (int i=0 ; i != end; i++) 
	 if (larg.tileAt(i)->Array<T, DIM, TRAIT>::isLocal())
	   TRAIT::TypeTileImpl::lmatmul (*larg.tileAt(i), *rarg.tileAt(i), *dst.tileAt(i));
     }
  }

  // matmul ----------------------------------------------------------------------

  static void matmul_leaf_ (const Array<T, DIM, TRAIT>& larg, 
		      const Array<T, DIM, TRAIT>& rarg, 
		      Array<T, DIM, TRAIT>& dst) 
  {
    assert (DIM == 2 || (DIM == 3 && larg.shape().size()[2] == 1));
    assert (larg.shape().size()[1] == rarg.shape().size()[0]);
    assert (dst.shape().size()[0] == larg.shape().size()[0]);
    assert (dst.shape().size()[1] == rarg.shape().size()[1]);

    int M = larg.shape().size()[0];
    int Q = larg.shape().size()[1];
    int N = rarg.shape().size()[1];
    
    const Tuple<DIM> step_dst = dst.memMapping__->linearStep() * dst.shape().step();
    const Tuple<DIM> step_larg = larg.memMapping__->linearStep() * larg.shape().step();
    const Tuple<DIM> step_rarg = rarg.memMapping__->linearStep() * rarg.shape().step();

    int step_larg_0 = step_larg[0];
    int step_rarg_0 = step_rarg[0];
    int step_dst_0 = step_dst[0];
    
    int step_larg_1 = step_larg[1];
    int step_rarg_1 = step_rarg[1];
    int step_dst_1 = step_dst[1];

    T* dst_raw = dst.raw();
    const T* larg_raw = larg.raw();
    const T* rarg_raw = rarg.raw();

    for (int i = 0; i < M; i++) {
      for (int k = 0; k < Q; k++) {
	for (int j = 0; j< N; j++) 
	  //	  dst.scalarAt(Tuple<2>(i,j)) += larg.scalarAt(Tuple<2>(i,k)) * rarg.scalarAt(Tuple<2>(k,j));
	  dst_raw[i * step_dst_0 + j * step_dst_1] += larg_raw[i * step_larg_0 + k * step_larg_1] 
	    * rarg_raw[k * step_rarg_0 + j * step_rarg_1];
      }
    }

  } 

 class matmul1Body {
    const Array<T, DIM, TRAIT>& my_larg;
    const Array<T, DIM, TRAIT>& my_rarg;
    Array<T, DIM, TRAIT>& my_dst;
    int Q;
    int N;
  public:
    void operator() ( const tbb::blocked_range<int>& range ) const {
      //      cout << "matmul1Body" << endl;
      int i_end = range.end();
      for (int i = range.begin(); i < i_end; i++) {
	for (int k = 0; k < Q; k++) {
	  for (int j = 0; j< N; j++) 
	    TRAIT::TypeTileImpl::matmul (*my_larg.tileAt(Tuple<2>(i,k)), *my_rarg.tileAt(Tuple<2>(k,j)), *my_dst.tileAt(Tuple<2>(i,j)));
	}
      }
    }
    matmul1Body(const Array<T, DIM, TRAIT>& larg, 
		const Array<T, DIM, TRAIT>& rarg, 
		Array<T, DIM, TRAIT>& dst, int q, int n) : my_larg(larg), my_rarg(rarg), my_dst(dst), Q(q), N(n)
    {}
  }; 

  /** 
   * block-recursive matrix multiplication 
   */
  static void matmul (const Array<T, DIM, TRAIT>& larg, 
				const Array<T, DIM, TRAIT>& rarg, 
				Array<T, DIM, TRAIT>& dst) 
  {
    if(larg.isLeaf()) {
      TRAIT::TypeTileImpl::matmul_leaf_(larg, rarg, dst);
    }
    else {
      assert (DIM == 2);      
      assert (larg.shape().size()[1] == rarg.shape().size()[0]);
      assert (dst.shape().size()[0] == larg.shape().size()[0]);
      assert (dst.shape().size()[1] == rarg.shape().size()[1]);
      
      int M = larg.shape().size()[0];
      int Q = larg.shape().size()[1];
      int N = rarg.shape().size()[1];
      
      tbb::parallel_for(tbb::blocked_range<int>(0,M), matmul1Body(larg,rarg,dst,Q,N));
      // for (int i = 0; i< M; i++) {
// 	for (int k = 0; k < Q; k++) {
// 	  for (int j = 0; j< N; j++) 
// 	    TRAIT::TypeTileImpl::matmul (*larg.tileAt(Tuple<2>(i,k)), *rarg.tileAt(Tuple<2>(k,j)), *dst.tileAt(Tuple<2>(i,j)));
// 	}
//       }
    }
  }

  virtual bool isAvailable_ () const 
  {
    assert (this->level() == 0); // this method is available for LeafHTAs only.
    return data_leaf_ != NULL;
  }
 

  /**
   * Map operator framework
   */
 template <typename OP>
  void mapReduce_leaf_ (OP& mr_op)
  { 
    switch (mr_op.keyType()) {
    case Operator::MR_NONE: {
      // TODO: specialize calls for DIM=1,2,3 like map
      // call map for all elements with argument key=tuple_scalarAt, val     
      switch (DIM) {
      case 1:
	for (int i = 0; i < this->shape().card(); i++) {
	  T& val = this->scalarAt_(i);
	  mr_op.map_none(val);
	}
	break;
      default:
	typename Shape<DIM>::iterator i = this->shape().begin();
	typename Shape<DIM>::iterator end = this->shape().end();
	for ( ; i != end; i++) {
	  Tuple<DIM> t = *i;
	  T& val = this->scalarAt(t);
	  mr_op.map_none(val);
	}
      }
      break;
    }
    case Operator::MR_PROC_OFF: {
      assert (false); // to be implemented
      // call map for all elements with argument key=(proc, idx), val
      break;
    }
    case Operator::MR_TUPLE: {
      // TODO: specialize calls for DIM=1,2,3 like map
      // call map for all elements with argument key=tuple_scalarAt, val
      typename Shape<DIM>::iterator i = this->shape().begin();
      typename Shape<DIM>::iterator end = this->shape().end();


      Tuple<DIM> pos; 
      //JG: modified for overlapped tiling
      if(this->actual_().ol()){
	Tuple<DIM> temp = this->memMap().leafPos();
	pos = temp/this->memMap().tiling()[0];
	pos=pos*this->shape().size();
      }else{
	pos = this->memMap().leafPos();
      }

      for ( ; i != end; i++) {
	Tuple<DIM> t = *i;
	T val = this->scalarAt(t);
	mr_op.map_tuple(t + pos, val);
      }
    }  
    } // switch
  }

  /**
   * Map operator framework
   */
 template <typename OP>
 void mapReduce_leaf_ (OP& mr_op,  
		       typename TRAIT::TypeImpl& arg1)
  { 
    int M = 1;
    switch (mr_op.keyType()) {
    case Operator::MR_NONE: {
      switch (DIM){
      case 1: {
	// TODO: specialize calls for DIM=1,2,3 like map	
	// call map for all elements with argument key=tuple_scalarAt, val
	Tuple<DIM> step = this->memMap().linearStep() * this->shape().step();
	int step_dst_0 = step[0];	
	for (int i = 0; i < this->shape().card(); i++) {
	  T& val = this->scalarAt_(i * step_dst_0);
	  mr_op.map_none(val, arg1.scalarAt_(i * step_dst_0));
	}
	break;
      }
      default:
      	typename Shape<DIM>::iterator i = this->shape().begin();
	typename Shape<DIM>::iterator end = this->shape().end();
	for ( ; i != end; i++) {
	  Tuple<DIM> t = *i;
	  T& val = this->scalarAt(t);
	  mr_op.map_none(val, arg1.scalarAt(t));
	}
      }
      break;
    }
    case Operator::MR_PROC_OFF: {
      assert (false); // to be implemented
      // call map for all elements with argument key=(proc, idx), val
      break;
    }
    case Operator::MR_TUPLE: {
      // TODO: specialize calls for DIM=1,2,3 like map
      // call map for all elements with argument key=tuple_scalarAt, val
      typename Shape<DIM>::iterator i = this->shape().begin();
      typename Shape<DIM>::iterator end = this->shape().end();
      Tuple<DIM> pos; 
      //JG: modified for overlapped tiling
      if(this->actual_().ol()){
	Tuple<DIM> temp = this->memMap().leafPos();
	pos = temp/this->memMap().tiling()[0];
	pos = pos*this->shape().size();

      }else{
	pos = this->memMap().leafPos();
      }
      for ( ; i != end; i++) {
	Tuple<DIM> t = *i;
	T& val = this->scalarAt(t);
	mr_op.map_tuple(t + pos, val, arg1.scalarAt(t));
      }
    }  
    } // switch
  }

  template <typename OP>
  class HTAmapReduce1Body {
    HTAImpl<T,DIM,TRAIT>* hta;
    OP& my_mr_op;
    const std::vector< Tuple<DIM> >& vec;
  public:
    void operator() ( const tbb::blocked_range<int>& range ) const {
      cout << "HTAmapReduce1Body" << endl;
      int i_end = range.end();
      for (int i = range.begin(); i < i_end; i++) {
	Tuple<DIM> t = vec[i];
	typename TRAIT::TypeTileImpl* tile = hta->tileAt(t);
	//I hope the compiler will peform dead code elimination 
	//if mr_op.init() is empty. 
	MapContext<T, DIM> context (t, hta, tile);
	my_mr_op.init(context); 
	tile->mapReduce(my_mr_op); // recurse
      }
    }
    HTAmapReduce1Body(HTAImpl<T,DIM,TRAIT>* a, OP& mr_op, const std::vector< Tuple<DIM> >& data) : hta(a), my_mr_op(mr_op), vec(data)
    {}
  };

  template <typename OP>
  void mapReduce (OP& mr_op)
  { 
    if(this->isLeaf()) {
      mapReduce_leaf_(mr_op);
    }
    else {
      assert (this->level() > 0);
      assert (this->isLocal());
      
      typename Shape<DIM>::iterator i = this->shape().begin();
      typename Shape<DIM>::iterator end = this->shape().end();
      std::vector< Tuple<DIM> > tuples;
      
      for ( ; i != end; i++) {
 	tuples.push_back(*i);
      }
      tbb::parallel_for(tbb::blocked_range<int>(0, tuples.size()), HTAmapReduce1Body<OP>(this, mr_op, tuples));
      
      // for ( ; i != end; i++) {
// 	Tuple<DIM> t = *i;
// 	typename TRAIT::TypeTileImpl* tile = this->tileAt(t);
// 	//I hope the compiler will peform dead code elimination 
// 	//if mr_op.init() is empty. 
// 	MapContext<T, DIM> context (t, this, tile);
// 	mr_op.init(context); 
// 	tile->mapReduce(mr_op); // recurse
//       }
    }
    // no reduction in the serial version.
  }


  template <typename OP>
  class HTAmapReduce2Body {
    HTAImpl<T, DIM, TRAIT>* hta;
    OP& my_mr_op;
    typename TRAIT::TypeImpl& my_arg1;
    std::vector< Tuple<DIM> >& vec;
  public:
    void operator() ( const tbb::blocked_range<int>& range ) const {
      int i_end = range.end();
      for (int i = range.begin(); i < i_end; i++) {
	Tuple<DIM> t = vec[i];
	typename TRAIT::TypeTileImpl* tile = hta->tileAt(t);
	//I hope the compiler will peform dead code elimination 
	//if mr_op.init() is empty. 
	MapContext<T, DIM> context (t, hta, tile);
	my_mr_op.init(context); 
	tile->mapReduce(my_mr_op, *my_arg1.tileAt(t)); // recurse
      }
    }
    HTAmapReduce2Body(HTAImpl<T,DIM,TRAIT>* a, OP& mr_op, typename TRAIT::TypeImpl& arg1, std::vector< Tuple<DIM> >& data) :
      hta(a), my_mr_op(mr_op), my_arg1(arg1), vec(data)
    {}
  };
  
  template <typename OP>
  void mapReduce (OP& mr_op, 
		  typename TRAIT::TypeImpl& arg1)
  { 
    if(this->isLeaf()) {
      mapReduce_leaf_(mr_op, arg1);
    }
    else {
      assert (this->level() > 0);
      assert (this->isLocal());
      
      typename Shape<DIM>::iterator i = this->shape().begin();
      typename Shape<DIM>::iterator end = this->shape().end();
      std::vector< Tuple<DIM> > tuples;
      for ( ; i != end; i++) {
	tuples.push_back(*i);
      }
      tbb::parallel_for(tbb::blocked_range<int>(0,tuples.size()), HTAmapReduce2Body<OP>(this, mr_op, arg1, tuples));
      
      // for ( ; i != end; i++) {
// 	Tuple<DIM> t = *i;
// 	typename TRAIT::TypeTileImpl* tile = this->tileAt(t);
// 	//I hope the compiler will peform dead code elimination 
// 	//if mr_op.init() is empty. 
// 	MapContext<T, DIM> context (t, this, tile);
// 	mr_op.init(context); 
// 	tile->mapReduce(mr_op, *arg1.tileAt(*i)); // recurse
//       }
    }

    // no reduction in the serial version.
  }


  //----------------------------------------
  //from leaf hta impl
  //----------------------------------------
  template <typename OP>
  static void map_leaf_ (const OP& nullary_op, 
		   Array<T, DIM, TRAIT>& dst)
  {
    assert (dst.isLocal());
   

    HTAImpl<T, DIM, TRAIT>* dst_actual = dynamic_cast<HTAImpl<T, DIM, TRAIT>*>(&dst); 

    if (DIM <= 3 && dst_actual) {
      switch (DIM) {
      case 1:
	map1D_ (nullary_op, dst);
	break;
      case 2:
	map2D_ (nullary_op, dst);
	break;
      case 3:
	map3D_ (nullary_op, *dst_actual);
	break;
      default:
	assert(false);
	break;
      }
    } else {
      assert (DIM > 0);
      if (dst.shape().isContiguous() && dst.memMapping__->isContiguous() && dst.ol()==NULL) {//JG: exclude the overlapped case.
	// TODO: use a mathkernel to do this
	// the iteration pattern on the linear-representation has step 1 for both arrays 
	int end = dst.shape().card();
	for (int i = 0 ; i < end; i++) 
	  dst.scalarAt_(i) = nullary_op();
      } else {
	// this is most inefficient!
	// different iteration patterns on both arrays
	typename Shape<DIM>::iterator i = dst.shape().begin();
	typename Shape<DIM>::iterator end = dst.shape().end();
	for ( ; i != end; i++) 
	  dst.scalarAt(*i) = nullary_op();
      }
    }
  }
  
  /** unary map */
  template <typename T2, typename OP, typename TRAIT2>
  static void map_leaf_ (const OP& unary_op, 
		   const Array<T, DIM, TRAIT>& arg, 
		   Array<T2, DIM, TRAIT2>& dst)
  {
    assert (arg.isAvailable_());
    assert (dst.isLocal());
    assert (Shape<DIM>::conformable(arg.shape(), dst.shape()));

    const HTAImpl<T, DIM, TRAIT>* arg_actual = dynamic_cast<const HTAImpl<T, DIM, TRAIT>*>(&arg); 
    HTAImpl<T, DIM, TRAIT>* dst_actual = dynamic_cast<HTAImpl<T, DIM, TRAIT>*>(&dst); 
 
    if (DIM <= 4 && arg_actual && dst_actual) {
      switch (DIM) {
      case 1:
	map1D_ (unary_op, arg, dst);
	break;
      case 2:
	map2D_ (unary_op, arg, dst);
	break;
      case 3: 
	map3D_ (unary_op, *arg_actual, *dst_actual);
	break;
      case 4:
	map4D_ (unary_op, *arg_actual, *dst_actual);
	break;
      default:
	assert (false);
	break;
      }
    } else {
      assert (DIM > 0);
      if (arg.shape().isContiguous() && dst.shape().isContiguous() &&
	  arg.memMapping__->isContiguous() && dst.memMapping__->isContiguous()&& !arg.ol() && !dst.ol()) {//JG: Exclude the overlapped case
	// TODO: use a mathkernel to do this
	// the iteration pattern on the linear-representation has step 1 for both arrays 
	int end = arg.shape().card();
	for (int i = 0; i < end; i++) 
	  dst.scalarAt_(i) =  static_cast<T2>(unary_op(arg.scalarAt_(i)));
      } else if (Shape<DIM>::iteratorConformable_(arg.shape(), dst.shape())) {
	// on the linearized representation, the iteration pattern is the same 
	// for both arrays, however not just step 1
	typename Shape<DIM>::iterator arg_it = arg.shape().begin();
	typename Shape<DIM>::iterator arg_end = arg.shape().end();
	for ( ; arg_it != arg_end; arg_it++) 
	  dst.scalarAt_(arg_it) = static_cast<T2>(unary_op(arg.scalarAt_(arg_it)));
      } else {
	// this is most inefficient!
	// different iteration patterns on both arrays
	typename Shape<DIM>::iterator arg_it = arg.shape().begin();
	typename Shape<DIM>::iterator arg_end = arg.shape().end();
	typename Shape<DIM>::iterator dst_it = dst.shape().begin();
	for ( ; arg_it != arg_end; arg_it++, dst_it++)
	  dst.scalarAt(*dst_it) = static_cast<T2>(unary_op(arg.scalarAt(*arg_it)));
      }
    }
  }
   
  /** unary map */
  template <typename T2, typename OP, typename OP2, typename TRAIT2>
  static void map_leaf_accum_ (const OP& unary_op, 
			 const OP2& acc_op,
			 const Array<T, DIM, TRAIT>& arg, 
			 Array<T2, DIM, TRAIT2>& dst)
  {
    assert (arg.isAvailable_());
    assert (dst.isLocal());
    assert (Shape<DIM>::conformable(arg.shape(), dst.shape()));

    const HTAImpl<T, DIM, TRAIT>* arg_actual = dynamic_cast<const HTAImpl<T, DIM, TRAIT>*>(&arg); 
    HTAImpl<T, DIM, TRAIT>* dst_actual = dynamic_cast<HTAImpl<T, DIM, TRAIT>*>(&dst); 
 
    //    cout << "here 1 " << endl;

    if (DIM <= 4 && arg_actual && dst_actual) {
      switch (DIM) {     
      case 3: 
	map3DAccum_ (unary_op, acc_op, *arg_actual, *dst_actual);
	break;
      case 4:
	map4DAccum_ (unary_op, acc_op, *arg_actual, *dst_actual);
	break;
      default:
	assert (false);
	break;
      }
    } else {
      assert (false);
    }
  }
   
  /** binary map */
  template <typename OP, typename T2, typename TRAIT2>
  static void map_leaf_ (const OP& binary_op, 
		   const Array<T, DIM, TRAIT>& larg, 
		   const Array<T2, DIM, TRAIT2>& rarg, 
		   Array<T, DIM, TRAIT>& dst) 
  {
    assert (larg.isAvailable_());
    assert (rarg.isAvailable_());
    assert (dst.isLocal());

    if (rarg.shape().card() == 1) {
      TRAIT::TypeImpl::map (binder2nd<OP>(binary_op, rarg.scalarAt_(0)), larg, dst);
      return;
    }

    assert (Shape<DIM>::conformable(rarg.shape(), dst.shape()) &&
	    Shape<DIM>::conformable(larg.shape(), dst.shape()));
    assert (Shape<DIM>::conformable(larg.shape(), rarg.shape()));

    const HTAImpl<T, DIM, TRAIT>* larg_actual = dynamic_cast<const HTAImpl<T, DIM, TRAIT>*>(&larg); 
    const HTAImpl<T2, DIM, TRAIT2>* rarg_actual = dynamic_cast<const HTAImpl<T2, DIM, TRAIT2>*>(&rarg); 
    HTAImpl<T, DIM, TRAIT>* dst_actual = dynamic_cast<HTAImpl<T, DIM, TRAIT>*>(&dst); 

    if (DIM <= 4 && larg_actual && rarg_actual && dst_actual) {
      switch (DIM) {
      case 1:
	map1D_ (binary_op, larg, rarg, dst);
	break;
      case 2:
	map2D_ (binary_op, *larg_actual, *rarg_actual,  *dst_actual);
	break;
      case 3:
	map3D_ (binary_op, *larg_actual, *rarg_actual, *dst_actual);
	break;     
      default:
	assert (false);
	break;
      }
    } else {

      // this is a limitation of the current implementation
      assert (Shape<DIM>::iteratorConformable_(larg.shape(), dst.shape()));

      if (larg.shape().isContiguous() && rarg.shape().isContiguous() && dst.shape().isContiguous() &&
	  larg.memMapping__->isContiguous() && rarg.memMapping__->isContiguous() && dst.memMapping__->isContiguous()) {
	// TODO: this might be done by an efficient mathkernel 
	int end = larg.shape().card();
	for (int idx = 0;  idx < end; idx++) {
	  // int idx = i.ord();
	  dst.scalarAt_(idx) = binary_op(larg.scalarAt_(idx), rarg.scalarAt_(idx));
	}
      } else if (Shape<DIM>::iteratorConformable_(larg.shape(), rarg.shape())) {
	typename Shape<DIM>::iterator i = larg.shape().begin();
	typename Shape<DIM>::iterator end = larg.shape().end();
	for ( ; i != end; i++)
	  dst.scalarAt_(i) = binary_op(larg.scalarAt_(i), rarg.scalarAt_(i));
      } else {
	// this is most inefficient!
	// different iteration patterns on both arrays
	typename Shape<DIM>::iterator larg_it = larg.shape().begin();
	typename Shape<DIM>::iterator larg_end = larg.shape().end();
	typename Shape<DIM>::iterator rarg_it = rarg.shape().begin();
	for ( ; larg_it != larg_end; larg_it++, rarg_it++)
	  dst.scalarAt_(larg_it) = binary_op(larg.scalarAt_(larg_it), rarg.scalarAt_(rarg_it));
      }
    }
  }


  // specialization of nullary map for DIM == 1. 
  template <typename OP> 
  static void map1D_ (const OP& nullary_op, 
			  Array<T, DIM, TRAIT>& dst) 
  {
    assert (DIM == 1);
    const Tuple<DIM>& size = dst.shape().size();
    Tuple<DIM> step = dst.memMapping__->linearStep() * dst.shape().step();

    int size_0 = size[0];
    int step_dst_0 = step[0];
    
    for (int i = 0; i < size_0; ++ i) 
      dst.scalarAt_(i * step_dst_0) = nullary_op();
  }

   // specialization of nullary map for DIM == 2. 
  template <typename OP> 
  static void map2D_ (const OP& nullary_op, 
			  Array<T, DIM, TRAIT>& dst) 
  {
    assert (DIM == 2);
    const Tuple<DIM>& size = dst.shape().size();
    const Tuple<DIM> step = dst.memMapping__->linearStep() * dst.shape().step();

    int size_0 = size[0];
    int step_dst_0 = step[0];

    int size_1 = size[1];
    int step_dst_1 = step[1];
    
    T* dst_raw = dst.raw();
    for (int j = 0; j < size_0; ++ j) 
      for (int i = 0; i < size_1; ++ i) 
	dst_raw [i * step_dst_1 + j * step_dst_0] = nullary_op();
  }

  // specialization of nullary map for DIM == 3. 
  template <typename OP> 
  static void map3D_ (const OP& nullary_op, 
		      HTAImpl<T, DIM, TRAIT>& dst) 
  {
    assert (DIM == 3);
    
    const Tuple<DIM>& size = dst.shape().size();
    const Tuple<DIM> step = dst.memMapping__->linearStep() * dst.shape().step();
    
    int size_0 = size[0];
    int step_dst_0 = step[0];
    
    int size_1 = size[1];
    int step_dst_1 = step[1];
    
    int size_2 = size[2];
    int step_dst_2 = step[2];

    T* dst_raw = dst.raw();

    for (int k = 0; k < size_0; ++ k)  
      for (int j = 0; j < size_1; ++ j)
	for (int i = 0; i < size_2; ++ i) 
	  dst_raw[i * step_dst_2 + j * step_dst_1 + k * step_dst_0] = nullary_op();
  }

  // specialization of unary map for DIM == 1. 
  template <typename T2, typename OP, typename TRAIT2> 
  static void map1D_ (const OP& unary_op, 
		      const Array<T, DIM, TRAIT>& arg, 
		      Array<T2, DIM, TRAIT2>& dst) 
  {
    assert (DIM == 1);
    const Tuple<DIM>& size = arg.shape().size();
    const Tuple<DIM> step_dst = dst.memMapping__->linearStep() * dst.shape().step();
    const Tuple<DIM> step_arg = arg.memMapping__->linearStep() * arg.shape().step();

    int size_0 = size[0];
    int step_arg_0 = step_arg[0];
    int step_dst_0 = step_dst[0];
    
    for (int i = 0; i < size_0; ++ i) 
      dst.scalarAt_(i * step_dst_0) = 
	static_cast<T2>(unary_op(arg.scalarAt_(i * step_arg_0)));
  }

  // specialization of unary map for DIM == 2. 
  template <typename T2, typename OP, typename TRAIT2> 
  static void map2D_ (const OP& unary_op, 
		      const Array<T, DIM, TRAIT>& arg, 
		      Array<T2, DIM, TRAIT2>& dst) 
  {
    assert (DIM == 2);
    const Tuple<DIM>& size = arg.shape().size();
    const Tuple<DIM> step_dst = dst.memMapping__->linearStep() * dst.shape().step();
    const Tuple<DIM> step_arg = arg.memMapping__->linearStep() * arg.shape().step();

    int size_0 = size[0];
    int step_arg_0 = step_arg[0];
    int step_dst_0 = step_dst[0];

    int size_1 = size[1];
    int step_arg_1 = step_arg[1];
    int step_dst_1 = step_dst[1];
    
    for (int j = 0; j < size_0; ++ j) 
      for (int i = 0; i < size_1; ++ i) 
	dst.scalarAt_(i * step_dst_1 + j * step_dst_0) = 
	  static_cast<T2>(unary_op(arg.scalarAt_(i * step_arg_1 + j * step_arg_0)));
  }

  // specialization of unary map for DIM == 3. 
  template <typename T2, typename OP, typename TRAIT2> 
  static void map3D_ (const OP& unary_op, 
		      const HTAImpl<T, DIM, TRAIT>& arg, 
		      HTAImpl<T2, DIM, TRAIT2>& dst) 
  {
    assert (DIM == 3);

    const Tuple<DIM>& size = arg.shape().size();
    const Tuple<DIM> step_dst = dst.memMapping__->linearStep() * dst.shape().step();
    const Tuple<DIM> step_arg = arg.memMapping__->linearStep() * arg.shape().step();

    int size_0 = size[0];
    int step_arg_0 = step_arg[0];
    int step_dst_0 = step_dst[0];

    int size_1 = size[1];
    int step_arg_1 = step_arg[1];
    int step_dst_1 = step_dst[1];

    int size_2 = size[2];
    int step_arg_2 = step_arg[2];
    int step_dst_2 = step_dst[2];

    T* dst_raw = dst.raw();
    const T* arg_raw = arg.raw();

    for (int k = 0; k < size_0; ++ k)  
      for (int j = 0; j < size_1; ++ j)
	for (int i = 0; i < size_2; ++ i) 
	  dst_raw[i * step_dst_2 + j * step_dst_1 + k * step_dst_0] = 
	    static_cast<T2>(unary_op(arg_raw[i * step_arg_2 + j * step_arg_1 + k * step_arg_0]));
  }

  // specialization of unary map for DIM == 3. 
  template <typename T2, typename OP, typename TRAIT2> 
  static void map4D_ (const OP& unary_op, 
		      const HTAImpl<T, DIM, TRAIT>& arg, 
		      HTAImpl<T2, DIM, TRAIT2>& dst) 
  {
    assert (DIM == 4);

    const Tuple<DIM>& size = arg.shape().size();
    const Tuple<DIM> step_dst = dst.memMapping__->linearStep() * dst.shape().step();
    const Tuple<DIM> step_arg = arg.memMapping__->linearStep() * arg.shape().step();
    int size_0 = size[0];
    int step_arg_0 = step_arg[0];
    int step_dst_0 = step_dst[0];

    int size_1 = size[1];
    int step_arg_1 = step_arg[1];
    int step_dst_1 = step_dst[1];

    int size_2 = size[2];
    int step_arg_2 = step_arg[2];
    int step_dst_2 = step_dst[2];

    int size_3 = size[3];
    int step_arg_3 = step_arg[3];
    int step_dst_3 = step_dst[3];

    T* dst_raw = dst.raw();
    const T* arg_raw = arg.raw();

    for (int k = 0; k < size_0; ++ k)  
      for (int j = 0; j < size_1; ++ j)
	for (int i = 0; i < size_2; ++ i) 
	  for (int m = 0; m < size_3; ++m)
	    dst_raw[m * step_dst_3 + i * step_dst_2 + j * step_dst_1 + k * step_dst_0] = 
	      static_cast<T2>(unary_op(arg_raw[m * step_arg_3 + i * step_arg_2 + j * step_arg_1 + k * step_arg_0]));
  }



 // specialization of unary map for DIM == 3. 
  template <typename T2, typename OP, typename OP2, typename TRAIT2> 
  static void map3DAccum_ (const OP& unary_op, 
		      const OP2& acc_op,
		      const HTAImpl<T, DIM, TRAIT>& arg, 
		      HTAImpl<T2, DIM, TRAIT2>& dst) 
  {
    assert (DIM == 3);

    const Tuple<DIM>& size = arg.shape().size();
    const Tuple<DIM> step_dst = dst.memMapping__->linearStep() * dst.shape().step();
    const Tuple<DIM> step_arg = arg.memMapping__->linearStep() * arg.shape().step();

    int size_0 = size[0];
    int step_arg_0 = step_arg[0];
    int step_dst_0 = step_dst[0];

    int size_1 = size[1];
    int step_arg_1 = step_arg[1];
    int step_dst_1 = step_dst[1];

    int size_2 = size[2];
    int step_arg_2 = step_arg[2];
    int step_dst_2 = step_dst[2];

    T* dst_raw = dst.raw();
    const T* arg_raw = arg.raw();

    //cout << "here " << endl;

    for (int k = 0; k < size_0; ++ k)  
      for (int j = 0; j < size_1; ++ j)
	for (int i = 0; i < size_2; ++ i) 
	  dst_raw[i * step_dst_2 + j * step_dst_1 + k * step_dst_0] = 
	    acc_op(dst_raw[i * step_dst_2 + j * step_dst_1 + k * step_dst_0],
		   static_cast<T2>(unary_op(arg_raw[i * step_arg_2 + j * step_arg_1 + k * step_arg_0])));
  }

// specialization of unary map for DIM == 3. 
  template <typename T2, typename OP, typename OP2, typename TRAIT2> 
  static void map4DAccum_ (const OP& unary_op, 
		      const OP2& acc_op,
		      const HTAImpl<T, DIM, TRAIT>& arg, 
		      HTAImpl<T2, DIM, TRAIT2>& dst) 
  {
    assert (DIM == 4);

    const Tuple<DIM>& size = arg.shape().size();
    const Tuple<DIM> step_dst = dst.memMapping__->linearStep() * dst.shape().step();
    const Tuple<DIM> step_arg = arg.memMapping__->linearStep() * arg.shape().step();

    int size_0 = size[0];
    int step_arg_0 = step_arg[0];
    int step_dst_0 = step_dst[0];

    int size_1 = size[1];
    int step_arg_1 = step_arg[1];
    int step_dst_1 = step_dst[1];

    int size_2 = size[2];
    int step_arg_2 = step_arg[2];
    int step_dst_2 = step_dst[2];

    int size_3 = size[3];
    int step_arg_3 = step_arg[3];
    int step_dst_3 = step_dst[3];

    T* dst_raw = dst.raw();
    const T* arg_raw = arg.raw();

    //cout << "here " << endl;

    for (int k = 0; k < size_0; ++ k)  
      for (int j = 0; j < size_1; ++ j)
	for (int i = 0; i < size_2; ++ i) 
	  for (int m = 0; m < size_3; ++ m) 	  
	    dst_raw[m * step_arg_3 + i * step_dst_2 + j * step_dst_1 + k * step_dst_0] = 
	      acc_op(dst_raw[m * step_arg_3 + i * step_dst_2 + j * step_dst_1 + k * step_dst_0],
		     static_cast<T2>(unary_op(arg_raw[m * step_arg_3 + i * step_arg_2 + j * step_arg_1 + k * step_arg_0])));
  }


  // specialization of binary map for DIM == 1. 
  template<typename OP, typename T2, typename TRAIT2>
  static void map1D_ (const OP& binary_op, 
		      const Array<T, DIM, TRAIT>& larg, 
		      const Array<T2, DIM, TRAIT2>& rarg, 
		      Array<T, DIM, TRAIT>& dst) 
  {
    assert(DIM == 1);

    const Tuple<DIM>& size = larg.shape().size();
    const Tuple<DIM> step_dst = dst.memMapping__->linearStep() * dst.shape().step();
    const Tuple<DIM> step_larg = larg.memMapping__->linearStep() * larg.shape().step();
    const Tuple<DIM> step_rarg = rarg.memMapping__->linearStep() * rarg.shape().step();

    int size_0 = size[0];
    int step_larg_0 = step_larg[0];
    int step_rarg_0 = step_rarg[0];
    int step_dst_0 = step_dst[0];

    for (int i = 0; i < size_0; ++ i) 
      dst.scalarAt_(i * step_dst_0) = 
	binary_op(larg.scalarAt_(i * step_larg_0), rarg.scalarAt_(i * step_rarg_0));
  } 

  // specialization of binary map for DIM == 2. 
  // iterate separately over each dimension
  template <typename OP, typename T2, typename TRAIT2> 
  static void map2D_ (const OP& binary_op, 
			  const HTAImpl<T, DIM, TRAIT>& larg, 
			  const HTAImpl<T2, DIM, TRAIT2>& rarg, 
			  HTAImpl<T, DIM, TRAIT>& dst) 
  {
    assert (DIM == 2);

    const Tuple<DIM>& size = larg.shape().size();   
    int size_0 = size[0];
    int size_1 = size[1];

    T* dst_raw = dst.raw();
    const T* larg_raw = larg.raw();   
    const T2* rarg_raw = rarg.raw();
    
    assert (dst_raw);
    assert (larg_raw);
    assert (rarg_raw);
  
    const Tuple<DIM> step_dst = dst.memMapping__->linearStep() * dst.shape().step();
    const Tuple<DIM> step_larg = larg.memMapping__->linearStep() * larg.shape().step();
    const Tuple<DIM> step_rarg = rarg.memMapping__->linearStep() * rarg.shape().step();
    
    int step_larg_0 = step_larg[0];
    int step_rarg_0 = step_rarg[0];
    int step_dst_0 = step_dst[0];
    
    int step_larg_1 = step_larg[1];
    int step_rarg_1 = step_rarg[1];
    int step_dst_1 = step_dst[1];      
    for (int k = 0; k < size_0; ++ k) 
      for (int j = 0; j < size_1; ++ j)  
	dst_raw[j * step_dst_1 + k * step_dst_0] = 
	  binary_op(larg_raw [j * step_larg_1 + k * step_larg_0], 
		    rarg_raw [j * step_rarg_1 + k * step_rarg_0]);		    
  }      
  
  // specialization of binary map for DIM == 3. 
  // iterate separately over each dimension
  template <typename OP, typename T2, typename TRAIT2> 
  static void map3D_ (const OP& binary_op, 
		      const HTAImpl<T, DIM, TRAIT>& larg, 
		      const HTAImpl<T2, DIM, TRAIT2>& rarg, 
		      HTAImpl<T, DIM, TRAIT>& dst) 
  {

    assert (DIM == 3);
    const Shape<DIM>& shpe = larg.shape();
    const Tuple<DIM>& size = shpe.size();
    const Tuple<DIM> step_dst = dst.memMapping__->linearStep() * dst.shape().step();
    const Tuple<DIM> step_larg = larg.memMapping__->linearStep() * larg.shape().step();
    const Tuple<DIM> step_rarg = rarg.memMapping__->linearStep() * rarg.shape().step();

    int size_0 = size[0];
    int step_larg_0 = step_larg[0];
    int step_rarg_0 = step_rarg[0];
    int step_dst_0 = step_dst[0];
    
    int size_1 = size[1];
    int step_larg_1 = step_larg[1];
    int step_rarg_1 = step_rarg[1];
    int step_dst_1 = step_dst[1];

    int size_2 = size[2];
    int step_larg_2 = step_larg[2];
    int step_rarg_2 = step_rarg[2];
    int step_dst_2 = step_dst[2];
    
    T* dst_raw = dst.raw();
    const T* larg_raw = larg.raw();
    const T2* rarg_raw = rarg.raw();
    if (larg.shape().isContiguous() && rarg.shape().isContiguous() && dst.shape().isContiguous() &&
	  larg.memMapping__->isContiguous() && rarg.memMapping__->isContiguous() && dst.memMapping__->isContiguous()
	&&(rarg.actual_().ol()==NULL) && (larg.actual_().ol()==NULL)){ //JG: add to exclude overlapped tiling case.
   
      for (int k = 0; k < size_0; ++ k) 
	for (int j = 0; j < size_1; ++ j)  
	  for (int i = 0; i < size_2; ++ i) {
	    //	 dst_raw[i * step_dst_2 + j * step_dst_1 + k * step_dst_0] = 
	    // binary_op(larg_raw [i * step_larg_2 + j * step_larg_1 + k * step_larg_0], 
	    //     rarg_raw [i * step_rarg_2 + j * step_rarg_1 + k * step_rarg_0]);	
	    *dst_raw++ = binary_op (*larg_raw++, *rarg_raw++);
	  }
    } else {
      for (int k = 0; k < size_0; ++ k) 
	for (int j = 0; j < size_1; ++ j)  
	  for (int i = 0; i < size_2; ++ i) {
	    dst_raw[i * step_dst_2 + j * step_dst_1 + k * step_dst_0] = 
	      binary_op(larg_raw [i * step_larg_2 + j * step_larg_1 + k * step_larg_0], 
			rarg_raw [i * step_rarg_2 + j * step_rarg_1 + k * step_rarg_0]);	
	    
	  }
    }
  }
  
  //-----------------------------------------------------
  //    end map_methods
  //-----------------------------------------------------
  template <typename OP>
  void reduceL (OP& mr_op) const
  {
    //do nothing
    //return T();
  }

  virtual void communicate_ (const Distribution<DIM>& target_dist, 
			    Array<T, DIM, TRAIT>* target_arr=NULL) const
  { 
    //GBDEBUG : some bug here!
    if (target_arr)
      target_arr->actual_() = this->actual_();
  }
  
  template <typename T2, typename TRAIT2>
  inline void communicate_ (Array<T2, DIM, TRAIT2>& target) const
  {
    // nothing to be done in the serial version
  } 
  
  inline void invalidate_ () const 
  {
    // nothing to be done in the serial version
  }


  typename TRAIT::TypeInternal raw() const //__attribute__((always_inline))
  {
    
    if(this->isLeaf())       
      return data_leaf_;
    else 
      return base::raw();
  }
  

protected:

  // all other template instance of this class should be friends (required when accessing print_)
  template<class T2, int DIM2, typename TRAIT2>
  friend class HTAImpl;

  // all other template instance of this class should be friends (required when accessing print_)
   template<class T2, int DIM2, typename TRAIT2>
  friend class HTASparseImpl;

  // to allow forwarding class to private methods form Array to this class.
  template<class T2, int DIM2, typename TRAIT2>
  friend class Array;

 /**
   * clone self
   * 
   * @param perm         optional permutation of dimensions
   * @param alloc        the allocator of the new created HTA
   * @param distc        the distribution of the new HTA
   * @param copy_values  flag that specifies if values should be 
   *                     copied from this HTA to new HTA.
   */

  typename TRAIT::TypeImpl* clone_leaf_ (const Tuple<DIM>* perm,
				    const typename TRAIT::TypeAllocator& alloc,
				    const MemMapping<DIM>& mmap, 
				    bool copy_values,
				    unsigned permute_mask = 0xFF) const
  {
    bool no_permute = (perm == NULL) || ((permute_mask & 1) == 0);
    typename TRAIT::TypeImpl* ret;
   
    const Shape<DIM> s = no_permute ? Shape<DIM> (this->shape().size()) : Shape<DIM>(this->shape_.size().permute(*perm));    
    ret = new typename TRAIT::TypeImpl(0, NULL, s, alloc, &mmap, this->home());
    
    if (this->Array<T, DIM, TRAIT>::isLocal() && copy_values) 
      this->copy_(perm, ret, permute_mask);
    
    return ret;
  }

  class HTAclone1Body {
    const HTAImpl<T,DIM,TRAIT>* const hta;
    const Tuple<DIM>* my_perm;
    const typename TRAIT::TypeAllocator& my_alloc;
    const MemMapping<DIM>& my_mmap;
    bool my_copy_values;
    unsigned my_permute_mask;
    typename TRAIT::TypeTileImpl** my_children;
  public:
    void operator() ( const tbb::blocked_range<int>& range ) const {
      int i_end = range.end();
      for (int i = range.begin(); i < i_end; i++) {
	my_children[i] = hta->data_[i]->clone_(my_perm, my_alloc, my_mmap[i], my_copy_values, my_permute_mask);
      }
    }
    HTAclone1Body(const HTAImpl<T,DIM,TRAIT>* const a, const Tuple<DIM>* perm,
		 const typename TRAIT::TypeAllocator& alloc,
		 const MemMapping<DIM>& mmap, 
		 bool copy_values,
		 unsigned permute_mask, typename TRAIT::TypeTileImpl** children) : hta(a), my_perm(perm), my_alloc(alloc), my_mmap(mmap),
											  my_copy_values(copy_values), my_permute_mask(permute_mask), 
											  my_children(children)
    {}
								 
  };

  class HTAclone2Body {
    const HTAImpl<T,DIM,TRAIT>* const hta;
    const Tuple<DIM>* my_perm;
    const typename TRAIT::TypeAllocator& my_alloc;
    const MemMapping<DIM>& my_mmap;
    bool my_copy_values;
    unsigned my_permute_mask;
    typename TRAIT::TypeTileImpl** my_children;
    std::vector< Tuple<DIM> >& vec;
  public:
    void operator() ( const tbb::blocked_range<int>& range ) const {
      int i_end = range.end();
      for (int i = range.begin(); i < i_end; i++) {
	Tuple<DIM> t = vec[i].permute(*my_perm);
	my_children[i] = hta->data_[i]->clone_(my_perm, my_alloc, my_mmap[i], my_copy_values, my_permute_mask);

      }
    }
    HTAclone2Body(const HTAImpl<T,DIM,TRAIT>* const a, const Tuple<DIM>* perm,
		  const typename TRAIT::TypeAllocator& alloc,
		  const MemMapping<DIM>& mmap, 
		  bool copy_values, unsigned permute_mask, 
		  typename TRAIT::TypeTileImpl** children, std::vector< Tuple<DIM> >& data) : 
      hta(a), my_perm(perm), my_alloc(alloc), my_mmap(mmap),
      my_copy_values(copy_values), my_permute_mask(permute_mask), 
      my_children(children), vec(data)
    {}
								 
  };


  typename TRAIT::TypeImpl* clone_ (const Tuple<DIM>* perm,
				    const typename TRAIT::TypeAllocator& alloc,
				    const MemMapping<DIM>& mmap, 
				    bool copy_values,
				    unsigned permute_mask = 0xFF) const
  {
    if(this->isLeaf()) {
      return clone_leaf_(perm,alloc,mmap,copy_values, permute_mask);
    }
    else {
      bool no_permute = (perm == NULL) || ((permute_mask & (1 << this->level())) == 0);
      typename TRAIT::TypeTileImpl** children;
      typename TRAIT::TypeImpl* ret;
      
      if (no_permute) {
	const Shape<DIM>& s = this->shape();
	int len = s.card();
	children = new typename TRAIT::TypeTileImpl*[len];

	tbb::parallel_for(tbb::blocked_range<int>(0, len), HTAclone1Body(this, perm, alloc, mmap, copy_values, permute_mask, children));
	
	// for (int i = 0; i < len; i++) {
// 	  children[i] = this->data_[i]->clone_(perm, alloc, mmap[i], copy_values, permute_mask);
// 	}
	ret = new typename TRAIT::TypeImpl(this->level(), children, s, alloc, &mmap, this->home()); 
      } else {
	const Tuple<DIM> t = this->shape().size().permute(*perm);
	const Shape<DIM> s = Shape<DIM>(t); 
	int len = s.card();
	children = new typename TRAIT::TypeTileImpl*[len];
	
	typename Shape<DIM>::iterator it = s.begin();
	typename Shape<DIM>::iterator end = s.end();
	std::vector< Tuple<DIM> > tuples;
	for (int i = 0; i < len; it++, i++) {
	  tuples.push_back(*it);
	}
	tbb::parallel_for(tbb::blocked_range<int>(0, len), HTAclone2Body(this, perm, alloc, mmap, copy_values, permute_mask, children, tuples));
	
	// for (int i = 0; i < len; it++, i++) {
// 	  Tuple<DIM> t = (*it).permute(*perm);
// 	  children[i] = this->tileAt(t)->clone_(perm, alloc, mmap[i], copy_values, permute_mask);
// 	}
	
	ret = new typename TRAIT::TypeImpl(this->level(), children, s, alloc, &mmap, this->home()); 
      }
     
      return ret;
    }
  }
  
  /**
   * COPY this to dst
   * re-factored from clone_ -- useful when copying without cloning.
   * works only for tile major storage layout
   */

  void copy3D_ (typename TRAIT::TypeImpl* dst, 
		const Tuple<DIM>*perm) const
  {
    const Tuple<DIM>& size = dst->shape().size();
    const Tuple<DIM> step_dst = (perm != NULL) ? 
      dst->memMapping__->linearStep().permute(*perm) * dst->shape().step().permute(*perm) :
      dst->memMapping__->linearStep() * dst->shape().step();

    const Tuple<DIM> step_larg = this->memMapping__->linearStep() * this->shape().step();
    
    int size_0 = size[0];
    int step_larg_0 = step_larg[0];
    int step_dst_0 = step_dst[0];
    
    int size_1 = size[1];
    int step_larg_1 = step_larg[1];
    int step_dst_1 = step_dst[1];
    
    int size_2 = size[2];
    int step_larg_2 = step_larg[2];
    int step_dst_2 = step_dst[2];
    
    T* dst_raw = dst->raw();

    // cout << "Here " << dst_raw << " " << dst->memMapping__->offset() << endl;

    const T* larg_raw = this->raw();
    
 //    if (false && this->memMapping__->isContiguous() && this->shape().isContiguous())
//       {
// 	int kBound = min (4, size_0);
// 	int jBound = min (4, size_1);
// 	int iBound = min (4, size_2);

// 	for (int kk = 0; kk < size_0; kk = kk + kBound)
// 	  for (int jj = 0; jj < size_1; jj = jj + jBound)
// 	    for (int ii = 0; ii < size_2; ii = ii + iBound)
	      
// 	      for (int k = kk; k < kk + kBound; ++ k) 
// 		for (int j = jj; j < jj + jBound; ++ j)  
// 		  for (int i = ii; i < ii + iBound; ++ i) 	    	     
// 		    //    dst_raw[i * step_dst_2 + j * step_dst_1 + k * step_dst_0] = *larg_raw++;
// 		    dst_raw[i * step_dst_2 + j * step_dst_1 + k * step_dst_0] = 
// 		      larg_raw [i * step_larg_2 + j * step_larg_1 + k * step_larg_0]; 		
	
//       } 
    
    for (int k = 0; k < size_0; ++ k) 
      for (int j = 0; j < size_1; ++ j)  
	for (int i = 0; i < size_2; ++ i) {
	  T tmp = larg_raw [i * step_larg_2 + j * step_larg_1 + k * step_larg_0]; 		
	  dst_raw[i * step_dst_2 + j * step_dst_1 + k * step_dst_0] = tmp;
	}	
 
  }

  void copy2D_ (typename TRAIT::TypeImpl* dst, 
		const Tuple<DIM>*perm) const
  {
    const Tuple<DIM>& size = dst->shape().size().permute(*perm);
    const Tuple<DIM> step_dst = perm != NULL ? dst->memMapping__->linearStep().permute(*perm) * dst->shape().step().permute(*perm):
      dst->memMapping__->linearStep() * dst->shape().step();
    const Tuple<DIM> step_larg = this->memMapping__->linearStep() * this->shape().step();
    
    int size_0 = size[0];
    int step_larg_0 = step_larg[0];
    int step_dst_0 = step_dst[0];
    
    int size_1 = size[1];
    int step_larg_1 = step_larg[1];
    int step_dst_1 = step_dst[1];
       
    T* dst_raw = dst->raw();
    const T* larg_raw = this->raw();

    
    if (this->memMapping__->isContiguous() && this->shape().isContiguous() &&
	dst->memMapping__->isContiguous() && dst->shape().isContiguous())
      {
	int kBound = min (4, size_0);
	int jBound = min (4, size_1);	
	
	for (int k = 0; k < size_0; ++ k) 
	  for (int j = 0; j < size_1; ++ j)  
	    //*dst_raw++ = *larg_raw++;
	    dst_raw[j * step_dst_1 + k * step_dst_0] = 
	      larg_raw [j * step_larg_1 + k * step_larg_0]; 		
	
      } else {
	for (int k = 0; k < size_0; ++ k) 
	  for (int j = 0; j < size_1; ++ j)  
	    dst_raw[j * step_dst_1 + k * step_dst_0] = 
	      larg_raw [j * step_larg_1 + k * step_larg_0]; 		
      }	
  }


  virtual void copy_leaf_ (const Tuple<DIM>* perm,
		   typename TRAIT::TypeImpl* dst,
		   unsigned permute_mask = 0xFF) const
  {
    bool no_permute = (perm == NULL) || ((permute_mask & 1) == 0);
    
    assert (dst->isLocal());
    assert (this->isAvailable_());
      
    switch (DIM) {
    case 2:
      this->copy2D_ (dst, perm);
      break;
    case 3:     
      this->copy3D_ (dst, perm);	
      break;
    default:      
      typename Shape<DIM>::iterator it = this->shape_.begin();
      typename Shape<DIM>::iterator end = this->shape_.end();
      for (; it != end; it++) {
	const Tuple<DIM> t = *it;
	const Tuple<DIM> tp = t.permute(*perm);
	dst->scalarAt(tp) = scalarAt(t);
      }   
    } 
  }
  
  class HTAcopy1Body {
    const HTAImpl<T,DIM,TRAIT>* const hta;
    const Tuple<DIM>* my_perm;
    typename TRAIT::TypeImpl* my_dst;
    unsigned my_permute_mask;
    public:
    void operator() ( const tbb::blocked_range<int>& range ) const {
      int i_end = range.end();
      for (int i = range.begin(); i < i_end; i++) {
	  if (hta->data_[i]->Array<T, DIM, TRAIT>::isLocal())
	    hta->data_[i]->copy_(my_perm, my_dst->data_[i], my_permute_mask);
      }
    }
    HTAcopy1Body(const HTAImpl<T,DIM,TRAIT>* const a, const Tuple<DIM>* perm, typename TRAIT::TypeImpl* dst,
		  unsigned permute_mask) : hta(a), my_perm(perm), my_dst(dst),
					   my_permute_mask(permute_mask)
											  
    {}
  };

  class HTAcopy2Body {
    const HTAImpl<T,DIM,TRAIT>* const hta;
    const Tuple<DIM>* my_perm;
    typename TRAIT::TypeImpl* my_dst;
    unsigned my_permute_mask;
    std::vector< Tuple<DIM> >& vec;
  public:
    void operator() ( const tbb::blocked_range<int>& range ) const {
      int i_end = range.end();
      for (int i = range.begin(); i < i_end; i++) {

      }
    }
    HTAcopy2Body(const HTAImpl<T,DIM,TRAIT>* const a, const Tuple<DIM>* perm, typename TRAIT::TypeImpl* dst,
		 unsigned permute_mask, std::vector< Tuple<DIM> >& data) : hta(a), my_perm(perm), my_dst(dst),
									   my_permute_mask(permute_mask), vec(data)
    {}
								 
  };

  void copy_ (const Tuple<DIM>* perm,
	      typename TRAIT::TypeImpl* dst,
	      unsigned permute_mask = 0xFF) const
  {   
    if(this->isLeaf()) {
      copy_leaf_(perm, dst, permute_mask);
    }
    else {
      bool no_permute = (perm == NULL) || ((permute_mask & (1 << this->level())) == 0);
     
      if (no_permute) {
	const Shape<DIM>& s = this->shape();
	int len = s.card();
	tbb::parallel_for(tbb::blocked_range<int>(0,len), HTAcopy1Body(this, perm, dst, permute_mask));
	
	// for (int i = 0; i < len; i++) {
// 	  if (this->data_[i]->Array<T, DIM, TRAIT>::isLocal())
// 	    this->data_[i]->copy_(perm, dst->data_[i], permute_mask);
// 	}
	
      } else {
	const Tuple<DIM> t = this->shape().size().permute(*perm);
	const Shape<DIM> s = Shape<DIM>(t); 
	int len = s.card();
	typename Shape<DIM>::iterator it = s.begin();
	std::vector< Tuple<DIM> > tuples;

	for (int i = 0; i < len; it++, i++) {
	  tuples.push_back(*it);
	}
	tbb::parallel_for(tbb::blocked_range<int>(0,len), HTAcopy2Body(this, perm, dst, permute_mask, tuples));

	// for (int i = 0; i < len; it++, i++) {
// 	  Tuple<DIM> t = (*it).permute(*perm);
// 	  //if (this->tileAt(t)->Array<T, DIM, TRAIT>::isLocal())
// 	  this->tileAt(t)->copy_(perm, dst->tileAt(*it), permute_mask);
// 	}
      }
    }
  }
  
  void inPlaceTrans (const Tuple<2>& perm) const
  {   
    if(this->isLeaf()) {

      Tuple<DIM> size = this->shape().size();
      switch (DIM) {
      case 3: 
	//one of the 3  possibilites
	if (perm[0] == 0 & perm[1] == 2){
	    for (int j = 0; j < size[1]; j++)
	      for (int i = 0; i < size[0]; i++)
		for (int k = i+1; k < size[2]; k++)
		  {
		  assert (this->home()!=-1);
		  T tmp = this->scalarAt (Tuple<3>(i,j,k));
		  this->scalarAt (Tuple<3>(i,j,k)) = this->scalarAt (Tuple<3>(k, j, i));
		  this->scalarAt (Tuple<3>(k, j, i)) = tmp;
		}
	} else {
	  assert (false);
	}
	break;
      case 2:
	for (int i = 0; i < size[0]; i++)
	  for (int j = i+1; j < size[1] - i; j++)
	    {
	      assert (this->home()!=-1);
	      T tmp = this->scalarAt (Tuple<2>(i,j));
	      this->scalarAt (Tuple<2>(i,j)) = this->scalarAt (Tuple<2>(j, i));
	      this->scalarAt (Tuple<2>(j, i)) = tmp;
	    }
	break;
	default:
	assert (false);
      }
    }
    else {            
      typename Shape<DIM>::iterator it = this->shape().begin();
      typename Shape<DIM>::iterator end = this->shape().end();
      for (; it != end; it++) {
	if (this->tileAt(*it)->Array<T, DIM, TRAIT>::isLocal())
	  this->tileAt(*it)->inPlaceTrans(perm);
      }
    }
  }
 
  //----------------------------------------
  /** reduction over all dimensions */
  template <typename OP> 
  T reduce_leaf_ (const OP& binary_op) const
  {
    assert (this->isAvailable_());
    T acc = TypeHelper<T>::identityVal(binary_op);

    // TODO: this should be done with a math kernel
    if (this->shape().isContiguous() && this->memMap().isContiguous() &&
	this->shape().card()==this->memMap().linearCard()) { // add this condition to exclude overlapped case
      
      int card = this->shape().card();
      for (int i = 0; i < card; i++)
	acc = binary_op(acc, this->data_leaf_[i]); 
    } else {
      int n = this->shape_.card();
      typename Shape<DIM>::iterator i = this->shape_.begin();
      typename Shape<DIM>::iterator end = this->shape_.end();
      for (; i != end; ++i) {
	acc = binary_op(acc, scalarAt (*i));
      }
    }
    HTA_DEBUG(5, "LeafHTAImpl::reduce res=" << acc);
    return acc;
  }
  
  template <typename OP> 
  void reduce_leaf_2D_ (const OP& binary_op, 
		     int dim, 
		     typename TRAIT::TypeImpl& dst) const
  {    
    // then do a the reduction
    Tuple<DIM> size = this->shape().size();
    int size_0 = size[0];
    int size_1 = size[1];

    const Tuple<DIM> step_dst = dst.memMapping__->linearStep() * dst.shape().step();
    int step_dst_0 = dim == 0 ? 0 : step_dst[0];
    int step_dst_1 = dim == 1 ? 0 : step_dst[1];

    const T* this_raw = this->raw();
    T* dst_raw = dst.raw();

    assert (this_raw);
    assert (dst_raw);

    for (int i = 0; i < size_0; i++)
      for (int j = 0; j < size_1; j++) {
	T& dst_var = dst_raw[i* step_dst_0 + j * step_dst_1];
	/* BBF: t undefined would not let compile
	 * HTA_DEBUG(5, "LeafHTAImpl::reduce_ dst=" << dst_var << " arg=" << scalarAt(t));
	 */
	dst_var = binary_op(dst_var, *this_raw++);
	HTA_DEBUG(5, "LeafHTAImpl::reduce_ res=" << dst_var);
      } 
  }
  
  /** reduction over one dimension */
  template <typename OP> 
  void reduce_leaf_ (const OP& binary_op, 
		     int dim, 
		     typename TRAIT::TypeImpl& dst) const
  {
    assert (this->isAvailable_());
    assert (dst.isLocal());

    switch (DIM) {

    case 2:
      reduce_leaf_2D_ (binary_op, dim, dst);
      break;
    default:
      Shape<DIM> masked = this->shape().mask(dim);
      
      /* GBDEBUG: I am intializing it in HTA.h, instead       
      // initialize the destination
      typename Shape<DIM>::iterator i = masked.begin();
      typename Shape<DIM>::iterator endi = masked.end();   
      T bottom = TypeHelper<T>::identityVal(binary_op);
      for ( ; i != endi; i++) 
      dst.scalarAt(*i) = bottom;
      */  
      
      // then do a the reduction
      typename Shape<DIM>::iterator j = this->shape().begin();
      typename Shape<DIM>::iterator endj = this->shape().end();
      for ( ; j != endj; j++) {
	Tuple<DIM> t = *j;
	T& dst_var = dst.scalarAt(t.mask(dim));
	HTA_DEBUG(5, "LeafHTAImpl::reduce_ dst=" << dst_var << " arg=" << scalarAt(t));
	dst_var = binary_op(dst_var, scalarAt(t));
	HTA_DEBUG(5, "LeafHTAImpl::reduce_ res=" << dst_var);
      }
      
      bool all_to_all = dst.shape().size()[dim] > 1;
      if (all_to_all) {
	//assert (dst.shape_.step() == Tuple<DIM>::one);
	//assert (dst.shape_.low() == Tuple<DIM>::zero);
	
	// propagate results from tile with dim=0 to other tiles ...
	Tuple<DIM> size_tuple = dst.shape().size();
	int dim_size = size_tuple[dim];
	typename Shape<DIM>::iterator k = masked.begin();
	typename Shape<DIM>::iterator endk = masked.end();
	for ( ; k != endk; k++) {
	  Tuple<DIM> t = *k;
	  T src = dst.scalarAt(t);
	  for (int j = 1; j < dim_size; ++j) {
	    t[dim] = j;
	    dst.scalarAt(t) = src;
	  }
	}
      }
    }
  }
  
  /** reduction over one dimension */
  template <typename OP> 
  void reduce_non_leaf_ (const OP& binary_op, 
		int dim, 
		typename TRAIT::TypeImpl& dst) const
  {
    assert (dst.isLocal());
    
    bool all_to_all = dst.shape().size()[dim] > 1;
    typename Shape<DIM>::iterator i = this->shape_.begin();
    typename Shape<DIM>::iterator end = this->shape_.end();

    //(1) all-to-one reduction
    for ( ; i != end; i++) {
      
      Tuple<DIM> t = *i;
      //src tile is the ith tile
      typename TRAIT::TypeTileImpl* tile = tileAt(t);
       
      //acc_tile is always the 0th tile
      typename TRAIT::TypeTileImpl* acc_tile = dst.tileAt(t.mask(dim,0));

      tile->reduce_(binary_op, dim, *acc_tile);               

    }
  
    //(2) replication, if all-to-all
    if (all_to_all) {
      assert (dst.shape_.step() == Tuple<DIM>::one);
      assert (dst.shape_.low() == Tuple<DIM>::zero);

      // propagate results from tile with dim=0 to other tiles ...
      Tuple<DIM> size_tuple = this->shape_.size();
      int dim_size = size_tuple[dim];
      Shape<DIM> red_shape (size_tuple.mask(dim, 1));
      typename Shape<DIM>::iterator k = red_shape.begin();
      typename Shape<DIM>::iterator end = red_shape.end();
      for ( ; k != end; k++) {
	Tuple <DIM> t = *k;
	typename TRAIT::TypeTileImpl* src_tile = dst.tileAt(t);
	for (int j = 1; j < dim_size; ++j) {
	  t[dim] = j;
	  typename TRAIT::TypeTileImpl* dst_tile = dst.tileAt(t);
	  *dst_tile = *src_tile;
	}
      }
    }
  }

 public:


  template <typename OP> 
  void reduce_ (const OP& binary_op, 
		int dim, 
		typename TRAIT::TypeImpl& dst) const
  {
    if(this->isLeaf()) {
      reduce_leaf_(binary_op, dim, dst);
    }
    else {
      reduce_non_leaf_(binary_op, dim, dst);
    }
  }


  /** reduction over one dimension */
  template <typename OP> 
  void reduce_leaf_ (int oplevel,
		const OP& binary_op, 
		int dim, 
		typename TRAIT::TypeImpl& dst) const
  {
    if (oplevel > 0)
      
      dst = (typename TRAIT::TypeImpl&)*this;

    else
      
      reduce_ (binary_op, dim, dst);
  }
  
  /** reduction over one dimension */
  template <typename OP> 
  void reduce_ (int oplevel,
		const OP& binary_op, 
		int dim, 
		typename TRAIT::TypeImpl& dst
		) const
  {
    if(this->isLeaf()) {
      reduce_leaf_(oplevel,binary_op,dim,dst);
    }
    else{    
      assert (dst.isLocal());
      bool all_to_all = dst.shape().size()[dim] > 1;
      typename Shape<DIM>::iterator i = this->shape_.begin();
      typename Shape<DIM>::iterator end = this->shape_.end();
      for ( ; i != end; i++) {
	Tuple<DIM> t = *i;
	typename TRAIT::TypeTileImpl* tile = tileAt(t);
	typename TRAIT::TypeTileImpl* dst_tile = dst.tileAt(t);
	tile->reduce_ (oplevel, binary_op, dim, *dst_tile);
	
	if (oplevel <= this->level()) 
	  {	  
	    typename TRAIT::TypeTileImpl* acc_tile = dst.tileAt(t.mask(dim));
	    
	    if (acc_tile != dst_tile)
	      TRAIT::TypeTileImpl::map (binary_op, *acc_tile, *dst_tile, *acc_tile);      
	  }
      }
      
      if (all_to_all && oplevel <= this->level()) {
	assert (dst.shape_.step() == Tuple<DIM>::one);
	assert (dst.shape_.low() == Tuple<DIM>::zero);
	
	// propagate results from tile with dim=0 to other tiles ...
	Tuple<DIM> size_tuple = this->shape_.size();
	int dim_size = size_tuple[dim];
	Shape<DIM> red_shape (size_tuple.mask(dim, 1));
	typename Shape<DIM>::iterator k = red_shape.begin();
	typename Shape<DIM>::iterator end = red_shape.end();
	for ( ; k != end; k++) {
	  Tuple <DIM> t = *k;
	  typename TRAIT::TypeTileImpl* src_tile = dst.tileAt(t);
	  for (int j = 1; j < dim_size; ++j) {
	    t[dim] = j;
	    typename TRAIT::TypeTileImpl* dst_tile = dst.tileAt(t);
	    *dst_tile = *src_tile;
	  }
	}
      }
    }
    
  }

/** print leaves*/   
    virtual void print_leaf_ (ostream& ost, int indent = 0) const 
  {
    assert (isAvailable_());
    
    HTA_DEBUG (2, "HTAImpl<LEVEL=0>::print_ this=" << this << " data_=" << data_);

    for (int j = 0; j < indent; ++j) 
      ost << " ";

    if (DIM == 2) {
      int printed = 0;
      int mod = this->shape_[1].size(); // column dimension has index 1
      for (typename Shape<DIM>::iterator i = this->shape_.begin(); i != this->shape_.end(); i++) {
	ost << scalarAt_(i) << "\t";
	if (++printed % mod == 0) {
	  ost << endl;
	  for (int j = 0; j < indent; ++j) ost << " ";
	}
      }
    } else {
      for (typename Shape<DIM>::iterator i = this->shape_.begin(); i != this->shape_.end(); i++)    
	ost << scalarAt(*i) << " "; 
      ost << endl;
    }
  }    


  /** print */
  virtual void print_non_leaf_(ostream& ost, int indent = 0) const 
  {
    assert (indent >= 0);

    ost << endl;
    // indent
    for (int j = 0; j < indent; ++j) ost << " ";
    
    if (this->isHomogeneous_() && DIM == 2 && this->level() == 1 && !this->isDistributed()) {
      Tuple<DIM> leaf_size = this->memMap().leafSize();
      int row_max = leaf_size[0]; 
      int col_max = leaf_size[1]; 
      int row_tile_sz = row_max / this->shape_.size()[0];
      int col_tile_sz = col_max / this->shape_.size()[1];
      
      // check if there is some step in the leaves ...
      int row_tile_step = this->tileAt(0)->shape().step()[0];
      int col_tile_step = this->tileAt(0)->shape().step()[1];
      
      assert (this->shape().step() == Tuple<DIM>::one);
      
      for (int i = 0; i < row_max; i += row_tile_step ) {
	// if necessary, print horizontal tile separator
	if (i > 0 && i % row_tile_sz == 0) { 
	  cout << "----------------------------------------------------------------" << endl;
	  for (int k = 0; k < indent; ++k) ost << " "; // indent
	}
	// print the row
	for (int j = 0; j < col_max; j += col_tile_step) {
	  int tmp[] = {i,j};
	  const Tuple<DIM> t = Tuple<DIM>(tmp);
	  // if necessary, print vertical tile separator
	  if (j > 0 && j % col_tile_sz == 0)
	    ost << "|\t" ;
	  ost << this->actual_().scalarAt(t) << "\t";
	}
	ost << endl;
	for (int k = 0; k < indent; ++k) ost << " "; // indent
      }
    } else { 
      for (typename Shape<DIM>::iterator i = this->shape_.begin(); i != this->shape_.end(); i++) {
	Tuple<DIM> t = *i;
	  this->actual_().tileAt(t)->print_(ost, 8);
	ost << endl;
      }
    }
  }
  
  void print_(ostream& ost, int indent = 0) const 
  {
    if(this->isLeaf())
      {
	print_leaf_(ost, indent);
      }
    else
      {
	print_non_leaf_(ost, indent);
      }
  }


  /** 
   * An array of length shape_->card() that holds
   * pointers to the tile HTAs. this->shape_ is the linearization 
   * function that maps tuples to the position in this array.
   * 
   * shape.low() is mapped to the index 0, shape.high()
   * is mapped to index shape.card() - 1. Any other point p
   * is found at index shape.ord(p).
   */
  typename TRAIT::TypeTileImpl* const * const data_;
  //if leaf store the real data
  typename TRAIT::TypeInternal data_leaf_;

public:

  /*mapReduceHTA -- should be reconciled with mapReduce */

  template <typename OP>
  void mapReduceHTA (OP& mr_op)
  { 
    if(this->isLeaf()) {
      mapReduce_leaf_(mr_op);
    }
    else {
      assert (this->level() > 0);
      //assert (this->Array<T, DIM, TRAIT>::isLocal());
      
      typename Shape<DIM>::iterator i = this->shape().begin();
      typename Shape<DIM>::iterator end = this->shape().end();
      for ( ; i != end; i++) {
	Tuple<DIM> t = *i;
	typename TRAIT::TypeTileImpl* tile = this->tileAt(t);
	//I hope the compiler will peform dead code elimination 
	//if mr_op.init() is empty. 
	
	if (!tile->Array<T, DIM, TRAIT>::isLocal()) continue;
	
	if (mr_op.level() == this->level())
	  { 
	    //cout << this->memMap().leafPos() << " " << this->shape().size() << " " << pos << endl;
	    Tuple<DIM> pos;
	    switch (mr_op.keyType()) {	      
	    case Operator::MR_TUPLE:
	      pos = this->shape().size() * mr_op.context_.position_;
	      //JG: modified to avoid using map_hta_tuple_intl()
	      mr_op.map_hta_tuple(t+pos, HTA<T, DIM,TRAIT>(tile->actual_()));	
	      break;
	    case Operator::MR_NONE:
	      //JG: modified to avoid using map_hta_tuple_intl()
	      mr_op.map_hta_none(HTA<T,DIM,TRAIT>(tile->actual_()));	
	      break;      
	    default:
	      assert (false);
	    }
	  } else {
	    MapContext<T, DIM> context (t, this, tile);
	    mr_op.init(context); 
	    tile->mapReduceHTA(mr_op); // recurse
	  }
      }
    }
        
  }

public:
  
  /**
   * Non-Homegenous tile At (used in LU)
   */
  typename TRAIT::TypeImpl* tileAt (int n, const Shape<DIM>* const* sv, const Tuple<DIM>& t) const 
  {
    assert (this->level() > 0);
    typename TRAIT::TypeImpl* ret;
    Shape<DIM> shape((*sv[0])[t].size());

    int len = shape.card();

    // (1) allocate / compute the children
    // responsibility to de-allocate is taken by the newly created 
    // MemMapping and HTA respectively
    typename TRAIT::TypeTileImpl** tiles = new typename TRAIT::TypeTileImpl* [len];
    typename Shape<DIM>::iterator i = (*sv[0])[t].begin();
    typename Shape<DIM>::iterator end = (*sv[0])[t].end();
    const MemMapping<DIM>** child_dist = new const MemMapping<DIM>* [len];
    for (int j = 0; i != end; i++, j++) {
      typename TRAIT::TypeTileImpl* tmp = tileAt(*i);
      if (n > 2) 
        // if the child is not a leaf hta, "forward" the next tileAt request to it 
        tiles[j] = tmp->tileAt(n-1, sv+1, *i);   
      else 
	tiles[j] = tmp->tileAt((*sv[1])[*i]);
      //      child_dist[j] = tiles[j]->memMapping__; 
      child_dist[j]= tmp->memMapping__;//JG: a temporary fix
    }
    // (2) allocate and initialize the result
    // compute mem-mapping
    Tuple<DIM> linear_size = shape.high() - shape.low() + Tuple<DIM>::one;

    const MemMapping<DIM>* mmap = this->memMapping__->select(child_dist, linear_size);
    
    // same allocator, i.e., the return HTA is a view on the current HTA
    ret = new typename TRAIT::TypeImpl (this->level(), tiles, shape, this->alloc_, mmap, this->home());

    //(3)
    //initialize the memMap in the router

    //(4) initialize the distribution
    if (ret->isDistributed())
      {
	// current limitation - we do not have a distribution in closed form for a 
	// selection of tiles on a distributed array
	const GenericDistribution<DIM>* old_dist = dynamic_cast<const GenericDistribution<DIM>*>(this->dist__);    
	
	//if old dist is a Cyclic distribution.
	if (!old_dist)
	  old_dist = new GenericDistribution<DIM> (this->shape_);
	
	//the new distribution is a Sub-distribution of old distribution
	ret->initDistribution (old_dist->subDist ((*sv[0])[t]));
	ret->initLocalTiles();
      } 

    //if (ret->isDistributed())
    //ret->initDistribution (NULL);
    
    return ret;
  }


};  // class HTAImpl






#endif /* __HTA_IMPL_H__ */
