#ifndef __HTA_O_IMPL_H__
#define __HTA_O_IMPL_H__

#include <cassert>
#include <iostream>
#include <functional>
#include "serial/HTAImpl.h"
#include "Overlap.h"
using namespace std;



/**
 * Overlapped sequential HTA implemenation.
 */

template<class T, int DIM, typename TRAIT>
class HTAOImpl : public HTAImpl<T, DIM, TRAIT>
{
private: 
  typedef HTAImpl<T,DIM,TRAIT> base;
  Overlap<DIM> * ol_;  
public:
  /**
   * Constructor.
   */
 HTAOImpl (size_t levels, 
	  typename TRAIT::TypeTileImpl** data, 
	  const Shape<DIM>& shape, 
	  const typename TRAIT::TypeAllocator& alloc,
	  const MemMapping<DIM>* mmap, 
	  int home,
	  Overlap<DIM>* ol=NULL) :
   HTAImpl<T,DIM,TRAIT>(levels, data, shape, alloc, mmap, home),
   ol_(ol){
   if(ol_){
     ol_->refCount_(1);
   }
 }

  ~HTAOImpl(){
    if(ol_ && ol_->refCount_(-1)==0){
      delete ol_;
    }
  }
	
  Overlap<DIM> * ol() const{
    return ol_;
  }

  
  typename TRAIT::TypeImpl& operator= (const typename TRAIT::TypeImpl & rhs) 
  {
    if(this->isLeaf()){
      if (this->isLocal()) { 
	//assert(rhs.isAvailable_());
	HTAImpl<T, DIM, TRAIT>::operator=(rhs);
      } 
      return *this;   
    }
    else{      
      // resort to implementation in the superclass
      Array<T, DIM, TRAIT>::operator=(rhs);
      TRAIT::barrier();
      return this->actual_();
    }
  }

  
  typename TRAIT::TypeImpl& operator= (const T& rhs) 
  {
    if (this->isLocal()) {
      // owner computes
      HTAImpl<T, DIM, TRAIT>::operator=(rhs);
      return *this;
    }
    else {
      // resort to implementation in the superclass
      Array<T, DIM, TRAIT>::map (Operator::cnst<T>(rhs), *this);
      TRAIT::barrier();
      return this->actual_();
    }
  }

  inline typename TRAIT::TypeImpl* scalarAt (const Shape<DIM>& s) const 
  {
    return base::scalarAt(s);
  }


  inline T& scalarAt (const Tuple<DIM>& t) const
  {
    if(this->isLeaf()){
      return base::scalarAt_leaf_(t);
    }
    else {
      assert (this->level() > 0);
      assert (this->shape().low() == Tuple<DIM>::zero);
      assert (this->shape().step() == Tuple<DIM>::one);
      Order ord = this->memMap().order();
      
      switch (ord) {
      case DIST:
      case TILE: {
	Tuple<DIM> size;
	if(ol_!=NULL){
	  size = this->memMap()[0].leafSize()-ol_->neg()-ol_->pos();
	}else{
	  size = this->memMap()[0].leafSize();
	}
	Tuple<DIM> tile_idx, rest_idx;
	for (int i = 0; i < DIM; ++i) {
	  
	  tile_idx[i]= t[i] / size[i];
	  rest_idx[i]= t[i] % size[i];
	} 
	typename TRAIT::TypeTileImpl* tile = tileAt(tile_idx);
	
	T& ret = tile->scalarAt(rest_idx);
	return ret;
      }
      case ROW: 
      case COL: 
      default: {
	assert (!this->isDistributed());
	assert (ord == ROW); 
	
	Tuple<DIM> linear_size_scalarAt = this->memMap().leafSize();
	//	assert (t >= Tuple<DIM>::zero&& t < linear_size_scalarAt); // bounds check
	T* data = this->raw();
	Tuple<DIM> linear_step_scalarAt = this->memMap().linearStep();
	int off =	VectorOps<Tuple<DIM>, DIM>::dotProduct (t, linear_step_scalarAt);
	return data[off];
      }
      }
    }
  }
  
  inline T& scalarAt (const int& i) const
  {
    return HTAImpl<T, DIM, TRAIT>::scalarAt_ (i);
  }

  void print_ (ostream& ost, int indent = 0)const
  {
    if(ol_!=NULL){
      if(this->isLeaf()){
	printOverlap_(ost,indent);
      }
      else { 
	int offset = VectorOps<Tuple<DIM>,DIM>::dotProduct(ol_->neg(),this->memMap().linearStep());
	printFlat(offset);
	// let all threads advance to this point, otherwise we might print data 
	// that is not yet initialized  (by other threads)
	/*for (typename Shape<DIM>::iterator i = this->shape_.begin(); i != this->shape_.end(); ++i) {
	  Tuple<DIM> t = *i;
	  typename TRAIT::TypeTileImpl* tile = tileAt(t);
	  tile->print_(ost, 8);
	  ost << endl;
	  }*/
	
      }
    }else{
      /* BBF: Jia had commented out base::print_ and was using the printFlat below,
          but for example printFlat was breaking for HTAs with level !=1, and I want
          to print level 0 HTAs. Thus I go back to base::print_, at least while it
          does not generate any problem!
       */
      base::print_(ost, indent);
      //printFlat(0);
    }
  }
      
  void printFlat(int offset) const{
    assert(this->level()==1);
    switch(DIM){
    case 1:{
      T* ptr = this->raw();
      ptr = ptr -offset;
      for(int i=0; i<this->memMap().leafSize()[0]; i++){
	cout<< ptr[i]<<"\t";
      }
      cout<<endl;
      break;
    }
    case 2: {
      int printed = 0;
      T * ptr = this->raw();
      ptr = ptr-offset;
      int mod = this->memMap().linearStep()[0];
      for(int i=0; i<this->memMap().leafSize()[0]; i++){
	for(int j=0; j<this->memMap().leafSize()[1]; j++){
	  int ord = i*mod+j;
	  cout << ptr[ord] << "\t";
	}
	cout<<endl;
      }
      cout<<endl;
      break;
    }
      
    case 3:{
      int printed = 0;
      T * ptr = this->raw();
      ptr = ptr-offset;
      int lda = this->memMap().linearStep()[0];
      int ldb = this->memMap().linearStep()[1];
      
      for(int i=0; i<this->memMap().leafSize()[0]; i++){
	for(int j=0; j<this->memMap().leafSize()[1]; j++){
	  for(int z=0; z<this->memMap().leafSize()[2]; z++){
	    
	    int ord = i*lda+j*ldb+z;
	    cout << ptr[ord] << "\t";
	  }
	  cout << endl;
	}
	cout << endl;
      }
      cout<<endl;
      break;
    }
    }
  }

  void printOverlap_ (ostream& ost, int indent=0) const{
    assert (this->isAvailable_());
    
    HTA_DEBUG (2, "HTAImpl<LEVEL=0>::print_ this=" << this << " data_=" << data_);

    for (int j = 0; j < indent; ++j) 
      ost << " ";
    Tuple<DIM> neg = ol_->neg();
    Tuple<DIM> pos = ol_->pos();
    switch(DIM){
    case 1:{
      T* ptr = this->raw();
                
      int offset = VectorOps<Tuple<DIM>,DIM>::dotProduct(ol_->neg(),this->memMap().linearStep());
      ptr = ptr -offset;
      for(int i=0; i<this->memMap().leafSize()[0]+neg[0]+pos[0]; i++){
	ost<< ptr[i]<<"\t";
      }
      break;
    }
    case 2:{
      int printed = 0;
      int mod = this->memMap().linearStep()[0];
      int offset = VectorOps<Tuple<DIM>,DIM>::dotProduct(ol_->neg(),this->memMap().linearStep());
      T * ptr = this->raw();
      ptr = ptr-offset;

      for(int i=0; i<this->memMap().leafSize()[0]+neg[0]+pos[0]; i++){
	for(int j=0; j<this->memMap().leafSize()[1]+neg[1]+pos[1]; j++){
	  int ord = i*mod+j;
	  ost << ptr[ord] << "\t";
	}
	ost << endl;
	for (int k = 0; k < indent; ++k) ost << " ";
      }
      break;
    }
    case 3:{
      int printed = 0;
      int offset = VectorOps<Tuple<DIM>,DIM>::dotProduct(ol_->neg(),this->memMap().linearStep());
      T * ptr = this->raw();
      ptr = ptr-offset;
      int lda = this->memMap().linearStep()[0];
      int ldb = this->memMap().linearStep()[1];
      
      for(int i=0; i<this->memMap().leafSize()[0]+neg[0]+pos[0]; i++){
	for(int j=0; j<this->memMap().leafSize()[1]+neg[1]+pos[1]; j++){
	  for(int z=0; z<this->memMap().leafSize()[2]+neg[1]+pos[1]; z++){
	    
	    int ord = i*lda+j*ldb+z;
	    ost << ptr[ord] << "\t";
	  }
	  ost << endl;
	  for (int k = 0; k < indent; ++k) ost << " ";
	}
	ost << endl;
	for (int k = 0; k < indent; ++k) ost << " ";
      }
      break;
    }
    }
  }
  
  /**
   * Overlapped version of lconv2D_
   */
  static void lconv2D_ (const T w0, const T w1, const T w2,
			const typename TRAIT::TypeImpl& arg,
			typename TRAIT::TypeImpl& dst) 
  {
    assert (DIM == 2);
    assert (arg.shape().step() == Tuple<DIM>::one);
    assert (dst.shape().step() == Tuple<DIM>::one);

    int size_0 = arg.shape().size()[0]; // most significant
    int size_1 = arg.shape().size()[1];
    T* a = arg.raw();
    T* d = dst.raw();

    const Tuple<DIM> step_arg = arg.memMapping__->linearStep() * arg.shape().step();
    int step_arg_0 = step_arg[0];

    const Tuple<DIM> step_dst = dst.memMapping__->linearStep() * dst.shape().step();
    int step_dst_0 = step_dst[0];

    for (int i = 0; i < size_0; ++ i) {
      for (int j = 0; j < size_1; ++ j) {
	T u0 = 
	  a[i     * step_arg_0 + j];
	T u1 = 
	  a[(i-1) * step_arg_0 + j] +
	  a[i     * step_arg_0 + (j-1)] +
	  a[(i+1) * step_arg_0 + j] +
	  a[i     * step_arg_0 + (j+1)];
	T u2 = 
	  a[(i-1) * step_arg_0 + (j-1)] + 
	  a[(i-1) * step_arg_0 + (j+1)] +
	  a[(i+1) * step_arg_0 + (j-1)] + 
	  a[(i+1) * step_arg_0 + (j+1)];
	d[i * step_dst_0 + j] = w0 * u0 + w1 * u1 + w2 * u2;
      }
    }
  }
  
  
  /**
   *  Overlapped version of lconv3D_
   */
   static void lconv3D_ (const T w0, const T w1, const T w2, const T w3,
			 const typename TRAIT::TypeImpl& arg,
			 typename TRAIT::TypeImpl& dst)
  {
    assert (DIM == 3);
    assert (arg.shape().step() == Tuple<DIM>::one);
    assert (dst.shape().step() == Tuple<DIM>::one);
    const Tuple<DIM> step_arg = arg.memMapping__->linearStep() * arg.shape().step();
    int step_arg_0 = step_arg[0];
    int step_arg_1 = step_arg[1];
    int step_arg_2 = step_arg[2];

    const Tuple<DIM> step_dst = dst.memMapping__->linearStep() * dst.shape().step();
    int step_dst_0 = step_dst[0];
    int step_dst_1 = step_dst[1];

    int size_0 = arg.shape().size()[0]; // most significant
    int size_1 = arg.shape().size()[1];
    int size_2 = arg.shape().size()[2];

    const T* a = arg.raw();
    T* d = dst.raw();

    for (int i = 0; i < size_0; i++) {
      for (int j = 0; j < size_1 ; j++) {
        for (int k = 0; k < size_2 ; k++) {
	  int indx = i * step_dst_0  + j * step_dst_1 + k;
	  int indxa = i * step_arg_0 + j * step_arg_1 + k;
	  d[indx] = w0 *(a[indxa-step_arg_0] + a[indxa-step_arg_1] + a[indxa-1] +
			 a[indxa+step_arg_0] + a[indxa+step_arg_1] + a[indxa+1]);
	}
      }
    }
  }
 /*
   * to manually update the shadow regions so that it is consistent 
   * this is now blank but think about the periodic case. 
   */
  void update (int myDim = -1) const
  {
  }
  void setNL(int numTiles){
   
  }

};


#endif //__HTA_O_IMPL_H__
