/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: LeafHTAImpl.h,v 1.114 2006/07/11 04:24:06 bikshand Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef __LEAF_HTA_IMPL_H__
#define __LEAF_HTA_IMPL_H__

#include <cassert>
#include <functional>
#include <iostream>

#include "Array.h"
#include "MemMapping.h"
#include "LeafMemMapping.h"
#include "MapReduce.h"
#include "Operators.h"
#include "Shape.h"
#include "Traits.h"
#include "serial/HTAImpl.h"
#include "util/TypeHelper.h"

/**
 * Leaf HTA - this is a template specialization of class 
 * HTAImpl for LEVEL=0
 */

template<class T, int DIM, typename TRAIT>
class HTAImpl<T, DIM, 0, TRAIT> : public Array<T, DIM, 0, TRAIT>
{
  
public:
  
  /** 
   * Constructor - this constructor is only called from HTA::alloc
   * 
   * @param data  not used, just for compatibility with the constructor for 
   *              template instances with LEVEL > 0
   */
  HTAImpl (typename TRAIT::TypeTileImpl** data, 
	   const Shape<DIM>& shape, 
	   const typename TRAIT::TypeAllocator& alloc, 
	   const MemMapping<DIM>* mmap, 
	   int home) :
    Array<T, DIM, 0, TRAIT> (shape, alloc, mmap, home) 
  {
    // we have to know about the mapping to memory in a leaf-hta
    assert (mmap != NULL);
      
    HTA_DEBUG(3, "HTAImpl<LEVEL=0>::init mmap=" << mmap << " mmap.place=" << home << " mmap.offset=" << mmap->offset());
    // check if the memory provided by the allocator tolerates access 
    // to the whole range of indices of the shape

    if (home == Traits::Default::myPlace()) {
      int offset = mmap->offset();
      data_ = alloc.get(offset);
#ifndef NDEBUG
      int ord_low = mmap->ord(shape.low());
      int ord_high = mmap->ord(shape.high());
      assert (this->alloc_.checkAddress_(offset + ord_low));
      assert (this->alloc_.checkAddress_(offset + ord_high));
#endif
    } else 
      data_ = NULL;
  }

  /**
   * The destructor - uses reference counting
   * as the data_ could be shared among PrivateLeafHTAs
   */  
  virtual ~HTAImpl () 
  {
    this->alloc_.recycle();
  }
  
  /**
   * return a clone of myself
   */
  typename TRAIT::TypeImpl* clone_ (const Tuple<DIM>* perm,
				    const typename TRAIT::TypeAllocator& alloc,
				    const MemMapping<DIM>& mmap, 
				    bool copy_values,
				    unsigned permute_mask = 0xFF) const
  {
    bool no_permute = (perm == NULL) || ((permute_mask & 1) == 0);
    typename TRAIT::TypeImpl* ret;
    
    const Shape<DIM> s = no_permute ? this->shape() : Shape<DIM>(this->shape_.size().permute(*perm));    
    ret = new typename TRAIT::TypeImpl(NULL, s, alloc, &mmap, this->home());
    
    if (copy_values) 
      this->copy_(perm, ret, permute_mask);
    
    return ret;
  }

  /**
   * COPY this to dst
   * re-factored from clone_ -- useful when copying without cloning.
   * works only for tile major storage layout
   */
  virtual void copy_ (const Tuple<DIM>* perm,
	      typename TRAIT::TypeImpl* dst,
	      unsigned permute_mask = 0xFF) const
  {
    bool no_permute = (perm == NULL) || ((permute_mask & 1) == 0);
    assert (dst->isLocal());
    assert (this->isLocal());
    
    if (no_permute)
      map (Operator::idem<T>(), *this, *dst);
    else {
      typename Shape<DIM>::iterator it = this->shape_.begin();
      typename Shape<DIM>::iterator end = this->shape_.end();
      for (; it != end; it++) {
	const Tuple<DIM> t = *it;
	const Tuple<DIM> tp = t.permute(*perm);
	dst->scalarAt(tp) = scalarAt(t);
      }   
    }
  }

  /**
   * assignment operator: elementwise assignment 
   * this is a specialization of the default implementation of 
   * this operator in the Array superclass.
   */
  typename TRAIT::TypeImpl& operator= (const typename TRAIT::TypeImpl& rhs) 
  {
    assert (this->isLocal());
    if (this != &rhs) {
      if (this->memMap().isContiguous() && rhs.memMap().isContiguous() &&
	  this->shape().isContiguous() && rhs.shape().isContiguous()) {
	assert (this->shape().card() == rhs.shape().card());
	memcpy(this->raw(), rhs.raw(), this->shape().card() * sizeof(T));
      } else 
	map (Operator::idem<T>(), rhs, *this);
    }
    return this->actual_();
  }
  
  typename TRAIT::TypeImpl& operator= (const T& rhs) 
  {
    map (Operator::cnst<T>(rhs), *this);
    return this->actual_();
  }

  // dotproduct ----------------------------------------------------------------------
  
  static void ldotproduct (const Array<T, DIM, 0, TRAIT>& larg, 
			   const Array<T, DIM, 0, TRAIT>& rarg,
			   Array<T, DIM, 0, TRAIT>& dst)
  {
    assert (DIM == 2); // current limitation
    dst.scalarAt_(0) = dotproduct2D_ (larg, rarg);
  }
  
  static typename TRAIT::TypeScalar dotproduct (const Array<T, DIM, 0, TRAIT>& larg, 
						const Array<T, DIM, 0, TRAIT>& rarg)
  {
    assert (larg.isLocal());
    assert (rarg.isLocal());
    
    T ret;
    if (DIM <= 2) {
      switch (DIM) {
      case 1:
	ret = dotproduct1D_ (larg, rarg);
	break;
      case 2:
	ret = dotproduct2D_ (larg, rarg);
	break;
      default:
	ret = TypeHelper<T>::nullVal();
	break;
      }
    } 
    HTA_DEBUG(5, "LeafHTAImpl::dotproduct larg=" << larg << endl << " rarg= " << rarg << endl << " ret=" << ret);
    return ret;
  }

  static T dotproduct1D_ (const Array<T, DIM, 0, TRAIT>& larg, 
			  const Array<T, DIM, 0, TRAIT>& rarg)
  {
    assert (larg.shape() == rarg.shape());
    assert (DIM == 1 || larg.shape().size()[0] == 1 || larg.shape().size()[1] == 1);

    T ret = TypeHelper<T>::nullVal();
    const Tuple<DIM>& size = larg.shape().size();
    const Tuple<DIM> step_larg = larg.memMap().linearStep() * larg.shape().step();
    const Tuple<DIM> step_rarg = rarg.memMap().linearStep() * rarg.shape().step();
 
    int size_0 = size[0];
    int step_larg_0 = step_larg[0];
    int step_rarg_0 = step_rarg[0];
    
    for (int i = 0; i < size_0; ++ i) 
      ret += larg.scalarAt_(i * step_larg_0) * rarg.scalarAt_(i * step_rarg_0);
    return ret;
  }

  static T dotproduct2D_ (const Array<T, DIM, 0, TRAIT>& larg, 
			  const Array<T, DIM, 0, TRAIT>& rarg)
  {
    assert (larg.shape() == rarg.shape());
    assert (DIM == 2);

    T ret = TypeHelper<T>::nullVal();
    const Tuple<DIM>& size = larg.shape().size();
    const Tuple<DIM> step_larg = larg.memMap().linearStep() * larg.shape().step();
    const Tuple<DIM> step_rarg = rarg.memMap().linearStep() * rarg.shape().step();

    int size_0 = size[0];
    int step_larg_0 = step_larg[0];
    int step_rarg_0 = step_rarg[0];

    int size_1 = size[1];
    int step_larg_1 = step_larg[1];
    int step_rarg_1 = step_rarg[1];
    
    for (int j = 0; j < size_0; ++ j) 
      for (int i = 0; i < size_1; ++ i) 
	ret += larg.scalarAt_(i * step_larg_1 + j * step_larg_0) *
	  rarg.scalarAt_(i * step_rarg_1 + j * step_rarg_0);

    return ret;
  }
 
  void initTile (const Tuple<DIM>& t, typename TRAIT::TypeTileImpl* tile)
  {
    assert (false);
  }

  void initSparse (T* values, const int* start, const int* idx) 
  {
    assert (false);
  }

  // leaf-matmul ----------------------------------------------------------------------

  static void lmatmul (const Array<T, DIM, 0, TRAIT>& larg, 
		       const Array<T, DIM, 0, TRAIT>& rarg, 
		       Array<T, DIM, 0, TRAIT>& dst)
  {
    return matmul(larg, rarg, dst);
  }

  // matmul ----------------------------------------------------------------------

  /*  static void matmul (const Array<T, DIM, 0, TRAIT>& larg, 
		      const Array<T, DIM, 0, TRAIT>& rarg, 
		      Array<T, DIM, 0, TRAIT>& dst) 
  {
    assert (DIM == 2);
    assert (larg.shape().size()[1] == rarg.shape().size()[0]);
    assert (dst.shape().size()[0] == larg.shape().size()[0]);
    assert (dst.shape().size()[1] == rarg.shape().size()[1]);

    int M = larg.shape().size()[0];
    int Q = larg.shape().size()[1];
    int N = rarg.shape().size()[1];

    for (int i = 0; i< M; i++) {
      for (int k = 0; k < Q; k++) {
	for (int j = 0; j< N; j++) 
	  dst.scalarAt(Tuple<2>(i,j)) += larg.scalarAt(Tuple<2>(i,k)) * rarg.scalarAt(Tuple<2>(k,j));
      }
    }
    } */
  
  /** 
   * scalarAt operation using a Tuple 
   * @return reference to a scalar
   */
  T& scalarAt (const Tuple<DIM>& t) const
  {
    assert (this->isAvailable_());
    assert (this->shape_.contains(t)); // bounds check;
    int ord = this->memMap().ord(t);
    assert (!this->isLocal() || // bounds check with allocator only for local tiles
	    this->alloc_.checkAddress_(this->memMap().offset() + ord));
    return data_[ord]; 
  }

  inline void communicate_ (const Distribution<DIM>& target_dist, 
			    Array<T, DIM, 0, TRAIT>* target_arr = NULL) const
  {
    if (target_arr)
      target_arr->actual_() = *this;
  }
  
  template <typename T2, int LEVEL2, typename TRAIT2>
  inline void communicate_ (Array<T2, DIM, LEVEL2, TRAIT2>& target) const
  {
    // nothing to be done in the serial version
  } 
  
  inline void invalidate_ () const 
  {
    // nothing to be done in the serial version
  }


private:

  /** 
   * This accessor allows other leaf-instances to efficiently access the current 
   * instance. 
   *
   * @see map
   */
  T& scalarAt_ (int idx) const 
  {
    assert (this->isAvailable_());

#ifndef NDEBUG
    if (this->isLocal())
      // if this is a local access, then the address must be within bounds....
      assert (this->alloc_.checkAddress_(this->memMap().offset() + idx)); 
    else 
      // else - access to cache
      assert (idx < this->memMap().linearCard()); 
    assert (!this->memMap().isContiguous() || (idx >= 0 && idx < this->memMap().linearCard())); 
#endif

    return data_[idx]; 
  }

  T& scalarAt_ (const typename Shape<DIM>::iterator& it) const 
  {
    assert (this->isAvailable_());

    // compute address
    int idx;
    if (this->memMap().isContiguous()) {
      // relative address corresponds to the ordinal number returned by th iterator
      // this is fast
      idx = it.ord();
    } else {
      Tuple<DIM> t = *it;
      idx = this->memMap().ord(t);
    }
    // if this is a local access, then the address must be within bounds....
    assert (!this->isLocal() || this->alloc_.checkAddress_(this->memMap().offset() + idx)); 
    
    // else - access to cache
    assert (this->isLocal() || idx < this->memMap().linearCard()); 
    
    // the following assertion may be violated if the shape covers only 
    // a part of the underlying array. It would still be OK to do the 
    // access
    // assert (idx < this->shape_.linearCard()); 
    
    HTA_DEBUG (3, "LeafHTAImpl::scalarAt_ = idx=" << idx);
    return data_[idx];
  }

public:

  /** 
   * scalarAt operation using a Shape
   * 
   * @return   HTA of the same level as this HTA which is a 'view' 
   *           restricted to the specified variables
   */
  typename TRAIT::TypeImpl* scalarAt (const Shape<DIM>& s) const
  { 
    assert (this->shape_.contains(s.low()));  // bounds check;
    assert (this->shape_.contains(s.high())); // bounds check;

    // compute the shape of the result
    int offset = this->memMap().offset() + this->memMap().ord(s.low());
    Shape<DIM> shape(s);
    shape.relocate(Tuple<DIM>::zero);
    Tuple<DIM> linear_size = s.high() - s.low() + Tuple<DIM>::one;
    MemMapping<DIM>* mmap = new LeafMemMapping<DIM> (offset, 0, linear_size,
						     Tuple<DIM>::zero, &this->memMap().linearStep()); 
    // same allocator, i.e., the return HTA is a view on the current HTA
    return new typename TRAIT::TypeImpl (NULL, shape, this->alloc_, mmap, this->home());
  }

  typename TRAIT::TypeImpl* tileAt (const Shape<DIM>& s) const 
  {
    return scalarAt(s);
  }
  
  typename TRAIT::TypeImpl* tileAt (int n, const Shape<DIM>* sv) const 
  {
    // this method is only available for HTA's with LEVEL > 0
    // it is implemented here only to satisfy a compile-time dependency.
    assert (false); 
    return NULL;
  }

  inline typename TRAIT::TypeTileImpl* tileAt (int tile_idx) const
  {
    assert (false);
    return NULL;
  }

  /** convolution operator */
  void lconv (const T* weights, 
	      Array<T, DIM, 0, TRAIT>& dst) const 
  {
    // this is a limitation of the current implementation
    assert (weights != NULL);
    assert (this->shape().size() == dst.shape().size());
    assert (this->shape().isContiguous());
    assert (this->memMap().isContiguous());
    assert (dst.shape().isContiguous());
    assert (dst.memMap().isContiguous());

    switch (DIM) {
    case 1:
      lconv1D_(weights[0], weights[1], this->actual_(), dst.actual_()); 
      break;
    case 2:
      lconv2D_(weights[0], weights[1], weights[2], this->actual_(), dst.actual_()); 
      break;
    case 3:
      lconv3D_(weights[0], weights[1], weights[2], weights[3], this->actual_(), dst.actual_()); 
      break;
    default:
      assert (false);
    } 
    return;
  }

  /** reduction over all dimensions */
  template <typename OP> 
  T reduce (const OP& binary_op) const
  {
    assert (this->isAvailable_());
    T acc = TypeHelper<T>::identityVal(binary_op);

    // TODO: this should be done with a math kernel
    if (this->shape().isContiguous() && this->memMap().isContiguous()) {
      int card = this->shape().card();
      for (int i = 0; i < card; i++)
	acc = binary_op(acc, this->data_[i]); 
    } else {
      typename Shape<DIM>::iterator i = this->shape_.begin();
      typename Shape<DIM>::iterator end = this->shape_.end();
      for ( ; i != end; i++) 
	acc = binary_op(acc, scalarAt(*i));
    }
    HTA_DEBUG(5, "LeafHTAImpl::reduce res=" << acc);
    return acc;
  }
  
  template <typename OP> 
  void reduce (const OP& binary_op, 
	       int dim,
	       typename TRAIT::TypeImpl& dst) const
  {
    // compiler asks for this method because another method wit the same name 
    // is defined in this class and shadown the inherited variant of this method.
    // call base-class version of method
    Array<T, DIM, 0, TRAIT>::reduce(binary_op, dim, dst);
  }
  
  template <typename OP>
  void mapReduce (OP& mr_op)
  { 
    switch (mr_op.keyType()) {
    case Operator::MR_NONE: {
      // TODO: specialize calls for DIM=1,2,3 like map
      // call map for all elements with argument key=tuple_scalarAt, val
      typename Shape<DIM>::iterator i = this->shape().begin();
      typename Shape<DIM>::iterator end = this->shape().end();
      for ( ; i != end; i++) {
	Tuple<DIM> t = *i;
	T& val = this->scalarAt(t);
	mr_op.map_none(val);
      }
      break;
    }
    case Operator::MR_PROC_OFF: {
      assert (false); // to be implemented
      // call map for all elements with argument key=(proc, idx), val
      break;
    }
    case Operator::MR_TUPLE: {
      // TODO: specialize calls for DIM=1,2,3 like map
      // call map for all elements with argument key=tuple_scalarAt, val
      typename Shape<DIM>::iterator i = this->shape().begin();
      typename Shape<DIM>::iterator end = this->shape().end();
      Tuple<DIM> pos = this->memMap().leafPos();
      for ( ; i != end; i++) {
	Tuple<DIM> t = *i;
	T val = this->scalarAt(t);
	mr_op.map_tuple(t + pos, val);
      }
    }  
    } // switch
  }

  template <typename OP>
  static void map (const OP& nullary_op, 
		   Array<T, DIM, 0, TRAIT>& dst)
  {
    assert (dst.isLocal());
    
    HTAImpl<T, DIM, 0, TRAIT>* dst_actual = dynamic_cast<HTAImpl<T, DIM, 0, TRAIT>*>(&dst); 

    if (DIM <= 3 && dst_actual) {
      switch (DIM) {
      case 1:
	map1D_ (nullary_op, dst);
	break;
      case 2:
	map2D_ (nullary_op, dst);
	break;
      case 3:
	map3D_ (nullary_op, *dst_actual);
	break;
      default:
	assert(false);
	break;
      }
    } else {
      assert (DIM > 0);
      if (dst.shape().isContiguous() && dst.memMap().isContiguous()) {
	// TODO: use a mathkernel to do this
	// the iteration pattern on the linear-representation has step 1 for both arrays 
	int end = dst.shape().card();
	for (int i = 0 ; i < end; i++) 
	  dst.scalarAt_(i) = nullary_op();
      } else {
	// this is most inefficient!
	// different iteration patterns on both arrays
	typename Shape<DIM>::iterator i = dst.shape().begin();
	typename Shape<DIM>::iterator end = dst.shape().end();
	for ( ; i != end; i++) 
	  dst.scalarAt(*i) = nullary_op();
      }
    }
  }
  
  /** unary map */
  template <typename T2, typename OP, typename TRAIT2>
  static void map (const OP& unary_op, 
		   const Array<T, DIM, 0, TRAIT>& arg, 
		   Array<T2, DIM, 0, TRAIT2>& dst)
  {
    assert (arg.isAvailable_());
    assert (dst.isLocal());
    assert (Shape<DIM>::conformable(arg.shape(), dst.shape()));

    const HTAImpl<T, DIM, 0, TRAIT>* arg_actual = dynamic_cast<const HTAImpl<T, DIM, 0, TRAIT>*>(&arg); 
    HTAImpl<T, DIM, 0, TRAIT>* dst_actual = dynamic_cast<HTAImpl<T, DIM, 0, TRAIT>*>(&dst); 
 
    if (DIM <= 3 && arg_actual && dst_actual) {
      switch (DIM) {
      case 1:
	map1D_ (unary_op, arg, dst);
	break;
      case 2:
	map2D_ (unary_op, arg, dst);
	break;
      case 3:
	map3D_ (unary_op, *arg_actual, *dst_actual);
	break;
      default:
	assert (false);
	break;
      }
    } else {
      assert (DIM > 0);
      if (arg.shape().isContiguous() && dst.shape().isContiguous() &&
	  arg.memMap().isContiguous() && dst.memMap().isContiguous()) {
	// TODO: use a mathkernel to do this
	// the iteration pattern on the linear-representation has step 1 for both arrays 
	int end = arg.shape().card();
	for (int i = 0; i < end; i++) 
	  dst.scalarAt_(i) =  static_cast<T2>(unary_op(arg.scalarAt_(i)));
      } else if (Shape<DIM>::iteratorConformable_(arg.shape(), dst.shape())) {
	// on the linearized representation, the iteration pattern is the same 
	// for both arrays, however not just step 1
	typename Shape<DIM>::iterator arg_it = arg.shape().begin();
	typename Shape<DIM>::iterator arg_end = arg.shape().end();
	for ( ; arg_it != arg_end; arg_it++) 
	  dst.scalarAt_(arg_it) = static_cast<T2>(unary_op(arg.scalarAt_(arg_it)));
      } else {
	// this is most inefficient!
	// different iteration patterns on both arrays
	typename Shape<DIM>::iterator arg_it = arg.shape().begin();
	typename Shape<DIM>::iterator arg_end = arg.shape().end();
	typename Shape<DIM>::iterator dst_it = dst.shape().begin();
	for ( ; arg_it != arg_end; arg_it++, dst_it++)
	  dst.scalarAt(*dst_it) = static_cast<T2>(unary_op(arg.scalarAt(*arg_it)));
      }
    }
  }
   
  /** binary map */
  template <typename OP, typename T2, typename TRAIT2>
  static void map (const OP& binary_op, 
		   const Array<T, DIM, 0, TRAIT>& larg, 
		   const Array<T2, DIM, 0, TRAIT2>& rarg, 
		   Array<T, DIM, 0, TRAIT>& dst) 
  {
    assert (larg.isAvailable_());
    assert (rarg.isAvailable_());
    assert (dst.isLocal());

    if (rarg.shape().card() == 1) {
      map (binder2nd<OP>(binary_op, rarg.scalarAt_(0)), larg, dst);
      return;
    }

    assert (Shape<DIM>::conformable(rarg.shape(), dst.shape()) &&
	    Shape<DIM>::conformable(larg.shape(), dst.shape()));
    assert (Shape<DIM>::conformable(larg.shape(), rarg.shape()));

    const HTAImpl<T, DIM, 0, TRAIT>* larg_actual = dynamic_cast<const HTAImpl<T, DIM, 0, TRAIT>*>(&larg); 
    const HTAImpl<T2, DIM, 0, TRAIT2>* rarg_actual = dynamic_cast<const HTAImpl<T2, DIM, 0, TRAIT2>*>(&rarg); 
    HTAImpl<T, DIM, 0, TRAIT>* dst_actual = dynamic_cast<HTAImpl<T, DIM, 0, TRAIT>*>(&dst); 

    if (DIM <= 3 && larg_actual && rarg_actual && dst_actual) {
      switch (DIM) {
      case 1:
	map1D_ (binary_op, larg, rarg, dst);
	break;
      case 2:
	map2D_ (binary_op, larg, rarg,  dst);
	break;
      case 3:
	map3D_ (binary_op, *larg_actual, *rarg_actual, *dst_actual);
	break;
      default:
	assert (false);
	break;
      }
    } else {

      // this is a limitation of the current implementation
      assert (Shape<DIM>::iteratorConformable_(larg.shape(), dst.shape()));

      if (larg.shape().isContiguous() && rarg.shape().isContiguous() && dst.shape().isContiguous() &&
	  larg.memMap().isContiguous() && rarg.memMap().isContiguous() && dst.memMap().isContiguous()) {
	// TODO: this might be done by an efficient mathkernel 
	int end = larg.shape().card();
	for (int idx = 0;  idx < end; idx++) {
	  // int idx = i.ord();
	  dst.scalarAt_(idx) = binary_op(larg.scalarAt_(idx), rarg.scalarAt_(idx));
	}
      } else if (Shape<DIM>::iteratorConformable_(larg.shape(), rarg.shape())) {
	typename Shape<DIM>::iterator i = larg.shape().begin();
	typename Shape<DIM>::iterator end = larg.shape().end();
	for ( ; i != end; i++)
	  dst.scalarAt_(i) = binary_op(larg.scalarAt_(i), rarg.scalarAt_(i));
      } else {
	// this is most inefficient!
	// different iteration patterns on both arrays
	typename Shape<DIM>::iterator larg_it = larg.shape().begin();
	typename Shape<DIM>::iterator larg_end = larg.shape().end();
	typename Shape<DIM>::iterator rarg_it = rarg.shape().begin();
	for ( ; larg_it != larg_end; larg_it++, rarg_it++)
	  dst.scalarAt_(larg_it) = binary_op(larg.scalarAt_(larg_it), rarg.scalarAt_(rarg_it));
      }
    }
  }

  typename TRAIT::TypeInternal raw() const //__attribute__((always_inline))
  {
    return data_;
  }

protected:

  // all other template instance of this class should be friends (required when accessing print_)
  template<class T2, int DIM2, int LEVEL2, typename TRAIT2>
  friend class HTAImpl;

  // to allow forwarding class to private methods from class Array to this class.
  template<class T2, int DIM2, int LEVEL2, typename TRAIT2>
  friend class Array;

  /**
   * @return  true if data is accessible locally. The tile might either be local
   *          or it might be cached from a remote location 
   */
  virtual bool isAvailable_() const 
  {
    return data_ != NULL;
  }


 /** reduction over one dimension */
  template <typename OP> 
  void reduce_ (int oplevel,
		const OP& binary_op, 
		int dim, 
		typename TRAIT::TypeImpl& dst) const
  {
    if (oplevel > 0)
      
      dst = *this;

    else
      
      reduce_ (binary_op, dim, dst);
  }
  
  /** reduction over one dimension */
  template <typename OP> 
  void reduce_ (const OP& binary_op, 
		int dim, 
		typename TRAIT::TypeImpl& dst) const
  {
    assert (this->isAvailable_());
    assert (dst.isLocal());
    
    // initialize the destination
    Shape<DIM> masked = this->shape().mask(dim);
    typename Shape<DIM>::iterator i = masked.begin();
    typename Shape<DIM>::iterator endi = masked.end();
    T bottom = TypeHelper<T>::identityVal(binary_op);
    for ( ; i != endi; i++) 
      dst.scalarAt(*i) = bottom;
      
    // then do a the reduction
    typename Shape<DIM>::iterator j = this->shape().begin();
    typename Shape<DIM>::iterator endj = this->shape().end();
    for ( ; j != endj; j++) {
      Tuple<DIM> t = *j;
      T& dst_var = dst.scalarAt(t.mask(dim));
      HTA_DEBUG(5, "LeafHTAImpl::reduce_ dst=" << dst_var << " arg=" << scalarAt(t));
      dst_var = binary_op(dst_var, scalarAt(t));
      HTA_DEBUG(5, "LeafHTAImpl::reduce_ res=" << dst_var);
    }

    bool all_to_all = dst.shape().size()[dim] > 1;
    if (all_to_all) {
      //assert (dst.shape_.step() == Tuple<DIM>::one);
      //assert (dst.shape_.low() == Tuple<DIM>::zero);

      // propagate results from tile with dim=0 to other tiles ...
      Tuple<DIM> size_tuple = dst.shape().size();
      int dim_size = size_tuple[dim];
      typename Shape<DIM>::iterator k = masked.begin();
      typename Shape<DIM>::iterator endk = masked.end();
      for ( ; k != endk; k++) {
	Tuple<DIM> t = *k;
	T src = dst.scalarAt(t);
	for (int j = 1; j < dim_size; ++j) {
	  t[dim] = j;
	  dst.scalarAt(t) = src;
	}
      }
     }
  }

  /** print */   
  virtual void print_ (ostream& ost, int indent = 0) const 
  {
    assert (isAvailable_());
    
    HTA_DEBUG (2, "HTAImpl<LEVEL=0>::print_ this=" << this << " data_=" << data_);

    for (int j = 0; j < indent; ++j) 
      ost << " ";

    if (DIM == 2) {
      int printed = 0;
      int mod = this->shape_[1].size(); // column dimension has index 1
      for (typename Shape<DIM>::iterator i = this->shape_.begin(); i != this->shape_.end(); i++) {
	ost << scalarAt_(i) << "\t";
	if (++printed % mod == 0) {
	  ost << endl;
	  for (int j = 0; j < indent; ++j) ost << " ";
	}
      }
    } else {
      for (typename Shape<DIM>::iterator i = this->shape_.begin(); i != this->shape_.end(); i++)    
	ost << scalarAt(*i) << " "; 
      ost << endl;
    }
  }    

  /** 
   * pointer to the array that holds the variables starting at this->memMap().offset() 
   * this is not const, because subclasses could change the pointer
   *
   */
  T* data_;

private:

  // specialization of convolution  for DIM==1 
  static void lconv1D_ (const T w0, const T w1, 
			const typename TRAIT::TypeImpl& arg,
			typename TRAIT::TypeImpl& dst) 
  {
    assert (DIM == 0);

    int size = dst.shape().size()[0];
    int step_dst = dst.memMap().linearStep()[0] * dst.shape().step()[0];
    int step_arg = arg.memMap().linearStep()[0] * arg.shape().step()[0];
    
    for (int i = 1; i < size-1; ++ i) {
      T u0 = arg.scalarAt_(i * step_arg);
      T u1 = arg.scalarAt_(i-1 * step_arg) + arg.scalarAt_(i+1 * step_arg);
      dst.scalarAt_(i * step_dst) = u0 * w0 +  u1 * w1;
    }
  }

  static void lconv2D_ (const T w0, const T w1, const T w2,
			const typename TRAIT::TypeImpl& arg,
			typename TRAIT::TypeImpl& dst) 
  {
    assert (DIM == 2);
    assert (arg.shape().step() == Tuple<DIM>::one);
    assert (dst.shape().step() == Tuple<DIM>::one);

    int size_0 = arg.shape().size()[0]; // most significant
    int size_1 = arg.shape().size()[1];
    T* a = arg.raw();
    T* d = dst.raw();

    for (int i = 1; i < size_0 - 1; ++ i) {
      for (int j = 1; j < size_1 - 1; ++ j) {
	T u0 = 
	  a[i     * size_1 + j];
	T u1 = 
	  a[(i-1) * size_1 + j] +
	  a[i     * size_1 + (j-1)] +
	  a[(i+1) * size_1 + j] +
	  a[i     * size_1 + (j+1)];
	T u2 = 
	  a[(i-1) * size_1 + (j-1)] + 
	  a[(i-1) * size_1 + (j+1)] +
	  a[(i+1) * size_1 + (j-1)] + 
	  a[(i+1) * size_1 + (j+1)];
	d[i * size_1 + j] = w0 * u0 + w1 * u1 + w2 * u2;
      }
    }
  }

  static void lconv3D_ (const T w0, const T w1, const T w2, const T w3,
			const typename TRAIT::TypeImpl& arg,
			typename TRAIT::TypeImpl& dst) 
  {
    assert (DIM == 3);
    assert (arg.shape().step() == Tuple<DIM>::one);
    assert (dst.shape().step() == Tuple<DIM>::one);

    int size_0 = arg.shape().size()[0]; // most significant
    int size_1 = arg.shape().size()[1];
    int size_2 = arg.shape().size()[2];
    
    double u1[size_2];
    double u2[size_2];
    
    T* a = arg.raw();
    T* d = dst.raw();
    assert (a);   
    assert (d);   
 
    for (int i = 1; i < size_0 - 1; i++) {
      for (int j = 1; j < size_1 - 1; j++) {
	
        for (int k = 0; k < size_2; k++) {
	  // can the compiler hoist + k from the index computation here ?
	  u1[k] = 
	    a[i * size_1 * size_2    + (j-1) * size_2 + k] + 
	    a[i * size_1 * size_2    + (j+1) * size_2 + k] + 
	    a[(i-1) * size_1 * size_2 + j * size_2    + k] + 
	    a[(i+1) * size_1 * size_2 + j * size_2    + k];
	  u2[k] = 
	    a[(i-1) * size_1 * size_2 + (j-1) * size_2 + k] +
	    a[(i-1) * size_1 * size_2 + (j+1) * size_2 + k] + 
	    a[(i+1) * size_1 * size_2 + (j-1) * size_2 + k] + 
	    a[(i+1) * size_1 * size_2 + (j+1) * size_2 + k];
	}
	
        for (int k = 1; k < size_2 - 1; k++) {
	  d[i * size_1 * size_2 + j * size_2 + k] = 
	    w0 * a[i * size_1 * size_2 + j * size_2 + k]  +
	    w1 * (u1[k] + a[i * size_1 * size_2 + j * size_2 + k-1] + a[i * size_1 * size_2 + j * size_2 + k+1]) +
	    w2 * (u2[k] + u1[k-1] + u1[k+1]) +
	    w3 * (u2[k-1] + u2[k+1]);
	}
      }
    }
  }
 


  // specialization of nullary map for DIM == 1. 
  template <typename OP> 
  static void map1D_ (const OP& nullary_op, 
		      Array<T, DIM, 0, TRAIT>& dst) 
  {
    assert (DIM == 1);
    const Tuple<DIM>& size = dst.shape().size();
    Tuple<DIM> step = dst.memMap().linearStep() * dst.shape().step();

    int size_0 = size[0];
    int step_dst_0 = step[0];
    
    for (int i = 0; i < size_0; ++ i) 
      dst.scalarAt_(i * step_dst_0) = nullary_op();
  }

   // specialization of nullary map for DIM == 2. 
  template <typename OP> 
  static void map2D_ (const OP& nullary_op, 
			  Array<T, DIM, 0, TRAIT>& dst) 
  {
    assert (DIM == 2);
    const Tuple<DIM>& size = dst.shape().size();
    const Tuple<DIM> step = dst.memMap().linearStep() * dst.shape().step();

    int size_0 = size[0];
    int step_dst_0 = step[0];

    int size_1 = size[1];
    int step_dst_1 = step[1];
    
    for (int j = 0; j < size_0; ++ j) 
      for (int i = 0; i < size_1; ++ i) 
	dst.scalarAt_(i * step_dst_1 + j * step_dst_0) = nullary_op();
  }

  // specialization of nullary map for DIM == 3. 
  template <typename OP> 
  static void map3D_ (const OP& nullary_op, 
		      HTAImpl<T, DIM, 0, TRAIT>& dst) 
  {
    assert (DIM == 3);
    
    const Tuple<DIM>& size = dst.shape().size();
    const Tuple<DIM> step = dst.memMap().linearStep() * dst.shape().step();
    
    int size_0 = size[0];
    int step_dst_0 = step[0];
    
    int size_1 = size[1];
    int step_dst_1 = step[1];
    
    int size_2 = size[2];
    int step_dst_2 = step[2];

    T* dst_raw = dst.raw();

    for (int k = 0; k < size_0; ++ k)  
      for (int j = 0; j < size_1; ++ j)
	for (int i = 0; i < size_2; ++ i) 
	  dst_raw[i * step_dst_2 + j * step_dst_1 + k * step_dst_0] = nullary_op();
  }

  // specialization of unary map for DIM == 1. 
  template <typename T2, typename OP, typename TRAIT2> 
  static void map1D_ (const OP& unary_op, 
		      const Array<T, DIM, 0, TRAIT>& arg, 
		      Array<T2, DIM, 0, TRAIT2>& dst) 
  {
    assert (DIM == 1);
    const Tuple<DIM>& size = arg.shape().size();
    const Tuple<DIM> step_dst = dst.memMap().linearStep() * dst.shape().step();
    const Tuple<DIM> step_arg = arg.memMap().linearStep() * arg.shape().step();

    int size_0 = size[0];
    int step_arg_0 = step_arg[0];
    int step_dst_0 = step_dst[0];
    
    for (int i = 0; i < size_0; ++ i) 
      dst.scalarAt_(i * step_dst_0) = 
	static_cast<T2>(unary_op(arg.scalarAt_(i * step_arg_0)));
  }

  // specialization of unary map for DIM == 2. 
  template <typename T2, typename OP, typename TRAIT2> 
  static void map2D_ (const OP& unary_op, 
		      const Array<T, DIM, 0, TRAIT>& arg, 
		      Array<T2, DIM, 0, TRAIT2>& dst) 
  {
    assert (DIM == 2);
    const Tuple<DIM>& size = arg.shape().size();
    const Tuple<DIM> step_dst = dst.memMap().linearStep() * dst.shape().step();
    const Tuple<DIM> step_arg = arg.memMap().linearStep() * arg.shape().step();

    int size_0 = size[0];
    int step_arg_0 = step_arg[0];
    int step_dst_0 = step_dst[0];

    int size_1 = size[1];
    int step_arg_1 = step_arg[1];
    int step_dst_1 = step_dst[1];
    
    for (int j = 0; j < size_0; ++ j) 
      for (int i = 0; i < size_1; ++ i) 
	dst.scalarAt_(i * step_dst_1 + j * step_dst_0) = 
	  static_cast<T2>(unary_op(arg.scalarAt_(i * step_arg_1 + j * step_arg_0)));
  }

  // specialization of unary map for DIM == 3. 
  template <typename T2, typename OP, typename TRAIT2> 
  static void map3D_ (const OP& unary_op, 
		      const HTAImpl<T, DIM, 0, TRAIT>& arg, 
		      HTAImpl<T2, DIM, 0, TRAIT2>& dst) 
  {
    assert (DIM == 3);

    const Tuple<DIM>& size = arg.shape().size();
    const Tuple<DIM> step_dst = dst.memMap().linearStep() * dst.shape().step();
    const Tuple<DIM> step_arg = arg.memMap().linearStep() * arg.shape().step();

    int size_0 = size[0];
    int step_arg_0 = step_arg[0];
    int step_dst_0 = step_dst[0];

    int size_1 = size[1];
    int step_arg_1 = step_arg[1];
    int step_dst_1 = step_dst[1];

    int size_2 = size[2];
    int step_arg_2 = step_arg[2];
    int step_dst_2 = step_dst[2];

    T* dst_raw = dst.raw();
    const T* arg_raw = arg.raw();

    for (int k = 0; k < size_0; ++ k)  
      for (int j = 0; j < size_1; ++ j)
	for (int i = 0; i < size_2; ++ i) 
	  dst_raw[i * step_dst_2 + j * step_dst_1 + k * step_dst_0] = 
	    static_cast<T2>(unary_op(arg_raw[i * step_arg_2 + j * step_arg_1 + k * step_arg_0]));
  }

  // specialization of binary map for DIM == 1. 
  template<typename OP, typename T2, typename TRAIT2>
  static void map1D_ (const OP& binary_op, 
		      const Array<T, DIM, 0, TRAIT>& larg, 
		      const Array<T2, DIM, 0, TRAIT2>& rarg, 
		      Array<T, DIM, 0, TRAIT>& dst) 
  {
    assert(DIM == 1);

    const Tuple<DIM>& size = larg.shape().size();
    const Tuple<DIM> step_dst = dst.memMap().linearStep() * dst.shape().step();
    const Tuple<DIM> step_larg = larg.memMap().linearStep() * larg.shape().step();
    const Tuple<DIM> step_rarg = rarg.memMap().linearStep() * rarg.shape().step();

    int size_0 = size[0];
    int step_larg_0 = step_larg[0];
    int step_rarg_0 = step_rarg[0];
    int step_dst_0 = step_dst[0];

    for (int i = 0; i < size_0; ++ i) 
      dst.scalarAt_(i * step_dst_0) = 
	binary_op(larg.scalarAt_(i * step_larg_0), rarg.scalarAt_(i * step_rarg_0));
  } 

  // specialization of binary map for DIM == 2. 
  // iterate separately over each dimension
  template <typename OP, typename T2, typename TRAIT2> 
  static void map2D_ (const OP& binary_op, 
			  const Array<T, DIM, 0, TRAIT>& larg, 
			  const Array<T2, DIM, 0, TRAIT2>& rarg, 
			  Array<T, DIM, 0, TRAIT>& dst) 
  {
    assert (DIM == 2);
    const Tuple<DIM>& size = larg.shape().size();
    const Tuple<DIM> step_dst = dst.memMap().linearStep() * dst.shape().step();
    const Tuple<DIM> step_larg = larg.memMap().linearStep() * larg.shape().step();
    const Tuple<DIM> step_rarg = rarg.memMap().linearStep() * rarg.shape().step();

    int size_0 = size[0];
    int step_larg_0 = step_larg[0];
    int step_rarg_0 = step_rarg[0];
    int step_dst_0 = step_dst[0];
    
    int size_1 = size[1];
    int step_larg_1 = step_larg[1];
    int step_rarg_1 = step_rarg[1];
    int step_dst_1 = step_dst[1];

    for (int j = 0; j < size_0; ++ j) 
      for (int i = 0; i < size_1; ++ i) 
	dst.scalarAt_(i * step_dst_1 + j * step_dst_0) = 
	  binary_op(larg.scalarAt_(i * step_larg_1 + j * step_larg_0), 
		    rarg.scalarAt_(i * step_rarg_1 + j * step_rarg_0));
  }

  // specialization of binary map for DIM == 3. 
  // iterate separately over each dimension
  template <typename OP, typename T2, typename TRAIT2> 
  static void map3D_ (const OP& binary_op, 
		      const HTAImpl<T, DIM, 0, TRAIT>& larg, 
		      const HTAImpl<T2, DIM, 0, TRAIT2>& rarg, 
		      HTAImpl<T, DIM, 0, TRAIT>& dst) 
  {
    assert (DIM == 3);
    const Shape<DIM>& shpe = larg.shape();
    const Tuple<DIM>& size = shpe.size();
    const Tuple<DIM> step_dst = dst.memMap().linearStep() * dst.shape().step();
    const Tuple<DIM> step_larg = larg.memMap().linearStep() * larg.shape().step();
    const Tuple<DIM> step_rarg = rarg.memMap().linearStep() * rarg.shape().step();

    int size_0 = size[0];
    int step_larg_0 = step_larg[0];
    int step_rarg_0 = step_rarg[0];
    int step_dst_0 = step_dst[0];
    
    int size_1 = size[1];
    int step_larg_1 = step_larg[1];
    int step_rarg_1 = step_rarg[1];
    int step_dst_1 = step_dst[1];

    int size_2 = size[2];
    int step_larg_2 = step_larg[2];
    int step_rarg_2 = step_rarg[2];
    int step_dst_2 = step_dst[2];
    
    T* dst_raw = dst.raw();
    const T* larg_raw = larg.raw();
    const T2* rarg_raw = rarg.raw();

    for (int k = 0; k < size_0; ++ k) 
     for (int j = 0; j < size_1; ++ j)  
       for (int i = 0; i < size_2; ++ i) {
	 dst_raw[i * step_dst_2 + j * step_dst_1 + k * step_dst_0] = 
	   binary_op(larg_raw [i * step_larg_2 + j * step_larg_1 + k * step_larg_0], 
		     rarg_raw [i * step_rarg_2 + j * step_rarg_1 + k * step_rarg_0]);
       }
  }

};

#endif /* __LEAF_HTA_IMPL_H__ */
