/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: LocalAllocator.h,v 1.11.4.1 2007/08/13 16:47:15 brodman2 Exp $
 * Authors: Christoph von Praun
 */

#ifndef  __LOCAL_ALLOCATOR_H__
#define  __LOCAL_ALLOCATOR_H__

#include <cassert>
#include <iostream>

#include "util/AbstractAllocator.h"
#include "debug/Tracing.h"

using namespace std;

template <typename T> 
class LocalAllocator : public AbstractAllocator<T> {


//protected:
public:
  struct LocalKernel : public AbstractAllocator<T>::Kernel {

    LocalKernel (T* data, int nblocks, int nelems, bool managed) : 
      AbstractAllocator<T>::Kernel(nblocks, nelems, managed),
      data_(data) 
    {
      assert (data);
      assert (nblocks > 0);
      assert (nelems > 0); 
      HTA_DEBUG(1, "LocalKernel create size=" << nelems * sizeof(T) << " bytes.");
    }

    virtual ~LocalKernel () 
    {
      assert (data_);
      if (this->managed_) {
	delete[] data_;
	HTA_DEBUG(1, "LocalKernel delete size=" << this->nelems_ * sizeof(T) << " bytes.");
      }
    }

    /** pointer to the raw array -s a valid pointer, never NULL or equivalent */
    T* data_;
  };
  
public:

  LocalAllocator () { }

  /**
   * Constructor.
   *
   * @param  size     the number of T elements to allocate - T must have a 
   *                  default constructor.
   * @param  init     an array that is used to initialize the allocated region
   * @param  managed  if true, then the pointer provided by init is eventally de-allocated 
   *                  by this allocator.  
   * @return          an allocator instance 
   */
  LocalAllocator (int size, bool managed = true, T* init = NULL) 
  { 
    T* data;
    if (init) 
      data = init;
    else {
      assert (managed);
      data = new T[size];
    }
    this->kernel_ = new LocalKernel (data, 1, size, managed);
  }

  /**
   * Destructor
   */
  virtual ~LocalAllocator ()
  {}

protected:

  LocalKernel* getKernel_() const 
  {
    return static_cast<LocalKernel*>(this->kernel_);
  }

  /**
   * New operator: forbidden, because allocators are passed by value and allocated inline.
   */
  void* operator new (size_t size) throw()
  {
    assert(0);
    return NULL;
  }

  /**
   * Delete operator: forbidden, because allocators are passed by value and allocated inline.
   */
  void operator delete (void* to_delete) 
  {
    assert (0);
  }

  virtual void print_ (ostream& ost) const
  {
    ost << "LocalAllocator(data=" << (void*) getKernel_()->data_ << " size=" << this->size() << " refCtr=" << getKernel_()->refCtr_ << ")";
  }

public:

  /**
   * Copy constructor
   */
  LocalAllocator (const LocalAllocator<T>& o) :
    AbstractAllocator<T>(o) 
  {}

  /**
   * Assignment operator. The current kernel (memory block) is dropped and 
   * de-allocated (if necessary). Then, the kernel of the rhs allocator is 
   * assigned to this allocator.
   */
  LocalAllocator<T>& operator=(const LocalAllocator<T>& rhs) 
  {
    LocalKernel* lk = getKernel_();
    if (lk != rhs.kernel_) {
      if (lk != NULL && --(lk->allocCtr_) == 0 && lk->managed_) 
	delete lk;
      // base-class will de-allocate the kernel if necessary
      this->AbstractAllocator<T>::operator=(rhs);
    }
    return *this;
  }
  
  /**
   * Creates a copy of the memory block managed by an allocator and returns 
   * a new allocator object that manages this copy.
   *
   * @return  allocator for the copied memory block
   */
  LocalAllocator<T> clone (bool copy_values = false) const
  {
    assert (getKernel_() != NULL);
    int nelems = getKernel_()->nelems_;
    T* new_buf = new T[nelems];
    
    if (copy_values)
      memcpy(new_buf, getKernel_()->data_, nelems * sizeof(T));

    LocalAllocator<T> ret(nelems, true, new_buf);
    return ret;
  }
    
  /**
   * Obtain pointer to the memory block managed by the allocator 
   * at a certain offset
   *
   * @param  offset  offset in the allocated block
   * @return         pointer to the memory block
   */
  T* get(int offset = 0) const 
  {
    T* ret;
    if (getKernel_() != NULL) {
      assert (offset >= 0);
      //cout << offset << " " << getKernel_()->nelems_ << endl;
      assert (offset < getKernel_()->nelems_);
      getKernel_()->refCtr_++;
      HTA_DEBUG(2, "SerialAlloc::get kernel=" << getKernel_() << " ctr=" << getKernel_()->refCtr_);
      ret = getKernel_()->data_ + offset;
      HTA_DEBUG(2, "SerialAlloc::get kernel BBF " << getKernel_()->data_ << ' ' << offset << ' ' << sizeof(*ret) << ' ' << sizeof(*(getKernel_()->data_)));
    } else 
      ret = NULL;
    return ret;
  }


  /** 
   * @return true if idx is a valid index in the managed by this allocation 
   */
  bool checkAddress_(int idx) const
  {
    assert (getKernel_() != NULL);
    bool ret = (idx >= 0 && idx < getKernel_()->nelems_);
    
    if (!ret)
      cout << "Here " << idx << " " << getKernel_()->nelems_;
    return ret;
  }

};

#endif  /* __LOCAL_ALLOCATOR_H__ */
