#ifndef  __O_LOCAL_ALLOCATOR_H__
#define  __O_LOCAL_ALLOCATOR_H__

#include <cassert>
#include <iostream>

#include "util/AbstractAllocator.h"
#include "debug/Tracing.h"
#include "serial/LocalAllocator.h"
#include "Overlap.h"

using namespace std;

template <typename T> 
class OLocalAllocator: public LocalAllocator<T>{
private:
  struct OLocalKernel: public LocalAllocator<T>::LocalKernel{
    OLocalKernel(T* zeroPos, T* data, int nblocks, int nelems, bool managed):
      LocalAllocator<T>::LocalKernel(data, nblocks, nelems, managed),
      zeroPos_(zeroPos)
    {
      assert(zeroPos);
    }
    
    virtual ~OLocalKernel(){}
    // point to the starting position of the non-ghost region.
    T * zeroPos_; 
  };

private:
  OLocalKernel* getKernel_() const
  {
    return static_cast<OLocalKernel*>(this->kernel_);
  }

public:
  OLocalAllocator () {}
  // same as the LocalAllocator
   OLocalAllocator (int size, bool managed = true, T* init = NULL) 
  { 
    T* data;
    if (init) 
      data = init;
    else {
      assert (managed);
      data = new T[size];      
      memset(data, 0, size*sizeof(int));

    }
    this->kernel_ = new OLocalKernel (data, data, 1, size, managed);
  }

  /** 
   * Constructor with Overlapped tiling specification.
   * size will be the total size including the overlapped region.
   */
  OLocalAllocator (int offset, int size, bool managed = true, T* init = NULL){ 
    T* data;
    T* zeroPos;
    if (init){
      data = init;
      zeroPos = init+offset;
    }
    else {
      assert (managed);
      data = new T[size];
      memset(data, 0, size*sizeof(int));
      zeroPos = data+offset;
    }
    this->kernel_ = new OLocalKernel (zeroPos, data, 1, size, managed);
  }

  virtual ~OLocalAllocator(){}

    /**
   * Creates a copy of the memory block managed by an allocator and returns 
   * a new allocator object that manages this copy.
   *
   * @return  allocator for the copied memory block
   */
  OLocalAllocator<T> clone (bool copy_values = false) const
  {
    assert (getKernel_() != NULL);
    int nelems = getKernel_()->nelems_;
    T* new_buf = new T[nelems];
    
    if (copy_values)
      memcpy(new_buf, getKernel_()->data_, nelems * sizeof(T));

    T* zeroPos = getKernel_()->zeroPos_;
    int offset = getKernel_()->data_ - zeroPos;
 
    OLocalAllocator<T> ret(offset, nelems, true, new_buf);
    return ret;
  }

    /**
   * Obtain pointer to the memory block managed by the allocator 
   * at a certain offset
   * NOTICE: get(0) will get the pointer to the zero position: zeroPos_
   * @param  offset  offset in the allocated block
   * @return         pointer to the memory block
   */
  T* get(int offset = 0) const 
  {
    T* ret;
    if (getKernel_() != NULL) {
      
      assert (offset >= (getKernel_()->data_ - getKernel_()->zeroPos_));
      //cout << offset << " " << getKernel_()->nelems_ << endl;
      assert (offset < getKernel_()->nelems_-(getKernel_()->zeroPos_-getKernel_()->data_));
      getKernel_()->refCtr_++;
      HTA_DEBUG(2, "SerialAlloc::get kernel=" << getKernel_() << " ctr=" << getKernel_()->refCtr_);
      T * zeroPos = getKernel_()->zeroPos_;
      ret = zeroPos + offset;
    } else 
      ret = NULL;
    return ret;
  }


  /** 
   * @return true if idx is a valid index in the managed by this allocation 
   */
  bool checkAddress_(int idx) const
  {
    assert (getKernel_() != NULL);
    int offset = (int)(getKernel_()->zeroPos_ - getKernel_()->data_);
    bool in = (idx >= -offset && idx < getKernel_()->nelems_-offset);
    if(!in){
      cout<<"idx: "<<idx<<" offset: "<<offset<<" nelems: "<<getKernel_()->nelems_<<endl;
    }
    return in;
  }


};
#endif /* __O_Local_ALLOCATOR_H__ */
