/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: SerialHelper.h,v 1.25.4.1 2007/08/13 16:47:15 brodman2 Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef __SERIAL_HELPER_H__
#define __SERIAL_HELPER_H__

#include <cassert>

#include "ConstDistribution.h"
#include "Distribution.h"
#include "LocalMemMapping.h"
#include "LeafMemMapping.h"
#include "Order.h"
#include "Tuple.h"
#include "debug/Timers.h"
#include "serial/Serial.h"
#include "util/VectorOps.h"
#include "Overlap.h"
/**
 * template class that implements functionality common to the 
 * serial leaf and non-leaf HTAs. All methods must follow a precise signature that 
 * is mandated by the callers of this template.
 */
template<typename TRAIT>
struct SerialHelper 
{

 /**
  * Allocates contiguous memory.
  *
  * @param  linear_size  total number of elements
  * @param  dist         number of places across which this aloc is spread
  * @return              allocator for local memory
  */
  static typename TRAIT::TypeAllocator getAllocator_ (int linear_size, 
						      int block_size,
						      typename TRAIT::TypeScalar* data = NULL) 
  {
    assert (linear_size > 0);
    assert (linear_size == block_size);
    typename TRAIT::TypeAllocator ret = typename TRAIT::TypeAllocator(linear_size, (data != NULL) ? false : true, data);
    return ret;
  }

  static typename TRAIT::TypeAllocator getOAllocator_ (int offset,
						      int linear_size, 
						      int block_size,
						      typename TRAIT::TypeScalar* data = NULL) 
  {
    assert (linear_size > 0);
    assert (linear_size == block_size);
    typename TRAIT::TypeAllocator ret = typename TRAIT::TypeAllocator(offset, linear_size, (data != NULL) ? false : true, data);
    return ret;
  }

  /**
   * @param dist_level  ignored resp. irrelevant for the serial helper. This paranmeter
   *                    is just here for compatibility with a distributed implementation.
   */
  static MemMapping<TRAIT::dim>* getMemMapping_ (Order order, 
						 const typename Tuple<TRAIT::dim>::Seq& tiling,
						 size_t levels,size_t dist_levels,
						 const Distribution<TRAIT::dim>* dist = NULL) 
  {
    int nlevels = (order == TILE) ? 0 : levels;
    //int nlevels = levels;
  
    MemMapping<TRAIT::dim>* ret;
    //if (TRAIT::level == 0) 
    if (levels == 0) 
      ret = new LeafMemMapping<TRAIT::dim>(0, nlevels, tiling, Tuple<TRAIT::dim>::zero);
    else {
      //ret = new LocalMemMapping<TRAIT::dim>(TRAIT::level, 0, nlevels, tiling, order, Tuple<TRAIT::dim>::zero);
      //cout<<"check this allocation because is not right"<<endl;
      ret = new LocalMemMapping<TRAIT::dim>(levels, 0, nlevels, tiling, order, Tuple<TRAIT::dim>::zero);
    }
  
    return ret;
  }

 /**
   * This is adapted to the overlapped version.
   * A few things to notice: if only leaf level HTA is constructed, 
   * for ROW layout, a level-1 leaf size equal to whole region for the current
   * level and true region for the leaf level. However, if it is tiled or distributed, 
   * the leaf size is the whole region for leaf level.
   * 
   */
  static MemMapping<TRAIT::dim>* getOMemMapping_ (const Overlap<TRAIT::dim> *ol,
						  Order order, 
						 const typename Tuple<TRAIT::dim>::Seq& tiling,
						 size_t levels,size_t dist_levels,
						 const Distribution<TRAIT::dim>* dist = NULL) 
  {

    int nlevels = (order == TILE) ? 0 : levels;
    int d = TRAIT::dim;

    Tuple<TRAIT::dim> neg = Tuple<TRAIT::dim>::zero;
    Tuple<TRAIT::dim> pos = Tuple<TRAIT::dim>::zero;

    typename Tuple<TRAIT::dim>::Seq t = tiling;


    Tuple<TRAIT::dim> leaf_size=tiling[0];
    if(ol){
      neg = ol->neg();
      pos = ol->pos();
    }
    
    if(order == TILE){
     
      leaf_size = leaf_size + pos + neg;
      t[0]=t[0]+pos+neg;       
    }else{ //ROW
      for (int i = 1; i <= levels; i++){ 
	leaf_size = leaf_size * tiling[i];
      }
      leaf_size = leaf_size + neg + pos;
    }

    MemMapping<TRAIT::dim>* ret;
    //if (TRAIT::level == 0) 
    if (levels == 0) 
      ret = new LeafMemMapping<TRAIT::dim>(0, nlevels, t, Tuple<TRAIT::dim>::zero);
    else {
      //ret = new LocalMemMapping<TRAIT::dim>(TRAIT::level, 0, nlevels, tiling, order, Tuple<TRAIT::dim>::zero);
      //cout<<"check this allocation because is not right"<<endl;
      ret = new LocalMemMapping<TRAIT::dim>(levels, 0, nlevels, t, leaf_size, order, Tuple<TRAIT::dim>::zero);
    }
    return ret;
  }


  static const Distribution<TRAIT::dim>* getDistribution_() 
  {
    return &ConstDistribution<TRAIT::dim>::get(Serial::myThread);
  }

};

#endif /* __SERIAL_HELPER_H__ */
