/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: HTASparseImpl.h,v 1.22 2007/06/26 10:41:07 fraguela Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef __HTA_SPARSE_IMPL_H__
#define __HTA_SPARSE_IMPL_H__

#include <cassert>
#include <iostream>
#include <functional>

#include "serial/HTAImpl.h"
//#include "serial/LeafHTAImpl.h"
#include "MemMapping.h"
#include "Shape.h"
#include "Traits.h"
#include "util/VectorOps.h"
#include "CompSparseCol.h"
#include "CompSparseRow.h"


using namespace std;

/**
 * Core implementation class for Sparse HTA instance with LEVEL > 0
 * The reason to call HTASparseImpl as opposed to SparseHTAImpl is 
 * to prefix all the HTA related classes with 'HTA'.
 */
template<class T, int DIM, typename TRAIT>
class HTASparseImpl : public HTAImpl<T, DIM, TRAIT>
{
  
public:
  
  /** 
   * Constructor.
   *
   * @param data   must be unaliased and heap allocated - the HTA object will take care of 
   *               de-allocating the memory
   */
  HTASparseImpl (size_t level, 
		 typename TRAIT::TypeTileImpl** data, 
		 const Shape<DIM>& s, 
		 const typename TRAIT::TypeAllocator& alloc,
		 const MemMapping<DIM>* dist, 
		 int home) :
    HTAImpl<T, DIM, TRAIT>(level, data, s, alloc, dist, home,true)
  {
    assert (typeid(T) == typeid(double)); // current limitation
    assert (dist);

    HTA_DEBUG(1, "HTASparseImpl<LEVEL=0>::init home=" << home << " dist->offset=" << dist->offset());
    // check if the memory provided by the allocator tolerates access 
    // to the whole range of indices of the shape
    int ord_low = this->shape_.ord(this->shape_.low());
    int ord_high = this->shape_.ord(this->shape_.high());

#ifndef NDEBUG
    // don't call isLocal() at this point, because the object is not entirely 
    // initialized 
    if (home == Traits::Default::myPlace()) { 
      assert (this->alloc_.checkAddress_(dist->offset() + ord_low));
      assert (this->alloc_.checkAddress_(dist->offset() + ord_high));
    }
#endif
  }

  virtual ~HTASparseImpl() 
  {
    //if(this->isLeaf())
    //this->alloc_.recycle();
  }


  
  /**
   * matmul with dense HTAs
   */
  static void lmatmul_leaf_ (const Array<T, DIM, TRAIT>& larg, 
			     const Array<double, DIM, typename TRAIT::TraitDense> & rarg, 
			     Array<double, DIM,typename TRAIT::TraitDense>& dst) 
  { 
    assert (DIM == 2);
    assert (larg.shape().size()[1] == rarg.shape().size()[0]);
    assert (dst.shape().size()[0] == larg.shape().size()[0]);
    assert (dst.shape().size()[1] == rarg.shape().size()[1]);

    // communicate if necessary
    //larg.actual_().communicate_(dst.dist());
    //rarg.actual_().communicate_(dst.dist());

    // owner computes
    assert (dst.isLocal()) ;
    assert (larg.isAvailable_());
    assert (rarg.isAvailable_());
    
    // current limitations...
    assert (larg.memMap().isContiguous());
    assert (rarg.memMap().isContiguous());
    assert (dst.memMap().isContiguous());
    
    // internals of the matrix structure
    typename TRAIT::TypeInternal mat = larg.actual_().raw();
    int M = larg.shape().size()[0];   
    double* val = mat.values();
    // cast away const - intel MKL wants it like this...
    int* idx = const_cast<int*>(mat.idx());
    int* ptr = const_cast<int*>(mat.ptr());
    // internals of the vector
    double* b = rarg.raw();
    // internals of the result
    double* result = dst.raw();
    
    if (typeid(typename TRAIT::TypeInternal) == typeid(CompSparseCol<T, DIM>)) {
      int N = 1;
      int K = larg.shape().size()[1];
      HTA_DEBUG(1, "LeafHTASparseImpl::lmatmul - dcscmm M=" << M << " N=" << N << " K=" << K);
      double alpha = 1.0;
      double beta = 1.0;
      // temporaries 
      char descra[] = {'G', 0, 0, 0, 0, 0};
      char transa[] = {'N'};
      TRAIT::TypeMathKernel::dcscmm(transa, &M, &N, &K, &alpha, descra, val, idx, ptr, ptr + 1, b , &K, &beta, result + 1, &M);	
    } else if (typeid(typename TRAIT::TypeInternal) == typeid(CompSparseRow<T, DIM>)) {
      for (int i = 0; i < M; i++) {
	T sum = 0.0;
	  for (int j = ptr[i]; j < ptr[i+1]; j++) {
	    sum = sum + val[j] * b[idx[j]];
	  }
	  result[i] = sum;
      }
    } else 
      assert (false);  
  
    // invalidate if necessary
    //  larg.actual_().invalidate_(); 
    // rarg.actual_().invalidate_(); 
  }

  
  /**
   * matvecmul with DenseHTAs
   */
  static void lmatmul (const HTAImpl<T, DIM, TRAIT>& larg, 
		       const HTAImpl<double, DIM, typename TRAIT::TraitDense>& rarg, 
		       HTAImpl<double, DIM, typename TRAIT::TraitDense>& dst) 
  {

    //cout << "Here " << endl;
    if(larg.isLeaf()){
      lmatmul_leaf_(larg,rarg,dst);
    }
    else{
      assert (larg.shape() == rarg.shape());
      assert (larg.shape() == dst.shape());
      
      //cout << "Here " << rarg.numTiles_ << endl;

      for (int i  = 0; i < dst.numTiles_; i++) {
	
	int j = dst.localTiles_[i];
	TRAIT::TypeTileImpl::lmatmul (*larg.tileAt(j), *rarg.tileAt(j), *dst.tileAt(j));
      }
      
    }
  }

  /** 
   * scalarAt operation using a Tuple 
   * returns a scalar
   */
  T& scalarAt_leaf_ (const Tuple<DIM>& t) const
  {
#ifndef NDEBUG
    assert (this->isAvailable_());
    assert (this->shape_.contains(t)); // bounds check;
    int ord = this->shape_.ord(t);
    assert (!this->isLocal() ||        // bounds check with allocator only for local tiles
	    this->alloc_.checkAddress_(this->memMap().offset() + ord));
#endif
    return data_leaf_(t[0], t[1]);
  }


  
  inline T& scalarAt (const Tuple<DIM>& t) const
  {
    if(this->isLeaf()){
      return scalarAt_leaf_(t);
    }
    else {
      assert (this->shape().low() == Tuple<DIM>::zero);
      assert (this->shape().step() == Tuple<DIM>::one);
      Order ord = this->memMap().order();
      
      switch (ord) {
      case DIST:
      case TILE: {
	// determine the tile where t falls in
	Tuple<DIM> linear_size_scalarAt = this->memMap().leafSize();
	Tuple<DIM> tile_idx, rest_idx;
	for (int i = 0; i < DIM; ++ i) {
	  int size = linear_size_scalarAt [i] / this->shape().size()[i];
	  tile_idx[i]= t[i] / size;
	  rest_idx[i]= t[i] % size;
	}
	// then call scalarAt recursively on that tile
	HTA_DEBUG (2, "HTAImpl::scalarAt t=" << t << "falls into tile " << tile_idx);
	typename TRAIT::TypeTileImpl* tile = tileAt(tile_idx);
	return tile->scalarAt(rest_idx);
      }
      case ROW:
      case COL: {
	assert(false);
      }
      }
    }
  }

  void initSparse (T* values, const int* ptr, const int* idx) 
  {
    this->data_leaf_ = typename TRAIT::TypeSparseInternal(values, ptr, idx);
  }

  virtual typename TRAIT::TypeInternal raw() const 
  {
    assert(this->level() == 0);
    return this->data_leaf_;
  }

  virtual typename TRAIT::TypeInternal& internal_()
  {
    assert(this->level() == 0);
    return this->data_leaf_;
  }

  typename TRAIT::TypeDenseInternal& internalDense_() //__attribute__((always_inline))
  {
    assert (false);
  }

  typename TRAIT::TypeSparseInternal& internalSparse_() //__attribute__((always_inline))
  {
    assert(this->level() == 0);
    return this->data_leaf_;
  }


  void mpiIssueRequests (const Distribution<TRAIT::dim>& dst) const
  {
    // not implemented
    assert (false);
  }
  
  inline void communicate_ (const Distribution<DIM>& target_dist, 
			    Array<T, DIM, TRAIT>* target_arr = NULL) const
  {
    assert (target_dist.place() == this->home());
    assert (target_arr == NULL); // current limitation
    // the case where real communication should occur is not implemented
  }

  template <typename T2, int LEVEL2, typename TRAIT2>
  inline void communicate_ (Array<T2, DIM, TRAIT2>& target) const
  {
    int sender = this->home();
    int receiver = target.home();
    assert (sender == receiver); 
    // the case where real communication should occur is not implemented
  } 
   
  void invalidate_ () const 
  {
    // as nothing is communicated, nothing to be done here ...
  }

  private:

  /** 
   * This accessor allows other leaf-instances to efficiently access the current 
   * instance. 
   *
   * @see map
   */
  T& scalarAt_(int idx) const 
  {
    assert (false);
    return TypeHelper<T>::nullVal();
  }

  T& scalarAt_ (const typename Shape<DIM>::iterator& it) const 
  {
    assert (false);
    return TypeHelper<T>::nullVal();
  }

 public:

  /** 
   * scalarAt operation using a Shape
   * 
   * @return   HTA of the same level as this HTA which is a 'view' 
   *           restricted to the specified tiles
   */
  inline typename TRAIT::TypeImpl* scalarAt (const Shape<DIM>& s) const
  { 
    assert (false);
    return NULL;
  }

  /*
  inline typename TRAIT::TypeImpl* tileAt (const Shape<DIM>& s) const 
  {
    return scalarAt(s);
  }
  
  inline typename TRAIT::TypeImpl* tileAt (int n, const Shape<DIM>* sv) const 
  {
    // this method is only available for HTA's with LEVEL > 0
    // it is implemented here only to satisfy a compile-time dependency.
    assert (false); 
    return NULL;
  }

  inline typename TRAIT::TypeTileImpl* tileAt (int tile_idx) const
  {
    assert (false);
    return NULL;
  }
  */



  virtual bool isAvailable_() const 
  {
    return this->data_leaf_ != typename TRAIT::TypeInternal();
  }


  /** reduction over all dimensions */
  template <typename OP> T 
  reduce (const OP& binary_op) const
  {
    assert (false); // not yet implemented
    return TypeHelper<T>::nullVal();
  }
  
  template <typename OP> 
  void reduce (const OP& binary_op, 
	       int dim,
	       typename TRAIT::TypeImpl& dst) const
  {
    // compiler asks for this method because another method wit the same name 
    // is defined in this class and shadown the inherited variant of this method.
    // call base-class version of method
    Array<T, DIM, TRAIT>::reduce(binary_op, dim, dst);
  }
  
  template <typename OP>
  static void map (const OP& nullary_op, 
		       Array<T, DIM, TRAIT>& dst)
  {
    assert (false);
  }
  
  /** unary map */
  template <typename OP>
  static void map (const OP& unary_op, 
		       const Array<T, DIM, TRAIT>& arg, 
		       Array<T, DIM, TRAIT>& dst)
  {
    assert (false);
  }

  /** binary map */
  template <typename OP>
  static void map (const OP& binary_op, 
		       const Array<T, DIM, TRAIT>& larg, 
		       const Array<T, DIM, TRAIT>& rarg, 
		       Array<T, DIM, TRAIT>& dst) 
  {
    assert (false);
  }

  virtual void type() const {
    cout << "LeafSparseHTAImpl " << endl;
  }


};  // class HTASparseImpl





#endif /* __HTA_SPARSE_IMPL_H__ */
