/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: LeafHTASparseImpl.h,v 1.32 2006/06/06 19:55:01 bikshand Exp $
 * Authors: Ganesh Bikshandi
 */

#ifndef __LEAF_HTA_SPARSE_IMPL_H__
#define __LEAF_HTA_SPARSE_IMPL_H__

#include <cassert>
#include <functional>
#include <iostream>
#include <typeinfo>

#include "Array.h"
#include "MemMapping.h"
#include "LeafMemMapping.h"
#include "Operators.h"
#include "Shape.h"
#include "Traits.h"
#include "sparse/CompSparseRow.h"
#include "sparse/CompSparseCol.h"
#include "sparse/HTASparseImpl.h"
#include "sparse/SparseTrait.h"


/**
 * Sparse Leaf HTA - this is a template specialization of class 
 * HTAImpl for LEVEL=0
 */

template<class T, int DIM, typename TRAIT>
class HTASparseImpl<T, DIM, 0, TRAIT> : public Array<T, DIM, 0, TRAIT>
{
  
public:
  
  /** 
   * Constructor - this constructor is only called from HTA::alloc
   * 
   * @param data  not used, just for compatibility with the constructor for 
   *              template instances with LEVEL > 0
   */
  HTASparseImpl (typename TRAIT::TypeTileImpl** data, 
	   const Shape<DIM>& shape, 
	   const typename TRAIT::TypeAllocator& alloc, 
	   const MemMapping<DIM>* dist, 
		 int home) :
    Array<T, DIM, 0, TRAIT> (shape, alloc, dist, home), 
    data_()
  {
    assert (typeid(T) == typeid(double)); // current limitation
    assert (dist);

    HTA_DEBUG(1, "HTASparseImpl<LEVEL=0>::init home=" << home << " dist->offset=" << dist->offset());
    // check if the memory provided by the allocator tolerates access 
    // to the whole range of indices of the shape
    int ord_low = this->shape_.ord(this->shape_.low());
    int ord_high = this->shape_.ord(this->shape_.high());

#ifndef NDEBUG
    // don't call isLocal() at this point, because the object is not entirely 
    // initialized 
    if (home == Traits::Default::myPlace()) { 
      assert (this->alloc_.checkAddress_(dist->offset() + ord_low));
      assert (this->alloc_.checkAddress_(dist->offset() + ord_high));
    }
#endif

  }

  /**
   * The destructor - uses reference counting
   * as the data_ could be shared among PrivateLeafHTAs
   */  
  virtual ~HTASparseImpl () 
  {
    this->alloc_.recycle();
  }
  
  /**
   * return a clone of myself
   */
  typename TRAIT::TypeImpl* clone_(const Tuple<DIM>* perm,
				   const typename TRAIT::TypeAllocator& alloc,
				   const MemMapping<DIM>& dist, 
				   bool copy_values) const
  {
    assert (false);
    return NULL;
  }
  
  /**
   * assignment operator: elementwise assignment 
   * this is a specialization of the default implementation of 
   * this operator in the Array superclass.
   */
  typename TRAIT::TypeImpl& operator= (const typename TRAIT::TypeImpl& rhs) 
  {
    assert (this->isLocal());
    if (this != &rhs) 
      map (Operator::idem<T>(), rhs, *this);
    return this->actual_();
  }
  
  typename TRAIT::TypeImpl& operator= (const T& rhs) 
  {
    map (Operator::cnst<T>(rhs), *this);
    return this->actual_();
  }

  // dotproduct ----------------------------------------------------------------------

  static T dotproduct (const Array<T, DIM, 0, TRAIT>& larg, 
		       const Array<T, DIM, 0, TRAIT>& rarg)
  {
    assert (larg.shape() == rarg.shape());
    T ret = 0;
    assert (false); // not yet implemented
    return ret;
  }
  
  // matmul ----------------------------------------------------------------------

  /*  static void matmul (const Array<T, DIM, 0, TRAIT>& larg, 
		      const Array<T, DIM, 0, TRAIT>& rarg, 
		      Array<T, DIM, 0, TRAIT>& dst) 
  {
    assert (DIM == 2);
    assert (larg.shape().size()[1] == rarg.shape().size()[0]);
    assert (dst.shape().size()[0] == larg.shape().size()[0]);
    assert (dst.shape().size()[1] == rarg.shape().size()[1]);

    int M = larg.shape().size()[0];
    int Q = larg.shape().size()[1];
    int N = rarg.shape().size()[1];
    
    assert (false); // not yet implemented
    }*/

  /**
   * matmul with dense HTAs
   */

  
  static void lmatmul (const Array<T, DIM, 0, TRAIT>& larg, 
		       const Array<double, DIM, 0, typename TRAIT::TraitDense> & rarg, 
		       Array<double, DIM, 0, typename TRAIT::TraitDense>& dst) 
  {
 
    assert (DIM == 2);
    assert (larg.shape().size()[1] == rarg.shape().size()[0]);
    assert (dst.shape().size()[0] == larg.shape().size()[0]);
    assert (dst.shape().size()[1] == rarg.shape().size()[1]);

    // communicate if necessary
    //larg.actual_().communicate_(dst.dist());
    //rarg.actual_().communicate_(dst.dist());

    // owner computes
    if (dst.isLocal()) {
      assert (larg.isAvailable_());
      assert (rarg.isAvailable_());
      
      // current limitations...
      assert (larg.memMap().isContiguous());
      assert (rarg.memMap().isContiguous());
      assert (dst.memMap().isContiguous());

      // internals of the matrix structure
      typename TRAIT::TypeInternal mat = larg.actual_().raw();
      int M = larg.shape().size()[0];
      int N = 1;
      int K = larg.shape().size()[1];
      double alpha = 1.0;
      double beta = 1.0;
      // temporaries 
      char descra[] = {'G', 0, 0, 0, 0, 0};
      char transa[] = {'N'};
      double* val = mat.values();
      // cast away const - intel MKL wants it like this...
      int* idx = const_cast<int*>(mat.idx());
      int* ptr = const_cast<int*>(mat.ptr());
      // internals of the vector
      double* b = rarg.raw();
      // internals of the result
      double* result = dst.raw();

      if (typeid(typename TRAIT::TypeInternal) == typeid(CompSparseCol<T, DIM>)) {
	HTA_DEBUG(1, "LeafHTASparseImpl::lmatmul - dcscmm M=" << M << " N=" << N << " K=" << K);
	TRAIT::TypeMathKernel::dcscmm(transa, &M, &N, &K, &alpha, descra, val, idx, ptr, ptr + 1, b , &K, &beta, result + 1, &M);
      } else if (typeid(typename TRAIT::TypeInternal) == typeid(CompSparseRow<T, DIM>)) {
	for (int i = 0; i < M; i++) {
	  T sum = 0.0;
	  for (int j = ptr[i]; j < ptr[i+1]; j++) {
	    sum = sum + val[j] * rarg.scalarAt_(idx[j]);
	  }
	  dst.scalarAt_(i) = sum;
	}
      } else 
	assert (false);
    }
    
    // invalidate if necessary
    larg.actual_().invalidate_(); 
    rarg.actual_().invalidate_(); 
    } 
  
  /** 
   * scalarAt operation using a Tuple 
   * returns a scalar
   */
  T& scalarAt (const Tuple<DIM>& t) const
  {
#ifndef NDEBUG
    assert (this->isAvailable_());
    assert (this->shape_.contains(t)); // bounds check;
    int ord = this->shape_.ord(t);
    assert (!this->isLocal() ||        // bounds check with allocator only for local tiles
	    this->alloc_.checkAddress_(this->memMap().offset() + ord));
#endif
    return data_(t[0], t[1]);
  }
  
  typename TRAIT::TypeImpl& operator= (const typename TRAIT::TypeDenseInternal& rhs) 
  {
    assert (rhs != NULL);
    assert (false); // not implemented
    return *this;
  }

  typename TRAIT::TypeImpl& operator= (const typename TRAIT::TypeSparseInternal& rhs) 
  {
    assert (rhs != NULL);
    *data_ = *rhs;
    return *this;
  }
 
  void initSparse (T* values, const int* ptr, const int* idx) 
  {
    data_ = typename TRAIT::TypeSparseInternal(values, ptr, idx);
  }

  virtual typename TRAIT::TypeInternal raw() const 
  {
    return data_;
  }

  virtual typename TRAIT::TypeInternal& internal_()
  {
    return data_;
  }

  typename TRAIT::TypeDenseInternal& internalDense_() //__attribute__((always_inline))
  {
    assert (false);
  }

  typename TRAIT::TypeSparseInternal& internalSparse_() //__attribute__((always_inline))
  {
    return data_;
  }

  void mpiIssueRequests (const Distribution<TRAIT::dim>& dst) const
  {
    // not implemented
    assert (false);
  }

  inline void communicate_ (const Distribution<DIM>& target_dist, 
			    Array<T, DIM, 0, TRAIT>* target_arr = NULL) const
  {
    assert (target_dist.place() == this->home());
    assert (target_arr == NULL); // current limitation
    
    if (target_dist.place() != Traits::Default::myPlace())
      assert (false); // not implemented yet
    /*    assert (sender >= 0 && sender < MPI::nThreads());
    assert (receiver >= 0 && receiver < MPI::nThreads());

    int this_rank = MPI::myThread();
    int dim[2];
    int nz;
    int size = sizeof(CompCol_Mat_double);
    MPI_Datatype type = MPI_INT;

    if (this_rank == sender) { 
      // send ---------------------------------------

      // first communicate the structure itself ...
      dim[0] = data->dim(0);
      dim[1] = data->dim(1);
      nz = data->NumNonzeros();
      MPI_Send (dim, 2, type, receiver, 0, MPI_COMM_WORLD); 
      MPI_Send (&nz, 1, type, receiver, 0, MPI_COMM_WORLD); 

      // then communicate values
      nz = data->NumNonzeros();
      type = MPI_DOUBLE;
      MPI_Send (&data->val(0), nz, type, receiver, 0, MPI_COMM_WORLD); 

      // then communicate the row indices
      type = MPI::type<int>();
      MPI_Send (&data->row_ind(0), nz, type, receiver, 0, MPI_COMM_WORLD); 

      // then communicate the column pointers
      size = data->dim(1) + 1;
      MPI_Send (&data->col_ptr(0), size, type, receiver, 0, MPI_COMM_WORLD); 
    } else if (this_rank == receiver) { 
      // receive ------------------------------------
      assert (data == NULL);

      // first communicate the structure itself ...
      MPI_Status status;
      MPI_Recv (dim, 2, type, sender, 0, MPI_COMM_WORLD, &status);   
      MPI_Recv (&nz, 1, type, sender, 0, MPI_COMM_WORLD, &status);

      // then communicate values
      double* vals = new double[nz];
      type = MPI::type<double>(); MPI_Recv (vals, nz, type, sender, 0, MPI_COMM_WORLD, &status);

      // then communicate the row indices
      int* rowind = new int[nz];
      type = MPI::type<int>();
      MPI_Recv (rowind, nz, type, sender, 0, MPI_COMM_WORLD, &status);   

      // then communicate the column pointers
      size = dim[1] + 1;
      int* colptr = new int[size];
      MPI_Recv (colptr, size, type, sender, 0, MPI_COMM_WORLD, &status);   

      // then allocate new sparse representation
      //data = new CompCol_Mat_double (dim[0], dim[1], nz, vals, rowind, colptr); 
      return true;
    } else
    */
  }

  template <typename T2, int LEVEL2, typename TRAIT2>
  inline void communicate_ (Array<T2, DIM, LEVEL2, TRAIT2>& target) const
  {
    int sender = this->home();
    int receiver = target.home();
    assert (sender == receiver); 
    // the case where real communication should occur is not implemented
  } 

  void invalidate_ () const 
  {
    // as nothing is communicated, nothing to be done here ...
  }

private:

  /** 
   * This accessor allows other leaf-instances to efficiently access the current 
   * instance. 
   *
   * @see map
   */
  T& scalarAt_(int idx) const 
  {
    assert (false);
    return dummy_;
  }

  T& scalarAt_ (const typename Shape<DIM>::iterator& it) const 
  {
    assert (false);
    return dummy_;
  }

  static T dummy_;

public:

  /** 
   * scalarAt operation using a Shape
   * 
   * @return   HTA of the same level as this HTA which is a 'view' 
   *           restricted to the specified tiles
   */
  inline typename TRAIT::TypeImpl* scalarAt (const Shape<DIM>& s) const
  { 
    assert (false);
    return NULL;
  }

  inline typename TRAIT::TypeImpl* tileAt (const Shape<DIM>& s) const 
  {
    return scalarAt(s);
  }
  
  inline typename TRAIT::TypeImpl* tileAt (int n, const Shape<DIM>* sv) const 
  {
    // this method is only available for HTA's with LEVEL > 0
    // it is implemented here only to satisfy a compile-time dependency.
    assert (false); 
    return NULL;
  }

  inline typename TRAIT::TypeTileImpl* tileAt (int tile_idx) const
  {
    assert (false);
    return NULL;
  }

  /** reduction over all dimensions */
  template <typename OP> T 
  reduce (const OP& binary_op) const
  {
    assert (false); // not yet implemented
    return TypeHelper<T>::nullVal();
  }
  
  template <typename OP> 
  void reduce (const OP& binary_op, 
	       int dim,
	       typename TRAIT::TypeImpl& dst) const
  {
    // compiler asks for this method because another method wit the same name 
    // is defined in this class and shadown the inherited variant of this method.
    // call base-class version of method
    Array<T, DIM, 0, TRAIT>::reduce(binary_op, dim, dst);
  }
  
  template <typename OP>
  static void map (const OP& nullary_op, 
		       Array<T, DIM, 0, TRAIT>& dst)
  {
    assert (false);
  }
  
  /** unary map */
  template <typename OP>
  static void map (const OP& unary_op, 
		       const Array<T, DIM, 0, TRAIT>& arg, 
		       Array<T, DIM, 0, TRAIT>& dst)
  {
    assert (false);
  }

  /** binary map */
  template <typename OP>
  static void map (const OP& binary_op, 
		       const Array<T, DIM, 0, TRAIT>& larg, 
		       const Array<T, DIM, 0, TRAIT>& rarg, 
		       Array<T, DIM, 0, TRAIT>& dst) 
  {
    assert (false);
  }

  virtual void type() const {
    cout << "LeafSparseHTAImpl " << endl;
  }

protected:

  // all other template instance of this class should be friends (required when accessing print_)
  template<class T2, int DIM2, int LEVEL2, typename TRAIT2>
  friend class HTASparseImpl;

  template<class T2, int DIM2, int LEVEL2, typename TRAIT2>
  friend class HTAImpl;

  // to allow forwarding class to private methods from class Array to this class.
  template<class T2, int DIM2, int LEVEL2, typename TRAIT2>
  friend class Array;
  
  /**
   * @return  true if data is accessible locally. The tile might either be local
   *          or it might be cached fro a remote location 
   */
  virtual bool isAvailable_() const 
  {
    return data_ != typename TRAIT::TypeInternal();
  }

  /** reduction over one dimension */
  template <typename OP> 
  void reduce_ (const OP& binary_op, 
		int dim, 
		typename TRAIT::TypeImpl& dst) const
  {
    assert (false);
  }

  /** print */   
  virtual void print_ (ostream& ost, int indent = 0) const 
  {
    assert (isAvailable_());
    
    for (int j = 0; j < indent; ++j) 
      ost << " ";
     if (DIM == 2) {
      int printed = 0;
      int mod = this->shape_[1].size(); // column dimension has index 1
      for (typename Shape<DIM>::iterator i = this->shape_.begin(); i != this->shape_.end(); i++) {
	ost << scalarAt_(i) << "\t";
	if (++printed % mod == 0) {
	  ost << endl;
	  for (int j = 0; j < indent; ++j) ost << " ";
	}
      }
    } else {
      for (typename Shape<DIM>::iterator i = this->shape_.begin(); i != this->shape_.end(); i++)    
	ost << scalarAt(*i) << " "; 
      ost << endl;
    }
  }    

  /* sparse data */
  typename TRAIT::TypeInternal data_;

private:

  // specialization of nullary map for DIM == 1. 
  template <typename OP> 
  static void map1D_ (const OP& nullary_op, 
			  Array<T, DIM, 0, TRAIT>& dst) 
  {
    assert (DIM == 1);

    int size = dst.shape().size()[0];
    int step_dst = dst.shape().linearStep()[0];
    
    for (int i = 0; i < size; ++ i) 
      dst.scalarAt_(i * step_dst) = nullary_op();
  }

   // specialization of nullary map for DIM == 2. 
  template <typename OP> 
  static void map2D_ (const OP& nullary_op, 
			  Array<T, DIM, 0, TRAIT>& dst) 
  {
    assert (DIM == 2);

    int size_0 = dst.shape().size()[0];
    int step_dst_0 = dst.shape().linearStep()[0];

    int size_1 = dst.shape().size()[1];
    int step_dst_1 = dst.shape().linearStep()[1];
    
    for (int i = 0; i < size_1; ++ i) 
      for (int j = 0; j < size_0; ++ j) 
	dst.scalarAt_(i * step_dst_1 + j * step_dst_0) = nullary_op();
  }

  // specialization of nullary map for DIM == 3. 
  template <typename OP> 
  static void map3D_ (const OP& nullary_op, 
			  Array<T, DIM, 0, TRAIT>& dst) 
  {
    assert (DIM == 3);

    int size_0 = dst.shape().size()[0];
    int step_dst_0 = dst.shape().linearStep()[0];
    
    int size_1 = dst.shape().size()[1];
    int step_dst_1 = dst.shape().linearStep()[1];
    
    int size_2 = dst.shape().size()[2];
    int step_dst_2 = dst.shape().linearStep()[2];
    
    for (int i = 0; i < size_2; ++ i) 
      for (int j = 0; j < size_1; ++ j)
	for (int k = 0; k < size_0; ++ k)  
	  dst.scalarAt_(i * step_dst_2 + j * step_dst_1 + k * step_dst_0) = nullary_op();

  }

  // specialization of unary map for DIM == 1. 
  template <typename OP> 
  static void map1D_ (const OP& unary_op, 
			  const Array<T, DIM, 0, TRAIT>& arg, 
			  Array<T, DIM, 0, TRAIT>& dst) 
  {
    assert (DIM == 1);

    int size = arg.shape().size()[0];
    int step_arg = arg.shape().linearStep()[0];
    int step_dst = dst.shape().linearStep()[0];
    
    for (int i = 0; i < size; ++ i) 
      dst.scalarAt_(i * step_dst) = 
	unary_op(arg.scalarAt_(i * step_arg));
  }

  // specialization of unary map for DIM == 2. 
  template <typename OP> 
  static void map2D_ (const OP& unary_op, 
			  const Array<T, DIM, 0, TRAIT>& arg, 
			  Array<T, DIM, 0, TRAIT>& dst) 
  {
    assert (DIM == 2);

    int size_0 = arg.shape().size()[0];
    int step_arg_0 = arg.shape().linearStep()[0];
    int step_dst_0 = dst.shape().linearStep()[0];

    int size_1 = arg.shape().size()[1];
    int step_arg_1 = arg.shape().linearStep()[1];
    int step_dst_1 = dst.shape().linearStep()[1];
    
    for (int i = 0; i < size_1; ++ i) 
      for (int j = 0; j < size_0; ++ j) 
	dst.scalarAt_(i * step_dst_1 + j * step_dst_0) = 
	  unary_op(arg.scalarAt_(i * step_arg_1 + j * step_arg_0));
  }

  // specialization of unary map for DIM == 3. 
  template <typename OP> 
  static void map3D_ (const OP& unary_op, 
			  const Array<T, DIM, 0, TRAIT>& arg, 
			  Array<T, DIM, 0, TRAIT>& dst) 
  {
    assert (DIM == 3);

    int size_0 = arg.shape().size()[0];
    int step_arg_0 = arg.shape().linearStep()[0];
    int step_dst_0 = dst.shape().linearStep()[0];
    
    int size_1 = arg.shape().size()[1];
    int step_arg_1 = arg.shape().linearStep()[1];
    int step_dst_1 = dst.shape().linearStep()[1];
    
    int size_2 = arg.shape().size()[2];
    int step_arg_2 = arg.shape().linearStep()[2];
    int step_dst_2 = dst.shape().linearStep()[2];
    
    
    for (int i = 0; i < size_2; ++ i) 
      for (int j = 0; j < size_1; ++ j)
	for (int k = 0; k < size_0; ++ k)  
	  dst.scalarAt_(i * step_dst_2 + j * step_dst_1 + k * step_dst_0) = 
	    unary_op(arg.scalarAt_(i * step_arg_2 + j * step_arg_1 + k * step_arg_0));
  }

  // specialization of binary map for DIM == 1. 
  template<typename OP>
  static void map1D_ (const OP& binary_op, 
		   const Array<T, DIM, 0, TRAIT>& larg, 
		   const Array<T, DIM, 0, TRAIT>& rarg, 
		   Array<T, DIM, 0, TRAIT>& dst) 
  {
    assert(DIM == 1);

    int size = larg.shape().size()[0];
    int step_larg = larg.shape().linearStep()[0];
    int step_rarg = rarg.shape().linearStep()[0];
    int step_dst = dst.shape().linearStep()[0];

    for (int i = 0; i < size; ++ i) 
      dst.scalarAt_(i * step_dst) = 
	binary_op(larg.scalarAt_(i * step_larg), rarg.scalarAt_(i * step_rarg));
  } 

  // specialization of binary map for DIM == 2. 
  // iterate separately over each dimension
  template <typename OP> 
  static void map2D_ (const OP& binary_op, 
			  const Array<T, DIM, 0, TRAIT>& larg, 
			  const Array<T, DIM, 0, TRAIT>& rarg, 
			  Array<T, DIM, 0, TRAIT>& dst) 
  {
    assert (DIM == 2);
    
    int size_0 = larg.shape().size()[0];
    int step_larg_0 = larg.shape().linearStep()[0];
    int step_rarg_0 = rarg.shape().linearStep()[0];
    int step_dst_0 = dst.shape().linearStep()[0];
    
    int size_1 = larg.shape().size()[1];
    int step_larg_1 = larg.shape().linearStep()[1];
    int step_rarg_1 = rarg.shape().linearStep()[1];
    int step_dst_1 = dst.shape().linearStep()[1];

    for (int i = 0; i < size_1; ++ i) 
      for (int j = 0; j < size_0; ++ j) 
	dst.scalarAt_(i * step_dst_1 + j * step_dst_0) = 
	  binary_op(larg.scalarAt_(i * step_larg_1 + j * step_larg_0), 
		    rarg.scalarAt_(i * step_rarg_1 + j * step_rarg_0));
  }

  // specialization of binary map for DIM == 3. 
  // iterate separately over each dimension
  template <typename OP> 
  static void map3D_ (const OP& binary_op, 
			  const Array<T, DIM, 0, TRAIT>& larg, 
			  const Array<T, DIM, 0, TRAIT>& rarg, 
			  Array<T, DIM, 0, TRAIT>& dst) 
  {
    assert (DIM == 3);
    
    int size_0 = larg.shape().size()[0];
    int step_larg_0 = larg.shape().linearStep()[0];
    int step_rarg_0 = rarg.shape().linearStep()[0];
    int step_dst_0 = dst.shape().linearStep()[0];
    
    int size_1 = larg.shape().size()[1];
    int step_larg_1 = larg.shape().linearStep()[1];
    int step_rarg_1 = rarg.shape().linearStep()[1];
    int step_dst_1 = dst.shape().linearStep()[1];

    int size_2 = larg.shape().size()[2];
    int step_larg_2 = larg.shape().linearStep()[2];
    int step_rarg_2 = rarg.shape().linearStep()[2];
    int step_dst_2 = dst.shape().linearStep()[2];
    
    for (int i = 0; i < size_2; ++ i) 
      for (int j = 0; j < size_1; ++ j)  
	for (int k = 0; k < size_0; ++ k) 
	  dst.scalarAt_(i * step_dst_2 + j * step_dst_1 + k * step_dst_0) = 
	    binary_op(larg.scalarAt_(i * step_larg_2 + j * step_larg_1 + k * step_larg_0), 
		      rarg.scalarAt_(i * step_rarg_2 + j * step_rarg_1 + k * step_rarg_0));
  }
  
};

template<class T, int DIM, typename TRAIT>
T HTASparseImpl<T, DIM, 0, TRAIT>::dummy_;

#endif /* __LEAF_HTA_SPARSE_IMPL_H__ */
