/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: AbstractAllocator.h,v 1.14.4.1 2007/08/13 16:47:15 brodman2 Exp $
 * Authors: Christoph von Praun
 */

#ifndef  __ABSTRACT_ALLOCATOR_H__
#define  __ABSTRACT_ALLOCATOR_H__

#include <iostream>

#include "debug/Tracing.h"

using namespace std;

/**
 * Base class of allocators - implements common functionality like reference 
 * counting.
 */
template <typename T> 
class AbstractAllocator {
  
  template <typename T2>
  friend ostream& operator << (ostream& ost, const AbstractAllocator<T2>& a);

//protected:
public:
  /**
   * Kernel class. Instances are shared among Allocator instances.
   * Instances of this class maintain reference counts and hold the 
   * pointer to the memory block.
   */
  struct Kernel {

    Kernel(int nblocks, int nelems, bool managed) :
      refCtr_(0),
      allocCtr_(1),
      nblocks_(nblocks),
      nelems_(nelems), 
      managed_(managed)
    {
      assert (nelems > 0 && nblocks > 0);
    }
    
    virtual ~Kernel() 
    {
      assert (allocCtr_ == 0);
      if (refCtr_ != 0) 
	HTA_WARN(1, "AbstractAllocator::~Kernel - free memory although there might be stale references (" << refCtr_ << ")");
    }

    /** 
     * number of C++ pointers to data - this is purely informational and can be 
     * to debug stale pointers 
     */
    int refCtr_;
    /** number of allocator objects pointing to this kernel */
    int allocCtr_;
    /** blocks */
    const int nblocks_;    
    /** elements per block */
    const int nelems_;
    /** 
     * flag that specifies if the memory shall be deallocated using delete[] if the
     * kernel is deallocated 
     */
    const bool managed_;
  };
  
  Kernel* kernel_;

protected:

  /** 
   * Default constructor
   * (this is only necessary because in some contexts, Allocators may not be 
   *  intialized immediately and initialization may happen later through assignment 
   */
  AbstractAllocator() : kernel_(NULL) 
  { 
    //  the initialization of the kernel_ is the responsibility of the subclass 
  }

public:

  /**
   * Copy constructor
   */
  AbstractAllocator(const AbstractAllocator<T>& o) :
    kernel_(o.kernel_) 
  { 
    if (kernel_)
      kernel_->allocCtr_++;
  }

  /**
   * Destructor. Might de-allocate kernel and with it the memory block.
   */
  virtual ~AbstractAllocator() 
  { 
    if (kernel_ != NULL) {
      // all reference to the array have to be cleaned up with 'recycle'
      kernel_->allocCtr_--;
      if (kernel_->allocCtr_ == 0) 
	delete kernel_;
    }
  }

  /**
   * Assignment operator. The current kernel (memory block) is dropped and 
   * de-allocated (if necessary). Then, the kernel of the rhs allocator is 
   * assigned to this allocator.
   */
  AbstractAllocator<T>& operator=(const AbstractAllocator<T>& rhs) 
  {
    if (this->kernel_ != rhs.kernel_) {
      if (kernel_ != NULL) {
	// drop the old kernel and delete it if necessary
	kernel_->allocCtr_--;
	if (kernel_->allocCtr_ == 0) 
	  delete kernel_;
      }
      
      // install the new kernel
      kernel_ = rhs.kernel_;
      if (kernel_ != NULL)
	kernel_->allocCtr_++;
    }

    return *this;
  }

  /**
   * Obtain pointer to the memory block managed by the allocator 
   * at a certain offset
   *
   * @param  offset  offset in the allocated block
   * @return         pointer to the memory block
   */
  virtual T* get(int offset = 0) const = 0;
  
  /**
   * @return number of blocks 
   */
  inline int nblocks() const
  {
    return (kernel_ != NULL) ? kernel_->nblocks_ : 0;
  } 

  /**
   * @return number of elements per block
   */
  inline int nelems() const
  {
    return (kernel_ != NULL) ? kernel_->nelems_ : 0;
  } 

  
  /**
   * @return the number of variables in the array that is managed by this allocator 
   */
  inline int size() const
  {
    return (kernel_ != NULL) ? kernel_->nelems_ *  kernel_->nblocks_ : 0;
  }

  /**
   * Specifies that a pointer, previously returned by get() is 
   * not used any longer. This is purely informational - 
   * it should help to debug memory allocation problems / stale pointers etc.
   *
   * @see ~Kernel()
   */
  void recycle() const 
  {
    if (kernel_ != NULL)
      -- (kernel_->refCtr_);
  }

  /** 
   * subclasses must implement this.
   *
   * @return true if idx is a valid offset (in size(T) steps) 
   *         in the buffer managed by this allocation 
   */
  virtual bool checkAddress_(int idx) const = 0;

};

template<typename T>
ostream& operator << (ostream& ost, const AbstractAllocator<T>& a) 
{
  a.print_(ost);
  return ost;
}

#endif  /* __ABSTRACT_ALLOCATOR_H__ */
