/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Test_HTA_ft3.cpp,v 1.2 2006/05/02 15:44:41 vonpraun Exp $
 * Authors: Ganesh Bikshandi
 */

#include <iostream>
#include <cmath>
#include <complex>
#include <stdlib.h>

#include "htalib_serial.h"
#include "mkl_dfti.h"

#define TESTNAME "Test_HTA_ft"
#ifdef NDEBUG
#warning Testing without assertions!
#endif

typedef HTA<complex<double>, 2, 2> HTA_2H;
typedef HTA<complex<double>, 2, 1> HTA_1H;
typedef HTA<complex<double>, 2, 0> HTA_0H;

using namespace std;


#define NX  4 
#define NY  4 

void ft_hta (complex<double> a[NX][NY], complex<double> b[NX][NY]);
void ft_array (complex<double> a[NX][NY]);
void print_array (const complex<double> a[NX][NY]);
void print_hta (const HTA_2H& hta, int nx, int ny);

int main( int argc, char** argv) {
  
  complex<double> in[NX][NY];
  // this Array holds the initialization data
  complex<double> init[NX][NY];

  // (0) initialize array
  for (int i = 0; i < NX; i++) {
    for (int j = 0; j < NY; j++) {
      in[i][j] = init[i][j]= complex<double>( ((double) rand()) / (double) RAND_MAX, ((double) rand()) / (double) RAND_MAX);
      }
    }
  
  ft_array(in);

  cout << "END fft array" << endl;
  cout << "=============================================================================" << endl;
  cout << "START fft hta" << endl;
  
  ft_hta(in, init);

  cout << TESTNAME << " PASSED" << endl;
}



void ft_array (complex<double> in[NX][NY]) 
{
  long status;
  DFTI_DESCRIPTOR_HANDLE handle;

  // (1) print initialized array
  cout << "Array before fft " << endl;
  print_array(in);

  /* FT along dim = 1 */
  status = DftiCreateDescriptor(&handle, DFTI_DOUBLE, DFTI_COMPLEX, 1, NY);

  // not necessary, but does not hurt ...
  int stride[2];
  stride[0] = 0;
  stride[1] = &(in[0][1]) - &(in[0][0]);
  cout << "Array fft-1 stride " << stride[1] << endl;
  status = DftiSetValue(handle, DFTI_INPUT_STRIDES, stride);
  status = DftiSetValue(handle, DFTI_OUTPUT_STRIDES, stride);

  status = DftiCommitDescriptor(handle);
  for (int i = 0; i < NX; i++) {
      cout << "DftiComputeForward tuple=" << Tuple<2>(i,0) << " addr=" << &in[i][0] << endl;
      status = DftiComputeForward(handle, &(in[i][0]));
  }
  
  // (4) print array after forward FFT on dim-1
  cout << "Array after fft-1 " << endl;
  print_array(in);


}

/** compute FFT on the HTA and compare the result to the FFT computed on the plain array */
void ft_hta (complex<double> a[NX][NY], complex<double> b[NX][NY]) 
{
  
  // create data set
  HTA_2H h = HTA_2H::alloc((Tuple<2>(NX/2, NY/2), Tuple<2>(2, 1), Tuple<2>(1,2)), NULL, ROW); // breaks
  // HTA_2H h = HTA_2H::alloc((Tuple<2>(NX, NY), Tuple<2>(1, 1), Tuple<2>(1, 1)), NULL, ROW); // works
  
  // (0) initialize
  for (int i = 0; i < NX; i++) {
    for (int j = 0; j < NY; j++) {
	h[Tuple<2>(i,j)] = b[i][j];
      }
  }
	
  // (1) print the initialized hta 
  cout << "HTA before ft " << endl;
  print_hta(h, NX, NY);

  // breaks because dimention 1 is tiled at the top-level
  // and fft must be done along a dim that is not tiled at the 
  // top-level!
  // h.ft(1, 1); 

  HTA_2H h_perm = h.dpermute(Tuple<2>(1,0), 1);
  cout << "HTA after dpermute " << endl;
  print_hta(h_perm, NY, NX);
  
  h_perm.ft(0, 1); 
  h = h_perm.dpermute(Tuple<2>(1, 0), 1);

  // (4) print after fft-1
  // cout << "HTA after FFT dim-1 " << endl;
  print_hta(h, NX, NY);

  // (5) validate
  for (int i = 0; i < NX; i++) {
    for (int j = 0; j < NY; j++) {
	assert (real(h[Tuple<2>(i,j)])- real(a[i][j]) < 1.0e-6);
	assert (imag(h[Tuple<2>(i,j)])- imag(a[i][j]) < 1.0e-6);
      }
  }
}

void print_array (const complex<double> in[NX][NY]) 
{
  for (int i = 0; i < NX; i++) {
    for (int j = 0; j < NY; j++) {
      cout << in[i][j] << "\t";
      }
      cout << endl;
  }
}

void print_hta (const HTA_2H& h, int nx, int ny) 
{
  for (int i = 0; i < nx; i++) {
    for (int j = 0; j < ny; j++) {
	cout << h[Tuple<2>(i,j)] << "\t" ;
      }
      cout << endl;
  }
}
