/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Test_HTA_matmul.cpp,v 1.12.2.1 2007/08/14 18:19:40 brodman2 Exp $
 * Authors: Ganesh Bikshandi
 */

#include <iostream>
#include <cmath>

#include "htalib_serial.h"

#define TESTNAME "Test_HTA_matmul"
#ifdef NDEBUG
#warning Testing without assertions!
#endif

typedef HTA<double, 2, Traits::SerialDense<double, 2> > HTA_0H;
typedef HTA<double, 2, Traits::SerialDense<double, 2> > HTA_1H;

using namespace std;

void matmul (const double& A, const double& B, double& C) 
{
  C += A * B;
}

void matmul (HTA<double, 2> A, 
	     HTA<double, 2> B, 
	     HTA<double, 2> C) 
{
  int M = A.shape()[0].size();
  int N = B.shape()[0].size();
  int Q = B.shape()[1].size();

  for (int i = 0; i< M; i++) {
    for (int k = 0; k < Q; k++) {
      for (int j = 0; j< N; j++) {
	matmul (A [ Tuple<2>(i,k) ], 
		B [ Tuple<2>(k,j) ], 
		C [ Tuple<2>(i,j) ]);
      }
    }
  }
}

void level_0() 
{
  cout << "level 0 start -----------------------------------" << endl;
  
  // allocate HTA
  HTA_0H a =  HTA_0H::alloc (0,Tuple<2>(4,4));
  HTA_0H b =  HTA_0H::alloc (0,Tuple<2>(4,4));
  HTA_0H c =  HTA_0H::alloc (0,Tuple<2>(4,4));

  // initialize array
  double* a_x = a.raw();
  double* b_x = b.raw();
  for(int i=0; i < 16; i++) {
    a_x[i] = i;
    b_x[i] = i;
  }
  
  // c is initiaized with 0
  
  cout << "a=" << endl << a << endl;
  cout << "b=" << endl << b << endl;

  cout << "-----------------------------------" << endl;
  
  // test of simple code above ...
  matmul (a, b, c);
  
  cout << "c (user matmul) =" << endl << c << endl;
  
  c = 0.0;

  cout << "c (zero'ed) =" << endl << c << endl;

  /* test of library version ... */
  HTA_0H::matmul (a, b, &c);
  
  cout << "c (library matmul) =" << endl << c << endl;

  cout << "level 0 end -----------------------------------" << endl;
}

void level_1() 
{
  cout << "level 1 start -----------------------------------" << endl;
  // allocate HTA
  HTA_1H a =  HTA_1H::alloc (1, (Tuple<2>(2,2), Tuple<2>(2,2)), NULL, TILE);
  HTA_1H b =  HTA_1H::alloc (1, (Tuple<2>(2,2), Tuple<2>(2,2)), NULL, TILE);
  HTA_1H c =  HTA_1H::alloc (1, (Tuple<2>(2,2), Tuple<2>(2,2)), NULL, TILE);

  // initialize array
  double* a_x = a.raw();
  double* b_x = b.raw();
  for(int i=0; i < 16; i++) {
    a_x[i] = i;
    b_x[i] = i;
  }
  
  // c is initiaized with 0
  
  cout << "a=" << endl << a << endl;
  cout << "b=" << endl << b << endl;

  cout << "-----------------------------------" << endl;
  
  // test of simple code above ...
  matmul (a, b, c);
  
  cout << "c=" << endl << c << endl;

  /* test of library version ... */
  c = HTA_1H::matmul (a, b, NULL);

  cout << "c (library matmul) =" << endl << c << endl;
  
  cout << "level 1 end -----------------------------------" << endl;
}

void level_1b() 
{
  cout << "level 1b start -----------------------------------" << endl;
  // allocate HTA
  HTA_1H a =  HTA_1H::alloc (1, (Tuple<2>(3,2), Tuple<2>(1,2)), NULL, TILE);
  HTA_1H b =  HTA_1H::alloc (1, (Tuple<2>(2,3), Tuple<2>(2,1)), NULL, TILE);

  // initialize array
  double* a_x = a.raw();
  double* b_x = b.raw();
  for(int i=0; i < 12; i++) {
    a_x[i] = i;
    b_x[i] = i;
  }
  
  // c is initialized with 0
  
  cout << "a=" << endl << a << endl;
  cout << "b=" << endl << b << endl;

  cout << "-----------------------------------" << endl;
  
  /* test of library version ... */
  HTA_1H c = HTA_1H::matmul (a, b, NULL);

  cout << "c (library matmul) =" << endl << c << endl;

  cout << "level 1b end -----------------------------------" << endl;
}

int main(int argc, char** argv) 
{
  Traits::Default::init(argc, argv);
  level_0();
  level_1();
  level_1b();
  cout << TESTNAME << " PASSED" << endl;
  Traits::Default::finalize();
}
